/* GHM PROGRAM FLOW IN JLA MODE

When ghm is in JLA mode, it does not use the allele frequencies in the
marker file, but instead estimates them from the data.  Afterwards,
ghm restarts itself using the newly created marker file and produces
the final result files.  The restart happens in finalprep.c.
    When the 'jla p value' option is used, the second ghm run uses
the GHMRunner to compute the MOD score based on the real data.  This
happens in modscore.c.  Then, it calculates MOD scores based on
simulated replicates.  Finally, it compares them to the MOD score
from the real data to obtain an empirical p value.

The following abbreviations are used in the detailed description of
the ghm program flow below.

    i/ .......... input directory
    f.* ......... basename of input files
    f.cmd ....... commands file
    f.dat ....... marker file
    f.dat.new ... marker file with EM allele frequencies
    repl/ ....... directory for replicates
    real/ ....... input and output files for real dataset
    results.* ... log and output file (.log, .txt)

Case 1: JLA run without empirical p value calculation

    ghm i/f.cmd  ( becomes )
    ghm i/f.cmd --datfile i/f.dat.new  ( produces )
    tmp/results.*

Case 2: JLA run with empirical p value calculation

    ghm i/f.cmd  ( becomes )
    ghm i/f.cmd --datfile i/f.dat.new  ( starts GHMRunner )
        ghm repl/real/i/f.cmd  ( becomes )
        ghm repl/real/i/f.cmd --datfile i/f.dat.new  ( produces )
        repl/real/tmp/results.*

    > The MOD score from the GHMRunner run represents the real data
    > MOD score.  Then, we fork the replicates and compare their MOD
    > scores with the real data MOD score to compute the empirical
    > p value.

        ghm repl/1/i/f.cmd  ( becomes )
        ghm repl/1/i/f.cmd --datfile repl/1/i/f.dat.new  ( produces )
        repl/1/tmp/results.*
        ghm repl/2/i/f.cmd  ( becomes )
        ghm repl/2/i/f.cmd --datfile repl/2/i/f.dat.new  ( produces )
        repl/2/tmp/results.*
        ...

    > The MOD scores and corresponding best-fitting models from all
    > replicates results.txt files are collected into tmp/results.txt.
    > The resulting empirical p value is written to tmp/results.log.

    tmp/results.*
*/

#include "Association.h"
#include "HaplotypeFrequencyGrid.h"
#include "InheritanceDistribution.h"
#include "InheritanceVector.h"
#include "Int.h"
#include "LinkageMarkerFile.h"
#include "Memory.h"
#include "ModelParamsFile.h"
#include "ModScoreList.h"
#include "ModScorer.h"
#include "NELEMS.h"
#include "ParseObject.h"
#include "PedigreeFile.h"
#include "Region.h"
#include "TestMarkerSetGenerator.h"
#include "UserAssociationRegionsFile.h"
#include "UserLinkageMarkerFile.h"

// Legacy includes.
#define INC_LIB
#define INC_SHELL
#include "npl.h"

// System includes.
#include <stddef.h>
#include <string.h>

#ifndef _XOPEN_SOURCE
#define _XOPEN_SOURCE 500   # Feature Test Macro for strdup(3) and realpath(3)
#endif
#include <dirent.h>      // opendir, closedir
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>   // opendir, closedir
#include <unistd.h>


#define MAX_FILENAME_LENGTH 1000

#define MAX(x, y) ((x) > (y) ? (x) : (y))

typedef int (*PersonAdder)(Family, Person);

const char *g_user_linkage_marker_file = NULL;
const char *g_user_association_regions_file = NULL;
const char *g_output_file = NULL;
const char *g_model_params_filename = NULL;
PedigreeFile g_ped_file = NULL;
ModelParamsFile g_model_params_file = NULL;
int g_association;  /* truthy for joint linkage and association, otherwise falsy */
int g_frequencies;  /* truthy for allele frequency estimation, otherwise falsy */
Vector g_jla_regions = NULL;
static long number_of_markers_per_test_marker_set = -1;
static double maximum_test_marker_distance = 0.0;
int g_jla_p_value = 0;
int g_jla_replicates = -1;
int g_jla_child_processes = -1;
int g_jla_delta_iterations = 100;
int g_jla_hapfreq_iterations = 100;
int g_jla_delta_modscore = 0;
int g_jla_store_replicates = 0;
int g_jla_debug_replicates = 0;
int g_jla_resume = 0;
int g_timeout_single_modscore = 60;
int nlopt_iterations = 0;
int g_jla_nlopt_iterations_free = 40000;
int g_jla_nlopt_iterations_free_old = 40000;
const char *g_jla_directory = NULL;
const char *g_ghm_executable = NULL;
const char *g_commands_file = NULL;
const char *g_tag = NULL;  // label to annotate ghm runs
const char *g_jla_datfile = NULL;
Vector g_test_marker_names = NULL;

Vector g_families = NULL;

static Family create_family(int family_number);
static void add_founders(Family, int family_index);
static void add_nonfounders(Family, int family_index);
static Person create_person(PED_MEMBER *ped_member);
static void set_sex(Person, int sex);
static void set_affection_status(Person, int affection_status);
static Vector find_inheritance_distributions(Vector families);

Vector create_families_from_global_state(int number_of_families)
{
    Vector families = Vector_CreateEmpty();
    for (int i = 0; i < number_of_families; i++) {
        Vector_PushBack(families, create_family(i));
    }

    int number_of_markers = LinkageMarkerFile_GetNumberOfMarkers(g_linkage_file);

    for (int i = 0; i < Vector_GetSize(families); i++) {
        Family family = Vector_Get(families, i);
        for (int j = 0; j < Family_GetNumberOfMembers(family); j++) {
            Person person = Family_GetMember(family, j);
            // famnum_pointer is crucial: without it, skipped peds
            // would be ignored and wrong family names taken from the
            // ped file.  Family_GetId(family) is nothing but the
            // famnum-based family index from 0 thru num_pedigrees-1.
            // Hence, we use the same "famnum_pointer[famnum]" that
            // appears also in many other places in the code.
            const char *family_name = PedigreeFile_GetFamilyName(g_ped_file, famnum_pointer[Family_GetId(family)]);
            const char *person_name = Person_GetId(person);
            const int *alleles = PedigreeFile_GetConvertedAlleles(g_ped_file, family_name, person_name, g_linkage_file, g_user_linkage_file);
            Person_SetMarkerAlleles2(person, alleles, number_of_markers);
        }
        // The famtree struct is famnum-based, i.e., if the 2nd of 3
        // pedigrees is skipped, the indexes for the two remaining
        // pedigrees are 0 and 1.  Note, however, that the famtree
        // struct gets (partially) filled for the skipped pedigree,
        // such that its entries are not destroyed but overwritten!
        // This means e.g. that if the skipped pedigree is the last
        // one of the dataset its famtree could actually be accessed
        // and printed.  The same holds true if the skipped pedigree
        // has more members then the next non-skipped pedigree, where
        // the additional pedigree members of the skipped pedigree
        // could be accessed in the famtree of the next pedigree.
        // Family_Show(family);
    }
    return families;
}

static Family create_family(int family_number)
{
    Family family = Family_Create(family_number);
    add_founders(family, family_number);
    add_nonfounders(family, family_number);
    Family_SetParents(family);
    Family_SetSingleton(family);
    return family;
}

static void add_founders(Family family, int family_index)
{
    int number_of_founders = save_bf_originals[family_index];
    PED_MEMBER *family_members = famtree[family_index];
    int *founder_indices = save_bf_original_list[family_index];
    for (int i = 0; i < number_of_founders; i++) {
        Person person = create_person(family_members + founder_indices[i]);
        Family_AddFounder(family, person);
    }
}

static void add_nonfounders(Family family, int family_index)
{
    int number_of_nonfounders = save_bf_non_originals[family_index];
    PED_MEMBER *family_members = famtree[family_index];
    int *nonfounder_indices = save_bf_non_original_list[family_index];

    // Add nonfounders according to their position in the inheritance
    // vector as defined by legacy ghm (inh_vec_pos) starting from 0.
    for (int i = 0; i < number_of_nonfounders; i++) {
        // Add nonfounder at inheritance vector position i.
        for (int j = 0; j < number_of_nonfounders; j++) {
            PED_MEMBER *nonfounder = family_members + nonfounder_indices[j];
            if (nonfounder->inh_vec_pos == i) {
                Person person = create_person(nonfounder);
                Family_AddNonfounder(family, person);
                break;
            }
        }
    }
}

static Person create_person(PED_MEMBER *ped_member)
{
    Person person = Person_Create(ped_member->indiv_ID);
    Person_SetFatherId(person, ped_member->dad);
    Person_SetMotherId(person, ped_member->mom);
    set_sex(person, ped_member->sex);
    set_affection_status(person, ped_member->affectation_status);
    if (ped_member->dad_bit_uninf)
        Person_FixPaternalBit(person);
    if (ped_member->mom_bit_uninf)
        Person_FixMaternalBit(person);
    // Note: alleles will be set later via Person_SetMarkerAlleles2
    return person;
}

static void set_sex(Person person, int sex)
{
    switch (sex) {
        case 1: Person_SetMale(person); break;
        case 2: Person_SetFemale(person); break;
        default:
            fprintf(stderr, "Person with missing sex!\n");
            exit(1);
    };
}

static void set_affection_status(Person person, int affection_status)
{
    switch (affection_status) {
        case 0: break;  // unknown affection status
        case 1: Person_SetUnaffected(person); break;
        case 2: Person_SetAffected(person); break;
        default:
            fprintf(stderr, "Person with invalid affection status!\n");
            exit(1);
    };
}

void compute_mod_scores(void) {
    if(modcalc == 2) {
        sf(ps, "Error - 'modcalc' must be set to 'global' for JLA analysis!\n");
        pr();
        exit(1);
    }
    Vector inheritance_distros = find_inheritance_distributions(g_families);
    ModScoreList mod_score_list = ModScoreList_Create();
    ModScorer mod_scorer = ModScorer_Create(
        g_families,
        inheritance_distros,
        number_of_markers_per_test_marker_set,
        maximum_test_marker_distance,
        g_linkage_file,
        g_user_linkage_file,
        g_model_params_file,
        mod_score_list
    );
    if (g_test_marker_names) {
        if (!LinkageMarkerFile_IsCompatibleWithTestMarkers(g_linkage_file, g_test_marker_names)) {
            sf(ps, "Test markers specified for JLA analysis cannot be found in the marker file.\n");
            pr();
            exit(1);
        }
        if (g_model_params_file) {
            ModScorer_ComputeModScoreForTestMarkersAndUserModels(mod_scorer, g_test_marker_names);
        } else {
            ModScorer_ComputeModScoreForTestMarkers(mod_scorer, g_test_marker_names);
        }
    } else {
        ModScorer_ComputeModScoresForRegions(mod_scorer, g_jla_regions);
    }
    FILE *fp = fopen(g_output_file, "wb");
    if (!fp) {
        sf(ps, "Failed to open file: \"%s\"\n", g_output_file);
        pr();
        exit(1);
    }
    ModScoreList_Write(mod_score_list, fp, number_of_markers_per_test_marker_set);
    fclose(fp);
    for (int i = 0; i < Vector_GetSize(inheritance_distros); i++) {
        InheritanceDistribution_Destroy(Vector_Get(inheritance_distros, i));
    }
    Vector_Destroy(inheritance_distros);
}

static Vector find_inheritance_distributions(Vector families)
{
    Vector distros = Vector_Create(Vector_GetSize(families));
    for (int i = 0; i < Vector_GetSize(families); i++) {
        Family family = Vector_Get(families, i);
        InheritanceDistribution distro = NULL;
        if (!Family_GetSingleton(family)) {
            distro = InheritanceDistribution_CreateFromMatrix(save_p_all_pos[i], num_mod_pos, Family_GetNumberOfInheritanceVectors(family));
        }
        Vector_PushBack(distros, distro);
    }
    return distros;
}

void read_user_linkage_marker_file(void)
{
    if (interactive) {
        sf(ps, "JLA-related commands are ignored in interactive mode.\n");
        pr();
        return;
    }
    if (scan_done) {
        sf(ps, "The JLA command to read the user linkage marker file must be evoked before the 'scan pedigrees' command.\n");
        pr();
        exit(1);
    }
    g_user_linkage_marker_file = strndup(args, MAX_FILENAME_LENGTH);
    g_user_linkage_file = UserLinkageMarkerFile_Create(g_user_linkage_marker_file);
    if (g_user_linkage_file == NULL) {
        sf(ps, "user linkage marker file does not exist: %s\n", g_user_linkage_marker_file);
        pr();
        exit(1);
    }
    sf(ps, "read user linkage marker file: %s\n", g_user_linkage_marker_file);
    pr();
}


void save_absolute_path_to_ghm_executable(const char *argv0, const char **g_ghm_executable)
{
    if (strchr(argv0, '/')) {  // absolute or relative path
        char *resolved_path = realpath(argv0, NULL);
        if (resolved_path == NULL) {
            sf(ps, "failed to resolve path: %s\n", argv0);
            pr();
            exit(1);
        } else {
            *g_ghm_executable = resolved_path;
        }
    } else {  // ghm executable must be on the $PATH
        *g_ghm_executable = strdup(argv0);
    }
}

void set_jla(void)
{
    if (interactive) {
        sf(ps, "JLA-related commands are ignored in interactive mode.\n");
        pr();
        return;
    }
    if (scan_done) {
        sf(ps, "The command to specify JLA analysis must be invoked before the 'scan pedigrees' command.\n");
        pr();
        exit(1);
    }

    if (g_scan_steps_changed == TRUE) {  // 'increment' command was specified
        sf(ps, "Setting scan positions between marker loci (via the 'increment' command) is only useful for linkage analysis, but will be ignored during JLA analysis.\n");
        pr();
    }

    if (g_use_map == TRUE) {
        sf(ps, "The 'use' command is useful for linkage analysis.  If you really want to do\na JLA analysis, take out the 'use' command.\n");
        pr();
        exit(1);
    }

    if (g_commands_file == NULL) {
        sf(ps, "For JLA analysis, please pass the path to the"
               " commands file via the --commands command-line"
               " option.\n");
        pr();
        exit(EXIT_FAILURE);
    }

    g_association = 1;
    scan_steps = 1;

    sf(ps, "Joint linkage and association analysis option is 'on'.\n");
    pr();
}

void read_association_regions(void)
{
    if (interactive) {
        sf(ps, "JLA-related commands are ignored in interactive mode.\n");
        pr();
        return;
    }
    if (scan_done) {
        sf(ps, "The JLA command to read the user association regions file must be invoked before the 'scan pedigrees' command.\n");
        pr();
        exit(1);
    }

    if (g_scan_steps_changed == TRUE) {  // 'increment' command was specified
        sf(ps, "Setting scan positions between marker loci (via the 'increment' command) is only useful for linkage analysis, but will be ignored during JLA analysis.\n");
        pr();
    }

    if (g_use_map == TRUE) {
        sf(ps, "The 'use' command is useful for linkage analysis.  If you really want to do\na JLA analysis, take out the 'use' command.\n");
        pr();
        exit(1);
    }

    if (number_of_markers_per_test_marker_set == -1) {
        sf(ps, "The 'markers per set' command must precede the 'read association regions' command.\n");
        pr();
        exit(1);
    }

    if (g_commands_file == NULL) {
        sf(ps, "For JLA analysis, please pass the path to the"
               " commands file via the --commands command-line"
               " option.\n");
        pr();
        exit(EXIT_FAILURE);
    }

    g_association = 1;
    scan_steps = 1;
    g_user_association_regions_file = strndup(args, MAX_FILENAME_LENGTH);

    sf(ps, "Joint linkage and association analysis option is 'on'.\n");
    pr();
    sf(ps, "User association regions file: %s\n", g_user_association_regions_file);
    pr();
}

void read_number_of_markers_per_test_marker_set(void)
{
    if (interactive) {
        sf(ps, "JLA-related commands are ignored in interactive mode.\n");
        pr();
        return;
    }
    ParseObject parser = ParseObject_Create(args);
    long n;
    if (!ParseObject_NextInteger(parser, &n)) {
        sf(ps, "Failed to parse number of test markers per test marker set: %s\n", args);
        pr();
        exit(1);
    }
    ParseObject_Destroy(parser);
    if (n < 1 || 3 < n) {
        sf(ps, "Invalid number of markers per test marker set: want 1, 2, or 3, got %ld\n", n);
        pr();
        exit(1);
    }
    if (number_of_markers_per_test_marker_set != -1
        && number_of_markers_per_test_marker_set != n) {
        sf(ps, "Something is wrong: Another command has already set the number of test markers per test marker set\n");
        pr();
        exit(1);
    }
    sf(ps, "Number of markers per set: %ld\n", n);
    pr();
    number_of_markers_per_test_marker_set = n;
}

void read_maximum_test_marker_distance(void)
{
    if (interactive) {
        sf(ps, "JLA-related commands are ignored in interactive mode.\n");
        pr();
        return;
    }
    ParseObject parser = ParseObject_Create(args);
    if (!ParseObject_NextDouble(parser, &maximum_test_marker_distance)) {
        sf(ps, "Failed to parse maximum test marker distance: %s\n", args);
        pr();
        exit(1);
    }
    ParseObject_Destroy(parser);
    sf(ps, "Maximum distance in *centiMorgan* between test markers: %.4f\n", maximum_test_marker_distance);
    pr();
}

void set_output_file(void)
{
    g_output_file = strndup(args, MAX_FILENAME_LENGTH);
    FILE *fp = fopen(g_output_file, "ab");
    if (!fp) {
        sf(ps, "Failed to open file: \"%s\"\n", g_output_file);
        pr();
        exit(1);
    } else {
        fclose(fp);
    }
    sf(ps, "Output file: %s\n", g_output_file);
    pr();
}

void read_model_params_file(void)
{
    if (interactive) {
        sf(ps, "JLA-related commands are ignored in interactive mode.\n");
        pr();
        return;
    }
    g_model_params_filename = strndup(args, MAX_FILENAME_LENGTH);
    g_model_params_file = ModelParamsFile_Create(g_model_params_filename);

    sf(ps, "read association models: %s\n", g_model_params_filename);
    pr();
    sf(ps, "Number of user-specified association models: %d\n", ModelParamsFile_GetNumberOfModels(g_model_params_file));
    pr();
}

void calculate_jla_p_value(void)
{
    if (interactive) {
        sf(ps, "JLA-related commands are ignored in interactive mode.\n");
        pr();
        return;
    }
    g_jla_p_value = 1;
    sf(ps, "Calculation of JLA p value is on.\n");
    pr();
}

void set_jla_replicates(void)
{
    if (interactive) {
        sf(ps, "JLA-related commands are ignored in interactive mode.\n");
        pr();
        return;
    }
    ParseObject parser = ParseObject_Create(args);
    long k;
    if (!ParseObject_NextInteger(parser, &k)) {
        sf(ps, "Failed to parse number of JLA replicates: %s\n", args);
        pr();
        exit(1);
    }
    g_jla_replicates = k;
    ParseObject_Destroy(parser);
    sf(ps, "Number of JLA replicates: %d\n", g_jla_replicates);
    pr();
}

void set_jla_child_processes(void)
{
    if (interactive) {
        sf(ps, "JLA-related commands are ignored in interactive mode.\n");
        pr();
        return;
    }
    ParseObject parser = ParseObject_Create(args);
    long k;
    if (!ParseObject_NextInteger(parser, &k)) {
        sf(ps, "Failed to parse number of JLA child processes: %s\n", args);
        pr();
        exit(1);
    }
    g_jla_child_processes = k;
    ParseObject_Destroy(parser);
    sf(ps, "Number of child JLA processes (simulations): %d\n", g_jla_child_processes);
    pr();
}

void set_jla_delta_iterations(void)
{
    if (interactive) {
        sf(ps, "JLA-related commands are ignored in interactive mode.\n");
        pr();
        return;
    }
    ParseObject parser = ParseObject_Create(args);
    long k;
    if (!ParseObject_NextInteger(parser, &k)) {
        sf(ps, "Please, specify the number of delta iterations\n");
        pr();
        exit(1);
    }
    g_jla_delta_iterations = k;
    ParseObject_Destroy(parser);
    sf(ps, "Number of delta iterations: %d\n", g_jla_delta_iterations);
    pr();
}

void set_jla_hapfreq_iterations(void)
{
    if (interactive) {
        sf(ps, "JLA-related commands are ignored in interactive mode.\n");
        pr();
        return;
    }
    ParseObject parser = ParseObject_Create(args);
    long k;
    if (!ParseObject_NextInteger(parser, &k)) {
        sf(ps, "Please, specify the number of haplotype frequency iterations\n");
        pr();
        exit(1);
    }
    g_jla_hapfreq_iterations = k;
    ParseObject_Destroy(parser);
    sf(ps, "Number of haplotype frequency iterations: %d\n", g_jla_hapfreq_iterations);
    pr();
}

void set_jla_nlopt_iterations(void)
{
    if (interactive) {
        sf(ps, "JLA-related commands are ignored in interactive mode.\n");
        pr();
        return;
    }
    ParseObject parser = ParseObject_Create(args);
    long k;
    if (!ParseObject_NextInteger(parser, &k)) {
        sf(ps, "Please, specify the number of maximum nlopt iterations\n");
        pr();
        exit(1);
    }
    g_jla_nlopt_iterations_free = k;
    g_jla_nlopt_iterations_free_old = k;
    ParseObject_Destroy(parser);
    sf(ps, "Number of maximum nlopt iterations: %d\n", g_jla_nlopt_iterations_free);
    pr();
}

void set_jla_delta_modscore(void)
{
    if (interactive) {
        sf(ps, "JLA-related commands are ignored in interactive mode.\n");
        pr();
        return;
    }
    char type[TOKLEN+1];
    get_one_arg(stoken, "", type);
    if (matches(type, "ON") || matches(type, "TRUE")) {
        g_jla_delta_modscore = 1;
    } else if (matches(type, "OFF") || matches(type, "FALSE")) {
        g_jla_delta_modscore = 0;
    } else if (!nullstr(type)) {
        usage_error(1);
    }
    sf(ps, "Delta modscore calculation mode is now '%s'.\n",
        g_jla_delta_modscore ? "on" : "off");
    pr();
}

void set_jla_directory(void)
{
    if (interactive) {
        sf(ps, "JLA-related commands are ignored in interactive mode.\n");
        pr();
        return;
    }
    char *s = strndup(args, MAX_FILENAME_LENGTH);
    if (s[strlen(s) - 1] == '/') {
        s[strlen(s) - 1] = '\0';
    }
    g_jla_directory = s;
    sf(ps, "JLA directory: %s\n", g_jla_directory);
    pr();
    // Fail unless replicates directory exists
    DIR *dp = opendir(g_jla_directory);
    if (dp == NULL) {
        sf(ps, "Failed to open jla directory: %s\nDid you create it?\n", g_jla_directory);
        pr();
        exit(EXIT_FAILURE);
    } else {
        closedir(dp);
    }
}

void set_jla_store_replicates(void)
{
    if (interactive) {
        sf(ps, "JLA-related commands are ignored in interactive mode.\n");
        pr();
        return;
    }
    g_jla_store_replicates = 1;
    sf(ps, "jla store replicates: on\n");
    pr();
}

void set_jla_debug_replicates(void)
{
    if (interactive) {
        sf(ps, "JLA-related commands are ignored in interactive mode.\n");
        pr();
        return;
    }
    g_jla_debug_replicates = 1;
    sf(ps, "jla debug replicates: on\n");
    pr();
}

void read_test_markers(void)
{
    if (interactive) {
        sf(ps, "JLA-related commands are ignored in interactive mode.\n");
        pr();
        return;
    }
    g_test_marker_names = Vector_CreateEmpty();
    ParseObject parser = ParseObject_Create(args);
    g_test_marker_names = ParseObject_SplitIntoWords(parser);
    ParseObject_Destroy(parser);
    if (Vector_GetSize(g_test_marker_names) > 3 || Vector_GetSize(g_test_marker_names) == 0) {
        sf(ps, "Error - invalid number of test markers: %d\n", Vector_GetSize(g_test_marker_names));
        pr();
        exit(1);
    }
    if (number_of_markers_per_test_marker_set != -1
        && number_of_markers_per_test_marker_set != Vector_GetSize(g_test_marker_names))
    {
        sf(ps, "Something is wrong: Another command has already set the number of test markers per test marker set\n");
        pr();
        exit(1);
    }
    sf(ps, "The following test markers were chosen for the analysis:\n");
    pr();
    for (int i = 0; i < Vector_GetSize(g_test_marker_names); i++) {
        sf(ps, "Test marker %s\n", (char *) Vector_Get(g_test_marker_names, i));
        pr();
    }
    number_of_markers_per_test_marker_set = Vector_GetSize(g_test_marker_names);
}

void set_timeout_single_modscore(void)
{
    if (interactive) {
        sf(ps, "JLA-related commands are ignored in interactive mode.\n");
        pr();
        return;
    }
    ParseObject parser = ParseObject_Create(args);
    long k;
    if (!ParseObject_NextInteger(parser, &k)) {
        sf(ps, "Please, specify a valid timeout [s] for a single modscore evaluation\n");
        pr();
        exit(1);
    }
    g_timeout_single_modscore = k;
    ParseObject_Destroy(parser);
    sf(ps, "Timeout for single modscore evaluation: %d seconds\n", g_timeout_single_modscore);
    pr();
}

int restart_ghm_with_new_datfile(const char *datfile)
{
    sf(ps, "\nStarting new GHM run to include EM-estimated linkage marker\nallele frequencies for calculation of the inheritance distribution.\n");
    pr();
    fclose(photo);
    char *arguments[9];
    int i = 0;
    // Note that strdup is needed because assigning a "const char *"
    // to a "char *" array would discard the const qualifier.
    arguments[i++] = strdup(g_ghm_executable);
    if (g_tag) {
        arguments[i++] = "--tag";
        arguments[i++] = strdup(g_tag);
    }
    if (g_jla_resume) {
        arguments[i++] = "--resume";
    }
    arguments[i++] = "--commands";
    arguments[i++] = strdup(g_commands_file);
    arguments[i++] = "--dat";
    arguments[i++] = strdup(datfile);
    arguments[i++] = NULL;
    return execvp(strdup(g_ghm_executable), arguments);
}

void set_allele_frequency_estimation(void)
{
    if (interactive) {
        sf(ps, "JLA-related commands are ignored in interactive mode.\n");
        pr();
        return;
    }
    if (scan_done) {
        sf(ps, "The command to specify the computation of allele frequencies must be invoked before the 'scan pedigrees' command.\n");
        pr();
        exit(1);
    }
    if (analysis_type == NPL_ANALYSIS) {
        sf(ps, "In the case of an NPL analysis, the computation of allele frequencies is not possible.\n");
        pr();
        exit(1);
    }
    if (g_use_map == TRUE) {
        sf(ps, "The 'use' command is not supported for the computation of allele frequencies. Please remove it from the commands file.\n");
        pr();
        exit(1);
    }

    if (g_commands_file == NULL) {
        sf(ps, "For the computation of allele frequencies, please pass the path to the"
               " commands file via the --commands command-line"
               " option.\n");
        pr();
        exit(EXIT_FAILURE);
    }

    g_frequencies = 1;
    scan_steps = 1;

    sf(ps, "Estimation of allele frequencies is now activated.\n");
    pr();
}

void save_allele_frequencies(void)
{
    if (!scan_done) {
        sf(ps, "This option must be preceded by the 'scan pedigrees' command.\n");
        pr();
        if (!interactive) {
            exit(1);
        } else {
            return;
        }
    }

    if (!g_association || !g_frequencies) {
        sf(ps, "This option only works with both the 'joint linkage and association' and\n'allele frequency estimation' options turned on.\n");
        pr();
        if (!interactive) {
            exit(1);
        } else {
            return;
        }
    }

    if (g_user_linkage_file == NULL) {
        sf(ps, "User linkage marker file does not exist, but is needed for the computation of allele frequencies.\nThe file should contain marker names (one per row) to indicate those markers,\nfor which allele frequencies are to be calculated.\n");
        pr();
        exit(1);
    }

    const char *filename = strndup(args, MAX_FILENAME_LENGTH);
    unsigned max_marker_width = strlen("Marker");
    sf(ps, "Saving allele frequencies in: %s\n", filename);
    pr();
    Vector families = create_families_from_global_state(num_pedigrees);
    Vector markers = LinkageMarkerFile_GetMarkers(g_linkage_file);
    Vector selected_markers = Vector_CreateEmpty();
    for (int i = 0; i < Vector_GetSize(markers); i++) {
        Marker marker = Vector_Get(markers, i);
        if (UserLinkageMarkerFile_Contains(g_user_linkage_file, Marker_GetName(marker))) {
            Vector_PushBack(selected_markers, marker);
        }
    }
    EM_ComputeMarkerFrequencies(selected_markers, families);
    int number_of_selected_markers = Vector_GetSize(selected_markers);
    FILE *fp = fopen(filename, "wb");
    if (!fp) {
        sf(ps, "Failed to open file: \"%s\"\n", filename);
        pr();
        exit(1);
    }
    for (int i = 0; i < number_of_selected_markers; i++) {
        Marker marker = Vector_Get(selected_markers, i);
        const char *marker_name = Marker_GetName(marker);
        max_marker_width = MAX(max_marker_width, strlen(marker_name));
    }

    fprintf(fp, "%*s ", max_marker_width, "Marker");
    fprintf(fp, "%*s ", 7, "Allele");
    fprintf(fp, "%*s\n", 10, "Frequency");
    for (int i = 0; i < number_of_selected_markers; i++) {
        Marker marker = Vector_Get(selected_markers, i);
        for (int j = 0; j < Marker_GetNumberOfAlleles(marker); j++) {
            fprintf(fp, "%*s %*d %*f\n", max_marker_width, Marker_GetName(marker), 7, j + 1, 10, Marker_GetAlleleFrequency(marker, j));
        }
    }
}
