// Legacy includes for print messages to log file
#define INC_LIB
#define INC_SHELL
#include "npl.h"

#undef command

#include "CommandReader.h"
#include "IO.h"
#include "Memory.h"
#include "Vector.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>      // int16_t, INT16_MAX

#define NELEMS(X) (sizeof(X) / sizeof((X)[0]))

struct CommandReaderStruct {
    Vector commands;
    const char *ped_file;
    const char *marker_file;
    const char *user_association_regions_file;
    const char *user_linkage_markers_file;
    const char *user_association_models_file;
    const char *log_file;
    const char *output_file;
};

struct FileCommand {
    const char *command;
    const char *file;
} FILE_COMMANDS[] = {
    {"log file ",                 "tmp/results.log"},
    {"photo ",                    "tmp/results.log"},  // alias for "log file"
    {"output file ",              "tmp/results.txt"},
    {"read association regions ", "input/user_association_regions.txt"},
    {"read linkage markers ",     "input/user_linkage_markers.txt"},
    {"read association models ",  "input/user_association_models.txt"},
    {"load markers ",             "input/sample.dat"},
    {"scan pedigrees ",           "input/sample.ped"}
};

CommandReader CommandReader_Create(const char *command_file)
{
    CommandReader reader = Memory_Malloc(sizeof(*reader));
    reader->commands = Vector_CreateEmpty();
    FILE *fp = fopen(command_file, "rb");
    if (!fp) {
        return NULL;
    }
    char *line;
    struct FileCommand {
        const char *command;
        const char **file;
    } file_commands[] = {
        {"scan pedigrees ", &reader->ped_file},
        {"load markers ", &reader->marker_file},
        {"read association regions ", &reader->user_association_regions_file},
        {"read linkage markers ", &reader->user_linkage_markers_file},
        {"read association models ", &reader->user_association_models_file},
        {"log file ", &reader->log_file},
        {"photo ", &reader->log_file},  // alias for "log file"
        {"output file ", &reader->output_file}
    };
    for (int i = 0; i < (int) NELEMS(file_commands); i++) {
        *file_commands[i].file = NULL;
    }
    while ((line = IO_GetLine(fp)) != NULL) {
        line[strlen(line) - 1] = '\0';  // remove trailing newline
        for (int i = 0; i < (int) NELEMS(file_commands); i++) {
            struct FileCommand fc = file_commands[i];
            if (strncmp(fc.command, line, strlen(fc.command)) == 0) {
                *fc.file = line + strlen(fc.command);
                break;
            }
        }
        Vector_PushBack(reader->commands, line);
    }
    fclose(fp);
    return reader;
}

void CommandReader_Write(CommandReader reader, const char *new_command_file, int16_t seed)
{
    FILE *fp = fopen(new_command_file, "wb");
    if (!fp) {
        sf(ps, "Failed to open file: \"%s\"\n", new_command_file);
        pr();
        exit(1);
    }
    for (int i = 0; i < Vector_GetSize(reader->commands); i++) {
        const char *command = Vector_Get(reader->commands, i);
        if (strncmp(command, "jla p value", 11) == 0
         || strncmp(command, "jla replicates", 14) == 0
         || strncmp(command, "jla child processes", 19) == 0
         || strncmp(command, "jla directory", 13) == 0
         || strncmp(command, "jla store replicates", 20) == 0
         || strncmp(command, "set random seed", 15) == 0
         || strncmp(command, "jla mod score", 13) == 0
         || strncmp(command, "jla debug replicates", 20) == 0)
        {
            continue;  // don't insert these simulation-specific jla commands
        }
        int handled = 0;
        for (int i = 0; i < (int) NELEMS(FILE_COMMANDS); i++) {
            struct FileCommand fc = FILE_COMMANDS[i];
            if (strncmp(fc.command, command, strlen(fc.command)) == 0) {
                fprintf(fp, "%s%s\n", fc.command, fc.file);
                if (i == 0)
                    fprintf(fp, "set random seed %u\n", seed);
                handled = 1;
                break;
            }
        }
        if (!handled) {
            fprintf(fp, "%s\n", command);
        }
    }
    fclose(fp);
}

const char *CommandReader_GetPedFile(CommandReader reader)
{
    return reader->ped_file;
}

const char *CommandReader_GetMarkerFile(CommandReader reader)
{
    return reader->marker_file;
}

const char *CommandReader_GetUserAssociationRegionsFile(CommandReader reader)
{
    return reader->user_association_regions_file;
}

const char *CommandReader_GetUserLinkageMarkersFile(CommandReader reader)
{
    return reader->user_linkage_markers_file;
}

const char *CommandReader_GetUserAssociationModelsFile(CommandReader reader)
{
    return reader->user_association_models_file;
}

const char *CommandReader_GetLogFile(CommandReader reader)
{
    return reader->log_file;
}

const char *CommandReader_GetOutputFile(CommandReader reader)
{
    return reader->output_file;
}

static const char *find_source_file(CommandReader reader, const char *command_name)
{
    const char *result = NULL;
    if (strcmp(command_name, "read association regions ") == 0) {
        result = reader->user_association_regions_file;
    } else if (strcmp(command_name, "read linkage markers ") == 0) {
        result = reader->user_linkage_markers_file;
    } else if (strcmp(command_name, "read association models ") == 0) {
        result = reader->user_association_models_file;
    } else if(strcmp(command_name, "load markers ") == 0) {
        result = reader->marker_file;
    } else if (strcmp(command_name, "scan pedigrees ") == 0) {
        result = reader->ped_file;
    }
    return result;
}

static const char *find_destination_file(const char *command_name)
{
    for (int i = 0; i < (int) NELEMS(FILE_COMMANDS); i++) {
        struct FileCommand fc = FILE_COMMANDS[i];
        if (strncmp(fc.command, command_name, strlen(fc.command)) == 0) {
            return fc.file;
        }
    }
    return NULL;
}

int CommandReader_CopyUserFile(CommandReader reader, const char *command_name, const char *directory)
{
    const char *source = find_source_file(reader, command_name);
    const char *relative_destination = find_destination_file(command_name);
    const int max_length = 200;
    char destination[max_length];
    int chars_written = snprintf(destination, max_length, "%s/%s", directory, relative_destination);
    if (chars_written >= max_length) {
        fprintf(stderr, "File path too long: %s/%s\n", directory, relative_destination);
        return 0;
    }
    FILE *fin = fopen(source, "rb");
    if (!fin) {
        sf(ps, "Failed to open file: \"%s\"\n", source);
        pr();
        exit(1);
    }
    FILE *fout = fopen(destination, "wb");
    if (!fout) {
        sf(ps, "Failed to open file: \"%s\"\n", destination);
        pr();
        exit(1);
    }
    char *line;
    while ((line = IO_GetLine(fin)) != NULL) {
        fprintf(fout, "%s", line);
        Memory_Free(line);
    }
    fclose(fin);
    fclose(fout);
    return 1;
}
