#include "AnnotatedFamily.h"
#include "EM.h"
#include "FounderHaplotypes.h"
#include "Marker.h"
#include "Memory.h"
#include "Vector.h"

#include <math.h>

#include <stdio.h>

// Legacy includes.
#define INC_LIB
#define INC_SHELL
#include "npl.h"

int EM_verbose = 1;

// static void print_triplet_and_exponents(FounderHaplotypes);

static double get_hap_freq_prod(double *hap_freqs, int *exponents, int n)
{
    double hap_freq_prod = 1.0;
    for (int i = 0; i < n; i++) {
        int e = exponents[i];
        while (e) {
            hap_freq_prod *= hap_freqs[i];
            e--;
        }
    }
    return hap_freq_prod;
}

double *EM_Allele_Frequencies(Vector annotated_families, int number_of_families, int total_founders, Marker marker, int which_marker) {
    double *em_freqs = Memory_Malloc(2 * sizeof(double));
    double *em_freqs_old = Memory_Malloc(2 * sizeof(double));
    double em_numerator;
    double em_denominator;
    double em_families;
    TripletHashTableIterator it = 0;
    double accuracy = 1e-8;
    const double constant = 1.0 / (2 * total_founders);
    double *all_equal = Memory_Malloc(2 * sizeof(double));
    double *all_equal_old = Memory_Malloc(2 * sizeof(double));
    int number_of_haplotypes = 2;  // one biallelic locus
    for (int i = 0; i < number_of_haplotypes; i++) {
        em_freqs[i] = Marker_GetAlleleFrequency(marker, i);
    }

    if (EM_verbose) {
        fprintf(stderr, "\nTest marker: %s\n", Marker_GetName(marker));
        fprintf(stderr, "Initial Values for EM:\n");
        for (int i = 0; i < number_of_haplotypes; i++) {
            fprintf(stderr, "%f ", em_freqs[i]);
        }
        fprintf(stderr, "\n");
    }
    double difference = accuracy + 1;
    double likelihood_old = 0.0;
    double likelihood_new = 0.0;
    int niterations = 0;
    while (difference > accuracy) {
        ++niterations;
        difference = 0.0;
        for (int i = 0; i < number_of_haplotypes; i++) {
            em_families = 0.0;
            em_freqs_old[i] = em_freqs[i];
            for (int j = 0; j < number_of_families; j++) {
                AnnotatedFamily annotated_family = Vector_Get(annotated_families, j);
                FounderHaplotypes founder_haplotypes = AnnotatedFamily_GetFounderHaplotypes(annotated_family);
                em_numerator = 0.0;
                em_denominator = 0.0;
                if (which_marker == 0) {
                    it = FounderHaplotypes_GetIterator_1(founder_haplotypes);
                }
                else if (which_marker == 1) {
                    it = FounderHaplotypes_GetIterator_2(founder_haplotypes);
                }
                else if (which_marker == 2) {
                    it = FounderHaplotypes_GetIterator_3(founder_haplotypes);
                }
                while (TripletHashTableIterator_HasNext(it)) {
                    TripletHashTableIterator_Next(it);
                    int *exponents = TripletHashTableIterator_GetValue(it);
                    double hap_freq_prod = get_hap_freq_prod(em_freqs, exponents, number_of_haplotypes);
                    em_numerator += exponents[number_of_haplotypes] * exponents[i] * hap_freq_prod;
                    em_denominator += exponents[number_of_haplotypes] * hap_freq_prod;
                }
                em_families += (em_numerator / em_denominator);
            }
            em_freqs[i] = constant * em_families;
        }
        if(niterations == 1)
            likelihood_old = EM_Allele_Frequencies_GetLikelihood(annotated_families, em_freqs_old, number_of_families, 2, which_marker);
        likelihood_new = EM_Allele_Frequencies_GetLikelihood(annotated_families, em_freqs, number_of_families, 2, which_marker);
        difference = fabs(likelihood_new - likelihood_old);
        likelihood_old = likelihood_new;
    }

    if (EM_verbose) {
        fprintf(stderr, "EM iterations using pre-specified marker allele frequencies as initial values: %d\n", niterations);
    }
    double total_likelihood = EM_Allele_Frequencies_GetLikelihood(annotated_families, em_freqs, number_of_families, number_of_haplotypes, which_marker);
    if (EM_verbose) {
        fprintf(stderr, "Log-likelihood of the allele frequencies using pre-specified allele frequencies as initial values: %.8f\n", total_likelihood);
    }

    for (int i = 0; i < number_of_haplotypes; i++) {
        all_equal[i] = 1.0 / 2.0;
    }
    if (EM_verbose) {
        fprintf(stderr, "Initial Values for EM:\n");
        for (int i = 0; i < number_of_haplotypes; i++) {
            fprintf(stderr, "%f ", all_equal[i]);
        }
        fprintf(stderr, "\n");
    }

    likelihood_old = 0.0;
    likelihood_new = 0.0;
    difference = accuracy + 1;
    niterations = 0;
    while (difference > accuracy) {
        ++niterations;
        difference = 0.0;
        for (int i = 0; i < number_of_haplotypes; i++) {
            em_families = 0.0;
            all_equal_old[i] = all_equal[i];
            for (int j = 0; j < number_of_families; j++) {
                AnnotatedFamily annotated_family = Vector_Get(annotated_families, j);
                FounderHaplotypes founder_haplotypes = AnnotatedFamily_GetFounderHaplotypes(annotated_family);
                em_numerator = 0.0;
                em_denominator = 0.0;
                if (which_marker == 0) {
                    it = FounderHaplotypes_GetIterator_1(founder_haplotypes);
                }
                else if (which_marker == 1) {
                    it = FounderHaplotypes_GetIterator_2(founder_haplotypes);
                }
                else if (which_marker == 2) {
                    it = FounderHaplotypes_GetIterator_3(founder_haplotypes);
                }
                while (TripletHashTableIterator_HasNext(it)) {
                    TripletHashTableIterator_Next(it);
                    int *exponents = TripletHashTableIterator_GetValue(it);
                    double hap_freq_prod = get_hap_freq_prod(all_equal, exponents, number_of_haplotypes);
                    em_numerator += exponents[number_of_haplotypes] * exponents[i] * hap_freq_prod;
                    em_denominator += exponents[number_of_haplotypes] * hap_freq_prod;
                }
                em_families += (em_numerator / em_denominator);
            }
            all_equal[i] = constant * em_families;
        }
        if(niterations == 1)
            likelihood_old = EM_Allele_Frequencies_GetLikelihood(annotated_families, all_equal_old, number_of_families, number_of_haplotypes, which_marker);
        likelihood_new = EM_Allele_Frequencies_GetLikelihood(annotated_families, all_equal, number_of_families, number_of_haplotypes, which_marker);
        difference = fabs(likelihood_new - likelihood_old);
        likelihood_old = likelihood_new;
    }
    if (EM_verbose) {
        fprintf(stderr, "EM iterations using equal initial values: %d\n", niterations);
    }
    double total_likelihood_equal = EM_Allele_Frequencies_GetLikelihood(annotated_families, all_equal, number_of_families, number_of_haplotypes, which_marker);
    if (EM_verbose) {
        fprintf(stderr, "Log-likelihood of the allele frequencies using equal initial values: %.8f\n", total_likelihood_equal);
    }
    if (total_likelihood > total_likelihood_equal) {
        if (EM_verbose) {
            fprintf(stderr, "Pre-specified allele frequencies win\n");
        }
    } else if(total_likelihood < total_likelihood_equal) {
        if (EM_verbose) {
            fprintf(stderr, "Equal initial values win\n");
        }
        em_freqs = all_equal;
    } else {
        if (EM_verbose) {
            fprintf(stderr, "Equally likely solutions with pre-specified and equal initial values\n");
        }
    }
    return em_freqs;
}

// Return vector of haplotype frequencies or NULL if any family
// contains a recombination event.
DoubleVector EM_Haplotype_H0(Vector annotated_families, Vector test_markers) {
    int number_of_families = Vector_GetSize(annotated_families);
    for (int i = 0; i < number_of_families; i++) {
        AnnotatedFamily family = Vector_Get(annotated_families, i);
        FounderHaplotypes founder_haplotypes = AnnotatedFamily_GetFounderHaplotypes(family);
        if (FounderHaplotypes_WithRecombination(founder_haplotypes)) {
            Family f = AnnotatedFamily_GetFamily(family);
            sf(ps, "Family %d has a recombination event or bad inheritance.\n", Family_GetId(f));
            pr();
            return NULL;
        }
    }
    AnnotatedFamily first_family = Vector_Get(annotated_families, 0);
    FounderHaplotypes first_founder_haplotypes = AnnotatedFamily_GetFounderHaplotypes(first_family);
    int number_of_haplotypes = FounderHaplotypes_GetNumberOfHaplotypes(first_founder_haplotypes);
    int number_of_test_markers = FounderHaplotypes_GetNumberOfTestMarkers(first_founder_haplotypes);
    double *em_freqs = Memory_Malloc(number_of_haplotypes * sizeof(double));
    double *em_freqs_old = Memory_Malloc(number_of_haplotypes * sizeof(double));
    double *all_equal = Memory_Malloc(number_of_haplotypes * sizeof(double));
    double *all_equal_old = Memory_Malloc(number_of_haplotypes * sizeof(double));
    double em_numerator;
    double em_denominator;
    double em_families;
    double accuracy = 1e-8;
    int total_founders = 0;
    DoubleVector hap_freqs = DoubleVector_Create(number_of_haplotypes);
    DoubleVector allele_freqs = DoubleVector_Create(2 * number_of_test_markers);
    for (int i = 0; i < number_of_families; i++) {
        AnnotatedFamily family = Vector_Get(annotated_families, i);
        FounderHaplotypes founder_haplotypes = AnnotatedFamily_GetFounderHaplotypes(family);
        int number_of_founders = FounderHaplotypes_GetNumberOfFounders(founder_haplotypes);
        total_founders += number_of_founders;
    }
    for (int i = 0; i < number_of_test_markers; i++) {
        Marker marker = Vector_Get(test_markers, i);
        double *freqs = Marker_GetEMAlleleFrequencies(marker);
        if (freqs == NULL) {
            freqs = EM_Allele_Frequencies(annotated_families, number_of_families, total_founders, marker, i);
            Marker_SetEMAlleleFrequencies(marker, freqs);
        }
        for (int j = 0; j < 2; j++) {
            DoubleVector_PushBack(allele_freqs, freqs[j]);
        }
        if (number_of_test_markers == 1) {
            double total_likelihood_1tm = EM_Haplotype_GetLikelihood(annotated_families, freqs, number_of_families, 2);
            if (EM_verbose) {
                fprintf(stderr, "\nLog-likelihood of the allele frequencies: %.8f\n", total_likelihood_1tm);
            }
        }
    }
    if (EM_verbose) {
        fprintf(stderr, "\nAllele frequencies: TM1.1 TM1.2 ...\n");
        for (int i = 0; i < DoubleVector_GetSize(allele_freqs); i++) {
            fprintf(stderr, "%f ", DoubleVector_Get(allele_freqs, i));
        }
        fprintf(stderr, "\n");
    }
    if (number_of_test_markers == 1) {
        return allele_freqs;
    }
    DoubleVector initial_hap_freqs = DoubleVector_Create(number_of_haplotypes);
    if (number_of_test_markers == 2) {
        DoubleVector_PushBack(initial_hap_freqs, DoubleVector_Get(allele_freqs, 0) * DoubleVector_Get(allele_freqs, 2));
        DoubleVector_PushBack(initial_hap_freqs, DoubleVector_Get(allele_freqs, 1) * DoubleVector_Get(allele_freqs, 2));
        DoubleVector_PushBack(initial_hap_freqs, DoubleVector_Get(allele_freqs, 0) * DoubleVector_Get(allele_freqs, 3));
        DoubleVector_PushBack(initial_hap_freqs, DoubleVector_Get(allele_freqs, 1) * DoubleVector_Get(allele_freqs, 3));
    } else if (number_of_test_markers == 3) {
        DoubleVector_PushBack(initial_hap_freqs, DoubleVector_Get(allele_freqs, 0) * DoubleVector_Get(allele_freqs, 2) * DoubleVector_Get(allele_freqs, 4));
        DoubleVector_PushBack(initial_hap_freqs, DoubleVector_Get(allele_freqs, 1) * DoubleVector_Get(allele_freqs, 2) * DoubleVector_Get(allele_freqs, 4));
        DoubleVector_PushBack(initial_hap_freqs, DoubleVector_Get(allele_freqs, 0) * DoubleVector_Get(allele_freqs, 3) * DoubleVector_Get(allele_freqs, 4));
        DoubleVector_PushBack(initial_hap_freqs, DoubleVector_Get(allele_freqs, 1) * DoubleVector_Get(allele_freqs, 3) * DoubleVector_Get(allele_freqs, 4));
        DoubleVector_PushBack(initial_hap_freqs, DoubleVector_Get(allele_freqs, 0) * DoubleVector_Get(allele_freqs, 2) * DoubleVector_Get(allele_freqs, 5));
        DoubleVector_PushBack(initial_hap_freqs, DoubleVector_Get(allele_freqs, 1) * DoubleVector_Get(allele_freqs, 2) * DoubleVector_Get(allele_freqs, 5));
        DoubleVector_PushBack(initial_hap_freqs, DoubleVector_Get(allele_freqs, 0) * DoubleVector_Get(allele_freqs, 3) * DoubleVector_Get(allele_freqs, 5));
        DoubleVector_PushBack(initial_hap_freqs, DoubleVector_Get(allele_freqs, 1) * DoubleVector_Get(allele_freqs, 3) * DoubleVector_Get(allele_freqs, 5));
    }
    const double constant = 1.0 / (2 * total_founders);
    for (int i = 0; i < number_of_haplotypes; i++) {
        em_freqs[i] = DoubleVector_Get(initial_hap_freqs, i);
    }
    if (EM_verbose) {
        fprintf(stderr, "Initial Values for EM:\n");
        for (int i = 0; i < number_of_haplotypes; i++) {
            fprintf(stderr, "%f ", em_freqs[i]);
        }
        fprintf(stderr, "\n");
    }

    double likelihood_old = 0.0;
    double likelihood_new = 0.0;
    double difference = accuracy + 1;
    int niterations = 0;
    while (difference > accuracy) {
        ++niterations;
        difference = 0.0;
        for (int i = 0; i < number_of_haplotypes; i++) {
            em_families = 0.0;
            em_freqs_old[i] = em_freqs[i];
            for (int j = 0; j < number_of_families; j++) {
                AnnotatedFamily annotated_family = Vector_Get(annotated_families, j);
                FounderHaplotypes founder_haplotypes = AnnotatedFamily_GetFounderHaplotypes(annotated_family);
                em_numerator = 0.0;
                em_denominator = 0.0;
                TripletHashTableIterator it = FounderHaplotypes_GetIterator(founder_haplotypes);
                while (TripletHashTableIterator_HasNext(it)) {
                    TripletHashTableIterator_Next(it);
                    int *exponents = TripletHashTableIterator_GetValue(it);
                    double hap_freq_prod = get_hap_freq_prod(em_freqs, exponents, number_of_haplotypes);
                    em_numerator += exponents[number_of_haplotypes] * exponents[i] * hap_freq_prod;
                    em_denominator += exponents[number_of_haplotypes] * hap_freq_prod;
                }
                em_families += (em_numerator / em_denominator);
            }
            em_freqs[i] = constant * em_families;
        }
        if (niterations == 1)
            likelihood_old = EM_Haplotype_GetLikelihood(annotated_families, em_freqs_old, number_of_families, number_of_haplotypes);
        likelihood_new = EM_Haplotype_GetLikelihood(annotated_families, em_freqs, number_of_families, number_of_haplotypes);
        // fprintf(stderr, "OLD LIKELIHOOD: %.8f\n", likelihood_old);
        // for (int i = 0; i < number_of_haplotypes; i++)
        //     fprintf(stderr, "OLD FREQS used for likelihood: %.8f\n", em_freqs_old[i]);
        // fprintf(stderr, "NEW LIKELIHOOD: %.8f\n", likelihood_new);
        // for (int i = 0; i < number_of_haplotypes; i++)
        //     fprintf(stderr, "NEW FREQS used for likelihood: %.8f\n", em_freqs[i]);
        difference = fabs(likelihood_new - likelihood_old);
        likelihood_old = likelihood_new;
    }
    if (EM_verbose) {
        fprintf(stderr, "EM iterations using product of EM allele frequencies as initial values: %d\n", niterations);
    }
    double total_likelihood = EM_Haplotype_GetLikelihood(annotated_families, em_freqs, number_of_families, number_of_haplotypes);
    if (EM_verbose) {
        fprintf(stderr, "Log-likelihood of the haplotype frequencies using product of EM allele frequencies as initial values: %.8f\n", total_likelihood);
    }

    for (int i = 0; i < number_of_haplotypes; i++) {
        all_equal[i] = 1.0 / number_of_haplotypes;
    }

    if (EM_verbose) {
        fprintf(stderr, "Initial Values for EM:\n");
        for (int i = 0; i < number_of_haplotypes; i++) {
            fprintf(stderr, "%f ", all_equal[i]);
        }
        fprintf(stderr, "\n");
    }

    likelihood_old = 0.0;
    likelihood_new = 0.0;
    difference = accuracy + 1;
    niterations = 0;
    while (difference > accuracy) {
        ++niterations;
        difference = 0.0;
        for (int i = 0; i < number_of_haplotypes; i++) {
            em_families = 0.0;
            all_equal_old[i] = all_equal[i];
            for (int j = 0; j < number_of_families; j++) {
                AnnotatedFamily annotated_family = Vector_Get(annotated_families, j);
                FounderHaplotypes founder_haplotypes = AnnotatedFamily_GetFounderHaplotypes(annotated_family);
                em_numerator = 0.0;
                em_denominator = 0.0;
                TripletHashTableIterator it = FounderHaplotypes_GetIterator(founder_haplotypes);
                while (TripletHashTableIterator_HasNext(it)) {
                    TripletHashTableIterator_Next(it);
                    int *exponents = TripletHashTableIterator_GetValue(it);
                    double hap_freq_prod = get_hap_freq_prod(all_equal, exponents, number_of_haplotypes);
                    em_numerator += exponents[number_of_haplotypes] * exponents[i] * hap_freq_prod;
                    em_denominator += exponents[number_of_haplotypes] * hap_freq_prod;
                }
                em_families += (em_numerator / em_denominator);
            }
            all_equal[i] = constant * em_families;
        }
        if (niterations == 1)
            likelihood_old = EM_Haplotype_GetLikelihood(annotated_families, all_equal_old, number_of_families, number_of_haplotypes);
        likelihood_new = EM_Haplotype_GetLikelihood(annotated_families, all_equal, number_of_families, number_of_haplotypes);
        // fprintf(stderr, "OLD LIKELIHOOD: %.8f\n", likelihood_old);
        // for (int i = 0; i < number_of_haplotypes; i++)
        //     fprintf(stderr, "OLD FREQS used for likelihood: %.8f\n", all_equal_old[i]);
        // fprintf(stderr, "NEW LIKELIHOOD: %.8f\n", likelihood_new);
        // for (int i = 0; i < number_of_haplotypes; i++)
        //     fprintf(stderr, "NEW FREQS used for likelihood: %.8f\n", all_equal[i]);
        difference = fabs(likelihood_new - likelihood_old);
        likelihood_old = likelihood_new;
    }
    if (EM_verbose) {
        fprintf(stderr, "EM iterations using equal initial values: %d\n", niterations);
    }
    double total_likelihood_equal = EM_Haplotype_GetLikelihood(annotated_families, all_equal, number_of_families, number_of_haplotypes);
    if (EM_verbose) {
        fprintf(stderr, "Log-likelihood of the haplotype frequencies using equal initial values: %.8f\n", total_likelihood_equal);
    }
    if (total_likelihood > total_likelihood_equal) {
        for (int i = 0; i < number_of_haplotypes; i++) {
            DoubleVector_PushBack(hap_freqs, em_freqs[i]);
        }
        if (EM_verbose) {
            fprintf(stderr, "Product of EM allele frequencies win\n");
        }
    } else if(total_likelihood < total_likelihood_equal) {
        for (int i = 0; i < number_of_haplotypes; i++) {
            DoubleVector_PushBack(hap_freqs, all_equal[i]);
        }
        if (EM_verbose) {
            fprintf(stderr, "Equal initial values win\n");
        }
    } else {
        for (int i = 0; i < number_of_haplotypes; i++) {
            DoubleVector_PushBack(hap_freqs, em_freqs[i]);
        }
        if (EM_verbose) {
            fprintf(stderr, "Equally likely solutions with product of EM allele frequencies and equal initial values\n");
        }
    }
    return hap_freqs;
}

// static void print_triplet_and_exponents(FounderHaplotypes founder_haplotypes)
// {
//     TripletHashTableIterator it = FounderHaplotypes_GetIterator(founder_haplotypes);
// 	int nhaps = FounderHaplotypes_GetNumberOfHaplotypes(founder_haplotypes);
//     while (TripletHashTableIterator_HasNext(it)) {
//         TripletHashTableIterator_Next(it);
//         int *exponents = TripletHashTableIterator_GetValue(it);
//         for (int i = 0; i < nhaps; i++) {
//             fprintf(stderr, " %2d", exponents[i]);
//         }
// 		fprintf(stderr, " coef %d\n", exponents[nhaps]);
//         Triplet triplet = TripletHashTableIterator_GetTriplet(it);
//         Triplet_Print(triplet);
//     }
// }

double EM_Haplotype_GetLikelihood(Vector annotated_families, double *em_freqs, int number_of_families, int number_of_haplotypes)
{
    double total_likelihood = 0.0;
    for (int i = 0; i < number_of_families; i++) {
        AnnotatedFamily annotated_family = Vector_Get(annotated_families, i);
        int number_of_inheritance_vectors = AnnotatedFamily_GetNumberOfInheritanceVectors(annotated_family);
        FounderHaplotypes founder_haplotypes = AnnotatedFamily_GetFounderHaplotypes(annotated_family);
        TripletHashTableIterator it = FounderHaplotypes_GetIterator(founder_haplotypes);
        double family_likelihood = 0.0;
        while (TripletHashTableIterator_HasNext(it)) {
            TripletHashTableIterator_Next(it);
            int *exponents = TripletHashTableIterator_GetValue(it);
            double hap_freq_prod = get_hap_freq_prod(em_freqs, exponents, number_of_haplotypes);
            family_likelihood += hap_freq_prod * exponents[number_of_haplotypes];
        }
        total_likelihood += log(family_likelihood / number_of_inheritance_vectors);
    }
    return total_likelihood;
}

double EM_Allele_Frequencies_GetLikelihood(Vector annotated_families, double *em_freqs, int number_of_families, int number_of_haplotypes, int which_marker)
{
    double total_likelihood = 0.0;
    for (int i = 0; i < number_of_families; i++) {
        AnnotatedFamily annotated_family = Vector_Get(annotated_families, i);
        int number_of_inheritance_vectors = AnnotatedFamily_GetNumberOfInheritanceVectors(annotated_family);
        FounderHaplotypes founder_haplotypes = AnnotatedFamily_GetFounderHaplotypes(annotated_family);
        // print_triplet_and_exponents(founder_haplotypes);
        TripletHashTableIterator it = 0;
        if (which_marker == 0) {
            it = FounderHaplotypes_GetIterator_1(founder_haplotypes);
        }
        else if (which_marker == 1) {
            it = FounderHaplotypes_GetIterator_2(founder_haplotypes);
        }
        else if (which_marker == 2) {
            it = FounderHaplotypes_GetIterator_3(founder_haplotypes);
        }
        double family_likelihood = 0.0;
        while (TripletHashTableIterator_HasNext(it)) {
            TripletHashTableIterator_Next(it);
            int *exponents = TripletHashTableIterator_GetValue(it);
            double hap_freq_prod = get_hap_freq_prod(em_freqs, exponents, number_of_haplotypes);
            family_likelihood += hap_freq_prod * exponents[number_of_haplotypes];
        }
        total_likelihood += log(family_likelihood / number_of_inheritance_vectors);
    }
    return total_likelihood;
}

// ==== Compute EM allele frequencies for linkage markers ============

void EM_ComputeMarkerFrequencies(Vector markers, Vector families) {
    int number_of_markers = Vector_GetSize(markers);
    int number_of_families = Vector_GetSize(families);
    int total_founders = 0;
    for (int i = 0; i < number_of_families; i++) {
        Family family = Vector_Get(families, i);
        total_founders += Family_GetNumberOfFounders(family);
    }
    for (int i = 0; i < number_of_markers; i++) {
        Marker marker = Vector_Get(markers, i);
        // allele_exponents is a Vector of Vectors of arrays (array
        // length = number of alleles + 1 for coef)
        // allele_exponents --> family 1 --> a1^2*a2*a3 [2 1 1 0 coef 1]  1st compatible founder haplotype combination
        //                               --> 2*(a1^4)   [4 0 0 0 coef 2]  2nd (coef 2 means that combo was possible for 2 inheritance vectors)
        //                   --> family 2 --> ...
        Vector allele_exponents = Vector_Create(Vector_GetSize(families));
        for (int j = 0; j < number_of_families; j++) {
            Vector_PushBack(allele_exponents, EM_ComputeAlleleExponents(Vector_Get(families, j), marker));
        }
        double *fs = EM_ComputeMarkerAlleleFrequencies(allele_exponents, families, number_of_families, total_founders, marker);
        Marker_SetAlleleFrequencies(marker, Marker_GetNumberOfAlleles(marker), fs);
    }
}

double *EM_ComputeMarkerAlleleFrequencies(Vector allele_exponents, Vector families, int number_of_families, int total_founders, Marker marker)
{
    int number_of_haplotypes = Marker_GetNumberOfAlleles(marker);
    double *em_freqs = Memory_Malloc(number_of_haplotypes * sizeof(double));
    double *em_freqs_old = Memory_Malloc(number_of_haplotypes * sizeof(double));
    double em_numerator;
    double em_denominator;
    double em_families;
    double accuracy = 1e-8;
    const double constant = 1.0 / (2 * total_founders);
    double *all_equal = Memory_Malloc(number_of_haplotypes * sizeof(double));
    double *all_equal_old = Memory_Malloc(number_of_haplotypes * sizeof(double));
    for (int i = 0; i < number_of_haplotypes; i++) {
        em_freqs[i] = Marker_GetAlleleFrequency(marker, i);
    }

    if (EM_verbose) {
        fprintf(stderr, "\nLinkage marker: %s\n", Marker_GetName(marker));
        fprintf(stderr, "Initial Values for EM:\n");
        for (int i = 0; i < number_of_haplotypes; i++) {
            fprintf(stderr, "%f ", em_freqs[i]);
        }
        fprintf(stderr, "\n");
    }
    double difference = accuracy + 1;
    double likelihood_old = 0.0;
    double likelihood_new = 0.0;
    int niterations = 0;
    while (difference > accuracy) {
        ++niterations;
        difference = 0.0;
        for (int i = 0; i < number_of_haplotypes; i++) {
            em_families = 0.0;
            em_freqs_old[i] = em_freqs[i];
            for (int j = 0; j < number_of_families; j++) {
                Vector exponents = Vector_Get(allele_exponents, j);
                em_numerator = 0.0;
                em_denominator = 0.0;
                for (int k = 0; k < Vector_GetSize(exponents); k++) {
                    int *exponents_array = Vector_Get(exponents, k);
                    double hap_freq_prod = get_hap_freq_prod(em_freqs, exponents_array, number_of_haplotypes);
                    //fprintf(stderr, "hap_freq_prod %d: %f\n", k+1 ,hap_freq_prod);
                    em_numerator += exponents_array[number_of_haplotypes] * exponents_array[i] * hap_freq_prod;
                    em_denominator += exponents_array[number_of_haplotypes] * hap_freq_prod;
                }
                em_families += (em_numerator / em_denominator);
            }
            em_freqs[i] = constant * em_families;
        }
        if(niterations == 1)
            likelihood_old = EM_ComputeLinkageMarkerLikelihood(allele_exponents, families, em_freqs_old, number_of_families, number_of_haplotypes);
        likelihood_new = EM_ComputeLinkageMarkerLikelihood(allele_exponents, families, em_freqs, number_of_families, number_of_haplotypes);
        difference = fabs(likelihood_new - likelihood_old);
        likelihood_old = likelihood_new;
    }

    if (EM_verbose) {
        fprintf(stderr, "EM iterations using pre-specified marker allele frequencies as initial values: %d\n", niterations);
    }
    double total_likelihood = EM_ComputeLinkageMarkerLikelihood(allele_exponents, families, em_freqs, number_of_families, number_of_haplotypes);
    if (EM_verbose) {
        fprintf(stderr, "Log-likelihood of the allele frequencies using pre-specified allele frequencies as initial values: %.8f\n", total_likelihood);
    }

    for (int i = 0; i < number_of_haplotypes; i++) {
        all_equal[i] = 1.0 / number_of_haplotypes;
    }
    if (EM_verbose) {
        fprintf(stderr, "Initial Values for EM:\n");
        for (int i = 0; i < number_of_haplotypes; i++) {
            fprintf(stderr, "%f ", all_equal[i]);
        }
        fprintf(stderr, "\n");
    }

    likelihood_old = 0.0;
    likelihood_new = 0.0;
    difference = accuracy + 1;
    niterations = 0;
    while (difference > accuracy) {
        ++niterations;
        difference = 0.0;
        for (int i = 0; i < number_of_haplotypes; i++) {
            em_families = 0.0;
            all_equal_old[i] = all_equal[i];
            for (int j = 0; j < number_of_families; j++) {
                Vector exponents = Vector_Get(allele_exponents, j);
                em_numerator = 0.0;
                em_denominator = 0.0;
                for (int k = 0; k < Vector_GetSize(exponents); k++) {
                    int *exponents_array = Vector_Get(exponents, k);
                    double hap_freq_prod = get_hap_freq_prod(all_equal, exponents_array, number_of_haplotypes);
                    em_numerator += exponents_array[number_of_haplotypes] * exponents_array[i] * hap_freq_prod;
                    em_denominator += exponents_array[number_of_haplotypes] * hap_freq_prod;
                }
                em_families += (em_numerator / em_denominator);
            }
            all_equal[i] = constant * em_families;
        }
        if(niterations == 1)
            likelihood_old = EM_ComputeLinkageMarkerLikelihood(allele_exponents, families, all_equal_old, number_of_families, number_of_haplotypes);
        likelihood_new = EM_ComputeLinkageMarkerLikelihood(allele_exponents, families, all_equal, number_of_families, number_of_haplotypes);
        difference = fabs(likelihood_new - likelihood_old);
        likelihood_old = likelihood_new;
    }
    if (EM_verbose) {
        fprintf(stderr, "EM iterations using equal initial values: %d\n", niterations);
    }
    double total_likelihood_equal = EM_ComputeLinkageMarkerLikelihood(allele_exponents, families, all_equal, number_of_families, number_of_haplotypes);
    if (EM_verbose) {
        fprintf(stderr, "Log-likelihood of the allele frequencies using equal initial values: %.8f\n", total_likelihood_equal);
    }
    if (total_likelihood > total_likelihood_equal) {
        if (EM_verbose) {
            fprintf(stderr, "Pre-specified allele frequencies win\n");
        }
    } else if(total_likelihood < total_likelihood_equal) {
        if (EM_verbose) {
            fprintf(stderr, "Equal initial values win\n");
        }
        em_freqs = all_equal;
    } else {
        if (EM_verbose) {
            fprintf(stderr, "Equally likely solutions with pre-specified and equal initial values\n");
        }
    }
    if (EM_verbose) {
        fprintf(stderr, "\nAllele frequencies: LM1.1 LM1.2 ...\n");
        for (int i = 0; i < number_of_haplotypes; i++) {
            fprintf(stderr, "%f ", em_freqs[i]);
        }
        fprintf(stderr, "\n");
    }
    return em_freqs;
}

// Allele pairs: (a1 a2) and (a2 a1), but if a1 == a2, then only (a1 a1).
static Vector collect_allele_pairs(Person founder, Marker marker)
{
    Vector allele_pairs = Vector_CreateEmpty();
    int number_of_alleles = Marker_GetNumberOfAlleles(marker);
    int a1 = Person_GetMarkerAllele1(founder, Marker_GetIndex(marker));
    int a2 = Person_GetMarkerAllele2(founder, Marker_GetIndex(marker));
    if (number_of_alleles == 2 && (a1 == MISSING_ALLELE || a2 == MISSING_ALLELE)) {
        for (int i = 0; i < number_of_alleles; i++) {
            for (int j = 0; j < number_of_alleles; j++) {
                int *allele_pair = Memory_Malloc(2 * sizeof(int));
                allele_pair[0] = i;
                allele_pair[1] = j;
                Vector_PushBack(allele_pairs, allele_pair);
            }
        }
    } else if (number_of_alleles > 2 && (a1 == 0 || a2 == 0)) {
        for (int i = 1; i <= number_of_alleles; i++) {
            for (int j = 1; j <= number_of_alleles; j++) {
                int *allele_pair = Memory_Malloc(2 * sizeof(int));
                allele_pair[0] = i;
                allele_pair[1] = j;
                Vector_PushBack(allele_pairs, allele_pair);
            }
        }
    } else if (a1 == a2) {
        int *allele_pair = Memory_Malloc(2 * sizeof(int));
        allele_pair[0] = a1;
        allele_pair[1] = a2;
        Vector_PushBack(allele_pairs, allele_pair);
    } else {  // a1 != a2
        int *allele_pair1 = Memory_Malloc(2 * sizeof(int));
        allele_pair1[0] = a1;
        allele_pair1[1] = a2;
        Vector_PushBack(allele_pairs, allele_pair1);
        int *allele_pair2 = Memory_Malloc(2 * sizeof(int));
        allele_pair2[0] = a2;
        allele_pair2[1] = a1;
        Vector_PushBack(allele_pairs, allele_pair2);
    }
    return allele_pairs;
}

static int *compute_exponents_from_combo(Vector combo, int ncompatible, int number_of_alleles)
{
    int *exponents = Memory_Calloc(number_of_alleles + 1, sizeof(int));  // +1 for coefficient
    for (int i = 0; i < Vector_GetSize(combo); i++) {
        // [1 2] [3 4] [5 5]
        // pair1 pair2 pair3
        // translates into:
        // 1 1 1 1 2 [coef=ncompatible]
        // exponents[0 1 2 3 4 5]
        int *allele_pair = Vector_Get(combo, i);
        int idx1 = allele_pair[0];
        int idx2 = allele_pair[1];
        if (number_of_alleles != 2) {
            idx1--; idx2--;  // alleles 1-based but index 0-based
        }
        ++exponents[idx1]; // allele 1 of allele pair
        ++exponents[idx2]; // allele 2 of allele pair
    }
    exponents[number_of_alleles] = ncompatible;  // coefficient
    return exponents;
}

static void update_allele_exponents(Vector allele_exponents, Vector combo, Family family, Marker marker)
{
    // static int ncalls = 0;
    // int *alleles;
    // fprintf(stderr, "update_allele_exponents: %d\n", ++ncalls);
    // fprintf(stderr, "combo %d:\n", ncalls);
    // for (int i = 0; i < Vector_GetSize(combo); i++) {
    //     alleles = Vector_Get(combo, i);
    //     fprintf(stderr, "%d", alleles[0]);
    //     fprintf(stderr, "%d", alleles[1]);
    // }
    // fprintf(stderr, "\n");
    // Loop over inheritance vectors and count for how many
    // inheritance vectors the combo is compatible with the
    // non-founders.
    Vector nonfounders = Family_GetNonfounders(family);
    int compatible;  // 1 if current inheritance vector is compatible with all nonfounders, otherwise 0
    int ncompatible = 0;
    int number_of_alleles = Marker_GetNumberOfAlleles(marker);
    int number_of_inheritance_vectors = Family_GetNumberOfInheritanceVectors(family);

    for (int i = 0; i < number_of_inheritance_vectors; i++) {
        Family_SetPhase(family, i);
        compatible = 1;
        // Founder1 Founder2
        // 0   1    2   3   // offset //
        // mat pat  mat pat // origin //
        // 1   2    3   1   // combo // correct alignment?!
        // a1  a2   a1  a2  //
        //  [0]      [1]    // index of allele pair in combo
        for (int j = 0; j < Vector_GetSize(nonfounders); j++) {
            Person nonfounder = Family_GetNonfounder(family, j);
            int paternal_allele_offset = Person_GetPaternalAlleleOffset(nonfounder);
            int *paternal_allele_pair = Vector_Get(combo, paternal_allele_offset / 2);
            int paternal_allele = paternal_allele_pair[paternal_allele_offset % 2];
            int maternal_allele_offset = Person_GetMaternalAlleleOffset(nonfounder);
            int *maternal_allele_pair = Vector_Get(combo, maternal_allele_offset / 2);
            int maternal_allele = maternal_allele_pair[maternal_allele_offset % 2];
            int a1 = Person_GetMarkerAllele1(nonfounder, Marker_GetIndex(marker));
            int a2 = Person_GetMarkerAllele2(nonfounder, Marker_GetIndex(marker));
            // Check if combo is compatible for given inheritance
            // vector by comparing assigned vs. actual allele pairs
            // for nonfounder j.
            if (number_of_alleles == 2 && Person_GetMarkerGenotype(nonfounder, Marker_GetIndex(marker)) == XX) {
                //compatible
            } else if (number_of_alleles > 2 && (a1 == 0 || a2 == 0)) {
                // compatible
            } else if ((a1 == maternal_allele && a2 == paternal_allele) ||
                    (a1 == paternal_allele && a2 == maternal_allele)) {
                // compatible
            } else {
                compatible = 0;
                break;
            }
        }
        ncompatible += compatible;  // compatible is 0 or 1
    }
    if (ncompatible > 0) {
        // fprintf(stderr, "ncompatible = %d\n", ncompatible);
        Vector_PushBack(allele_exponents, compute_exponents_from_combo(combo, ncompatible, number_of_alleles));
    }
}

Vector EM_ComputeAlleleExponents(Family family, Marker marker)
{
    Vector allele_exponents = Vector_Create(1);
    // Loop over founders and collect pairs of compatible alleles into
    // a founder-specific Vector.
    Vector founders = Family_GetFounders(family);
    Vector founder_allele_pairs = Vector_Create(Vector_GetSize(founders));
    for (int i = 0; i < Vector_GetSize(founders); i++) {
        Person founder = Vector_Get(founders, i);
        Vector_PushBack(founder_allele_pairs, collect_allele_pairs(founder, marker));
    }
    // Loop over all combinations of founder alleles and collect those
    // combinations that are compatible with the non-founder alleles.
    int *indexes = Memory_Malloc(Vector_GetSize(founders) * sizeof(int));
    for (int i = 0; i < Vector_GetSize(founders); i++) {
        indexes[i] = 0;
    }
    int fi = 0;  // index of founder whose allele pair index we will try to increment next
    for (;;) {
        Vector combo = Vector_Create(Vector_GetSize(founders));
        for (int i = 0; i < Vector_GetSize(founder_allele_pairs); i++) {
            Vector_PushBack(combo, Vector_Get(Vector_Get(founder_allele_pairs, i), indexes[i]));
        }
        update_allele_exponents(allele_exponents, combo, family, marker);
        while (fi < Vector_GetSize(founders) && indexes[fi] == Vector_GetSize(Vector_Get(founder_allele_pairs, fi)) - 1) {
            fi++;
        }
        if (fi == Vector_GetSize(founders)) {
            break;
        }
        indexes[fi]++;
        for (int i = 0; i < fi; i++) {
            indexes[i] = 0;
        }
        fi = 0;
    }
    return allele_exponents;
}

double EM_ComputeLinkageMarkerLikelihood(Vector allele_exponents, Vector families, double *em_freqs, int number_of_families, int number_of_haplotypes)
{
    double total_likelihood = 0.0;
    double family_likelihood;
    for (int i = 0; i < number_of_families; i++) {
        Family family = Vector_Get(families, i);
        int number_of_inheritance_vectors = Family_GetNumberOfInheritanceVectors(family);
        family_likelihood = 0.0;
        Vector exponents = Vector_Get(allele_exponents, i);
        for (int j = 0; j < Vector_GetSize(exponents); j++) {
            int *exponents_array = Vector_Get(exponents, j);
            double hap_freq_prod = get_hap_freq_prod(em_freqs, exponents_array, number_of_haplotypes);
            //fprintf(stderr, "hap_freq_prod: %f\n", hap_freq_prod);
            family_likelihood += hap_freq_prod * exponents_array[number_of_haplotypes] / number_of_inheritance_vectors;
            //fprintf(stderr, "LM family_likelihood %d: %.4f\n", j+1, family_likelihood);
        }
        total_likelihood += log(family_likelihood);
    }
    return total_likelihood;
}
