#include "Family.h"
#include "FailWith.h"
#include "Memory.h"
#include "Person.h"
#include "Vector.h"
#include "Penetrances.h"
#include <stddef.h>
#include <string.h>
#include <stdio.h>

struct FamilyStruct {
    int id;
    Vector members;
    Vector founders;
    Vector nonfounders;
    Matrix penetrance_matrix;
    Vector pqterms;
    int singleton;
};

static Person get_person(Vector v, int i);
static Person find_person_by_id(const char *id, Vector persons);
static int has_allele_offset(Person, AlleleOrigin);
static int get_allele_offset(Person, AlleleOrigin);
static AlleleOrigin get_origin_of_parental_allele(Person, AlleleOrigin);
static Person get_parent(Person, AlleleOrigin);
static void set_allele_offset(Person, AlleleOrigin, int allele_offset);
static int get_parental_allele_offset(Person, AlleleOrigin);
static void show_persons(Vector persons);

static IntSet Family_FindCompatiblePhasedFounderGenotypesImpl(Family, int marker);
static int Family_GetNumberOfFoundersImpl(Family);
static int Family_GetNumberOfNonfoundersImpl(Family);

static Person get_person(Vector v, int i)
{
    return (Person) Vector_Get(v, i);
}

Family Family_Create(int id)
{
    Family family = Memory_Malloc(sizeof *family);
    family->id = id;
    family->members = Vector_CreateEmpty();
    family->founders = Vector_CreateEmpty();
    family->nonfounders = Vector_CreateEmpty();
    family->penetrance_matrix = NULL;
    family->pqterms = Vector_CreateEmpty();
    family->singleton = 0;
    return family;
}

void Family_Destroy(Family family)
{
    if (family == NULL)
        return;
    for (int i = 0; i < Family_GetNumberOfMembers(family); i++)
        Person_Destroy(Family_GetMember(family, i));
    Vector_Destroy(family->members);
    Vector_Destroy(family->founders);
    Vector_Destroy(family->nonfounders);
    if (family->penetrance_matrix) {
        for (int i = 0; i < Matrix_GetNumberOfRows(family->penetrance_matrix); i++)
            for (int j = 0; j < Matrix_GetNumberOfColumns(family->penetrance_matrix); j++)
                Penetrances_Destroy(Matrix_Get(family->penetrance_matrix, i, j));
    }
    Matrix_Destroy(family->penetrance_matrix);
    for (int i = 0; i < Vector_GetSize(family->pqterms); i++)
        PQTerm_Destroy(Vector_Get(family->pqterms, i));
    Vector_Destroy(family->pqterms);
    family->members = NULL;
    family->founders = NULL;
    family->nonfounders = NULL;
}

int Family_GetId(Family family)
{
    return family->id;
}

static int Family_GetNumberOfFoundersImpl(Family family)
{
    return Vector_GetSize(family->founders);
}

int (*Family_GetNumberOfFounders)(Family family) = Family_GetNumberOfFoundersImpl;

static int Family_GetNumberOfNonfoundersImpl(Family family)
{
    return Vector_GetSize(family->nonfounders);
}

int (*Family_GetNumberOfNonfounders)(Family family) = Family_GetNumberOfNonfoundersImpl;

int Family_GetNumberOfMembers(Family family)
{
    return Vector_GetSize(family->members);
}

int Family_AddFounder(Family family, Person p)
{
    Vector_PushBack(family->members, p);
    Vector_PushBack(family->founders, p);

    return 1;
}

int Family_AddNonfounder(Family family, Person p)
{
    Vector_PushBack(family->members, p);
    Vector_PushBack(family->nonfounders, p);

    return 1;
}

Person Family_GetFounder(Family family, int i)
{
    return get_person(family->founders, i);
}

Person Family_GetNonfounder(Family family, int i)
{
    return get_person(family->nonfounders, i);
}

Vector Family_GetFounders(Family family)
{
    return family->founders;
}

Vector Family_GetNonfounders(Family family)
{
    return family->nonfounders;
}

Vector Family_GetMembers(Family family)
{
    return family->members;
}

Person Family_GetMember(Family family, int i)
{
    return get_person(family->members, i);
}

void Family_SetParents(Family family)
{
    Vector nonfounders = Family_GetNonfounders(family);
    Vector members = Family_GetMembers(family);
    int number_of_nonfounders = Vector_GetSize(nonfounders);
    for (int i = 0; i < number_of_nonfounders; i++) {
        Person person = Vector_Get(nonfounders, i);
        Person father = find_person_by_id(Person_GetFatherId(person), members);
        Person mother = find_person_by_id(Person_GetMotherId(person), members);
        Person_SetFather(person, father);
        Person_SetMother(person, mother);
    }
}

static Person find_person_by_id(const char *id, Vector persons)
{
    Person found = NULL;
    int number_of_persons = Vector_GetSize(persons);
    for (int i = 0; i < number_of_persons; i++) {
        Person person = Vector_Get(persons, i);
        if (strcmp(id, Person_GetId(person)) == 0) {
            found = person;
            break;
        }
    }
    if (!found)
        FailWith("find_person_by_id: can't find person %s", id);
    return found;
}

void Family_ClearPhase(Family family)
{
    int number_of_members = Family_GetNumberOfMembers(family);
    Vector members = Family_GetMembers(family);
    int default_value = -1;
    for (int i = 0; i < number_of_members; i++) {
        Person person = Vector_Get(members, i);
        Person_SetMaternalAlleleOffset(person, default_value);
        Person_SetPaternalAlleleOffset(person, default_value);
    }
}

void Family_SetAlleleOffsetsForFounders(Vector founders)
{
    int number_of_founders = Vector_GetSize(founders);
    for (int i = 0; i < number_of_founders; i++) {
        Person person = Vector_Get(founders, i);
        int maternal_allele_offset = 2*i;
        int paternal_allele_offset = 2*i + 1;
        Person_SetMaternalAlleleOffset(person, maternal_allele_offset);
        Person_SetPaternalAlleleOffset(person, paternal_allele_offset);
    }
}

void Family_SetAlleleOriginsForNonfounders(Vector nonfounders, InheritanceVector inheritance_vector)
{
    InheritanceVector iv = inheritance_vector;
    int number_of_nonfounders = Vector_GetSize(nonfounders);
    for (int i = number_of_nonfounders - 1; i >= 0; i--) {
        Person person = Vector_Get(nonfounders, i);
        // Maternal must come before paternal because maternal allele
        // is closer to the least significant bit (LSB).
        if (Person_HasMaternalBitFixed(person)) {
            Person_SetMaternalAlleleOrigin(person, 0);
        } else {
            Person_SetMaternalAlleleOrigin(person, iv & 0x1);  // pick LSB
            iv >>= 1;  // remove LSB
        }
        if (Person_HasPaternalBitFixed(person)) {
            Person_SetPaternalAlleleOrigin(person, 0);
        } else {
            Person_SetPaternalAlleleOrigin(person, iv & 0x1);  // pick LSB
            iv >>= 1;  // remove LSB
        }
    }
}

void Family_SetPhase(Family family, InheritanceVector inheritance_vector)
{
    Vector founders = Family_GetFounders(family);
    Vector nonfounders = Family_GetNonfounders(family);
    Family_ClearPhase(family);
    Family_SetAlleleOffsetsForFounders(founders);
    Family_SetAlleleOriginsForNonfounders(nonfounders, inheritance_vector);
    Family_SetAlleleOffsetsForNonfounders(nonfounders);
}

void Family_SetAlleleOffsetsForNonfounders(Vector nonfounders)
{
    int number_of_nonfounders = Vector_GetSize(nonfounders);
    for (int i = 0; i < number_of_nonfounders; i++) {
        Person person = Vector_Get(nonfounders, i);
        get_parental_allele_offset(person, PATERNAL);
        get_parental_allele_offset(person, MATERNAL);
    }
}

static int get_parental_allele_offset(Person person, AlleleOrigin allele_origin)
{
    if (has_allele_offset(person, allele_origin)) {
        return get_allele_offset(person, allele_origin);
    } else {
        AlleleOrigin parental_allele_origin = get_origin_of_parental_allele(person, allele_origin);
        int allele_offset =  get_parental_allele_offset(get_parent(person, allele_origin), parental_allele_origin);
        set_allele_offset(person, allele_origin, allele_offset);
        return allele_offset;
    }
}

static int has_allele_offset(Person person, AlleleOrigin allele_origin)
{
    if (allele_origin == PATERNAL)
        return Person_HasPaternalAlleleOffset(person);
    else
        return Person_HasMaternalAlleleOffset(person);
}

static int get_allele_offset(Person person, AlleleOrigin allele_origin)
{
    if (allele_origin == PATERNAL)
        return Person_GetPaternalAlleleOffset(person);
    else
        return Person_GetMaternalAlleleOffset(person);
}

static AlleleOrigin get_origin_of_parental_allele(Person person, AlleleOrigin allele_origin)
{
    if (allele_origin == PATERNAL)
        return Person_GetPaternalAlleleOrigin(person);
    else
        return Person_GetMaternalAlleleOrigin(person);
}

static Person get_parent(Person person, AlleleOrigin allele_origin)
{
    if (allele_origin == PATERNAL)
        return Person_GetFather(person);
    else
        return Person_GetMother(person);
}

static void set_allele_offset(Person person, AlleleOrigin allele_origin, int allele_offset)
{
    if (allele_origin == PATERNAL)
        Person_SetPaternalAlleleOffset(person, allele_offset);
    else
        Person_SetMaternalAlleleOffset(person, allele_offset);
}

PhasedGenotypes Family_GetPhasedGenotypesForNonfounders(Family family, PhasedGenotypes phased_founder_genotypes)
{
    PhasedGenotypes phased_nonfounder_genotypes = 0x0;
    int last_nonfounder = Family_GetNumberOfNonfounders(family) - 1;
    int i;

    for (i = last_nonfounder; i >= 0; i--) {
        Person nonfounder = Family_GetNonfounder(family, i);
        int paternal_allele_offset = Person_GetPaternalAlleleOffset(nonfounder);
        int maternal_allele_offset = Person_GetMaternalAlleleOffset(nonfounder);
        int paternal_allele = (phased_founder_genotypes >> paternal_allele_offset) & 0x1;
        int maternal_allele = (phased_founder_genotypes >> maternal_allele_offset) & 0x1;
        Genotype genotype = Genotype_FromAlleles(paternal_allele, maternal_allele);

        phased_nonfounder_genotypes <<= 2;
        phased_nonfounder_genotypes |= genotype;
    }
    return phased_nonfounder_genotypes;
}

static IntSet Family_FindCompatiblePhasedFounderGenotypesImpl(Family family, int marker)
{
    int number_of_phased_founder_genotypes = Family_GetNumberOfDiseaseGenotypes(family);
    IntSet compatible_phased_founder_genotypes = IntSet_Create(number_of_phased_founder_genotypes / 4);
    Vector founders = Family_GetFounders(family);
    Vector nonfounders = Family_GetNonfounders(family);
    PhasedGenotypes phased_founder_genotypes, phased_nonfounder_genotypes;

    for (phased_founder_genotypes = 0; (int) phased_founder_genotypes < number_of_phased_founder_genotypes; phased_founder_genotypes++) {
        if (!Person_AreActualGenotypesCompatibleWithGenotypes(founders, marker, phased_founder_genotypes))
            continue;
        phased_nonfounder_genotypes = Family_GetPhasedGenotypesForNonfounders(family, phased_founder_genotypes);
        if (!Person_AreActualGenotypesCompatibleWithGenotypes(nonfounders, marker, phased_nonfounder_genotypes))
            continue;
        IntSet_Put(compatible_phased_founder_genotypes, phased_founder_genotypes);
    }
    return compatible_phased_founder_genotypes;
}

IntSet (*Family_FindCompatiblePhasedFounderGenotypes)(Family, int marker) = Family_FindCompatiblePhasedFounderGenotypesImpl;

Matrix Family_GetPenetranceMatrix(Family family)
{
    return family->penetrance_matrix;
}

void Family_ComputePenetrances(Family family)
{
    extern Matrix PenetranceMatrix_Create(Family family);
    family->penetrance_matrix = PenetranceMatrix_Create(family);
}

Vector Family_GetPQTerms(Family family)
{
    return family->pqterms;
}

void Family_ComputePQTerms(Family family)
{
    int number_of_disease_genotypes = Family_GetNumberOfDiseaseGenotypes(family);
    int number_of_founders = Family_GetNumberOfFounders(family);
    for (int dg = 0; dg < number_of_disease_genotypes; dg++)
        Vector_PushBack(family->pqterms, PQTerm_Create(dg, number_of_founders));
}

int Family_GetNumberOfDiseaseGenotypes(Family family)
{
    return PhasedGenotypes_GetNumberOfCombinations(Family_GetNumberOfFounders(family));
}

int Family_GetNumberOfInheritanceVectors(Family family)
{
    if (!Family_GetSingleton(family))
        return 1 << (2 * Family_GetNumberOfNonfounders(family) - Family_GetNumberOfFounders(family));
    else
        return 1;
}

void Family_Show(Family family)
{
    fprintf(stderr, "Founders:\n");
    show_persons(Family_GetFounders(family));
    fprintf(stderr, "Nonfounders:\n");
    show_persons(Family_GetNonfounders(family));
}

static void show_persons(Vector persons)
{
    for (int i = 0; i < Vector_GetSize(persons); i++)
        Person_Show(Vector_Get(persons, i));
}

void Family_SetSingleton(Family family)
{
    if (Family_GetNumberOfFounders(family) == 1 && Family_GetNumberOfMembers(family) == 1)
        family->singleton = 1;
}

int Family_GetSingleton(Family family)
{
    return family->singleton;
}
