#include "CommandReader.h"
#include "GHMRunner.h"
#include "IO.h"
#include "Memory.h"
#include "ParseObject.h"

#include <errno.h>
#include <stdlib.h>
#include <sys/stat.h>    // mkdir
#include <sys/types.h>   // opendir, closedir, mkdir

// Legacy includes.
#define INC_LIB
#define INC_SHELL
#include "npl.h"


struct GHMRunnerStruct {
    CommandReader reader;
    int number_of_test_markers;
};

static void mkdir_or_fail(const char *directory)
{
    errno = 0;
    int ok = mkdir(directory, 0755) == 0 || errno == EEXIST;
    if (!ok) {
        sf(ps, "mkdir failed: %s\n", directory);
        pr();
        exit(1);
    }
}

static void copy_input_files(CommandReader reader, const char *directory)
{
    if (!CommandReader_CopyUserFile(reader, "read association regions ", directory)) {
        fprintf(stderr, "Failed to copy file: %s\n", CommandReader_GetUserAssociationRegionsFile(reader));
        exit(EXIT_FAILURE);
    }
    if (!CommandReader_CopyUserFile(reader, "read linkage markers ", directory)) {
        fprintf(stderr, "Failed to copy file: %s\n", CommandReader_GetUserLinkageMarkersFile(reader));
        exit(EXIT_FAILURE);
    }
    if (g_model_params_file) {
        if (!CommandReader_CopyUserFile(reader, "read association models ", directory)) {
            fprintf(stderr, "Failed to copy file: %s\n", CommandReader_GetUserAssociationModelsFile(reader));
            exit(EXIT_FAILURE);
        }
    }
    if (!CommandReader_CopyUserFile(reader, "load markers ", directory)) {
        fprintf(stderr, "Failed to copy file: %s\n", CommandReader_GetMarkerFile(reader));
        exit(EXIT_FAILURE);
    }
    if (!CommandReader_CopyUserFile(reader, "scan pedigrees ", directory)) {
        fprintf(stderr, "Failed to copy file: %s\n", CommandReader_GetPedFile(reader));
        exit(EXIT_FAILURE);
    }
}

static void create_commands_file(CommandReader reader, const char *directory, int16_t seed)
{
    char filename[250];
    snprintf(filename, sizeof(filename), "%s/sample.commands", directory);
    CommandReader_Write(reader, filename, seed);
}

static Vector split_into_fields(const char *line)
{
    ParseObject parser = ParseObject_Create(line);
    Vector fields = ParseObject_SplitIntoWords(parser);
    ParseObject_Destroy(parser);
    return fields;
}

static int find_index(Vector v, const char *s)
{
    for (int i = 0; i < Vector_GetSize(v); i++) {
        const char *item = Vector_Get(v, i);
        if (strcmp(item, s) == 0) {
            return i;
        }
    }
    return -1;
}

static void find_hapfreq_indices(Vector fields, int number_of_hapfreqs, int *hapfreq_indices)
{
    int found = 0;
    for (int i = 0; i < Vector_GetSize(fields); i++) {
        const char *item = Vector_Get(fields, i);
        if (strncmp(item, "EMH", 3) == 0) {
            if (found > number_of_hapfreqs) {
                sf(ps, "Too many test marker haplotypes in results.txt. Expected %d frequencies.\n", number_of_hapfreqs);
                pr();
                exit(1);
            }
            hapfreq_indices[found++] = i;
        }
    }
    if (found == 0) {
        // Every results file contains at least two haplotype
        // frequencies.  If we don't find any, then the file doesn't
        // contain any results.
        sf(ps, "No results were produced in results.txt.  See the associated log file.\n");
        pr();
        exit(1);
    }
    if (found != number_of_hapfreqs) {
        sf(ps, "Number of test marker haplotypes in results.txt (%d) do not match with those in the command file (%d)\n", found, number_of_hapfreqs);
        pr();
        exit(1);
    }
}

static void extract_mod_score_and_hapfreqs(
    const char *directory,
    int number_of_test_markers,
    double *mod_score,
    DoubleVector *hapfreqs)
{
    char txt[200];
    double hap_freq;
    int number_of_hapfreqs = 1 << number_of_test_markers;
    int *hapfreq_indices = Memory_Malloc(number_of_hapfreqs * sizeof(int));
    snprintf(txt, sizeof txt, "%s/tmp/results.txt", directory);
    FILE *fp = fopen(txt, "rb");
    if (!fp) {
        sf(ps, "Failed to open file: \"%s\"\n", txt);
        pr();
        exit(1);
    }
    char *header = IO_GetLine(fp);
    Vector fields = split_into_fields(header);
    int mod_score_column = find_index(fields, "mod");
    find_hapfreq_indices(fields, number_of_hapfreqs, hapfreq_indices);
    Memory_Free(header);
    char *line = IO_GetLine(fp);
    fclose(fp);
    Vector_Destroy(fields);
    fields = split_into_fields(line);
    const char *mod_score_string = Vector_Get(fields, mod_score_column);
    *mod_score = atof(mod_score_string);
    DoubleVector haps = DoubleVector_Create(number_of_hapfreqs);
    for (int i = 0; i < number_of_hapfreqs; i++) {
        char *hap_freq_string = Vector_Get(fields, hapfreq_indices[i]);
        hap_freq = atof(hap_freq_string);
        DoubleVector_PushBack(haps, hap_freq);
    }
    *hapfreqs = haps;
}

static struct GHMResults extract_results(const char *directory, int number_of_test_markers)
{
    struct GHMResults results;
    extract_mod_score_and_hapfreqs(
        directory, number_of_test_markers, &results.mod_score, &results.hapfreqs);
    return results;
}

static struct GHMResults run_ghm(const char *directory, int number_of_test_markers)
{
    char cmd[500];
    const char *fmt = "cd %s && %s --commands input/sample.commands >/dev/null 2>&1";
    // const char *fmt = "cd %s && %s --commands input/sample.commands";
    snprintf(cmd, sizeof cmd, fmt, directory, g_ghm_executable);
    int status = system(cmd);
    (void) status;
    return extract_results(directory, number_of_test_markers);
}

GHMRunner GHMRunner_Create(CommandReader reader, int number_of_test_markers)
{
    GHMRunner runner = Memory_Malloc(sizeof *runner);
    runner->reader = reader;
    runner->number_of_test_markers = number_of_test_markers;
    return runner;
}

struct GHMResults GHMRunner_Run(GHMRunner runner, const char *directory, int16_t seed)
{
    int number_of_test_markers = runner->number_of_test_markers;
    char input_dir[200];
    char output_dir[200];
    snprintf(input_dir, sizeof(input_dir), "%s/input", directory);
    snprintf(output_dir, sizeof(output_dir), "%s/tmp", directory);
    mkdir_or_fail(input_dir);
    mkdir_or_fail(output_dir);
    copy_input_files(runner->reader, directory);
    create_commands_file(runner->reader, input_dir, seed);
    return run_ghm(directory, number_of_test_markers);
}
