#include "AnnotatedFamily.h"
#include "EM.h"
#include "GridModScoreFunction.h"
#include "HaplotypeFrequencyGrid.h"
#include "HaplotypeTerm.h"
#include "IntHashTable.h"
#include "LinkageMarkerFile.h"
#include "Marker.h"
#include "MathFunctions.h"
#include "Memory.h"
#include "ModScorer.h"
#include "PQTerm.h"
#include "TestMarkerSetGenerator.h"

#include <stdio.h>
#include <math.h>

#include "dSFMT.h"

// Legacy includes.
#define INC_LIB
#define INC_SHELL
#include "npl.h"

typedef void (*ModScoreComputer)(
    Vector test_markers,
    Vector annotated_families,
    double disease_allele_frequency,
    Vector penetrances,
    DoubleVector em_hapfreqs,
    LinkageMarkerFile linkage_file,
    ModScore *mod_score,
    ModScore *best_mod_score
);

// HapfreqPool
//
// HapfreqPool helps to reduce the number of dynamic allocations of
// hapfreqs DoubleVectors in draw_deltas_and_compute_modscore.
// Before, every call to draw_deltas_and_compute_modscore would
// dynamically allocate a new hapfreqs DoubleVector which would
// steadily increase memory consumption probably due to memory
// fragmentation.
//
// Every ResultTableRow refers to a DoubleVector of hapfreqs.  Thus,
// at most max_row hapfreqs DoubleVectors are referenced by the
// ResultTable.  In addition, there is the hapfreqs DoubleVector that
// might be inserted into the ResultTable during the current call to
// draw_deltas_and_compute_modscore.  Therefore, as long as we
// release any hapfreqs DoubleVector that is no longer needed, either
// because it never made it into the ResultTable or because it was
// pushed out again, we don't need more than (max_rows + 1) hapfreqs
// DoubleVectors.

typedef struct {
    DoubleVector *pool;
    int *free;
    int size;
} *HapfreqPool;

static HapfreqPool HAPFREQPOOL = NULL;


// Create a pool of size DoubleVectors where each DoubleVector has
// space for nhapfreqs haplotype frequencies.
HapfreqPool HapfreqPool_Create(int size, int nhapfreqs)
{
    HapfreqPool pool = Memory_Malloc(sizeof(*pool));
    pool->pool = Memory_Malloc(size * sizeof(DoubleVector *));
    pool->free = Memory_Malloc(size * sizeof(int *));
    pool->size = size;
    for (int i = 0; i < size; i++) {
        pool->pool[i] = DoubleVector_Create(nhapfreqs);
        pool->free[i] = 1;
    }
    return pool;
}


DoubleVector HapfreqPool_Take(HapfreqPool pool)
{
    for (int i = 0; i < pool->size; i++) {
        if (pool->free[i]) {
            pool->free[i] = 0;
            return pool->pool[i];
        }
    }
    fprintf(stderr, "HapfreqPool of size %d is empty\n", pool->size);
    exit(1);
}


void HapfreqPool_Release(HapfreqPool pool, DoubleVector hapfreqs)
{
    for (int i = 0; i < pool->size; i++) {
        if (pool->pool[i] == hapfreqs) {
            pool->free[i] = 1;
            return;
        }
    }
    fprintf(stderr, "HapfreqPool: can't release non-pool element\n");
    exit(1);
}


static Vector create_penetrance_combinations(void);

static int draw_deltas_and_compute_modscore(
    Vector test_markers,
    Vector annotated_families,
    double disease_allele_frequency,
    Vector penetrances,
    DoubleVector em_hapfreqs,
    LinkageMarkerFile linkage_file,
    ModScore *mod_score);

static int draw_hapfreqs_and_compute_modscore(
    Vector test_markers,
    Vector annotated_families,
    double disease_allele_frequency,
    Vector penetrances,
    DoubleVector em_hapfreqs,
    LinkageMarkerFile linkage_file,
    ModScore *mod_score);

static void compute_modscore_with_deltas(
    Vector test_markers,
    Vector annotated_families,
    double disease_allele_frequency,
    Vector penetrances,
    DoubleVector em_hapfreqs,
    LinkageMarkerFile linkage_file,
    ModScore *mod_score,
    ModScore *best_mod_score)
{
    for (int j = 0; j < g_jla_delta_iterations; j++) {
        // Draw dxyz.  Then compute haplotype freqs and plug in.
        int mod_score_valid = draw_deltas_and_compute_modscore(
            test_markers, annotated_families,
            disease_allele_frequency,
            penetrances, em_hapfreqs, linkage_file, mod_score
        );
        if (mod_score_valid && mod_score->mod_score > best_mod_score->mod_score) {
            ModScore_Update(best_mod_score, mod_score);
        }
    }
}

static void compute_modscore_without_deltas(
    Vector test_markers,
    Vector annotated_families,
    double disease_allele_frequency,
    Vector penetrances,
    DoubleVector em_hapfreqs,
    LinkageMarkerFile linkage_file,
    ModScore *mod_score,
    ModScore *best_mod_score)
{
    for (int j = 0; j < g_jla_hapfreq_iterations; j++) {
        int mod_score_valid = draw_hapfreqs_and_compute_modscore(
            test_markers, annotated_families,
            disease_allele_frequency,
            penetrances, em_hapfreqs, linkage_file, mod_score
        );
        if (mod_score_valid && mod_score->mod_score > best_mod_score->mod_score) {
            ModScore_Update(best_mod_score, mod_score);
        }
    }
}

ModScore *GridModScoreFunction(
    Vector test_markers,
    Vector annotated_families,
    DoubleVector em_hapfreqs,
    LinkageMarkerFile linkage_file,
    ModelParams user_model,
    int number_of_test_marker_sets,
    ModScore *grid_mod_score)
{
    (void) grid_mod_score;
    static Vector penetrances1 = NULL;
    static ModScore *best_mod_score = NULL;
    static ModScore *mod_score = NULL;
    static Vector prev_test_markers = NULL;
    if (!best_mod_score) {
        best_mod_score = ModScore_Create();
    }
    if (!mod_score) {
        mod_score = ModScore_Create();
    }
    if (test_markers != prev_test_markers) {
        ModScore_Clear(best_mod_score);
        prev_test_markers = test_markers;
    }
    if (HAPFREQPOOL == NULL) {
        HAPFREQPOOL = HapfreqPool_Create(number_of_test_marker_sets + 1, 2 * DoubleVector_GetSize(em_hapfreqs));
    }
    if (!user_model) {
        const int nps = 9;
        double ps[9] = {
            0.01, 0.05, 0.1, 0.25, 0.5, 0.75, 0.9, 0.95, 0.99
        };
        Vector penetrances = create_penetrance_combinations();
        DoubleVector disease_allele_frequencies = DoubleVector_Create(1);
        if (allfreq_restriction) {
            for (int i = 0; i < nps; i++) {
                if (ps[i] <= highest_allfreq) {
                    DoubleVector_PushBack(disease_allele_frequencies, ps[i]);
                }
            }
        } else {
            for (int i = 0; i < nps; i++) {
                DoubleVector_PushBack(disease_allele_frequencies, ps[i]);
            }
        }
        int grid_size = DoubleVector_GetSize(disease_allele_frequencies);
        int pen_grid_size = Vector_GetSize(penetrances);
        ModScoreComputer compute_modscore;
        if (g_jla_delta_modscore) {
            int total_grid_size = grid_size * pen_grid_size * g_jla_delta_iterations;
            fprintf(stderr, "Size of allele frequency grid: %d\n", grid_size);
            fprintf(stderr, "Size of penetrance grid: %d\n", pen_grid_size);
            fprintf(stderr, "Total pre-defined delta-based grid JLA evaluations: %d\n", total_grid_size);
            compute_modscore = compute_modscore_with_deltas;
        } else {
            int total_grid_size = grid_size * pen_grid_size * g_jla_hapfreq_iterations;
            fprintf(stderr, "Size of allele frequency grid: %d\n", grid_size);
            fprintf(stderr, "Size of penetrance grid: %d\n", pen_grid_size);
            fprintf(stderr, "Total pre-defined grid JLA evaluations: %d\n", total_grid_size);
            compute_modscore = compute_modscore_without_deltas;
        }
        for (int i = 0; i < grid_size; i++) {
            compute_modscore(
                test_markers,
                annotated_families,
                DoubleVector_Get(disease_allele_frequencies, i),
                penetrances,
                em_hapfreqs,
                linkage_file,
                mod_score,
                best_mod_score
            );
        }
    } else {
        DoubleVector pens = DoubleVector_Create(10);
        DoubleVector_PushBack(pens, user_model->f0);
        DoubleVector_PushBack(pens, user_model->f1_pat);
        DoubleVector_PushBack(pens, user_model->f1_mat);
        DoubleVector_PushBack(pens, user_model->f2);
        DoubleVector_PushBack(pens, (1 - user_model->f0));
        DoubleVector_PushBack(pens, (1 - user_model->f1_pat));
        DoubleVector_PushBack(pens, (1 - user_model->f1_mat));
        DoubleVector_PushBack(pens, (1 - user_model->f2));
        DoubleVector_PushBack(pens, (0.5 * (user_model->f1_pat + user_model->f1_mat)));
        DoubleVector_PushBack(pens, (0.5 * ((1 - user_model->f1_pat) + (1 - user_model->f1_mat))));
        if (penetrances1 == NULL) {
            penetrances1 = Vector_Create(1);
        }
        Vector_Set(penetrances1, 0, pens);
        int mod_score_valid;
        if (g_jla_delta_modscore) {
            mod_score_valid = draw_deltas_and_compute_modscore(
                    test_markers, annotated_families,
                    user_model->disease_allele_frequency,
                    penetrances1, em_hapfreqs, linkage_file, mod_score);
        } else {
            mod_score_valid = draw_hapfreqs_and_compute_modscore(
                    test_markers, annotated_families,
                    user_model->disease_allele_frequency,
                    penetrances1, em_hapfreqs, linkage_file, mod_score);
        }
        if (mod_score_valid && mod_score->mod_score > best_mod_score->mod_score) {
            ModScore_Update(best_mod_score, mod_score);
        }
    }
    ModScore *result = ModScore_Create();
    ModScore_Update(result, best_mod_score);
    return result;
}

static Vector create_penetrance_combinations(void)
{
    int pat_first = genrand_open_open() < .5;
    const int nvalues = 3;
    double values[3] = { .01, .5, .99 };
    double *f0, *f1_pat, *f1_mat, *f2;
    f0 = f1_pat = f1_mat = f2 = values;
    Vector result = Vector_Create(nvalues * nvalues * nvalues * nvalues);
    for (int i = 0; i < nvalues; i++)
        for (int j = 0; j < nvalues; j++)
            for (int k = 0; k < nvalues; k++)
                for (int m = 0; m < nvalues; m++) {
                    // No imprinting means f1_pat == f1_mat
                    if (!imprinting && j != k) {
                        continue;
                    }
                    // Allow exactly one model where all penetrances
                    // are equal, i.e., where there is no genetic
                    // effect.  (The choice of i == 0 is arbitrary.)
                    if (i != 0 && i == j && i == k && i == m) {
                        continue;
                    }
                    if (penetrance_restriction && (i > j || i > k || i > m || j > k || j > m || k > m)) {
                        continue;
                    }
                    int mat_idx, pat_idx;
                    if (pat_first) {
                        pat_idx = j;
                        mat_idx = k;
                    } else {
                        pat_idx = k;
                        mat_idx = j;
                    }
                    DoubleVector combination = DoubleVector_With(10,
                        f0[i], f1_pat[pat_idx], f1_mat[mat_idx], f2[m],
                        1 - f0[i], 1 - f1_pat[pat_idx], 1 - f1_mat[mat_idx], 1 - f2[m],
                        .5 * (f1_pat[pat_idx] + f1_mat[mat_idx]),
                        .5 * ((1 - f1_pat[pat_idx]) + (1 - f1_mat[mat_idx])));
                    Vector_PushBack(result, combination);
                }
    return result;
}


static int draw_deltas_and_compute_modscore(
    Vector test_markers,
    Vector annotated_families,
    double disease_allele_frequency,
    Vector penetrances,
    DoubleVector em_hapfreqs,
    LinkageMarkerFile linkage_file,
    ModScore *mod_score)
{
    static double *ds = NULL;
    static double *lower_indiv = NULL;
    static double *upper_indiv = NULL;
    static double best_mod_score = -HUGE_VAL;
    static Vector prev_test_markers = NULL;
    int best_mod_score_improved = 0;
    if (test_markers != prev_test_markers) {
        best_mod_score = -HUGE_VAL;
        prev_test_markers = test_markers;
    }
    int number_of_ds = 1 << Vector_GetSize(test_markers);  // 2 ^ number of test markers
    if (ds == NULL) {
        ds = Memory_Malloc(number_of_ds * sizeof(double));
    }
    if (lower_indiv == NULL) {
        lower_indiv = Memory_Malloc(number_of_ds * sizeof(double));
    }
    if (upper_indiv == NULL) {
        upper_indiv = Memory_Malloc(number_of_ds * sizeof(double));
    }
    double wildtype_allele_frequency = 1 - disease_allele_frequency;
    // Compute individual constraints (example for 3 test markers xyz)
    // max(-fxyz * p, fxyz * (1-p) - 1) <= dxyz <= min(1 - fxyz * p, fxyz * (1-p))
    for (int i = 0; i < number_of_ds; i++) {
        lower_indiv[i] = max(-DoubleVector_Get(em_hapfreqs, i) * disease_allele_frequency, DoubleVector_Get(em_hapfreqs, i) * wildtype_allele_frequency - 1);
        upper_indiv[i] = min(1 - DoubleVector_Get(em_hapfreqs, i) * disease_allele_frequency, DoubleVector_Get(em_hapfreqs, i) * wildtype_allele_frequency);
    }
    // Draw d's iteratively until all d's are drawn.
    for (int i = 0; i < number_of_ds; i++) {
        // Compute group constraint.
        // -sum(dxyz_max) <= d <= -sum(dxyz_min)   (sums without d)
        double lower_group = 0;
        double upper_group = 0;
        for (int j = 0; j < number_of_ds; j++) {
            if (j < i) {
                // Note: ds[j] was initialized in previous iteration
                lower_group -= ds[j];
                upper_group -= ds[j];
            } else if (j == i) {
                continue;
            } else {            // j > i
                lower_group -= upper_indiv[j];
                upper_group -= lower_indiv[j];
            }
        }
        // Combine group with individual constraint.
        double lower = max(lower_indiv[i], lower_group);
        double upper = min(upper_indiv[i], upper_group);
        // Draw d from combined constraint.
        ds[i] = lower + (upper - lower) * genrand_open_open();
    }
    // Compute hapfreqs from p, q, ds, and em_hapfreqs.
    DoubleVector hapfreqs = HapfreqPool_Take(HAPFREQPOOL);
    DoubleVector deltas = DoubleVector_Create(number_of_ds);
    for (int i = 0; i < number_of_ds; i++) {
        // The order of hapfreqs is always H0, H1, ..., where hapfreqs
        // H0, H2, etc. contain the wildtype allele and hapfreqs H1,
        // H3, etc. contain the mutant allele.  To get this order the
        // even indexes must be wildtype hapfreqs with wildtype allele
        // frequency q and the odd indexes must be mutant hapfreqs
        // with mutant allele frequency p.
        DoubleVector_Set(hapfreqs, 2*i,     DoubleVector_Get(em_hapfreqs, i) * wildtype_allele_frequency - ds[i]);
        DoubleVector_Set(hapfreqs, 2*i + 1, DoubleVector_Get(em_hapfreqs, i) * disease_allele_frequency + ds[i]);
        DoubleVector_PushBack(deltas, ds[i]);
    }
    for (int i = 0; i < Vector_GetSize(penetrances); i++) {
        double mod_score_value = ModScorer_ComputeModScore(
            0, annotated_families, test_markers, disease_allele_frequency, hapfreqs,
            Vector_Get(penetrances, i), em_hapfreqs);
        if (mod_score_value > best_mod_score) {
            best_mod_score = mod_score_value;
            best_mod_score_improved = 1;
            mod_score->test_markers = test_markers;
            mod_score->position = LinkageMarkerFile_GetDistanceBetween(linkage_file, Marker_GetIndex(Vector_Get(test_markers, 0)), 0);
            mod_score->mod_score = mod_score_value;
            mod_score->disease_allele_frequency = disease_allele_frequency;
            mod_score->penetrances = Vector_Get(penetrances, i);
            if (mod_score->hapfreqs && mod_score->hapfreqs != hapfreqs) {
                HapfreqPool_Release(HAPFREQPOOL, mod_score->hapfreqs);
            }
            mod_score->hapfreqs = hapfreqs;
            mod_score->em_hapfreqs = em_hapfreqs;
            mod_score->deltas = deltas;
        }
        if (mod_score->hapfreqs != hapfreqs) {
            HapfreqPool_Release(HAPFREQPOOL, hapfreqs);
        }
    }
    // Release hapfreqs unless they were inserted into results table
    if (!best_mod_score_improved) {
        HapfreqPool_Release(HAPFREQPOOL, hapfreqs);
    }
    return best_mod_score_improved;
}

static int draw_hapfreqs_and_compute_modscore(
    Vector test_markers,
    Vector annotated_families,
    double disease_allele_frequency,
    Vector penetrances,
    DoubleVector em_hapfreqs,
    LinkageMarkerFile linkage_file,
    ModScore *mod_score)
{
    static double best_mod_score = -HUGE_VAL;
    static Vector prev_test_markers = NULL;
    int best_mod_score_improved = 0;
    if (test_markers != prev_test_markers) {
        best_mod_score = -HUGE_VAL;
        prev_test_markers = test_markers;
    }
    int number_of_hapfreqs = 2 * DoubleVector_GetSize(em_hapfreqs);

    // Draw hapfreqs iteratively.
    DoubleVector hapfreqs = HapfreqPool_Take(HAPFREQPOOL);
    double p = disease_allele_frequency;
    double sum_even_so_far = 0;
    double sum_odd_so_far = 0;
    for (int i = 0; i < number_of_hapfreqs - 2; i++) {
        // Group constraints: sum H(even) = 1-p and sum H(odd) = p
        // The constraint that H(x) in [0, 1] is automatically
        // satisfied by the below code.
        double hap;
        if (i % 2 == 0) {  // even
            hap = ((1 - p) - sum_even_so_far) * genrand_open_open();
            sum_even_so_far += hap;
        } else {  // odd
            hap = (p - sum_odd_so_far) * genrand_open_open();
            sum_odd_so_far += hap;
        }
        DoubleVector_Set(hapfreqs, i, hap);
    }
    DoubleVector_Set(hapfreqs, number_of_hapfreqs - 2, (1 - p) - sum_even_so_far);
    DoubleVector_Set(hapfreqs, number_of_hapfreqs - 1, p - sum_odd_so_far);

    // Set EM-estimated test marker haplotype frequencies for likelihood denominator

    // DoubleVector dummy_hapfreqs = DoubleVector_Create(4);
    // DoubleVector_PushBack(dummy_hapfreqs, 0.25);
    // DoubleVector_PushBack(dummy_hapfreqs, 0.25);
    // DoubleVector_PushBack(dummy_hapfreqs, 0.25);
    // DoubleVector_PushBack(dummy_hapfreqs, 0.25);
    // DoubleVector dummy_em_hapfreqs = DoubleVector_Create(2);
    // DoubleVector_PushBack(dummy_em_hapfreqs, 0.5);
    // DoubleVector_PushBack(dummy_em_hapfreqs, 0.5);
    // double dummy_disease_allele_frequency = 0.5;
    // DoubleVector dummy_penetrances = DoubleVector_With(10,
    //                    0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0, 1.0);
    // DoubleVector dummy_penetrances = DoubleVector_With(10,
    //                     0.01, 0.01, 0.01, 0.01, 0.99, 0.99, 0.99, 0.01, 0.01, 0.99);
    for (int i = 0; i < Vector_GetSize(penetrances); i++) {
        // double mod_score_value = ModScorer_ComputeModScore(
        //     0, annotated_families, test_markers, dummy_disease_allele_frequency, dummy_hapfreqs,
        //     dummy_penetrances, dummy_em_hapfreqs); // Note em_hapfreqs
        double mod_score_value = ModScorer_ComputeModScore(
            0, annotated_families, test_markers, disease_allele_frequency, hapfreqs,
            Vector_Get(penetrances, i), em_hapfreqs); // Note em_hapfreqs
        // sf(ps, "mod_score: %.15f\n", mod_score_value);
        // pr();
        // exit(1);
        if (mod_score_value > best_mod_score) {
            best_mod_score = mod_score_value;
            best_mod_score_improved = 1;
            mod_score->test_markers = test_markers;
            mod_score->position = LinkageMarkerFile_GetDistanceBetween(linkage_file, Marker_GetIndex(Vector_Get(test_markers, 0)), 0);
            mod_score->mod_score = mod_score_value;
            mod_score->disease_allele_frequency = disease_allele_frequency;
            mod_score->penetrances = Vector_Get(penetrances, i);
            if (mod_score->hapfreqs && mod_score->hapfreqs != hapfreqs) {
                HapfreqPool_Release(HAPFREQPOOL, mod_score->hapfreqs);
            }
            mod_score->hapfreqs = hapfreqs;
            mod_score->em_hapfreqs = em_hapfreqs;
            mod_score->deltas = NULL;
        }
        if (mod_score->hapfreqs != hapfreqs) {
            HapfreqPool_Release(HAPFREQPOOL, hapfreqs);
        }
    }
    // Release hapfreqs unless they were inserted into results table
    if (!best_mod_score_improved) {
        HapfreqPool_Release(HAPFREQPOOL, hapfreqs);
    }
    return best_mod_score_improved;
}
