#include "HaplotypeMatrix.h"
#include "Memory.h"
#include "Marker.h"
#include "HaplotypeTerm.h"

typedef struct {
    HaplotypeTerm term;
    int number_of_founders;
    PhasedGenotypes phased_disease_genotypes;
    PhasedGenotypes *phased_genotypes_of_test_markers;
    int number_of_test_markers;
    IntSet *sets_of_compatible_test_marker_genotypes;
} *Arguments;

static Arguments set_additional_arguments(HaplotypeTerm, int number_of_founders, PhasedGenotypes phased_disease_genotypes, PhasedGenotypes *phased_genotypes_of_test_markers, int number_of_test_markers, IntSet *sets_of_compatible_test_marker_genotypes);
static void collect_haplotype_terms(int test_marker, Arguments args);

HaplotypeMatrix HaplotypeMatrix_Create(Family family, Vector test_marker_set)
{
    IntSet *compatible_phased_test_marker_genotypes = Memory_Malloc(Vector_GetSize(test_marker_set) * sizeof(IntSet));;
    HaplotypeMatrix haplotype_matrix = HaplotypeMatrix_CreateEmpty(Family_GetNumberOfInheritanceVectors(family), Family_GetNumberOfDiseaseGenotypes(family));
    for (int inheritance_vector = 0; inheritance_vector < Family_GetNumberOfInheritanceVectors(family); inheritance_vector++) {
        Family_SetPhase(family, inheritance_vector);
        for (int i = 0; i < Vector_GetSize(test_marker_set); i++)
            compatible_phased_test_marker_genotypes[i] = Family_FindCompatiblePhasedFounderGenotypes(family, Marker_GetIndex(Vector_Get(test_marker_set, i)));
        for (int phased_disease_genotypes = 0; phased_disease_genotypes < Family_GetNumberOfDiseaseGenotypes(family); phased_disease_genotypes++) {
            int number_of_haplotypes = 1 << (Vector_GetSize(test_marker_set) + 1);
            HaplotypeTerm term = HaplotypeTerm_Create(number_of_haplotypes);
            HaplotypeMatrix_CollectHaplotypeTerms(term, Family_GetNumberOfFounders(family), phased_disease_genotypes, Vector_GetSize(test_marker_set), compatible_phased_test_marker_genotypes);
            HaplotypeMatrix_Set(haplotype_matrix, inheritance_vector, phased_disease_genotypes, term);
        }
        for (int i = 0; i < Vector_GetSize(test_marker_set); i++)
            IntSet_Destroy(compatible_phased_test_marker_genotypes[i]);
    }
    return haplotype_matrix;
}

HaplotypeMatrix HaplotypeMatrix_RemoveDiseaseLocus(HaplotypeMatrix matrix)
{
    int rows = Matrix_GetNumberOfRows(matrix);
    int columns = Matrix_GetNumberOfColumns(matrix);
    HaplotypeMatrix new_matrix = Matrix_Create(rows, columns);
    for (int i = 0; i < rows; i++) {
        for (int j = 0; j < columns; j++) {
            HaplotypeTerm term = Matrix_Get(matrix, i, j);
            Matrix_Set(new_matrix, i, j, HaplotypeTerm_RemoveDiseaseLocus(term));
        }
    }
    return new_matrix;
}

HaplotypeMatrix (*HaplotypeMatrix_CreateEmpty)(int number_of_rows, int number_of_columns) = Matrix_Create;
void (*HaplotypeMatrix_Set)(HaplotypeMatrix, int row, int column, void *value) = Matrix_Set;

void HaplotypeMatrix_CollectHaplotypeTerms(HaplotypeTerm term, int number_of_founders, PhasedGenotypes phased_disease_genotypes, int number_of_test_markers, IntSet *sets_of_compatible_test_marker_genotypes)
{
    PhasedGenotypes *phased_genotypes_of_test_markers = Memory_Malloc(number_of_test_markers * sizeof(PhasedGenotypes));
    Arguments arguments = set_additional_arguments(term, number_of_founders, phased_disease_genotypes, phased_genotypes_of_test_markers, number_of_test_markers, sets_of_compatible_test_marker_genotypes);
    collect_haplotype_terms(0, arguments);
    Memory_Free(arguments);
    Memory_Free(phased_genotypes_of_test_markers);
}

static Arguments set_additional_arguments(HaplotypeTerm term, int number_of_founders, PhasedGenotypes phased_disease_genotypes, PhasedGenotypes *phased_genotypes_of_test_markers, int number_of_test_markers, IntSet *sets_of_compatible_test_marker_genotypes)
{
    Arguments arguments = Memory_Malloc(sizeof(*arguments));
    arguments->term = term;
    arguments->number_of_founders = number_of_founders;
    arguments->phased_disease_genotypes = phased_disease_genotypes;
    arguments->phased_genotypes_of_test_markers = phased_genotypes_of_test_markers;
    arguments->number_of_test_markers = number_of_test_markers;
    arguments->sets_of_compatible_test_marker_genotypes = sets_of_compatible_test_marker_genotypes;
    return arguments;
}

// This recursive function does much more work than hits the eye.  It
// solves the following problem: We would like to run some code for
// every n-tuple of genotypes from n test markers.  With 2 test
// markers this would naturally translate into 2 nested for-loops.
// The problem is that the number of test markers and thus the number
// of nested for-loops is not fixed in advance.  So what we need is a
// way to express a variable number of nested for-loops.  This problem
// is solved by using a recursive function that contains a for-loop.
// If enough nested for-loops were "emulated", the function runs the
// code that would otherwise run in the innermost loop.  Otherwise
// the function uses a for-loop in which it calls itself recursively.
static void collect_haplotype_terms(int test_marker, Arguments args)
{
    if (test_marker == args->number_of_test_markers) {
        // Use n-tuple of compatible genotypes from n test markers.
        int *exponents = HaplotypeTerm_ComputeExponents(
            args->term,
            args->number_of_founders,
            args->phased_disease_genotypes,
            args->phased_genotypes_of_test_markers,
            args->number_of_test_markers);
        HaplotypeTerm_Add(args->term, exponents);
        return;
    }
    // Collect n-tuple of compatible genotypes from n test markers.
    for (PhasedGenotypes i = 0; (int) i < PhasedGenotypes_GetNumberOfCombinations(args->number_of_founders); i++) {
        if (!IntSet_Contains(args->sets_of_compatible_test_marker_genotypes[test_marker], i))
            continue;
        args->phased_genotypes_of_test_markers[test_marker] = i;
        collect_haplotype_terms(test_marker + 1, args);
    }
}
