#include "HaplotypeTerm.h"
#include "Memory.h"
#include "Vector.h"
#include "MathFunctions.h"
#include <math.h>
#include <stdio.h>

// Legacy includes.
#define INC_LIB
#define INC_SHELL
#include "npl.h"

struct HaplotypeTermStruct {
    Vector terms;
    int number_of_haplotypes;
};

static int build_haplotype_index(int left, int index_of_person, PhasedGenotypes disease_genotypes, PhasedGenotypes *test_marker_genotypes, int number_of_test_markers);
static int get_index_of_haplotype(int left, Genotype disease_genotype, Genotype *test_marker_genotype, int number_of_test_markers);

HaplotypeTerm HaplotypeTerm_Create(int number_of_haplotypes)
{
    HaplotypeTerm term = Memory_Malloc(sizeof(*term));
    term->terms = Vector_CreateEmpty();
    term->number_of_haplotypes = number_of_haplotypes;
    return term;
}

void HaplotypeTerm_Destroy(HaplotypeTerm term)
{
    for (int i = 0; i < Vector_GetSize(term->terms); i++)
        Memory_Free(Vector_Get(term->terms, i));
    Vector_Destroy(term->terms);
    Memory_Free(term);
}

void HaplotypeTerm_Add(HaplotypeTerm term, int *exponents)
{
    Vector_PushBack(term->terms, exponents);
}

double HaplotypeTerm_Evaluate(HaplotypeTerm term, DoubleVector frequencies, double **cache)
{
    double sum = 0.0;
    double *freqs = DoubleVector_GetInternalArray(frequencies);
    int nfreqs = DoubleVector_GetSize(frequencies);
    int number_of_terms = Vector_GetSize(term->terms);
    void **terms = Vector_GetInternalArray(term->terms);
    for (int i = 0; i < number_of_terms; i++) {
        int *exponents = terms[i];
        double product = 1.0;
        for (int j = 0; j < nfreqs; j++) {
            int exp = exponents[j];
            if (exp == 0) {
                continue;
            }
            double f = cache[j][exp];
            if (f < 0) {
                f = cache[j][exp] = Math_IntegerPower(freqs[j], exp);
            }
            if (f == 0) {
                product = 0;
                break;
            }
            product *= f;
        }
        if (product == 0) {
            continue;
        }
        sum += product;
    }
    return sum;
}

double HaplotypeTerm_EvaluateUnderH0(HaplotypeTerm term, DoubleVector haplotype_frequencies)
{
    double p = HaplotypeTerm_GetMutantFrequency(haplotype_frequencies);
    double q = 1 - p;
    int *exponents = Vector_Get(term->terms, 0);
    double product = 1.0;
    for (int i = 0; i < term->number_of_haplotypes; i++)
        product *= pow((i % 2) ? p : q, exponents[i]);
    return product;
}

int *HaplotypeTerm_ComputeExponents(
    HaplotypeTerm term,
    int number_of_founders,
    PhasedGenotypes disease_genotypes,
    PhasedGenotypes *test_marker_genotypes,
    int number_of_test_markers)
{
    int *exponents = Memory_Calloc(term->number_of_haplotypes, sizeof(int));
    for (int i = 0; i < number_of_founders; i++) {
        ++exponents[build_haplotype_index(0, i, disease_genotypes, test_marker_genotypes, number_of_test_markers)];
        ++exponents[build_haplotype_index(1, i, disease_genotypes, test_marker_genotypes, number_of_test_markers)];
    }
    return exponents;
}

static int build_haplotype_index(int left, int index_of_person, PhasedGenotypes disease_genotypes, PhasedGenotypes *test_marker_genotypes, int number_of_test_markers)
{
    Genotype genotype_of_test_markers[MAX_NUMBER_OF_TEST_MARKERS] = { 0 };
    Genotype disease_genotype = PhasedGenotypes_Get(disease_genotypes, index_of_person);
    for (int i = 0; i < number_of_test_markers; i++)
        genotype_of_test_markers[i] = PhasedGenotypes_Get(test_marker_genotypes[i], index_of_person);
    return get_index_of_haplotype(left, disease_genotype, genotype_of_test_markers, number_of_test_markers);
}

// The bit pattern of a haplotype's index (read "number") defines the
// alleles in the haplotype.  The least significant bit corresponds to
// the allele of the disease genotype.  The next bit corresponds to
// the allele of the first test marker.  The next bit corresponds to
// the allele of the second test marker, and so on.  The most
// significant bit corresponds to the last test marker.
//
// The following figure assumes 2 test markers and shows the alleles
// of the haplotype with (0-based) index 3, that is, of the 4th
// haplotype.
//
//                     +------ most significant bit
//                     |   +-- least significant bit
//                     |   |
// haplotype index 3 = 0 1 1
//                     | | |
//                     | | +-- allele 2 of disease locus (mutant allele)
//                     | +---- allele 2 of test marker 1
//                     +------ allele 1 of test marker 2
//
static int get_index_of_haplotype(int left, Genotype disease_genotype, Genotype *test_marker_genotype, int number_of_test_markers)
{
    int shift_width = left ? 1 : 0;
    int result = 0x0;
    for (int test_marker = number_of_test_markers - 1; test_marker >= 0; test_marker--) {
        result <<= 1;
        result |= (test_marker_genotype[test_marker] >> shift_width) & 0x1;
    }
    result <<= 1;
    result |= (disease_genotype >> shift_width) & 0x1;
    return result;
}

DoubleVector HaplotypeTerm_GetWildtypeFrequencies(DoubleVector haplotype_frequencies)
{
    int number_of_haplotypes = DoubleVector_GetSize(haplotype_frequencies);
    int number_of_alleles = HaplotypeTerm_GetNumberOfAlleles(number_of_haplotypes);
    DoubleVector result = DoubleVector_Create(number_of_alleles);
    for (int i = 0; i < number_of_alleles; i++) {
        double wildtype_freq = 0.0;
        for (int j = 0; j < number_of_haplotypes; j++) {
            int has_wildtype_allele = (((j >> i) & 0x1) == 0);
            if (has_wildtype_allele)
                wildtype_freq += DoubleVector_Get(haplotype_frequencies, j);
        }
        DoubleVector_PushBack(result, wildtype_freq);
    }
    return result;
}

int HaplotypeTerm_GetNumberOfAlleles(int number_of_haplotypes)
{
    if (number_of_haplotypes == 0)
        return 0;
    return Math_Log2(number_of_haplotypes);
}

double HaplotypeTerm_GetMutantFrequency(DoubleVector haplotype_frequencies)
{
    double result = 0.0;
    for (int i = 1; i < DoubleVector_GetSize(haplotype_frequencies); i += 2)
        result += DoubleVector_Get(haplotype_frequencies, i);
    return result;
}

int HaplotypeTerm_IsEmpty(HaplotypeTerm term)
{
    return Vector_GetSize(term->terms) == 0;
}

void HaplotypeTerm_Show(HaplotypeTerm term)
{
    fprintf(stderr, "\n");
    for (int i = 0; i < term->number_of_haplotypes; i++)
        fprintf(stderr, "H%d ", i);
    fprintf(stderr, "\n");
    for (int i = 0; i < Vector_GetSize(term->terms); i++) {
        int *exponents = Vector_Get(term->terms, i);
        for (int j = 0; j < term->number_of_haplotypes; j++)
            fprintf(stderr, "%2d ", exponents[j]);
        fprintf(stderr, "\n");
    }
}

HaplotypeTerm HaplotypeTerm_RemoveDiseaseLocus(HaplotypeTerm term)
{
    int new_number_of_haplotypes = term->number_of_haplotypes / 2;
    HaplotypeTerm new_term = HaplotypeTerm_Create(new_number_of_haplotypes);
    for (int i = 0; i < Vector_GetSize(term->terms); i++) {
        int *old_exponents = Vector_Get(term->terms, i);
        int *new_exponents = Memory_Calloc(new_number_of_haplotypes, sizeof(int));
        // Sum consecutive pairs of exponents.
        for (int j = 0; j < term->number_of_haplotypes; j++)
            new_exponents[j/2] += old_exponents[j];
        HaplotypeTerm_Add(new_term, new_exponents);
    }
    return new_term;
}
