#include "Haplotypes.h"
#include "Memory.h"
#include "Marker.h"

static int build_haplotype_index(int left, int index_of_person, PhasedGenotypes phased_founder_genotypes, PhasedGenotypes *phased_genotypes_of_test_markers, int number_of_test_markers);
static int get_index_of_haplotype(int left, Genotype disease_genotype, Genotype *genotype_of_test_markers, int number_of_test_markers);
static void Haplotypes_GetForFoundersImpl(Haplotypes, int number_of_founders, PhasedGenotypes phased_founder_genotypes, PhasedGenotypes *phased_genotypes_of_test_markers, int number_of_test_markers);
static Vector collect_marker_frequencies(Vector markers);
static DoubleVector multiply_frequencies_in_tuples(Vector frequency_tuples);

Haplotypes Haplotypes_Create(void)
{
    int i;
    Haplotypes haplotypes = (Haplotypes) Memory_Malloc(MAX_NUMBER_OF_POSSIBLE_HAPLOTYPES * sizeof(char));

    for (i = 0; i < MAX_NUMBER_OF_POSSIBLE_HAPLOTYPES; i++)
        haplotypes[i] = 0;

    return haplotypes;
}

static void Haplotypes_GetForFoundersImpl(Haplotypes haplotypes, int number_of_founders, PhasedGenotypes phased_founder_genotypes, PhasedGenotypes *phased_genotypes_of_test_markers, int number_of_test_markers)
{
    int i;

    for (i = 0; i < number_of_founders; i++) {
        ++haplotypes[build_haplotype_index(0, i, phased_founder_genotypes, phased_genotypes_of_test_markers, number_of_test_markers)];
        ++haplotypes[build_haplotype_index(1, i, phased_founder_genotypes, phased_genotypes_of_test_markers, number_of_test_markers)];
    }
}

void (*Haplotypes_GetForFounders)(Haplotypes, int number_of_founders, PhasedGenotypes, PhasedGenotypes *phased_genotypes_of_test_markers, int number_of_test_markers) = Haplotypes_GetForFoundersImpl;

static int build_haplotype_index(int left, int index_of_person, PhasedGenotypes phased_founder_genotypes, PhasedGenotypes *phased_genotypes_of_test_markers, int number_of_test_markers)
{
    Genotype genotype_of_test_markers[MAX_NUMBER_OF_TEST_MARKERS] = { 0 };
    Genotype disease_genotype = PhasedGenotypes_Get(phased_founder_genotypes, index_of_person);
    int i;

    for (i = 0; i < number_of_test_markers; i++)
        genotype_of_test_markers[i] = PhasedGenotypes_Get(phased_genotypes_of_test_markers[i], index_of_person);

    return get_index_of_haplotype(left, disease_genotype, genotype_of_test_markers, number_of_test_markers);
}

int Haplotypes_GetIndexOfLeftHaplotype(Genotype disease_genotype, Genotype *genotype_of_test_markers, int number_of_test_markers)
{
    return get_index_of_haplotype(1, disease_genotype, genotype_of_test_markers, number_of_test_markers);
}

int Haplotypes_GetIndexOfRightHaplotype(Genotype disease_genotype, Genotype *genotype_of_test_markers, int number_of_test_markers)
{
    return get_index_of_haplotype(0, disease_genotype, genotype_of_test_markers, number_of_test_markers);
}

static int get_index_of_haplotype(int left, Genotype disease_genotype, Genotype *genotype_of_test_markers, int number_of_test_markers)
{
    int test_marker, haplotype_index = 0x0;
    int shift_width = left ? 1 : 0;

    for (test_marker = number_of_test_markers - 1; test_marker >= 0; test_marker--) {
        haplotype_index <<= 1;
        haplotype_index |= (genotype_of_test_markers[test_marker] >> shift_width) & OI;
    }
    haplotype_index <<= 1;
    haplotype_index |= (disease_genotype >> shift_width) & OI;

    return haplotype_index;
}

void Haplotypes_Destroy(Haplotypes haplotypes)
{
    Memory_Free(haplotypes);
}

DoubleVector Haplotypes_GetExpectedFrequencies(Vector markers)
{
    Vector marker_frequencies = collect_marker_frequencies(markers);
    Vector frequency_tuples = Vector_FindCombinations(marker_frequencies);
    DoubleVector expected_haplotype_frequencies = multiply_frequencies_in_tuples(frequency_tuples);
    Vector_Destroy(marker_frequencies);
    Vector_Destroy(frequency_tuples);
    return expected_haplotype_frequencies;
}

static Vector collect_marker_frequencies(Vector markers)
{
    Vector result = Vector_Create(Vector_GetSize(markers));
    for (int i = 0; i < Vector_GetSize(markers); i++)
        Vector_PushBack(result, Marker_GetAlleleFrequencies(Vector_Get(markers, i)));
    return result;
}

static DoubleVector multiply_frequencies_in_tuples(Vector frequency_tuples)
{
    DoubleVector result = DoubleVector_Create(Vector_GetSize(frequency_tuples));
    for (int i = 0; i < Vector_GetSize(frequency_tuples); i++)
        DoubleVector_PushBack(result, Vector_ProductOfDoubles(Vector_Get(frequency_tuples, i)));
    return result;
}

double Haplotypes_GetExpectedDiseaseAlleleFrequency(DoubleVector haplotype_frequencies)
{
    double result = 0;
    for (int i = 1; i < DoubleVector_GetSize(haplotype_frequencies); i++)
        if (Haplotypes_ContainsDiseaseAllele(i))
            result += DoubleVector_Get(haplotype_frequencies, i);
    return result;
}

int Haplotypes_ContainsDiseaseAllele(int haplotype)
{
    return haplotype & 0x1;
}
