#define INC_LIB
#define INC_SHELL
#include "npl.h"
#include "dSFMT.h"

#include "CommandReader.h"
#include "IO.h"
#include "Memory.h"
#include "ParseObject.h"

#include <errno.h>
#include <signal.h>
#include <stdint.h>      // int16_t, INT16_MAX
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>    // mkdir
#include <sys/types.h>   // mkdir, setpgid
#include <sys/wait.h>
#include <time.h>
#include <unistd.h>      // setpgid

#define MAX_CONSECUTIVE_FAILURES 3
#define NELEMS(X) (sizeof(X) / sizeof((X)[0]))

#define MAKE_STRING(S, FMT, ...) do { \
    int chars_written = 0; \
    const int max_length = 500; \
    (S) = Memory_Malloc(max_length); \
    if ((S) == NULL) { \
        fprintf(stderr, "out of memory: " FMT "\n", __VA_ARGS__); \
        exit(EXIT_FAILURE); \
    } \
    chars_written = snprintf((S), max_length, (FMT), __VA_ARGS__); \
    if (chars_written >= max_length) { \
        fprintf(stderr, "string too long: " FMT "\n", __VA_ARGS__); \
        exit(EXIT_FAILURE); \
    } \
} while (0)


struct ChildProcess {
    pid_t pid;
    int replicate;
    time_t start_time;
    int killed;  // 1 = killed, 0 = not killed
    char cmd[200];  // shell command run by child process
};

typedef void (*sighandler_t)(int);

static volatile int stopped = 0;  // 1 if simulation was stopped

// Array of up to g_jla_child_processes 1-based replicate numbers that
// were "pushed back" because they failed.  When the next replicate
// number is needed, it will be taken from the stack of pushed back
// replicate numbers.
static int *pushed_back_replicates = NULL;

// The seeds array contains the random number generator seeds for
// all replicates, MAX_CONSECUTIVE_FAILURES seeds per replicate.
// The seed_numbers array has the same length as the seeds array and
// contains 1-based, absolute seed indexes, where "absolute" means
// that indexes keep incrementing across --resume invocations.
// Absolute seed indexes are used to give unique names to the
// directories of failed replicates.
static int16_t *seeds = NULL;
static int *seed_numbers = NULL;

// seeds[replicate_seed_indexes[i]] is the next seed to be used for
// replicate i, where i is 1-based, i.e., 1 <= i <= g_jla_replicates.
static int *replicate_seed_indexes = NULL;

// replicate_successful[i] is 1 if replicate i was successful and
// otherwise 0, where i is 1-based, i.e., 1 <= i <= g_jla_replicates.
static int *replicate_successful = NULL;

// killed_pids contains the process IDs of child processes that were
// forcibly killed after stop_simulation was called.  The maximum
// number of PIDs in killed_pids is g_jla_child_processes, the maximum
// number of child processes that are allowed to run in parallel.
static int *killed_pids = NULL;


static void mkdir_or_fail(const char *directory)
{
    errno = 0;
    int ok = mkdir(directory, 0755) == 0 || errno == EEXIST;
    if (!ok) {
        sf(ps, "mkdir failed: %s\n", directory);
        pr();
        exit(1);
    }
}


static void create_folders_for_replicate(const char *directory)
{
    mkdir_or_fail(directory);

    char input_dir[200];
    size_t n;
    n = snprintf(input_dir, sizeof input_dir, "%s/input", directory);
    if (n >= sizeof input_dir) {
        sf(ps, "path name too long: %s/input", directory);
        pr();
        exit(1);
    }
    mkdir_or_fail(input_dir);

    char tmp_dir[200];
    n = snprintf(tmp_dir, sizeof tmp_dir, "%s/tmp", directory);
    if (n >= sizeof tmp_dir) {
        sf(ps, "path name too long: %s/tmp", directory);
        pr();
        exit(1);
    }
    mkdir_or_fail(tmp_dir);
}


static void copy_user_input_files(const char *directory, CommandReader reader, PedigreeFileSimulator simulator, int16_t seed)
{
    char *input_directory;
    MAKE_STRING(input_directory, "%s/input", directory);
    char *commands_file;
    MAKE_STRING(commands_file, "%s/sample.commands", input_directory);
    CommandReader_Write(reader, commands_file, seed);
    Memory_Free(commands_file);
    char *ped_file;
    MAKE_STRING(ped_file, "%s/sample.ped", input_directory);
    PedigreeFileSimulator_Simulate(simulator, ped_file);
    Memory_Free(ped_file);
    Memory_Free(input_directory);
    if (!CommandReader_CopyUserFile(reader, "read association regions ", directory)) {
        fprintf(stderr, "Failed to copy file: %s\n", CommandReader_GetUserAssociationRegionsFile(reader));
        exit(EXIT_FAILURE);
    }
    if (!CommandReader_CopyUserFile(reader, "read linkage markers ", directory)) {
        fprintf(stderr, "Failed to copy file: %s\n", CommandReader_GetUserLinkageMarkersFile(reader));
        exit(EXIT_FAILURE);
    }
    if (g_model_params_file) {
        if (!CommandReader_CopyUserFile(reader, "read association models ", directory)) {
            fprintf(stderr, "Failed to copy file: %s\n", CommandReader_GetUserAssociationModelsFile(reader));
            exit(EXIT_FAILURE);
        }
    }
    if (!CommandReader_CopyUserFile(reader, "load markers ", directory)) {
        fprintf(stderr, "Failed to copy file: %s\n", CommandReader_GetMarkerFile(reader));
        exit(EXIT_FAILURE);
    }
}


static void delete_old_output_files(const char *directory)
{
    char *result_file;
    MAKE_STRING(result_file, "%s/tmp/results.txt", directory);
    unlink(result_file);
    Memory_Free(result_file);
    char *log_file;
    MAKE_STRING(log_file, "%s/tmp/results.log", directory);
    unlink(log_file);
    Memory_Free(log_file);
}


// The int parameter is needed because we (also) use stop_simulation
// as a signalhandler_t which has the int parameter in its signature.
static void stop_simulation(int ignored)
{
    (void) ignored;
    stopped = 1;
}


static int copy_best_mod_score(int replicate, int nreplicates, const char *directory, FILE *results)
{
    static int header_copied = 0;
    static int max_rep_width = 0;
    char *results_file;
    MAKE_STRING(results_file, "%s/%d/tmp/results.txt", directory, replicate);
    FILE *fp = fopen(results_file, "rb");
    Memory_Free(results_file);
    if (fp == NULL) {
        return 0;
    }
    char *header = IO_GetLine(fp);
    if (header == NULL) {
        fclose(fp);
        return 0;
    }
    // Compute width of column with replicate number:
    // width = max(digits in highest replicate number,
    //             characters in column name)
    const char *rep_column = "rep";
    if (max_rep_width == 0) {
        while (nreplicates > 0) {
            max_rep_width++;
            nreplicates /= 10;
        }
        if (max_rep_width < (int) strlen(rep_column)) {
            max_rep_width = strlen(rep_column);
        }
    }
    if (!header_copied && !g_jla_resume) {
        fprintf(results, "%*s %s", max_rep_width, rep_column, header);
        header_copied = 1;
    }
    char *line = IO_GetLine(fp);
    fclose(fp);
    if (line == NULL) {
        return 0;
    }
    fprintf(results, "%*d %s", max_rep_width, replicate, line);
    fflush(results);
    return 1;
}


static void delete_replicate_directory(int replicate, const char *directory)
{
    char *cmd;
    MAKE_STRING(cmd, "rm -rf %s/%d", directory, replicate);
    int status = system(cmd);   // catch return value to silence warning
    (void) status;              // then ignore explicitly
    Memory_Free(cmd);
}


static void rename_failed_replicate_directory(int replicate, const char *directory)
{
    // We do -1 because get_replicate_and_seed increments the index
    // every time it returns a seed.  Since a failure can only happen
    // after a seed was returned, -1 always yields a valid index.
    int seed_number = seed_numbers[replicate_seed_indexes[replicate] - 1];
    char *cmd;
    MAKE_STRING(cmd, "rm -rf %s/%d_failed.%d; mv %s/%d %s/%d_failed.%d",
        directory, replicate, seed_number,
        directory, replicate,
        directory, replicate, seed_number);
    int status = system(cmd);   // catch return value to silence warning
    (void) status;              // then ignore explicitly
    Memory_Free(cmd);
}


static void init_pushed_back_replicates(void)
{
    pushed_back_replicates = Memory_Malloc(g_jla_child_processes * sizeof(int));
    for (int i = 0; i < g_jla_child_processes; i++) {
        pushed_back_replicates[i] = -1;
    }
}


static void get_replicate_and_seed(int *replicate, int16_t *seed)
{
    static int prev_rep = 0;  // highest replicate number returned so far
    if (!pushed_back_replicates) {
        init_pushed_back_replicates();
    }
    // If there is a replicate number that was "pushed back" using
    // unget_replicate, it means that the replicate failed, and we
    // can reuse it.
    int rep = -1;
    for (int i = 0; i < g_jla_child_processes; i++) {
        if (pushed_back_replicates[i] > 0) {
            rep = pushed_back_replicates[i];
            // Clear pushed back replicate number.
            pushed_back_replicates[i] = -1;
            break;
        }
    }
    // If no replicate number was pushed back, use next higher
    // replicate number (and update prev_rep accordingly).
    if (rep == -1) {
        for (int i = prev_rep + 1; i <= g_jla_replicates; i++) {
            if (replicate_successful[i] <= 0) {
                prev_rep = rep = i;
                break;
            }
        }
    }
    if (rep == -1) {
        sf(ps, "Error: no next replicate, all replicates are done\n");
        pr();
        exit(1);
    }
    *replicate = rep;
    *seed = seeds[replicate_seed_indexes[rep]++];
}


static void unget_replicate(int replicate)
{
    if (!pushed_back_replicates) {
        init_pushed_back_replicates();
    }
    for (int i = 0; i < g_jla_child_processes; i++) {
        if (pushed_back_replicates[i] == -1) {
            pushed_back_replicates[i] = replicate;
            return;
        }
    }
    sf(ps, "Error: impossible: more consecutive failed replicates than JLA child processes.\n");
    pr();
    exit(1);
}

static int is_killed_pid(int pid)
{
    for (int i = 0; i < g_jla_child_processes; i++) {
        if (pid == killed_pids[i]) {
            return 1;
        }
    }
    return 0;
}

static void wait_for_children(
    struct ChildProcess *children,
    int npids,
    int options,
    int nreplicates,
    const char *directory,
    FILE *results,
    FILE *resume_fp,
    int *nsuccessful,
    int *nfinished)
{
    static int consecutive_failures = 0;
    static int finished_so_far = 0;
    *nsuccessful = 0;
    *nfinished = 0;
    int status;
    for (int i = 0; i < npids; i++) {
        if (children[i].pid < 0) {
            continue;
        }
        pid_t pid = waitpid(children[i].pid, &status, options);
        if (pid > 0) {
            int replicate = -1;
            // Free slot of finished child and find replicate number
            for (int j = 0; j < npids; j++) {
                if (children[j].pid == pid) {
                    children[j].pid = -1;
                    replicate = children[j].replicate;
                    break;
                }
            }
            int ok = copy_best_mod_score(replicate, nreplicates, directory, results);
            if (ok) {
                ++*nsuccessful;
                replicate_successful[replicate] = 1;
                fprintf(stderr, ".");
                fprintf(photo, ".");
                fflush(photo);
                if (!g_jla_store_replicates) {
                    delete_replicate_directory(replicate, directory);
                }
                consecutive_failures = 0;
            } else {
                fprintf(stderr, "x");
                fprintf(photo, "x");
                fflush(photo);
                unget_replicate(replicate);
                if (g_jla_debug_replicates && !is_killed_pid(pid)) {
                    rename_failed_replicate_directory(replicate, directory);
                } else {
                    delete_replicate_directory(replicate, directory);
                }
                if (consecutive_failures >= 0) {
                    ++consecutive_failures;
                }
            }
            ++finished_so_far;
            if (finished_so_far % 100 == 0 && finished_so_far > 0) {
                fprintf(stderr, "|\n|");
                fprintf(photo, "|\n|");
                fflush(photo);
            } else if (finished_so_far % 10 == 0 && finished_so_far > 0) {
                fprintf(stderr, "|");
                fprintf(photo, "|");
                fflush(photo);
            }
            if (!is_killed_pid(pid)) {
                fprintf(resume_fp, " %d", (ok ? 1 : -1) * replicate);
            }
            ++*nfinished;
        }
    }
    if (!stopped && consecutive_failures >= MAX_CONSECUTIVE_FAILURES) {
        sf(ps, "\n%d consecutive replicates failed ... aborting p value simulation\n", consecutive_failures);
        pr();
        consecutive_failures = -1;  // ignore failures from now on
        stop_simulation(0);
    }
}

// Kill child processes by sending them a SIGTERM signal.
static void kill_children(struct ChildProcess *children, int npids)
{
    int k = 0;  // index into killed_pids
    for (int i = 0; i < npids; i++) {
        if (children[i].pid < 0 || children[i].killed) {
            continue;
        }
        sf(ps, "aborting replicate %d (pid %ld)\n",
            children[i].replicate, (long) children[i].pid);
        pr();
        kill(-children[i].pid, SIGTERM);  // pgid == pid
        children[i].killed = 1;
        killed_pids[k++] = children[i].pid;
    }
}

static struct ChildProcess *prepare_replicate(
    int replicate,
    struct ChildProcess *children,
    int max_children,
    const char *parent_dir,
    int16_t seed,
    CommandReader reader,
    PedigreeFileSimulator simulator)
{
    // Find next free child process
    struct ChildProcess *child = NULL;
    for (int i = 0; i < max_children; i++) {
        if (children[i].pid < 0) {
            child = &children[i];
            break;
        }
    }
    child->replicate = replicate;
    child->start_time = time(0);
    child->killed = 0;

    // Prepare directories for replicate.
    char dir[200];
    size_t n = snprintf(dir, sizeof dir, "%s/%d", parent_dir, replicate);
    if (n >= sizeof dir) {
        sf(ps, "path name too long: %s/%d\n", parent_dir, replicate);
        pr();
        exit(1);
    }
    create_folders_for_replicate(dir);
    copy_user_input_files(dir, reader, simulator, seed);
    delete_old_output_files(dir);

    // Create command-line run by replicate
    //
    // Include number of replicate as an unused command line argument
    // to ghm.  This allows the user to follow simulation progress
    // using programs like top(1) (toggle to full command lines by
    // pressing c inside top), ps(1), or pgrep(1) (pgrep -a ghm).
    n = snprintf(child->cmd, sizeof child->cmd,
        "cd %s && %s --tag %d --commands input/sample.commands >/dev/null 2>&1",
        dir, g_ghm_executable, replicate);
    if (n >= sizeof child->cmd) {
        sf(ps, "command too long: \"cd %s && %s --tag %d"
            " --commands input/sample.commands >/dev/null 2>&1\"\n",
            dir, g_ghm_executable, replicate);
        pr();
        exit(1);
    }

    return child;
}


static int run_all_replicates(
    int nreplicates,
    int max_children,
    const char *directory,
    FILE *results_fp,
    FILE *resume_fp,
    CommandReader reader,
    PedigreeFileSimulator simulator)
{
    sf(ps, "Replicates: %d\n", nreplicates);
    pr();
    sf(ps, ". ok\nx failed\n");
    pr();
    fprintf(stderr, "|");
    fprintf(photo, "|");
    // Unless we flush the file buffer at this point, the forked
    // children will inherit an unflushed file buffer and will flush
    // it on exit leading to repeated prints in the log file.
    fflush(photo);

    sighandler_t old_sigint_handler = signal(SIGINT, stop_simulation);
    sighandler_t old_sigterm_handler = signal(SIGTERM, stop_simulation);
    int nchildren = 0;   // number of currently active child processes
    struct ChildProcess *children = Memory_Malloc(max_children * sizeof(struct ChildProcess));
    for (int i = 0; i < max_children; i++) {
        children[i].pid = -1;  // mark free slots
    }
    int nsuccessful = 0;  // number of successful replicates
    for (int i = 1; i <= nreplicates; i++) {
        nsuccessful += (replicate_successful[i] == 1);
    }
    int n, nfinished;
    while (nsuccessful < nreplicates && !stopped) {
        if (nchildren < max_children
            && nsuccessful + nchildren < nreplicates)
        {
            int replicate;  // 1-based replicate number
            int16_t seed;
            get_replicate_and_seed(&replicate, &seed);
            struct ChildProcess *child = prepare_replicate(
                replicate, children, max_children, directory,
                seed, reader, simulator);
            pid_t pid = fork();
            if (pid == 0) {  // This code runs in the forked child.
                close(fileno(photo));
                close(fileno(results_fp));
                close(fileno(resume_fp));
                close(fileno(stdout));
                close(fileno(stderr));
                exit(system(child->cmd));
            }
            if (pid == -1) {
                perror("fork");
                stop_simulation(0);
                break;
            }
            child->pid = pid;
            // Make forked child process group leader so that we can
            // kill it and its descendants via its process group id.
            if (setpgid(pid, pid) == -1) {
                perror("setpgid");
                stop_simulation(0);
                break;
            }
            nchildren++;
        }
        if (nchildren > 0) {
            wait_for_children(children, max_children, WNOHANG,
                nreplicates, directory, results_fp, resume_fp, &n, &nfinished);
            nsuccessful += n;
            nchildren -= nfinished;
        }
    }
    if (stopped) {
        sf(ps, "stopped\n");
        pr();
        kill_children(children, max_children);
        wait_for_children(children, max_children, 0, nreplicates,
            directory, results_fp, resume_fp, &n, &nfinished);
    }
    Memory_Free(children);
    signal(SIGINT, old_sigint_handler);
    signal(SIGTERM, old_sigterm_handler);
    return 1;
}


static void verify_prerequisites_or_exit(void)
{
    if (g_jla_replicates <= 0) {
        sf(ps, "Number of JLA replicates must be at least 1\n");
        pr();
        exit(EXIT_FAILURE);
    }
    if (g_jla_child_processes <= 0) {
        sf(ps, "Number of JLA child processes must be at least 1\n");
        pr();
        exit(EXIT_FAILURE);
    }
    if (g_jla_directory == NULL) {
        sf(ps, "JLA directory is missing in commands file\n");
        pr();
        exit(EXIT_FAILURE);
    }
    if (g_jla_resume) {
        char *filename;
        MAKE_STRING(filename, "%s/.ghm_resume", g_jla_directory);
        FILE *fp = fopen(filename, "r");
        if (fp == NULL) {
            sf(ps, "JLA resume file does not exist: %s\n", filename);
            pr();
            exit(EXIT_FAILURE);
        }
        fclose(fp);
    }
}


static double compute_empirical_p_value(const char *results_file, double mod_score)
{
    FILE *fp = fopen(results_file, "rb");
    if (!fp) {
        sf(ps, "Failed to open file: \"%s\"\n", results_file);
        pr();
        exit(1);
    }
    char *header = IO_GetLine(fp);
    if (header == NULL) {
        fclose(fp);
        return -1.0;
    }
    // Find index of mod score column
    int mod_score_field_index = -1;
    ParseObject header_object = ParseObject_Create(header);
    Vector column_names = ParseObject_SplitIntoWords(header_object);
    for (int i = 0; i < Vector_GetSize(column_names); i++) {
        if (strcmp(Vector_Get(column_names, i), "mod") == 0) {
            mod_score_field_index = i;
        }
    }
    if (mod_score_field_index == -1) {
        fclose(fp);
        return -1.0;
    }
    Memory_Free(header);
    int bigger_mod_scores = 1; // the real data mod score
    int nmod_scores = 1; // the real data mod score
    char *line = IO_GetLine(fp);
    while (line != NULL) {
        ParseObject parser = ParseObject_Create(line);
        Vector fields = ParseObject_SplitIntoWords(parser);
        const char *this_mod_score_field = Vector_Get(fields, mod_score_field_index);
        double this_mod_score = atof(this_mod_score_field);
        if (this_mod_score >= mod_score) {
            bigger_mod_scores++;
        }
        nmod_scores++;
        Memory_Free(line);
        ParseObject_Destroy(parser);
        line = IO_GetLine(fp);
    }
    fclose(fp);
    return (double) bigger_mod_scores / nmod_scores;
}


// A ghm p value simulation run should never fail just because all
// seeds for a replicate were exhausted.  Seeds are never exhausted;
// we can always generate more.  The trick is to generate the seeds in
// a way that no matter which replicates fail and in what order, when
// re-running the same setup, every run of every replicate has to get
// the same seed even when the actual running order of the replicates
// changes.  This ensures reproducibility.  We can achieve this by
// using "virtual" seeds arrays.  Every virtual seeds array contains
// MAX_CONSECUTIVE_FAILURES seeds for every replicate.  If a seed from
// the current virtual seeds array fails, we replace it by a seed from
// the next virtual seeds array.

// Return the 0-based number of the virtual seeds array from which the
// random seed at seeds[seeds_index] should be taken given that the
// associated replicate failed failures times.
static int get_virtual_array(int seeds_index, int failures)
{
    static const int M = MAX_CONSECUTIVE_FAILURES;
    return (failures + (seeds_index % M)) / M;
}


// Return the 0-based index into the current virtual seeds array from
// which the random seed at seeds[seeds_index] should be taken given
// that the associated replicate failed failure times.
static int get_virtual_array_index(int seeds_index, int failures)
{
    static const int M = MAX_CONSECUTIVE_FAILURES;
    int replicate = (seeds_index / M) + 1;
    return (replicate - 1) * M + ((failures + seeds_index) % M);
}


// Initialize seeds array with from virtual seeds arrays taking into
// account previous replicate failures from replicate_successful.
static void init_seeds_for_replicates(void)
{
    static const int M = MAX_CONSECUTIVE_FAILURES;
    int needed_seeds = M * g_jla_replicates;
    int virtual_array = 0, virtual_array_index = 0;
    while (1) {
        int seed = INT16_MAX * genrand_open_open();
        // Does any element of the seeds array want this random seed?
        for (int i = 0; i < g_jla_replicates * M; i++) {
            int replicate = (i / M) + 1;
            int failures = replicate_successful[replicate];
            if (failures < 0)
                failures = -failures;
            else if (failures > 0)
                failures = 0;
            int va = get_virtual_array(i, failures);
            int vai = get_virtual_array_index(i, failures);
            if (va == virtual_array && vai == virtual_array_index) {
                seeds[i] = seed;
                seed_numbers[i] = va * M + (vai % M) + 1;  // 1-based
                needed_seeds--;
                break;
            }
        }
        if (needed_seeds == 0)
            return;
        virtual_array_index++;
        if (virtual_array_index == g_jla_replicates * M) {
            virtual_array++;
            virtual_array_index = 0;
        }
    }
}


void simulate_jla_p_value(PedigreeFileSimulator simulator, CommandReader reader, double mod_score)
{
    int nreplicates = g_jla_replicates;
    verify_prerequisites_or_exit();
    killed_pids = Memory_Calloc(g_jla_child_processes, sizeof(int));
    if (killed_pids == NULL) {
        sf(ps, "simulate_jla_p_value: out of memory\n");
        pr();
    }
    // Note: +1 to allow 1-based indexing via replicate numbers
    replicate_successful = Memory_Calloc(nreplicates + 1, sizeof(int));
    if (!replicate_successful) {
        sf(ps, "Failed to allocate memory for replicate_successful\n");
        pr();
        exit(EXIT_FAILURE);
    }
    // Note that MAX_CONSECUTIVE_FAILURES (= N) * nreplicates seeds
    // can never be exhausted due to failed replicates.  Either the
    // simulation finishes with nreplicates successful replicates or
    // it is stopped short by N consecutive failed replicates.  Since
    // if we were to exhaust the seeds, it would mean that at most
    // ((N-1)/N) * (N * nreplicates) = (N-1) * nreplicates replicates
    // failed.  But that would imply that at least (1/N) * (N *
    // nreplicates) = nreplicates replicates were successful, i.e.,
    // the simulation would have been successful and the seeds would
    // not have been exhausted.  This is a contradiction and thus N *
    // nreplicates will always be enough.
    seeds = Memory_Malloc(MAX_CONSECUTIVE_FAILURES * nreplicates * sizeof(int16_t));
    if (seeds == NULL) {
        sf(ps, "simulate_jla_p_value: out of memory\n");
        pr();
    }
    seed_numbers = Memory_Malloc(MAX_CONSECUTIVE_FAILURES * nreplicates * sizeof(int));
    if (seed_numbers == NULL) {
        sf(ps, "simulate_jla_p_value: out of memory\n");
        pr();
    }
    // seeds[replicate_seed_indexes[i]] is the next seed for replicate
    // i.  Whenever a seed was used replicate_seed_index[i] must be
    // incremented.
    // Note: +1 to allow 1-based indexing via replicate numbers
    replicate_seed_indexes = Memory_Malloc((nreplicates + 1) * sizeof(int));
    if (!replicate_seed_indexes) {
        sf(ps, "Failed to allocate memory for replicate_seed_indexes\n");
        pr();
        exit(EXIT_FAILURE);
    }
    // Why will we never need more then MAX_CONSECUTIVE_FAILURES seeds
    // per replicate?
    // If there is a systematic problem with the ped file, then it
    // will be shared by all replicates, MAX_CONSECUTIVE_FAILURES will
    // be reached quickly, and the ghm p value simulation will abort.
    // Otherwise, since all replicates are equally likely to fail, the
    // chances of any one replicate failing MAX_CONSECUTIVE_FAILURES
    // times while the other replicates succeed are negligible.
    //
    // Access to replicate_seed_indexes uses the 1-based replicate
    // numbers.  However, the resulting indexes into the "seeds" array
    // are 0-based.  Thus, the loop starts at 1 but the assigned
    // indexes start at 0.
    for (int i = 1; i <= nreplicates; i++) {
        replicate_seed_indexes[i] = (i-1) * MAX_CONSECUTIVE_FAILURES;
    }
    char *resume_file;
    MAKE_STRING(resume_file, "%s/.ghm_resume", g_jla_directory);
    FILE *resume_fp = NULL;
    if (g_jla_resume) {
        resume_fp = fopen(resume_file, "rb");
        if (resume_fp == NULL) {
            sf(ps, "Failed to read file: %s\n", resume_file);
            pr();
            exit(EXIT_FAILURE);
        }
        // Remember which replicates were successful.  In resume_fp,
        // failed replicates are coded as the replicate number with
        // a leading minus sign.  Successful replicates are coded as
        // positive replicate numbers.  The number of times that a
        // replicate appears with a leading minus sign corresponds to
        // the number of seeds that failed for that replicate.
        int rep;  // 1-based replicate numbers from resume_fp
        while (fscanf(resume_fp, "%d", &rep) == 1) {
            if (rep < 0) {  // negative means "failed"
                --replicate_successful[-rep];  // count failed seeds
            } else if (rep > 0) {  // positive means "successful"
                replicate_successful[rep] = 1;
            }
        }
        fclose(resume_fp);
        int nsuccessful = 0;
        for (int i = 1; i <= nreplicates; i++) {
            nsuccessful += (replicate_successful[i] == 1);
        }
        if (nsuccessful == nreplicates) {
            sf(ps, "Nothing to --resume, all replicates completed.\n");
            pr();
            exit(EXIT_FAILURE);
        }
        resume_fp = fopen(resume_file, "ab");
        if (resume_fp == NULL) {
            sf(ps, "Failed to open file: %s\n", resume_file);
            pr();
            exit(EXIT_FAILURE);
        }
    } else {
        resume_fp = fopen(resume_file, "wb");
        if (resume_fp == NULL) {
            sf(ps, "Failed to create file: %s\n", resume_file);
            pr();
            exit(EXIT_FAILURE);
        }
    }
    init_seeds_for_replicates();
    const char *output_file = CommandReader_GetOutputFile(reader);
    FILE *results_fp = NULL;
    if (g_jla_resume) {
        results_fp = fopen(output_file, "ab");
        if (results_fp == NULL) {
            sf(ps, "Failed to open file: %s\n", output_file);
            pr();
            fclose(resume_fp);
            exit(EXIT_FAILURE);
        }
    } else {
        results_fp = fopen(output_file, "wb");
        if (results_fp == NULL) {
            sf(ps, "Failed to create file: %s\n", output_file);
            pr();
            fclose(resume_fp);
            exit(EXIT_FAILURE);
        }
    }
    run_all_replicates(
        nreplicates,
        g_jla_child_processes,
        g_jla_directory,
        results_fp,
        resume_fp,
        reader,
        simulator);
    fclose(results_fp);
    fclose(resume_fp);
    double p_value = compute_empirical_p_value(output_file, mod_score);
    if (p_value >= 0) {
        sf(ps, "\nEmpirical p value: %f\n", p_value);
        pr();
    }
}
