#include "IntVector.h"
#include "IO.h"
#include "LinkageMarkerFile.h"
#include "Marker.h"
#include "Memory.h"
#include "ParseObject.h"
#include "PedigreeFile.h"
#include "StringHashTable.h"
#include "Vector.h"
#include <string.h>
#include <stdlib.h>
// Legacy includes.
#define INC_LIB
#define INC_SHELL
#include "npl.h"

struct LinkageMarkerFileStruct {
    int number_of_markers;
    IntVector map_order;
    double disease_locus_allele_frequency[2];
    int number_of_liability_classes;
    int number_of_penetrances;
    double penetrances[4];
    Vector markers;
    int is_sex_specific;
    int is_ready;
    double *distances;
    double *distances_m;
    double *distances_f;
    int *marker_indices;
    StringHashTable map;
};

static int read_number_of_markers(LinkageMarkerFile, FILE *);
static void skip_line(FILE *);
static void read_map_order(LinkageMarkerFile, FILE *);
static void read_disease_locus(LinkageMarkerFile, FILE *);
static void read_number_of_liability_classes(LinkageMarkerFile, FILE *);
static void read_penetrances(LinkageMarkerFile, FILE *);
static void read_all_markers(LinkageMarkerFile, FILE *);
static Marker read_marker(FILE *, int marker_index);
static Marker read_number_of_alleles_and_maybe_marker_name(FILE *, int marker_index);
static const char *find_marker_name(int marker_index, ParseObject);
static const char *read_marker_name(ParseObject);
static double *read_allele_frequencies(FILE *, int number_of_alleles);
static void read_sex_averaged_or_sex_specific(LinkageMarkerFile, FILE *);
static void read_distances(LinkageMarkerFile);
static void read_distances_old(LinkageMarkerFile, FILE *);
static void map_marker_names_to_marker_indices(LinkageMarkerFile);
static int is_disease_locus_in_map_order(IntVector map_order);
static void sort_observed_alleles(LinkageMarkerFile, PedigreeFile);

LinkageMarkerFile LinkageMarkerFile_Create(const char *filename, PedigreeFile ped_file)
{
    LinkageMarkerFile linkage_file = Memory_Malloc(sizeof *linkage_file);
    linkage_file->is_ready = 0;
    FILE *fp = fopen(filename, "rb");
    if (fp == NULL) {
        Memory_Free(linkage_file);
        return NULL;
    }
    if (!read_number_of_markers(linkage_file, fp)) {
        Memory_Free(linkage_file);
        return NULL;
    }
    skip_line(fp);  // skip mutation line
    read_map_order(linkage_file, fp);
    read_disease_locus(linkage_file, fp);
    read_number_of_liability_classes(linkage_file, fp);
    if (linkage_file->number_of_liability_classes != 1) {
        sf(ps, "Number of liability classes must be 1 for JLA analysis\n");
        error(ps);
    }
    read_penetrances(linkage_file, fp);
    read_all_markers(linkage_file, fp);
    read_sex_averaged_or_sex_specific(linkage_file, fp);
    read_distances(linkage_file);
    fclose(fp);
    map_marker_names_to_marker_indices(linkage_file);
    sort_observed_alleles(linkage_file, ped_file);
    linkage_file->is_ready = 1;
    return linkage_file;
}

LinkageMarkerFile LinkageMarkerFile_Create_old(const char *filename)
{
    LinkageMarkerFile linkage_file = Memory_Malloc(sizeof *linkage_file);
    FILE *fp = fopen(filename, "rb");
    if (fp == NULL) {
        Memory_Free(linkage_file);
        return NULL;
    }
    if (!read_number_of_markers(linkage_file, fp)) {
        Memory_Free(linkage_file);
        return NULL;
    }
    skip_line(fp);  // skip mutation line
    read_map_order(linkage_file, fp);
    read_disease_locus(linkage_file, fp);
    read_number_of_liability_classes(linkage_file, fp);
    if (linkage_file->number_of_liability_classes != 1) {
        sf(ps, "Number of liability classes must be 1 for JLA analysis\n");
        error(ps);
    }
    read_penetrances(linkage_file, fp);
    read_all_markers(linkage_file, fp);
    read_sex_averaged_or_sex_specific(linkage_file, fp);
    read_distances_old(linkage_file, fp);
    fclose(fp);
    map_marker_names_to_marker_indices(linkage_file);
    linkage_file->is_ready = 1;
    return linkage_file;
}

int LinkageMarkerFile_GetNumberOfMarkers(LinkageMarkerFile linkage_file)
{
    return linkage_file->number_of_markers;
}

IntVector LinkageMarkerFile_GetMapOrder(LinkageMarkerFile linkage_file)
{
    return linkage_file->map_order;
}

static int read_number_of_markers(LinkageMarkerFile linkage_file, FILE *fp)
{
    char *line = IO_GetLine(fp);
    if (line == NULL) {
        return 0;
    }
    ParseObject parser = ParseObject_Create(line);
    long number_of_markers_including_disease_locus;

    if (!ParseObject_NextInteger(parser, &number_of_markers_including_disease_locus)) {
        return 0;
    }
    ParseObject_Destroy(parser);
    free(line);

    linkage_file->number_of_markers = number_of_markers_including_disease_locus - 1;
    return 1;
}

static void read_map_order(LinkageMarkerFile linkage_file, FILE *fp)
{
    IntVector map_order = IntVector_Create(0);
    char *line = IO_GetLine(fp);
    ParseObject parser = ParseObject_Create(line);
    long k;

    while (ParseObject_NextInteger(parser, &k))
        IntVector_PushBack(map_order, k);
    ParseObject_Destroy(parser);
    free(line);

    int len, first;
    if (is_disease_locus_in_map_order(map_order)) {
        len = linkage_file->number_of_markers + 1;
        first = 1;
    } else {
        len = linkage_file->number_of_markers;
        first = 2;
    }
    int i = -1;
    if (IntVector_GetSize(map_order) == len) {
        for (i = first; i < len; i++) {
            if (i != IntVector_Get(map_order, i - first)) {
                break;
            }
        }
    }
    if (i != len) {  // for-loop did not finish
        sf(ps, "Please make sure that the map order is [1] 2 3 4 5 ... "
               "where 1 is optional and reserved for the disease locus.");
        error(ps);
    }
    linkage_file->map_order = map_order;
}

static void skip_line(FILE *fp)
{
    free(IO_GetLine(fp));
}

void LinkageMarkerFile_Destroy(LinkageMarkerFile linkage_file)
{
    if (linkage_file == NULL) {
        return;
    }
    if (linkage_file->is_ready) {
        linkage_file->number_of_markers = -1;
        IntVector_Destroy(linkage_file->map_order);
        linkage_file->map_order = NULL;
        Vector_Destroy(linkage_file->markers);
        linkage_file->markers = NULL;
        Memory_Free(linkage_file->distances);
        linkage_file->distances = NULL;
        Memory_Free(linkage_file->distances_m);
        linkage_file->distances_m = NULL;
        Memory_Free(linkage_file->distances_f);
        linkage_file->distances_f = NULL;
        Memory_Free(linkage_file->marker_indices);
        linkage_file->marker_indices = NULL;
        StringHashTable_Destroy(linkage_file->map);
        linkage_file->map = NULL;
    }
    linkage_file->is_ready = 0;
    Memory_Free(linkage_file);
}

double LinkageMarkerFile_GetDiseaseLocusAlleleFrequency1(LinkageMarkerFile linkage_file)
{
    return linkage_file->disease_locus_allele_frequency[0];
}

double LinkageMarkerFile_GetDiseaseLocusAlleleFrequency2(LinkageMarkerFile linkage_file)
{
    return linkage_file->disease_locus_allele_frequency[1];
}

static void read_disease_locus(LinkageMarkerFile linkage_file, FILE *fp)
{
    skip_line(fp);  // skip disease locus identifier and number of alleles
    char *line = IO_GetLine(fp);
    ParseObject parser = ParseObject_Create(line);

    for (int freq = 0; freq < 2; freq++) {
        if (!ParseObject_NextDouble(parser, &linkage_file->disease_locus_allele_frequency[freq])) {
            sf(ps, "Failed to read disease locus frequencies in line: %s\n", line);
            pr();
            exit(1);
        }
    }
    ParseObject_Destroy(parser);
    free(line);
}

int LinkageMarkerFile_GetNumberOfLiabilityClasses(LinkageMarkerFile linkage_file)
{
    return linkage_file->number_of_liability_classes;
}

static void read_number_of_liability_classes(LinkageMarkerFile linkage_file, FILE *fp)
{
    char *line = IO_GetLine(fp);
    ParseObject parser = ParseObject_Create(line);
    long k;

    if (!ParseObject_NextInteger(parser, &k)) {
        sf(ps, "Failed to parse number of liability classes in line: %s\n", line);
        pr();
        exit(1);
    }
    ParseObject_Destroy(parser);
    free(line);

    linkage_file->number_of_liability_classes = k;
}

int LinkageMarkerFile_GetNumberOfPenetrances(LinkageMarkerFile linkage_file)
{
    return linkage_file->number_of_penetrances;
}

static void read_penetrances(LinkageMarkerFile linkage_file, FILE *fp)
{
    char *line = IO_GetLine(fp);
    ParseObject parser = ParseObject_Create(line);
    const int max_number_of_penetrances = 4;

    int i = 0;
    while (i < max_number_of_penetrances) {
        double d;
        if (!ParseObject_NextDouble(parser, &d))
            break;
        else
            linkage_file->penetrances[i++] = d;
    }
    linkage_file->number_of_penetrances = i;

    ParseObject_Destroy(parser);
    free(line);
}

double LinkageMarkerFile_GetPenetrance(LinkageMarkerFile linkage_file, int which)
{
    return linkage_file->penetrances[which];
}

int LinkageMarkerFile_GetNumberOfAlleles(LinkageMarkerFile linkage_file, int marker_index)
{
    return Marker_GetNumberOfAlleles(Vector_Get(linkage_file->markers, marker_index));
}

static void read_all_markers(LinkageMarkerFile linkage_file, FILE *fp)
{
    int number_of_markers = linkage_file->number_of_markers;

    linkage_file->markers = Vector_CreateEmpty();

    int marker_index = 0;
    while (marker_index < number_of_markers) {
        Marker marker = read_marker(fp, marker_index);
        Vector_PushBack(linkage_file->markers, marker);
        ++marker_index;
    }
}

static Marker read_marker(FILE *fp, int marker_index)
{
    Marker marker = read_number_of_alleles_and_maybe_marker_name(fp, marker_index);
    int number_of_alleles = Marker_GetNumberOfAlleles(marker);
    double *allele_frequencies = read_allele_frequencies(fp, number_of_alleles);
    Marker_SetAlleleFrequencies(marker, number_of_alleles, allele_frequencies);

    return marker;
}

static Marker read_number_of_alleles_and_maybe_marker_name(FILE *fp, int marker_index)
{
    char *line = IO_GetLine(fp);
    ParseObject parser = ParseObject_Create(line);

    long marker_code = -1;
    if (!ParseObject_NextInteger(parser, &marker_code)) {
        sf(ps, "Failed to parse marker code in line: %s\n", line);
        pr();
        exit(1);
    }

    long number_of_alleles = -1;
    if (!ParseObject_NextInteger(parser, &number_of_alleles)) {
        sf(ps, "Failed to parse number of alleles in line: %s\n", line);
        pr();
        exit(1);
    }
    const char *marker_name = find_marker_name(marker_index, parser);
    Marker marker = Marker_Create(marker_name, marker_index);
    Marker_SetNumberOfAlleles(marker, number_of_alleles);

    ParseObject_Destroy(parser);
    free(line);

    return marker;
}

static const char *find_marker_name(int marker_index, ParseObject parser)
{
    const char *marker_name = read_marker_name(parser);

    if (marker_name)
        return marker_name;

    // Make sure that marker_index + 1 is in { 1, 2, ..., 9999 }.
    if (! (marker_index + 1 >= 1 && marker_index + 1 <= 9999)) {
        fprintf(stderr, "LinkageMarkerFile: marker_index + 1 is not in { 1, 2, ..., 9999 }.");
        exit(1);
    }
    int number_of_bytes = 8;  // LOC + up to 4 digits + nul byte
    char *default_marker_name = Memory_Malloc(number_of_bytes * sizeof(char));
    snprintf(default_marker_name, number_of_bytes, "loc%d", marker_index + 1);
    return default_marker_name;
}

static const char *read_marker_name(ParseObject parser)
{
    char *word = NULL;

    if (!ParseObject_NextWord(parser, &word))
        return NULL;

    if (strcmp("#", word) != 0) {
        Memory_Free(word);
        return NULL;
    }

    if (!ParseObject_NextWord(parser, &word))
        return NULL;

    return word;
}

const char *LinkageMarkerFile_GetMarkerName(LinkageMarkerFile linkage_file, int marker_index)
{
    return Marker_GetName(Vector_Get(linkage_file->markers, marker_index));
}

static double *read_allele_frequencies(FILE *fp, int number_of_alleles)
{
    char *line = IO_GetLine(fp);
    ParseObject parser = ParseObject_Create(line);

    double *frequencies = (double *) Memory_Malloc(number_of_alleles * sizeof(double));
    for (int i = 0; i < number_of_alleles; i++) {
        double f;
        if (!ParseObject_NextDouble(parser, &f)) {
            sf(ps, "Failed to parse allele frequencies in line: %s\n", line);
            pr();
            exit(1);
        }
        frequencies[i] = f;
    }

    ParseObject_Destroy(parser);
    free(line);

    return frequencies;
}

double LinkageMarkerFile_GetAlleleFrequency(LinkageMarkerFile linkage_file, int marker_index, int allele)
{
    return Marker_GetAlleleFrequency(Vector_Get(linkage_file->markers, marker_index), allele);
}

int LinkageMarkerFile_IsSexSpecific(LinkageMarkerFile linkage_file)
{
    return linkage_file->is_sex_specific;
}

static void read_sex_averaged_or_sex_specific(LinkageMarkerFile linkage_file, FILE *fp)
{
    char *line = IO_GetLine(fp);
    ParseObject parser = ParseObject_Create(line);

    long sex_code = -1;
    if (!ParseObject_NextInteger(parser, &sex_code)) {
        sf(ps, "Failed to parse sex code in line: %s\n", line);
        pr();
        exit(1);
    }

    linkage_file->is_sex_specific = sex_code == 2;

    ParseObject_Destroy(parser);
    free(line);
}

double LinkageMarkerFile_GetDistanceBetween(LinkageMarkerFile linkage_file, int marker_1, int marker_2)
{
    int start, end;

    if (marker_1 < marker_2) {
        start = marker_1;
        end = marker_2;
    } else {
        start = marker_2;
        end = marker_1;
    }
    double d = 0.0;
    for (int i = start; i < end; i++)
        d += linkage_file->distances[i];

    return d;
}

double LinkageMarkerFile_GetMaleDistanceBetween(LinkageMarkerFile linkage_file, int marker_1, int marker_2)
{
    int start, end;

    if (marker_1 < marker_2) {
        start = marker_1;
        end = marker_2;
    } else {
        start = marker_2;
        end = marker_1;
    }
    double d = 0.0;
    for (int i = start; i < end; i++)
        d += linkage_file->distances_m[i];

    return d;
}

double LinkageMarkerFile_GetFemaleDistanceBetween(LinkageMarkerFile linkage_file, int marker_1, int marker_2)
{
    int start, end;

    if (marker_1 < marker_2) {
        start = marker_1;
        end = marker_2;
    } else {
        start = marker_2;
        end = marker_1;
    }
    double d = 0.0;
    for (int i = start; i < end; i++)
        d += linkage_file->distances_f[i];

    return d;
}

static void read_distances(LinkageMarkerFile linkage_file)
{
    // always read from global struct map_distance
    int number_of_distances = linkage_file->number_of_markers - 1;
    double *distances = (double *) Memory_Malloc(number_of_distances * sizeof(double));
    double *distances_m = (double *) Memory_Malloc(number_of_distances * sizeof(double));
    double *distances_f = (double *) Memory_Malloc(number_of_distances * sizeof(double));
    for (int i = 0; i < number_of_distances; i++) {
        distances[i] = map_distance[i];
        distances_m[i] = map_distance_m[i];
        distances_f[i] = map_distance_f[i];
    }
    linkage_file->distances = distances;
    linkage_file->distances_m = distances_m;
    linkage_file->distances_f = distances_f;
}

static void ensure_distances_in_cM(LinkageMarkerFile linkage_file)
{
    double *dist_m = linkage_file->distances_m;
    double *dist_f = linkage_file->distances_f;
    double *dist = linkage_file->distances;
    int number_of_distances = linkage_file->number_of_markers - 1;
    int is_cM = 0;
    for (int i = 0; i < number_of_distances; i++) {
        if (dist_m[i] >= .5 || dist_f[i] >= .5) {
            is_cM = 1;
            break;
        }
    }
    if (!is_cM) {
        for (int i = 0; i < number_of_distances; i++) {
            dist_m[i] = rec_to_dist(dist_m[i]);
            dist_f[i] = rec_to_dist(dist_f[i]);
            dist[i] = (dist_m[i] + dist_f[i]) / 2;
        }
    }
}

static void read_distances_old(LinkageMarkerFile linkage_file, FILE *fp)
{
    // read from linkage marker file
    char *line = IO_GetLine(fp);
    ParseObject parser = ParseObject_Create(line);

    if (is_disease_locus_in_map_order(linkage_file->map_order)) {
        // skip first distance
        double f;
        if (!ParseObject_NextDouble(parser, &f)) {
            sf(ps, "Failed to parse disease locus distance in line: %s\n", line);
            pr();
            exit(1);
        }
    }

    int number_of_distances = linkage_file->number_of_markers - 1;
    double *distances = (double *) Memory_Malloc(number_of_distances * sizeof(double));
    double *distances_m = (double *) Memory_Malloc(number_of_distances * sizeof(double));
    double *distances_f = (double *) Memory_Malloc(number_of_distances * sizeof(double));

    if (LinkageMarkerFile_IsSexSpecific(linkage_file)) {
        for (int i = 0; i < number_of_distances; i++) {
            double f;
            if (!ParseObject_NextDouble(parser, &f)) {
                sf(ps, "Failed to parse male marker locus distance in line: %s\n", line);
                pr();
                exit(1);
            }
            distances_m[i] = f;
        }
        linkage_file->distances_m = distances_m;

        ParseObject_Destroy(parser);
        free(line);
        // next line with female distances
        char *line = IO_GetLine(fp);
        ParseObject parser = ParseObject_Create(line);

        if (is_disease_locus_in_map_order(linkage_file->map_order)) {
            // skip first distance
            double f;
            if (!ParseObject_NextDouble(parser, &f)) {
                sf(ps, "Failed to parse female disease locus distance in line: %s\n", line);
                pr();
                exit(1);
            }
        }
        for (int i = 0; i < number_of_distances; i++) {
            double f;
            if (!ParseObject_NextDouble(parser, &f)) {
                sf(ps, "Failed to parse female marker locus distance in line: %s\n", line);
                pr();
                exit(1);
            }
            distances_f[i] = f;
        }
        linkage_file->distances_f = distances_f;
        for (int i = 0; i < number_of_distances; i++)
            distances[i] = (distances_m[i] + distances_f[i]) / 2;
        linkage_file->distances = distances;

        ParseObject_Destroy(parser);
        free(line);
    } else {
        for (int i = 0; i < number_of_distances; i++) {
            double f;
            if (!ParseObject_NextDouble(parser, &f)) {
                sf(ps, "Failed to parse marker locus distance in line: %s\n", line);
                pr();
                exit(1);
            }
            distances[i] = f;
            distances_m[i] = f;
            distances_f[i] = f;
        }
        linkage_file->distances = distances;
        linkage_file->distances_m = distances_m;
        linkage_file->distances_f = distances_f;

        ParseObject_Destroy(parser);
        free(line);
    }
    ensure_distances_in_cM(linkage_file);
}

static void map_marker_names_to_marker_indices(LinkageMarkerFile linkage_file)
{
    int number_of_markers = Vector_GetSize(linkage_file->markers);
    int size = (number_of_markers == 1) ? 1 : number_of_markers / 2;
    StringHashTable map = StringHashTable_Create(size);
    int *marker_indices = Memory_Malloc(sizeof(int) * number_of_markers);

    for (int i = 0; i < number_of_markers; i++) {
        Marker marker = Vector_Get(linkage_file->markers, i);
        marker_indices[i] = Marker_GetIndex(marker);
        StringHashTable_Put(map, Marker_GetName(marker), &marker_indices[i]);
    }

    linkage_file->marker_indices = marker_indices;
    linkage_file->map = map;
}

static int is_disease_locus_in_map_order(IntVector map_order)
{
    return IntVector_Get(map_order, 0) == 1;
}

int LinkageMarkerFile_GetNumberOfPhenotypes(void)
{
    return 0;
}

int LinkageMarkerFile_GetNumberOfCovariates(void)
{
    return 0;
}

int LinkageMarkerFile_GetMarkerIndex(LinkageMarkerFile linkage_file, const char *marker_name)
{
    int *marker_index = NULL;
    int found = StringHashTable_Get(linkage_file->map, marker_name, (void **) &marker_index);

    return found ? *marker_index : -1;
}

int LinkageMarkerFile_IsReady(LinkageMarkerFile linkage_file)
{
    return linkage_file->is_ready;
}

Marker LinkageMarkerFile_GetMarkerByIndex(LinkageMarkerFile linkage_file, int marker_index)
{
    return Vector_Get(linkage_file->markers, marker_index);
}

Marker LinkageMarkerFile_GetMarkerByName(LinkageMarkerFile linkage_file, const char *marker_name)
{
    int marker_index = LinkageMarkerFile_GetMarkerIndex(linkage_file, marker_name);
    if (marker_index == -1) {
        return NULL;
    }
    return LinkageMarkerFile_GetMarkerByIndex(linkage_file, marker_index);
}

int LinkageMarkerFile_WithinMaxDistance(LinkageMarkerFile linkage_file, Vector markers, double maximum_marker_distance)
{
    Marker first_marker = Vector_Get(markers, 0);
    Marker last_marker = Vector_Get(markers, Vector_GetSize(markers) - 1);
    double distance_m = LinkageMarkerFile_GetMaleDistanceBetween(
            linkage_file, Marker_GetIndex(first_marker), Marker_GetIndex(last_marker));
    double distance_f = LinkageMarkerFile_GetFemaleDistanceBetween(
            linkage_file, Marker_GetIndex(first_marker), Marker_GetIndex(last_marker));
    if (distance_m < distance_f)
        return distance_f <= maximum_marker_distance;
    else
        return distance_m <= maximum_marker_distance;
}

void print_distances(LinkageMarkerFile linkage_file)
{
    double dist, dist_male, dist_female;
    const char *marker_name;
    for (int i = 0; i < linkage_file->number_of_markers; i++) {
        marker_name = LinkageMarkerFile_GetMarkerName(linkage_file, i);
        if (ind_sp) {
            dist_male = LinkageMarkerFile_GetMaleDistanceBetween(linkage_file, i, 0);
            dist_female = LinkageMarkerFile_GetFemaleDistanceBetween(linkage_file, i, 0);
            if (units == RECFRACS) {
                dist_male = dist_to_rec(dist_male);
                dist_female = dist_to_rec(dist_female);
            }
            sf(ps, "Position on male map of marker %s: %.2f\n", marker_name, dist_male);
            pr();
            sf(ps, "Position on female map of marker %s: %.2f\n", marker_name, dist_female);
            pr();
        } else {
            // Male and female distances are the same = averaged distances
            dist = LinkageMarkerFile_GetMaleDistanceBetween(linkage_file, i, 0);
            sf(ps, "Position on sex-averaged map of marker %s: %.2f\n", marker_name, dist);
            pr();
        }
    }
}

Vector LinkageMarkerFile_GetMarkers(LinkageMarkerFile linkage_file)
{
    return linkage_file->markers;
}

static void sort_observed_alleles(LinkageMarkerFile linkage_file, PedigreeFile ped_file)
{
    int number_of_markers = LinkageMarkerFile_GetNumberOfMarkers(linkage_file);
    IntSet *marker_alleles = Memory_Malloc(number_of_markers * sizeof(IntSet));
    for (int i = 0; i < number_of_markers; i++) {
        marker_alleles[i] = IntSet_Create(2);
    }
    for (int i = 0; i < PedigreeFile_GetNumberOfRecords(ped_file); i++) {
        PedigreeFileRecord record = PedigreeFile_GetRecordByIndex(ped_file, i);
        if (PedigreeFileRecord_GetNumberOfMarkers(record) != number_of_markers) {
            sf(ps, "Error: different number of markers in pedigree and marker file: %d vs %d\n",
                PedigreeFileRecord_GetNumberOfMarkers(record), number_of_markers);
            pr();
            exit(1);
        }
        for (int j = 0; j < number_of_markers; j++) {
            int a1 = atoi(PedigreeFileRecord_GetAllele1(record, j));
            int a2 = atoi(PedigreeFileRecord_GetAllele2(record, j));
            if (a1 != 0 && a2 != 0) {
                IntSet_Put(marker_alleles[j], a1);
                IntSet_Put(marker_alleles[j], a2);
            }
        }
    }
    for (int i = 0; i < number_of_markers; i++) {
        Marker marker = LinkageMarkerFile_GetMarkerByIndex(linkage_file, i);
        int err = Marker_SetAlleleCodes(marker, IntSet_Sort(marker_alleles[i]), IntSet_GetSize(marker_alleles[i]));
        if (err == 1) {
            sf(ps, "Error: marker %s has more alleles (%d) than expected (%d)\n", Marker_GetName(marker), IntSet_GetSize(marker_alleles[i]), Marker_GetNumberOfAlleles(marker));
            pr();
            exit(1);
        } else if (err == 2) {
            sf(ps, "Error: marker %s has allele code > number of"
                   " alleles (%d), allele codes should be numbered"
                   " 1, 2, ..., %d\n",
                Marker_GetName(marker),
                Marker_GetNumberOfAlleles(marker),
                Marker_GetNumberOfAlleles(marker));
            pr();
            exit(1);
        }
    }
}

void LinkageMarkerFile_WriteWithNewLinkageFreqs(LinkageMarkerFile linkage_file, UserLinkageMarkerFile user_linkage_file, Vector linkage_markers, const char *filename)
{
    FILE *fp = fopen(filename, "wb");
    if (!fp) {
        sf(ps, "Failed to open file: \"%s\"\n", filename);
        pr();
        exit(1);
    }
    int nmarkers = linkage_file->number_of_markers; // always without disease locus
    fprintf(fp, "%d 0 0 5\n", nmarkers + 1);
    fprintf(fp, "0 0.0 0.0 0\n");
    if (is_disease_locus_in_map_order(linkage_file->map_order)) {
        for (int i = 0; i < nmarkers + 1; i++) {
            if (i > 0) {
                fprintf(fp, " ");
            }
            fprintf(fp, "%d", i + 1);
        }
    } else {
        for (int i = 1; i < nmarkers + 1; i++) {
            if (i > 1) {
                fprintf(fp, " ");
            }
            fprintf(fp, "%d", i + 1);
        }
    }
    fprintf(fp, "\n");
    fprintf(fp, "1 2\n");
    fprintf(fp, "%f %f\n",
        linkage_file->disease_locus_allele_frequency[0],
        linkage_file->disease_locus_allele_frequency[1]);
    fprintf(fp, "1\n");  // must be 1 for JLA analysis
    if (linkage_file->number_of_penetrances == 4) {
        fprintf(fp, "%f %f %f %f\n",
                linkage_file->penetrances[0],
                linkage_file->penetrances[1],
                linkage_file->penetrances[2],
                linkage_file->penetrances[3]);
    } else if (linkage_file->number_of_penetrances == 3) {
        fprintf(fp, "%f %f %f\n",
        linkage_file->penetrances[0],
        linkage_file->penetrances[1],
        linkage_file->penetrances[2]);
    }
    int linkage_marker_index = 0;
    for (int i = 0; i < nmarkers; i++) {
        Marker marker = Vector_Get(linkage_file->markers, i);
        fprintf(fp, "3 %d", Marker_GetNumberOfAlleles(marker));
        fprintf(fp, " # %s\n", Marker_GetName(marker));
        if (UserLinkageMarkerFile_Contains(user_linkage_file, Marker_GetName(marker))) {
            for (int j = 0; j < Marker_GetNumberOfAlleles(marker); j++) {
                if (j > 0) {
                    fprintf(fp, " ");
                }
                Marker linkage_marker = Vector_Get(linkage_markers, linkage_marker_index);
                fprintf(fp, "%f", Marker_GetAlleleFrequency(linkage_marker, j));
            }
            linkage_marker_index++;
        } else {
            for (int j = 0; j < Marker_GetNumberOfAlleles(marker); j++) {
                if (j > 0) {
                    fprintf(fp, " ");
                }
                fprintf(fp, "%f", Marker_GetAlleleFrequency(marker, j));
            }
        }
        fprintf(fp, "\n");
    }
    fprintf(fp, "%d 0\n", linkage_file->is_sex_specific ? 2 : 0);
    if(linkage_file->is_sex_specific) {
        if (is_disease_locus_in_map_order(linkage_file->map_order)) {
            fprintf(fp, "-9");
            for (int i = 0; i < nmarkers - 1; i++) {
                fprintf(fp, " %g", dist_to_rec(LinkageMarkerFile_GetMaleDistanceBetween(linkage_file, i, i + 1)));
            }
            fprintf(fp, "\n");
            fprintf(fp, "-9");
            for (int i = 0; i < nmarkers - 1; i++) {
                fprintf(fp, " %g", dist_to_rec(LinkageMarkerFile_GetFemaleDistanceBetween(linkage_file, i, i + 1)));
            }
        }
        else {
            for (int i = 0; i < nmarkers - 1; i++) {
                if (i < nmarkers - 2)
                    fprintf(fp, "%g ", dist_to_rec(LinkageMarkerFile_GetMaleDistanceBetween(linkage_file, i, i + 1)));
                else
                    fprintf(fp, "%g", dist_to_rec(LinkageMarkerFile_GetMaleDistanceBetween(linkage_file, i, i + 1)));

            }
            fprintf(fp, "\n");
            for (int i = 0; i < nmarkers - 1; i++) {
                if (i < nmarkers - 2)
                    fprintf(fp, "%g ", dist_to_rec(LinkageMarkerFile_GetFemaleDistanceBetween(linkage_file, i, i + 1)));
                else
                    fprintf(fp, "%g", dist_to_rec(LinkageMarkerFile_GetFemaleDistanceBetween(linkage_file, i, i + 1)));
            }
        }
    } else {
        if (is_disease_locus_in_map_order(linkage_file->map_order)) {
            fprintf(fp, "-9");
            for (int i = 0; i < nmarkers - 1; i++) {
                fprintf(fp, " %g", dist_to_rec(LinkageMarkerFile_GetDistanceBetween(linkage_file, i, i + 1)));
            }
        } else {
            for (int i = 0; i < nmarkers - 1; i++) {
                if (i < nmarkers - 2 )
                    fprintf(fp, "%g ", dist_to_rec(LinkageMarkerFile_GetDistanceBetween(linkage_file, i, i + 1)));
                else
                    fprintf(fp, "%g", dist_to_rec(LinkageMarkerFile_GetDistanceBetween(linkage_file, i, i + 1)));
            }
        }
    }
    fprintf(fp, "\n");
    fprintf(fp, "1 0.01000 0.50000\n");
    fclose(fp);
}

int LinkageMarkerFile_IsCompatibleWith(LinkageMarkerFile linkage_file, UserLinkageMarkerFile user_linkage_file)
{
    Vector marker_names = UserLinkageMarkerFile_GetMarkerNames(user_linkage_file);
    for (int i = 0; i < Vector_GetSize(marker_names); i++) {
        const char *marker_name = Vector_Get(marker_names, i);
        if (!LinkageMarkerFile_GetMarkerByName(linkage_file, marker_name)) {
            return 0;
        }
    }
    return 1;
}

int LinkageMarkerFile_IsCompatibleWithTestMarkers(LinkageMarkerFile linkage_file, Vector test_marker_names)
{
    for (int i = 0; i < Vector_GetSize(test_marker_names); i++) {
        const char *marker_name = Vector_Get(test_marker_names, i);
        if (!LinkageMarkerFile_GetMarkerByName(linkage_file, marker_name)) {
            return 0;
        }
    }
    return 1;
}
