#include "Marker.h"
#include "Memory.h"
#include <stdio.h>
// Legacy includes.
#define INC_LIB
#define INC_SHELL
#include "npl.h"

struct MarkerStruct {
    const char* name;
    Vector allele_frequencies;
    int index;
    int number_of_alleles;
    int *allele_codes;
    int number_of_allele_codes;
    // only for biallelic markers
    double em_freqs[2];
};

Marker Marker_Create(const char *marker_name, int marker_index)
{
    Marker marker = Memory_Malloc(sizeof(*marker));
    marker->name = marker_name;
    marker->index = marker_index;
    marker->number_of_alleles = -1;
    marker->allele_codes = NULL;
    marker->em_freqs[0] = -1;
    marker->em_freqs[1] = -1;
    return marker;
}

void Marker_Destroy(Marker marker)
{
    if (marker)
        Memory_Free(marker);
}

const char *Marker_GetName(Marker marker)
{
    return marker->name;
}

int Marker_GetIndex(Marker marker)
{
    return marker->index;
}

void Marker_SetNumberOfAlleles(Marker marker, int number_of_alleles)
{
    marker->number_of_alleles = number_of_alleles;
}

void Marker_SetAlleleFrequencies(Marker marker, int number_of_alleles, double *allele_frequencies)
{
    marker->number_of_alleles = number_of_alleles;
    marker->allele_frequencies = Vector_FromArray(allele_frequencies, number_of_alleles, sizeof(double));
}

int Marker_GetNumberOfAlleles(Marker marker)
{
    return marker->number_of_alleles;
}

double Marker_GetAlleleFrequency(Marker marker, int allele)
{
    return * (double *) Vector_Get(marker->allele_frequencies, allele);
}

Vector Marker_GetAlleleFrequencies(Marker marker)
{
    return marker->allele_frequencies;
}

void Marker_Show(Marker marker)
{
    fprintf(stderr, "marker: %s at index %d with %d alleles (freqs:",
            marker->name, marker->index, marker->number_of_alleles);
    for (int i = 0; i < marker->number_of_alleles; i++)
        fprintf(stderr, " %f", * (double *) Vector_Get(marker->allele_frequencies, i));
    fprintf(stderr, ")\n");
}

double Marker_GetSumOfAlleleFrequencies(Marker marker)
{
    double allele_sum = 0.0;
    for (int i = 0; i < Marker_GetNumberOfAlleles(marker); i++) {
        allele_sum += Marker_GetAlleleFrequency(marker, i);
    }

    return allele_sum;
}

double *Marker_GetEMAlleleFrequencies(Marker marker)
{
    return marker->em_freqs[0] == -1 ? NULL : marker->em_freqs;
}

void Marker_SetEMAlleleFrequencies(Marker marker, double *freqs)
{
    marker->em_freqs[0] = freqs[0];
    marker->em_freqs[1] = freqs[1];
}

// Map allele codes to allele frequencies.
int Marker_SetAlleleCodes(Marker marker, int *allele_codes, int number_of_allele_codes)
{
    if (number_of_allele_codes > marker->number_of_alleles) {
        return 1;  // something is wrong
    }
    for (int i = 0; i < number_of_allele_codes; i++) {
        if (allele_codes[i] > marker->number_of_alleles) {
            return 2;
        }
    }
    if (number_of_allele_codes == marker->number_of_alleles) {
        marker->allele_codes = allele_codes;
        marker->number_of_allele_codes = number_of_allele_codes;
    } else {  // # observed alleles < # actual alleles
        // We have to replace actual alleles that were not observed by
        // a dummy value (-1) so that allele codes are aligned with
        // allele frequencies.
        int *alleles = Memory_Malloc(marker->number_of_alleles * sizeof(int));
        for (int i = 0; i < marker->number_of_alleles; i++) {
            alleles[i] = -1;
        }
        for (int i = 0; i < number_of_allele_codes; i++) {
            alleles[allele_codes[i]-1] = allele_codes[i];
        }
        Memory_Free(allele_codes);  // not needed
        marker->allele_codes = alleles;
        marker->number_of_allele_codes = number_of_allele_codes;
    }
    return 0;
}

int Marker_GetNumberOfAlleleCodes(Marker marker)
{
    return marker->number_of_allele_codes;
}

int Marker_ConvertTo01(Marker marker, int allele_code)
{
    (void) marker;
    int result;
    if (allele_code == 0) {
        result = -1;
    } else if (allele_code == 1) {
        result = 0;
    } else if (allele_code == 2) {
        result = 1;
    } else {
        result = -999;
    }
    if (result == -999) {
        sf(ps, "Error: Allele code (%d) not found.\n", allele_code);
        pr();
        exit(1);
    }
    return result;
}

int Marker_ConvertFrom01(Marker marker, int allele01)
{
    int result;
    if (allele01 == -1) {
        result = 0;
    } else if (allele01 == 0) {
        result = 1;
    } else if (allele01 == 1) {
        result = 2;
    } else {
        result = -999;
    }
    if (result == -999) {
        sf(ps, "Error in allele conversion: Allele (%d) not found for marker %s.\n", allele01, Marker_GetName(marker));
        pr();
        exit(1);
    }
    return result;
}

int Marker_GetAlleleCodeByIndex(Marker marker, int allele_index)
{
    return marker->allele_codes[allele_index];
}
