#include "GridModScoreFunction.h"
#include "MathFunctions.h"
#include "Memory.h"
#include "ModelParamsFile.h"
#include "ModScorer.h"
#include "NloptModScoreFunction.h"
#include "Region.h"
#include "TestMarkerSetGenerator.h"

#include <math.h>
#include <stdio.h>

// Legacy includes.
#define INC_LIB
#define INC_SHELL
#include "npl.h"

static void print_matrix_cache(MatrixCache cache)
{
    for (int i = 0; i < cache->nrow; i++) {
        printf("[%d] ", i+1);
        for (int j = 0; j < cache->ncol; j++) {
            printf("%.2f ", cache->matrix[i][j]);
        }
        printf("\n");
    }
}


static MatrixCache create_or_reuse_matrix_cache(MatrixCache cache, int nrow, int ncol)
{
    if (cache != NULL && (cache->nrow < nrow || cache->ncol < ncol)) {
        Memory_Free(&cache->matrix[0][0]); // free cells
        Memory_Free(cache->matrix);
        Memory_Free(cache);
        cache = NULL;
    }
    if (cache == NULL) {
        cache = Memory_Malloc(sizeof(struct MatrixCacheStruct));
        double *cells = Memory_Malloc(nrow * ncol * sizeof(double));
        double **matrix = Memory_Malloc(nrow * sizeof(double *));
        for (int i = 0; i < nrow; i++) {
            matrix[i] = cells + i * ncol;
        }
        cache->matrix = matrix;
        cache->nrow = nrow;
        cache->ncol = ncol;
    }
    return cache;
}


static void initialize_cache(double **cache, int nrows, int ncols, DoubleVector hapfreqs) {
    for (int i = 0; i < nrows; i++) {
        double *p = cache[i];
        double *p_end = p + ncols;
        *p++ = 1.0;
        *p++ = DoubleVector_Get(hapfreqs, i);
        while (p < p_end)
            *p++ = -1.0;
    }
}


static double compute_product_of_penetrance_powers(unsigned char *exponents, DoubleVector penetrances, double **cache)
{
    double result = 1.0;
    double *penetrances10 = DoubleVector_GetInternalArray(penetrances);
    double f;
    int exponent;

    // Unroll loop for improved performance

#define FUNC(i) \
    exponent = exponents[i]; \
    if (exponent > 0) { \
        f = cache[i][exponent]; \
        if (f < 0.0) { \
            f = cache[i][exponent] = Math_IntegerPower(penetrances10[i], exponent); \
        } \
        if (f == 0) return 0.0; \
        result *= f; \
    }

    // Calls sorted by decreasing chance of being 0 (= early return).
    FUNC(0); FUNC(1); FUNC(2); FUNC(3); FUNC(7);
    FUNC(6); FUNC(5); FUNC(4); FUNC(8); FUNC(9);

#undef FUNC

    return result;
}


static DoubleVector create_uniform_weights(int n)
{
    static IntHashTable weights = NULL;
    if (weights == NULL) {
        weights = IntHashTable_Create(10);
        DoubleVector singleton_weight = DoubleVector_Create(1);
        DoubleVector_Set(singleton_weight, 0, 1.0);
        IntHashTable_Put(weights, 0, singleton_weight);
    }
    DoubleVector ws;
    if (!IntHashTable_Get(weights, n, (void **) &ws)) {
        ws = DoubleVector_Create(n);
        for (int i = 0; i < n; i++) {
            DoubleVector_Set(ws, i, 1.0 / n);
        }
        IntHashTable_Put(weights, n, ws);
    }
    return ws;
}


static DoubleVector find_weights_under_h1(Vector test_markers, InheritanceDistribution distro)
{
    static IntHashTable weights = NULL;
    if (weights == NULL) {
        weights = IntHashTable_Create(10);
        DoubleVector singleton_weight = DoubleVector_Create(1);
        DoubleVector_Set(singleton_weight, 0, 1.0);
        IntHashTable_Put(weights, 0, singleton_weight);
    }
    DoubleVector ws;
    if (distro == NULL) {  // singleton case
        IntHashTable_Get(weights, 0, (void **) &ws);
        return ws;
    }
    int number_of_inheritance_vectors = InheritanceDistribution_GetNumberOfInheritanceVectors(distro);
    if (!IntHashTable_Get(weights, number_of_inheritance_vectors, (void **) &ws)) {
        ws = DoubleVector_Create(number_of_inheritance_vectors);
        IntHashTable_Put(weights, number_of_inheritance_vectors, ws);
    }
    Marker marker = Vector_Get(test_markers, 0);
    int marker_index = Marker_GetIndex(marker);
    for (int i = 0; i < number_of_inheritance_vectors; i++) {
        DoubleVector_Set(ws, i, InheritanceDistribution_Get(distro, marker_index, i));
    }
    return ws;
}


static double compute_mod_score_for_family(
    Vector test_markers,
    AnnotatedFamily annotated_family,
    double disease_allele_frequency,
    DoubleVector haplotype_frequencies,
    DoubleVector penetrances,
    DoubleVector em_hapfreqs)
{
    static MatrixCache cache_pen = NULL;
    int npen = DoubleVector_GetSize(penetrances);
    int max_exponent_pen = Family_GetNumberOfMembers(AnnotatedFamily_GetFamily(annotated_family));
    cache_pen = create_or_reuse_matrix_cache(cache_pen, npen, max_exponent_pen + 1);
    initialize_cache(cache_pen->matrix, npen, max_exponent_pen + 1, penetrances);
    InheritanceDistribution distro = AnnotatedFamily_GetInheritanceDistribution(annotated_family);
    DoubleVector weights_under_h1 = find_weights_under_h1(test_markers, distro);
    DoubleVector weights_under_h0 = create_uniform_weights(InheritanceDistribution_GetNumberOfInheritanceVectors(distro));
    double score_under_h1 = ModScorer_ScoreUnderH1(
        penetrances,
        haplotype_frequencies,
        weights_under_h1,
        AnnotatedFamily_GetPenetranceMatrix(annotated_family),
        AnnotatedFamily_GetHaplotypeMatrix(annotated_family),
        Family_GetNumberOfFounders(AnnotatedFamily_GetFamily(annotated_family)),
        cache_pen->matrix
    );
    if (score_under_h1 <= 0.0) {
        return -1e4;
    }
    double score_under_h0 = ModScorer_ScoreUnderH0(
        penetrances,
        disease_allele_frequency,
        weights_under_h0,
        weights_under_h1,
        em_hapfreqs,
        AnnotatedFamily_GetPenetranceMatrix(annotated_family),
        AnnotatedFamily_GetHaplotypeMatrixWithoutDiseaseLocus(annotated_family),
        AnnotatedFamily_GetPQTerms(annotated_family),
        Family_GetNumberOfFounders(AnnotatedFamily_GetFamily(annotated_family)),
        cache_pen->matrix
    );
    if (score_under_h0 <= 0.0) {
        return -1e4;
    }
    return log(score_under_h1 / score_under_h0) / log(10);
}

ModScorer ModScorer_Create(
    Vector families,
    Vector inheritance_distros,
    int test_markers_per_set,
    double max_marker_distance,
    LinkageMarkerFile linkage_file,
    UserLinkageMarkerFile user_linkage_file,
    ModelParamsFile model_params_file,
    ModScoreList mod_score_list)
{
    ModScorer mod_scorer = Memory_Malloc(sizeof(*mod_scorer));
    mod_scorer->families            = families;
    mod_scorer->inheritance_distros = inheritance_distros;
    mod_scorer->test_markers_per_set = test_markers_per_set;
    mod_scorer->max_marker_distance = max_marker_distance;
    mod_scorer->linkage_file        = linkage_file;
    mod_scorer->user_linkage_file   = user_linkage_file;
    mod_scorer->mod_score_list      = mod_score_list;
    mod_scorer->model_params_file   = model_params_file;
    return mod_scorer;
}

static Vector create_annotated_families(Vector families, Vector test_markers, Vector inheritance_distros)
{
    Vector annotated_families = Vector_Create(Vector_GetSize(families));
    for (int i = 0; i < Vector_GetSize(families); i++) {
        Family family = Vector_Get(families, i);
        AnnotatedFamily annotated_family = AnnotatedFamily_CreateWithInheritanceDistro(family, test_markers, Vector_Get(inheritance_distros, i));
        Vector_PushBack(annotated_families, annotated_family);
    }
    return annotated_families;
}

static void destroy(Vector annotated_families)
{
    for (int i = 0; i < Vector_GetSize(annotated_families); i++) {
        AnnotatedFamily_Destroy(Vector_Get(annotated_families, i));
    }
    Vector_Destroy(annotated_families);
}

static void compute_mod_scores_for_region(ModScorer mod_scorer, Region region, int region_id, ModelParams user_model, Vector user_model_em_haplotypes, int model_id)
{
    TestMarkerSetGenerator generator = TestMarkerSetGenerator_Create(
        region, mod_scorer->test_markers_per_set, mod_scorer->max_marker_distance);
    int test_marker_sets = 0;
    Vector test_markers = NULL;
    DoubleVector em_haplotypes = NULL;
    ModScoreList mod_score_list = ModScorer_GetModScoreList(mod_scorer);
    while ((test_markers = TestMarkerSetGenerator_Next(generator, mod_scorer->linkage_file, mod_scorer->user_linkage_file)) != NULL) {
        sf(ps, "  set %d:", ++test_marker_sets);
        pr();
        for (int i = 0; i < Vector_GetSize(test_markers); i++) {
            Marker marker = Vector_Get(test_markers, i);
            sf(ps, " %s", Marker_GetName(marker));
            pr();
        }
        print("\n");
        Vector annotated_families = create_annotated_families(mod_scorer->families, test_markers, mod_scorer->inheritance_distros);
        if (user_model != NULL && model_id == 0)  {
            em_haplotypes = EM_Haplotype_H0(annotated_families, test_markers);
            if (em_haplotypes == NULL) {
                sf(ps, "Skipping current test marker set due to recombination between test markers.\n");
                pr();
                sf(ps, "One might try to decrease the maximum test marker distance.\n");
                pr();
                destroy(annotated_families);
                continue;
            }
            Vector_PushBack(user_model_em_haplotypes, em_haplotypes);
            ModScore *mod_score_grid = GridModScoreFunction(test_markers, annotated_families, em_haplotypes, mod_scorer->linkage_file, user_model, TestMarkerSetGenerator_GetNumberOfSets(generator), NULL);
            ModScore *mod_score_nlopt = NloptModScoreFunction(test_markers, annotated_families, em_haplotypes, mod_scorer->linkage_file, user_model, TestMarkerSetGenerator_GetNumberOfSets(generator), mod_score_grid);
            if (mod_score_nlopt != NULL) {
                if (mod_score_grid->mod_score > mod_score_nlopt->mod_score) {
                    sf(ps, "No better nlopt result, reporting globally optimized result first.\n");
                    pr();
                    ModScoreList_Add(mod_score_list, mod_score_grid);
                    ModScoreList_Add(mod_score_list, mod_score_nlopt);
                } else {
                    ModScoreList_Add(mod_score_list, mod_score_nlopt);
                    ModScoreList_Add(mod_score_list, mod_score_grid);
                }
            } else {
                sf(ps, "No valid nlopt result, reporting globally optimized result twice.\n");
                pr();
                ModScoreList_Add(mod_score_list, mod_score_grid);
                ModScoreList_Add(mod_score_list, mod_score_grid);
            }
        } else if (user_model != NULL && model_id > 0) {
            em_haplotypes = Vector_Get(user_model_em_haplotypes, test_marker_sets - 1);
            ModScore *mod_score_grid = GridModScoreFunction(test_markers, annotated_families, em_haplotypes, mod_scorer->linkage_file, user_model, TestMarkerSetGenerator_GetNumberOfSets(generator), NULL);
            ModScore *mod_score_nlopt = NloptModScoreFunction(test_markers, annotated_families, em_haplotypes, mod_scorer->linkage_file, user_model, TestMarkerSetGenerator_GetNumberOfSets(generator), mod_score_grid);
            if (mod_score_nlopt != NULL) {
                if (mod_score_grid->mod_score > mod_score_nlopt->mod_score) {
                    sf(ps, "No better nlopt result, reporting globally optimized result first.\n");
                    pr();
                    ModScoreList_Add(mod_score_list, mod_score_grid);
                    ModScoreList_Add(mod_score_list, mod_score_nlopt);
                } else {
                    ModScoreList_Add(mod_score_list, mod_score_nlopt);
                    ModScoreList_Add(mod_score_list, mod_score_grid);
                }
            } else {
                sf(ps, "No valid nlopt result, reporting globally optimized result twice.\n");
                pr();
                ModScoreList_Add(mod_score_list, mod_score_grid);
                ModScoreList_Add(mod_score_list, mod_score_grid);
            }
        } else {
            em_haplotypes = EM_Haplotype_H0(annotated_families, test_markers);
            if (em_haplotypes == NULL) {
                sf(ps, "Skipping current test marker set due to recombination between test markers.\n");
                pr();
                sf(ps, "One might try to decrease the maximum test marker distance.\n");
                pr();
                destroy(annotated_families);
                continue;
            }
            ModScore *mod_score_grid = GridModScoreFunction(test_markers, annotated_families, em_haplotypes, mod_scorer->linkage_file, user_model, TestMarkerSetGenerator_GetNumberOfSets(generator), NULL);
            ModScore *mod_score_nlopt = NloptModScoreFunction(test_markers, annotated_families, em_haplotypes, mod_scorer->linkage_file, user_model, TestMarkerSetGenerator_GetNumberOfSets(generator), mod_score_grid);
            if (mod_score_nlopt != NULL) {
                if (mod_score_grid->mod_score > mod_score_nlopt->mod_score) {
                    sf(ps, "No better nlopt result, reporting globally optimized result first.\n");
                    pr();
                    ModScoreList_Add(mod_score_list, mod_score_grid);
                    ModScoreList_Add(mod_score_list, mod_score_nlopt);
                } else {
                    ModScoreList_Add(mod_score_list, mod_score_nlopt);
                    ModScoreList_Add(mod_score_list, mod_score_grid);
                }
            } else {
                sf(ps, "No valid nlopt result, reporting globally optimized result twice.\n");
                pr();
                ModScoreList_Add(mod_score_list, mod_score_grid);
                ModScoreList_Add(mod_score_list, mod_score_grid);
            }
        }
        destroy(annotated_families);
    }
    if (test_marker_sets == 0) {
        sf(ps, "There are no test marker sets of size %d in region %d.\n", mod_scorer->test_markers_per_set, region_id + 1);
        pr();
        sf(ps, "There can be two reasons for this.  For every test "
               "marker set, at least one of the actual test marker "
               "distances is larger than the maximum test marker "
               "distance.  Or the region only contains linkage "
               "markers.\n");
        pr();
    }
}

void ModScorer_ComputeModScoreForTestMarkers(ModScorer mod_scorer, Vector test_marker_names)
{
    for (int i = 0; i < Vector_GetSize(mod_scorer->families); i++) {
        Family family = Vector_Get(mod_scorer->families, i);
        Family_ComputePenetrances(family);
        Family_ComputePQTerms(family);
    }
    Vector test_markers = Vector_Create(Vector_GetSize(test_marker_names));
    for (int i = 0; i < Vector_GetSize(test_marker_names); i++) {
        Vector_PushBack(test_markers, LinkageMarkerFile_GetMarkerByName(mod_scorer->linkage_file, Vector_Get(test_marker_names, i)));
    }
    if (!LinkageMarkerFile_WithinMaxDistance(mod_scorer->linkage_file, test_markers, mod_scorer->max_marker_distance)) {
        sf(ps, "The test marker set is not valid.  At least one of the "
               "actual test marker distances is larger than the maximum "
               "test marker distance.\n");
        pr();
        return;
    }
    if (UserLinkageMarkerFile_CountLinkageMarkers(mod_scorer->user_linkage_file, test_markers) > 0) {
        sf(ps, "The test marker set is not valid.  It contains a linkage marker.\n");
        pr();
        return;
    }
    Vector annotated_families = create_annotated_families(mod_scorer->families, test_markers, mod_scorer->inheritance_distros);
    DoubleVector em_haplotypes = EM_Haplotype_H0(annotated_families, test_markers);
    if (em_haplotypes == NULL) {
        sf(ps, "The test marker set is invalid due to a recombination between test markers\n"
               "or bad inheritance.\n");
        pr();
        sf(ps, "In the case of a recombination event, one might try to decrease the maximum\n"
               "test marker distance.\n");
        pr();
        destroy(annotated_families);
        return;
    }
    ModScoreList mod_score_list = ModScorer_GetModScoreList(mod_scorer);
    ModScore *mod_score_grid = GridModScoreFunction(test_markers, annotated_families, em_haplotypes, mod_scorer->linkage_file, NULL, 1, NULL);
    ModScore *mod_score_nlopt = NloptModScoreFunction(test_markers, annotated_families, em_haplotypes, mod_scorer->linkage_file, NULL, 1, mod_score_grid);
    if (mod_score_nlopt != NULL) {
        if (mod_score_grid->mod_score > mod_score_nlopt->mod_score) {
            sf(ps, "No better nlopt result, reporting globally optimized result first.\n");
            pr();
            ModScoreList_Add(mod_score_list, mod_score_grid);
            ModScoreList_Add(mod_score_list, mod_score_nlopt);
        } else {
            ModScoreList_Add(mod_score_list, mod_score_nlopt);
            ModScoreList_Add(mod_score_list, mod_score_grid);
        }
    } else {
        sf(ps, "No valid nlopt result, reporting globally optimized result twice.\n");
        pr();
        ModScoreList_Add(mod_score_list, mod_score_grid);
        ModScoreList_Add(mod_score_list, mod_score_grid);
    }
    destroy(annotated_families);
}


void ModScorer_ComputeModScoresForRegions(ModScorer mod_scorer, Vector regions)
{
    for (int i = 0; i < Vector_GetSize(mod_scorer->families); i++) {
        Family family = Vector_Get(mod_scorer->families, i);
        Family_ComputePenetrances(family);
        Family_ComputePQTerms(family);
    }
    if (mod_scorer->model_params_file) {
        int nmodels = ModelParamsFile_GetNumberOfModels(mod_scorer->model_params_file);
        Vector user_model_em_haplotypes = Vector_CreateEmpty();
        for (int i = 0; i < nmodels; i++) {
            ModelParams user_model = ModelParamsFile_GetModel(mod_scorer->model_params_file, i);
            for (int j = 0; j < Vector_GetSize(regions); j++) {
                Region region = Vector_Get(regions, j);
                sf(ps, "Processing model %d and region %d\n", i + 1, j + 1);
                pr();
                compute_mod_scores_for_region(mod_scorer, region, j, user_model, user_model_em_haplotypes, i);
            }
        }
    } else {
        for (int i = 0; i < Vector_GetSize(regions); i++) {
            Region region = Vector_Get(regions, i);
            sf(ps, "region %d:\n", i + 1);
            pr();
            compute_mod_scores_for_region(mod_scorer, region, i, NULL, NULL, 0);
        }
    }
}

void ModScorer_ComputeModScoreForTestMarkersAndUserModels(ModScorer mod_scorer, Vector test_marker_names)
{
    for (int i = 0; i < Vector_GetSize(mod_scorer->families); i++) {
        Family family = Vector_Get(mod_scorer->families, i);
        Family_ComputePenetrances(family);
        Family_ComputePQTerms(family);
    }
    Vector test_markers = Vector_Create(Vector_GetSize(test_marker_names));
    for (int i = 0; i < Vector_GetSize(test_marker_names); i++) {
        Vector_PushBack(test_markers, LinkageMarkerFile_GetMarkerByName(mod_scorer->linkage_file, Vector_Get(test_marker_names, i)));
    }
    if (!LinkageMarkerFile_WithinMaxDistance(mod_scorer->linkage_file, test_markers, mod_scorer->max_marker_distance)) {
        sf(ps, "The test marker set is not valid.  At least one of the "
               "actual test marker distances is larger than the maximum "
               "test marker distance.\n");
        pr();
        return;
    }
    if (UserLinkageMarkerFile_CountLinkageMarkers(mod_scorer->user_linkage_file, test_markers) >= 1) {
        sf(ps, "The test marker set is not valid.  It contains a linkage marker.\n");
        pr();
        return;
    }
    ModScoreList mod_score_list = ModScorer_GetModScoreList(mod_scorer);
    Vector annotated_families = create_annotated_families(mod_scorer->families, test_markers, mod_scorer->inheritance_distros);
    DoubleVector em_haplotypes = EM_Haplotype_H0(annotated_families, test_markers);
    if (em_haplotypes == NULL) {
        sf(ps, "The test marker set is invalid due to a recombination between test markers.\n");
        pr();
        sf(ps, "One might try to decrease the maximum test marker distance.\n");
        pr();
        destroy(annotated_families);
        return;
    }
    int nmodels = ModelParamsFile_GetNumberOfModels(mod_scorer->model_params_file);
    for (int i = 0; i < nmodels; i++) {
        // Copy test marker vector so that the comparison against the
        // previous set of test markers returns false and resets the
        // current best mod score.
        Vector test_markers_copy = Vector_Create(Vector_GetSize(test_marker_names));
        for (int j = 0; j < Vector_GetSize(test_marker_names); j++) {
            Vector_PushBack(test_markers_copy, LinkageMarkerFile_GetMarkerByName(mod_scorer->linkage_file, Vector_Get(test_marker_names, j)));
        }
        ModelParams user_model = ModelParamsFile_GetModel(mod_scorer->model_params_file, i);
        sf(ps, "Processing model %d\n", i + 1);
        pr();
        ModScore *mod_score_grid = GridModScoreFunction(test_markers_copy, annotated_families, em_haplotypes, mod_scorer->linkage_file, user_model, 1, NULL);
        ModScore *mod_score_nlopt = NloptModScoreFunction(test_markers_copy, annotated_families, em_haplotypes, mod_scorer->linkage_file, user_model, 1, mod_score_grid);
        if (mod_score_nlopt != NULL) {
            if (mod_score_grid->mod_score > mod_score_nlopt->mod_score) {
                sf(ps, "No better nlopt result, reporting globally optimized result first.\n");
                pr();
                ModScoreList_Add(mod_score_list, mod_score_grid);
                ModScoreList_Add(mod_score_list, mod_score_nlopt);
            } else {
                ModScoreList_Add(mod_score_list, mod_score_nlopt);
                ModScoreList_Add(mod_score_list, mod_score_grid);
            }
        } else {
            sf(ps, "No valid nlopt result, reporting globally optimized result twice.\n");
            pr();
            ModScoreList_Add(mod_score_list, mod_score_grid);
            ModScoreList_Add(mod_score_list, mod_score_grid);
        }
    }
    destroy(annotated_families);
}

double ModScorer_ScoreUnderH0(
    DoubleVector penetrances,
    double disease_allele_frequency,
    DoubleVector weights_h0,
    DoubleVector weights_h1,
    DoubleVector em_hapfreqs,
    PenetranceMatrix penetrance_matrix,
    HaplotypeMatrix haplotype_matrix_without_disease_locus,
    Vector pqterms,
    int number_of_founders,
    double **cache_pen_matrix)
{
    static MatrixCache cache_hap = NULL;
    int nhap = DoubleVector_GetSize(em_hapfreqs);
    int max_exponent_hap = 2 * number_of_founders;
    cache_hap = create_or_reuse_matrix_cache(cache_hap, nhap, max_exponent_hap + 1);
    initialize_cache(cache_hap->matrix, nhap, max_exponent_hap + 1, em_hapfreqs);
    double result = 0;
    int number_of_inheritance_vectors = PenetranceMatrix_GetNumberOfRows(penetrance_matrix);
    int number_of_disease_genotypes = PenetranceMatrix_GetNumberOfColumns(penetrance_matrix);
    double dl_score = 0;  // disease locus
    double tm_score = 0;  // test marker
    for (int i = 0; i < number_of_inheritance_vectors; i++) {
        HaplotypeTerm haplotype_term = HaplotypeMatrix_Get(haplotype_matrix_without_disease_locus, i, 0);
        // HaplotypeTerm_Show(haplotype_term);
        double haps = HaplotypeTerm_Evaluate(haplotype_term, em_hapfreqs, cache_hap->matrix);
        // sf(ps, "H0 haps: %f\n", haps);
        // pr();
        tm_score += haps * DoubleVector_Get(weights_h1, i);
        // sf(ps, "H0 tm_score: %f\n", tm_score);
        // pr();
        for (int j = 0; j < number_of_disease_genotypes; j++) {
            unsigned char *penetrance_exponents = PenetranceMatrix_Get(penetrance_matrix, i, j);
            PQTerm pqterm = Vector_Get(pqterms, j);
            double prod = compute_product_of_penetrance_powers(penetrance_exponents, penetrances, cache_pen_matrix);
            // sf(ps, "H0 prod: %f\n", prod);
            // pr();
            if (prod == 0) continue;
            double pq = PQTerm_Evaluate(pqterm, disease_allele_frequency);
            // sf(ps, "H0 pq: %f\n", pq);
            // pr();
            if (pq == 0) continue;
            dl_score += prod * pq;
            // sf(ps, "H0 dl_score: %f\n", dl_score);
            // pr();
        }
    }
    dl_score *= DoubleVector_Get(weights_h0, 0);
    // sf(ps, "H0 dl_score weighted: %f\n", dl_score);
    // pr();
    result = tm_score * dl_score;
    // sf(ps, "H0 result: %f\n", result);
    // pr();
    return result;
}

double ModScorer_ScoreUnderH1(
    DoubleVector penetrances,
    DoubleVector haplotype_frequencies,
    DoubleVector weights_h1,
    PenetranceMatrix penetrance_matrix,
    HaplotypeMatrix haplotype_matrix,
    int number_of_founders,
    double **cache_pen_matrix)
{
    static MatrixCache cache_hap = NULL;
    int nhap = DoubleVector_GetSize(haplotype_frequencies);
    int max_exponent_hap = 2 * number_of_founders;
    cache_hap = create_or_reuse_matrix_cache(cache_hap, nhap, max_exponent_hap + 1);
    initialize_cache(cache_hap->matrix, nhap, max_exponent_hap + 1, haplotype_frequencies);
    double result = 0;
    int number_of_inheritance_vectors = PenetranceMatrix_GetNumberOfRows(penetrance_matrix);
    int number_of_disease_genotypes = PenetranceMatrix_GetNumberOfColumns(penetrance_matrix);
    for (int i = 0; i < number_of_inheritance_vectors; i++) {
        double w = DoubleVector_Get(weights_h1, i);
        // sf(ps, "w: %f\n", w);
        // pr();
        if (w == 0) {
            continue;
        }
        double x = 0;
        for (int j = 0; j < number_of_disease_genotypes; j++) {
            unsigned char *penetrance_exponents = penetrance_matrix->matrix[i][j];
            HaplotypeTerm haplotype_term = haplotype_matrix->matrix[i][j];
            // HaplotypeTerm_Show(haplotype_term);
            double haps = HaplotypeTerm_Evaluate(haplotype_term, haplotype_frequencies, cache_hap->matrix);
            // sf(ps, "haps: %.32f\n", haps);
            // pr();
            if (haps == 0) continue;
            double prod = compute_product_of_penetrance_powers(penetrance_exponents, penetrances, cache_pen_matrix);
            // sf(ps, "prod: %g\n", prod);
            // pr();
            if (prod == 0) continue;
            x += prod * haps;
            // sf(ps, "x: %g\n", x);
            // pr();
        }
        // sf(ps, "x*w: %g\n", x*w);
        // pr();
        result += x * w;
        // sf(ps, "result: %g\n", result);
        // pr();
    }
    return result;
}

double ModScorer_ComputeModScore(
    ModScorer mod_scorer,
    Vector annotated_families,
    Vector test_markers,
    double disease_allele_frequency,
    DoubleVector hapfreqs,
    DoubleVector penetrances,
    DoubleVector em_hapfreqs)
{
    (void) mod_scorer;
    double score = 0.0;
    for (int k = 0; k < Vector_GetSize(annotated_families); k++) {
        AnnotatedFamily annotated_family = Vector_Get(annotated_families, k);
        score += compute_mod_score_for_family(test_markers, annotated_family, disease_allele_frequency, hapfreqs, penetrances, em_hapfreqs);
    }
    if (isnan(score)) {
    return -HUGE_VAL;
    } else {
    return score;
    }
}

ModScoreList ModScorer_GetModScoreList(ModScorer mod_scorer)
{
    return mod_scorer->mod_score_list;
}
