#include "PedigreeFile.h"
#include "PedigreeFileRecord.h"
#include "LinkageMarkerFile.h"
#include "UserLinkageMarkerFile.h"
#include "Vector.h"
#include "Str.h"
#include "IO.h"
#include "Memory.h"
#include "Int.h"
#include <ctype.h>
#include <string.h>
#include <stdlib.h>

// Legacy includes.
#define INC_LIB
#define INC_SHELL
#include "npl.h"

struct PedigreeFileStruct {
    Vector records;
    Vector family_names;
};

static void collect_family_ids(PedigreeFile ped_file);
static int is_whitespace_only(const char *);

PedigreeFile PedigreeFile_Create(const char *filename)
{
    PedigreeFile ped_file = Memory_Malloc(sizeof(*ped_file));
    ped_file->records = Vector_CreateEmpty();
    ped_file->family_names = Vector_CreateEmpty();
    FILE *fp = fopen(filename, "rb");
    if (fp == NULL) {
        Memory_Free(ped_file);
        return NULL;
    }
    char *line;
    while ((line = IO_GetLine(fp)) != NULL) {
        if (!is_whitespace_only(line)) {
            Vector_PushBack(ped_file->records, PedigreeFileRecord_Create(line));
        }
        free(line);
    }
    fclose(fp);
    collect_family_ids(ped_file);
    return ped_file;
}

void PedigreeFile_Destroy(PedigreeFile ped_file)
{
    for (int i = 0; i < Vector_GetSize(ped_file->records); i++)
        PedigreeFileRecord_Destroy(Vector_Get(ped_file->records, i));
    Vector_Destroy(ped_file->records);
    Memory_Free(ped_file);
}

int PedigreeFile_GetNumberOfFamilies(PedigreeFile ped_file)
{
    return Vector_GetSize(ped_file->family_names);
}

static void collect_family_ids(PedigreeFile ped_file)
{
    int number_of_families = 0;
    const char *previous_family_id = NULL;

    for (int i = 0; i < Vector_GetSize(ped_file->records); i++) {
        const char *family_id = PedigreeFileRecord_GetFamilyId(Vector_Get(ped_file->records, i));
        if (previous_family_id == NULL
            || !String_Equal(family_id, previous_family_id))
        {
            ++number_of_families;
            previous_family_id = family_id;
            Vector_PushBack(ped_file->family_names, (void *) family_id);
        }
    }
}

PedigreeFileRecord PedigreeFile_GetRecord(PedigreeFile ped_file, const char *family_name, const char *person_name)
{
    for (int i = 0; i < Vector_GetSize(ped_file->records); i++) {
        PedigreeFileRecord record = Vector_Get(ped_file->records, i);
        const char *this_family_name = PedigreeFileRecord_GetFamilyId(record);
        const char *this_person_name = PedigreeFileRecord_GetIndividualId(record);
        if (String_Equal(this_family_name, family_name) && String_Equal(this_person_name, person_name)) {
            return record;
        }
    }
    return NULL;
}

const char *PedigreeFile_GetFamilyName(PedigreeFile ped_file, int family_index)
{
    return Vector_Get(ped_file->family_names, family_index);
}

const int *PedigreeFile_GetConvertedAlleles(
        PedigreeFile ped_file,
        const char *family_name,
        const char *person_name,
        LinkageMarkerFile linkage_file,
        UserLinkageMarkerFile user_linkage_file)
{
    PedigreeFileRecord record = PedigreeFile_GetRecord(ped_file, family_name, person_name);
    char **alleles1, **alleles2;
    PedigreeFileRecord_GetAlleles(record, &alleles1, &alleles2);
    int number_of_markers = PedigreeFileRecord_GetNumberOfMarkers(record);
    int *alleles = Memory_Malloc(2 * number_of_markers * sizeof(int));
    for (int i = 0; i < number_of_markers; i++) {
        int a1 = atoi(alleles1[i]);
        int a2 = atoi(alleles2[i]);
        // Set genotype to missing if either allele is missing.
        if (a1 == 0 || a2 == 0) {
            a1 = a2 = 0;
        }
        Marker marker = LinkageMarkerFile_GetMarkerByIndex(linkage_file, i);
        if (UserLinkageMarkerFile_Contains(user_linkage_file, Marker_GetName(marker))) {
            if (Marker_GetNumberOfAlleles(marker) > 2) {
                // Multiallelic linkage marker
                alleles[2*i]     = a1;
                alleles[2*i + 1] = a2;
            } else {
                // Biallelic linkage and potential test marker
                alleles[2*i]     = Marker_ConvertTo01(marker, a1);
                alleles[2*i + 1] = Marker_ConvertTo01(marker, a2);
            }
        } else if (Marker_GetNumberOfAlleles(marker) <= 2) {
            // Biallelic test marker
            alleles[2*i]     = Marker_ConvertTo01(marker, a1);
            alleles[2*i + 1] = Marker_ConvertTo01(marker, a2);
        } else {
            sf(ps, "Marker %s is supposed to be a biallelic test marker, but has > 2 alleles\n",
                Marker_GetName(marker));
            pr();
            exit(1);
        }
        if (Marker_GetNumberOfAlleles(marker) > 2
                && UserLinkageMarkerFile_Contains(user_linkage_file, Marker_GetName(marker))) {
            alleles[2*i] = a1;
            alleles[2*i + 1] = a2;
        } else {
            alleles[2*i] = Marker_ConvertTo01(marker, a1);
            alleles[2*i + 1] = Marker_ConvertTo01(marker, a2);
        }
    }
    return alleles;
}

static int is_whitespace_only(const char *line)
{
    for (int i = 0; i < (int) strlen(line); i++) {
        if (!isspace(line[i])) {
            return 0;
        }
    }
    return 1;
}

int PedigreeFile_GetNumberOfRecords(PedigreeFile ped_file)
{
    return Vector_GetSize(ped_file->records);
}

PedigreeFileRecord PedigreeFile_GetRecordByIndex(PedigreeFile ped_file, int i)
{
    return Vector_Get(ped_file->records, i);
}
