#include "PedigreeFileRecord.h"
#include "ParseObject.h"
#include "Memory.h"
#include <stddef.h>
#include <string.h>

struct PedigreeFileRecordStruct {
    Vector fields;
    char *family_id;
    char *individual_id;
    char *father_id;
    char *mother_id;
    char *sex;
    char *affection_status;
    char **allele1;
    char **allele2;
    int number_of_markers;
};

PedigreeFileRecord PedigreeFileRecord_Create(char *line)
{
    ParseObject parser = ParseObject_Create(line);
    Vector fields = ParseObject_SplitIntoWords(parser);
    ParseObject_Destroy(parser);
    int number_of_fields = Vector_GetSize(fields);
    PedigreeFileRecord record = (PedigreeFileRecord) Memory_Malloc(sizeof(*record));
    record->fields = fields;
    record->family_id = Vector_Get(fields, 0);
    record->individual_id = Vector_Get(fields, 1);
    record->father_id = Vector_Get(fields, 2);
    record->mother_id = Vector_Get(fields, 3);
    record->sex = Vector_Get(fields, 4);
    record->affection_status = Vector_Get(fields, 5);
    // Parse *optional* liability class
    record->number_of_markers = (number_of_fields - 6) / 2;
    record->allele1 = (char **) Memory_Malloc(record->number_of_markers * sizeof(char *));
    record->allele2 = (char **) Memory_Malloc(record->number_of_markers * sizeof(char *));
    for (int i = 0; i < record->number_of_markers; i++) {
        record->allele1[i] = Vector_Get(fields, 6 + 2*i + 0);
        record->allele2[i] = Vector_Get(fields, 6 + 2*i + 1);
    }
    // Parse *optional* phenotype data
    // Parse *optional* covariate data
    return record;
}

const char *PedigreeFileRecord_GetFamilyId(PedigreeFileRecord record)
{
    return record->family_id;
}

const char *PedigreeFileRecord_GetIndividualId(PedigreeFileRecord record)
{
    return record->individual_id;
}

const char *PedigreeFileRecord_GetFatherId(PedigreeFileRecord record)
{
    return record->father_id;
}

const char *PedigreeFileRecord_GetMotherId(PedigreeFileRecord record)
{
    return record->mother_id;
}

const char *PedigreeFileRecord_GetSex(PedigreeFileRecord record)
{
    return record->sex;
}

const char *PedigreeFileRecord_GetAffectionStatus(PedigreeFileRecord record)
{
    return record->affection_status;
}

const char *PedigreeFileRecord_GetAllele1(PedigreeFileRecord record, int marker)
{
    return record->allele1[marker];
}

const char *PedigreeFileRecord_GetAllele2(PedigreeFileRecord record, int marker)
{
    return record->allele2[marker];
}

const char *PedigreeFileRecord_GetGenotypes(PedigreeFileRecord record)
{
    int length = 0;
    for (int i = 0; i < record->number_of_markers; i++)
       length += strlen(record->allele1[i]) + strlen(record->allele2[i]) + 2;  // + 2 for spaces between alleles (or final newline)
    char *genotypes = Memory_Malloc(length * sizeof(char));
    int pos = 0;
    for (int i = 0; i < record->number_of_markers; i++) {
        strcpy(genotypes + pos, record->allele1[i]);
        pos += strlen(record->allele1[i]);
        genotypes[pos++] = ' ';
        strcpy(genotypes + pos, record->allele2[i]);
        pos += strlen(record->allele2[i]);
        genotypes[pos++] = ' ';
    }
    genotypes[length - 1] = '\0';
    return genotypes;
}

void PedigreeFileRecord_Destroy(PedigreeFileRecord record)
{
    for (int i = 0; i < Vector_GetSize(record->fields); i++)
        Memory_Free(Vector_Get(record->fields, i));
    Vector_Destroy(record->fields);
    Memory_Free(record->allele1);
    Memory_Free(record->allele2);
}

void PedigreeFileRecord_GetAlleles(PedigreeFileRecord record, char ***a1, char ***a2)
{
    *a1 = record->allele1;
    *a2 = record->allele2;
}

int PedigreeFileRecord_GetNumberOfMarkers(PedigreeFileRecord record)
{
    return record->number_of_markers;
}
