#include <time.h>
#include "dSFMT.h"

#include "AnnotatedFamily.h"
#include "DoubleVector.h"
#include "EM.h"
#include "Family.h"
#include "IntSet.h"
#include "LinkageMarkerFile.h"
#include "Marker.h"
#include "Memory.h"
#include "ModelParamsFile.h"
#include "PedigreeFileSimulator.h"
#include "Person.h"
#include "UserLinkageMarkerFile.h"

// Legacy includes.
#define INC_LIB
#define INC_SHELL
#include "npl.h"

struct PedigreeFileSimulatorStruct {
    Vector families;
    UserLinkageMarkerFile user_linkage_file;
    LinkageMarkerFile linkage_file;
    Vector test_markers;
    IntSet marker_indexes;
    IntSet test_marker_indexes;
    DoubleVector hapfreqs;
    ModelParamsFile user_association_models_file;
};

static void simulate_linkage_markers_in_founders(Family family, PedigreeFileSimulator simulator);
static void simulate_test_marker_haplotypes_in_founders(Family family, PedigreeFileSimulator simulator);
static void drop_marker_alleles_down_to_nonfounders(Family family, PedigreeFileSimulator simulator);
static void print_family(FILE *fp, PedigreeFileSimulator simulator, Family family);

PedigreeFileSimulator PedigreeFileSimulator_Create(Vector families, PedigreeFile ped, LinkageMarkerFile linkage_file, UserLinkageMarkerFile user_linkage_file, Vector test_markers, ModelParamsFile user_association_models_file, DoubleVector hapfreqs)
{
    (void) ped;
    PedigreeFileSimulator simulator = Memory_Malloc(sizeof(*simulator));
    simulator->families = families;
    simulator->user_linkage_file = user_linkage_file;
    simulator->linkage_file = linkage_file;
    simulator->test_markers = test_markers;
    simulator->marker_indexes = PedigreeFileSimulator_GetMarkerIndexesToKeep(simulator);
    simulator->test_marker_indexes = PedigreeFileSimulator_GetTestMarkerIndexes(simulator);
    simulator->user_association_models_file = user_association_models_file;
    simulator->hapfreqs = hapfreqs;
    if (simulator->hapfreqs == NULL) {
        sf(ps, "The test marker set is invalid due to a recombination between test markers.\n");
        pr();
        Memory_Free(simulator);
        return NULL;
    }
    return simulator;
}

static void mark_founders_as_simulated(Vector founders) {
    for (int i = 0; i < Vector_GetSize(founders); i++) {
        Person founder = Vector_Get(founders, i);
        Person_SetSimulated(founder);
    }
}

static void mark_person_as_simulated(Person person) {
    Person_SetSimulated(person);
}

void PedigreeFileSimulator_Simulate(PedigreeFileSimulator simulator, const char *simulated_pedfile)
{
    FILE *fp = fopen(simulated_pedfile, "wb");
    if (!fp) {
        sf(ps, "Failed to open file: \"%s\"\n", simulated_pedfile);
        pr();
        exit(1);
    }
    for (int i = 0; i < Vector_GetSize(simulator->families); i++) {
        Family family = Vector_Get(simulator->families, i);
        simulate_linkage_markers_in_founders(family, simulator);
        simulate_test_marker_haplotypes_in_founders(family, simulator);
        mark_founders_as_simulated(Family_GetFounders(family));
        drop_marker_alleles_down_to_nonfounders(family, simulator);
        print_family(fp, simulator, family);
    }
    fclose(fp);
}

static void simulate_linkage_markers_in_founders(Family family, PedigreeFileSimulator simulator) {
    Vector founders = Family_GetFounders(family);
    int number_of_markers = LinkageMarkerFile_GetNumberOfMarkers(simulator->linkage_file);
    for (int i = 0; i < Vector_GetSize(founders); i++) {
        Person person = Vector_Get(founders, i);
        int frequency_index = 0;
        for (int j = 0; j < number_of_markers; j++) {
            if (!IntSet_Contains(simulator->marker_indexes, j)) {
                continue;
            }
            Marker marker = LinkageMarkerFile_GetMarkerByIndex(simulator->linkage_file, j);
            int is_linkage_marker = UserLinkageMarkerFile_Contains(simulator->user_linkage_file, Marker_GetName(marker));
            int is_test_marker = IntSet_Contains(simulator->test_marker_indexes, j);
            if (!is_linkage_marker || is_test_marker) {
                continue;
            }
            int number_of_alleles = Marker_GetNumberOfAlleles(marker);
            if (number_of_alleles == 2 && Person_GetMarkerGenotype(person, j) == XX) {
                Person_MarkGenotypeAsMissing(person, j);
            }
            if (number_of_alleles > 2 && (Person_GetMarkerAllele1(person, j) == 0 || Person_GetMarkerAllele2(person, j) == 0)) {
                Person_MarkGenotypeAsMissing(person, j);
            }
            double allele_sum = Marker_GetSumOfAlleleFrequencies(marker);
            double random_number_allele1 = genrand_open_open() * allele_sum;
            double random_number_allele2 = genrand_open_open() * allele_sum;
            double lower, upper = 0.0;
            int allele = 0;
            for (int k = 0; k < number_of_alleles; k++) {
                if (number_of_alleles <= 2) {
                    allele = k;
                } else {
                    allele = k + 1;
                }
                lower = upper;
                upper += Marker_GetAlleleFrequency(marker, k);
                if (lower < random_number_allele1 && random_number_allele1 <= upper) {
                    Person_SetSimulatedMarkerAllele1(person, j, allele);
                }
                if (lower < random_number_allele2 && random_number_allele2 <= upper) {
                    Person_SetSimulatedMarkerAllele2(person, j, allele);
                }
            }
            ++frequency_index;
        }
    }
}

static void simulate_test_marker_haplotypes_in_founders(Family family, PedigreeFileSimulator simulator) {
    Vector founders = Family_GetFounders(family);
    int number_of_test_markers = Vector_GetSize(simulator->test_markers);
    DoubleVector hapfreqs = simulator->hapfreqs;
    for (int i = 0; i < Vector_GetSize(founders); i++) {
        Person person = Vector_Get(founders, i);
        double random_number_hap1 = genrand_open_open();
        double random_number_hap2 = genrand_open_open();
        double lower, upper = 0.0;
        for (int hap = 0; hap < DoubleVector_GetSize(hapfreqs); hap++) {
            lower = upper;
            upper += DoubleVector_Get(hapfreqs, hap);
            if (lower < random_number_hap1 && random_number_hap1 <= upper) {
                for (int k = 0; k < number_of_test_markers; k++) {
                    int marker_index = Marker_GetIndex(Vector_Get(simulator->test_markers, k));
                    int number_of_alleles = Marker_GetNumberOfAlleles(Vector_Get(simulator->test_markers, k));
                    if (number_of_alleles == 2 && Person_GetMarkerGenotype(person, marker_index) == XX) {
                        Person_MarkGenotypeAsMissing(person, marker_index);
                    }
                    if (number_of_alleles > 2 && (Person_GetMarkerAllele1(person, marker_index) == 0 || Person_GetMarkerAllele2(person, marker_index) == 0)) {
                        Person_MarkGenotypeAsMissing(person, marker_index);
                    }
                    int allele = ((hap >> k) & 0x1);
                    Person_SetSimulatedMarkerAllele1(person, marker_index, allele);
                }
            }
            if (lower < random_number_hap2 && random_number_hap2 <= upper) {
                for (int k = 0; k < number_of_test_markers; k++) {
                    int marker_index = Marker_GetIndex(Vector_Get(simulator->test_markers, k));
                    int number_of_alleles = Marker_GetNumberOfAlleles(Vector_Get(simulator->test_markers, k));
                    if (number_of_alleles == 2 && Person_GetMarkerGenotype(person, marker_index) == XX) {
                        Person_MarkGenotypeAsMissing(person, marker_index);
                    }
                    if (number_of_alleles > 2 && (Person_GetMarkerAllele1(person, marker_index) == 0 || Person_GetMarkerAllele2(person, marker_index) == 0)) {
                        Person_MarkGenotypeAsMissing(person, marker_index);
                    }
                    int allele = (hap >> k) & 0x1;
                    Person_SetSimulatedMarkerAllele2(person, marker_index, allele);
                }
            }
        }
    }
}

static void drop_marker_alleles_down_to_nonfounders(Family family, PedigreeFileSimulator simulator) {
    int nmarkers = LinkageMarkerFile_GetNumberOfMarkers(simulator->linkage_file);
    Vector nonfounders = Family_GetNonfounders(family);
    int nnonfounders = Vector_GetSize(nonfounders);
    for (int i = 0; i < nnonfounders; i++) {
        Person person = Vector_Get(nonfounders, i);
        Person_SetNotSimulated(person);
    }
    int nsimulated = 0;
    for (int i = 0; nsimulated < nnonfounders; i = (i + 1) % nnonfounders) {
        Person person = Vector_Get(nonfounders, i);
        if (Person_IsSimulated(person)) {
            continue;
        }
        if (!Person_SimulatedParents(person)) {
            continue;
        }
        Person father = Person_GetFather(person);
        Person mother = Person_GetMother(person);
        int saw_first_marker = 0;
        int last_seen_marker = 0;
        int test_markers_simulated = 0;
        double dist_male = 0.0;
        double dist_female = 0.0;
        double r1, r2;  // random numbers for choosing 1st and 2nd allele
        for (int j = 0; j < nmarkers; j++) {
            int allele1, allele2;
            AlleleOrigin paternal_origin, maternal_origin;
            if (!IntSet_Contains(simulator->marker_indexes, j)) {
                continue;
            }
            Marker marker = LinkageMarkerFile_GetMarkerByIndex(simulator->linkage_file, j);
            int number_of_alleles = Marker_GetNumberOfAlleles(marker);
            if (number_of_alleles == 2 && Person_GetMarkerGenotype(person, j) == XX) {
                Person_MarkGenotypeAsMissing(person, j);
            }
            if (number_of_alleles > 2 && (Person_GetMarkerAllele1(person, j) == 0 || Person_GetMarkerAllele2(person, j) == 0)) {
                Person_MarkGenotypeAsMissing(person, j);
            }
            int is_linkage_marker = UserLinkageMarkerFile_Contains(simulator->user_linkage_file, Marker_GetName(marker));
            int is_test_marker = IntSet_Contains(simulator->test_marker_indexes, j);
            if (is_linkage_marker && !is_test_marker) {  // pure linkage marker
                if (!saw_first_marker) {
                    r1 = genrand_open_open();
                    if (r1 < 0.5) {  // take paternal allele
                        allele1 = Person_GetMarkerAllele1(father, j);
                        paternal_origin = PATERNAL;
                    } else {  // take maternal allele
                        allele1 = Person_GetMarkerAllele2(father, j);
                        paternal_origin = MATERNAL;
                    }
                    r2 = genrand_open_open();
                    if (r2 < 0.5) {  // take paternal allele
                        allele2 = Person_GetMarkerAllele1(mother, j);
                        maternal_origin = PATERNAL;
                    } else {  // take maternal allele
                        allele2 = Person_GetMarkerAllele2(mother, j);
                        maternal_origin = MATERNAL;
                    }
                    Person_SetSimulatedMarkerAllele1(person, j, allele1);
                    Person_SetSimulatedMarkerAllele2(person, j, allele2);
                    Person_SetPaternalAlleleOrigin(person, paternal_origin);
                    Person_SetMaternalAlleleOrigin(person, maternal_origin);
                    saw_first_marker = 1;
                } else {  // already saw first marker
                    if (LinkageMarkerFile_IsSexSpecific(simulator->linkage_file)) {
                        dist_male = dist_to_rec(LinkageMarkerFile_GetMaleDistanceBetween(simulator->linkage_file, j, last_seen_marker));
                        dist_female = dist_to_rec(LinkageMarkerFile_GetFemaleDistanceBetween(simulator->linkage_file, j, last_seen_marker));
                    } else {
                        dist_male = dist_to_rec(LinkageMarkerFile_GetMaleDistanceBetween(simulator->linkage_file, j, last_seen_marker));
                        dist_female = dist_male;
                    }
                    r1 = genrand_open_open();
                    if (r1 <= dist_male) {  // recombination happened: origin changes
                        if (Person_GetPaternalAlleleOrigin(person) == PATERNAL) {
                            allele1 = Person_GetMarkerAllele2(father, j);
                            paternal_origin = MATERNAL;
                        } else {
                            allele1 = Person_GetMarkerAllele1(father, j);
                            paternal_origin = PATERNAL;
                        }
                        Person_SetSimulatedMarkerAllele1(person, j, allele1);
                        Person_SetPaternalAlleleOrigin(person, paternal_origin);
                    } else {  // no recombination: same origin as previous allele
                        if (Person_GetPaternalAlleleOrigin(person) == PATERNAL) {
                            allele1 = Person_GetMarkerAllele1(father, j);
                        } else {
                            allele1 = Person_GetMarkerAllele2(father, j);
                        }
                        Person_SetSimulatedMarkerAllele1(person, j, allele1);
                    }
                    r2 = genrand_open_open();
                    if (r2 <= dist_female) {  // recombination happened: origin changes
                        if (Person_GetMaternalAlleleOrigin(person) == PATERNAL) {
                            allele2 = Person_GetMarkerAllele2(mother, j);
                            maternal_origin = MATERNAL;
                        } else {
                            allele2 = Person_GetMarkerAllele1(mother, j);
                            maternal_origin = PATERNAL;
                        }
                        Person_SetSimulatedMarkerAllele2(person, j, allele2);
                        Person_SetMaternalAlleleOrigin(person, maternal_origin);
                    } else {  // no recombination: same origin as previous allele
                        if (Person_GetMaternalAlleleOrigin(person) == PATERNAL) {
                            allele2 = Person_GetMarkerAllele1(mother, j);
                        } else {
                            allele2 = Person_GetMarkerAllele2(mother, j);
                        }
                        Person_SetSimulatedMarkerAllele2(person, j, allele2);
                    }
                }
                last_seen_marker = j;
            } else if (!test_markers_simulated) {
                int number_of_test_markers = Vector_GetSize(simulator->test_markers);
                int first_test_marker_index = Marker_GetIndex(Vector_Get(simulator->test_markers, 0));
                if (!saw_first_marker) {  // --> 1st marker is test marker
                    dist_male = dist_female = 0.5;
                    saw_first_marker = 1;
                } else {  // first test marker is not the first marker overall
                    if (LinkageMarkerFile_IsSexSpecific(simulator->linkage_file)) {
                        dist_male = dist_to_rec(LinkageMarkerFile_GetMaleDistanceBetween(simulator->linkage_file, first_test_marker_index, last_seen_marker));
                        dist_female = dist_to_rec(LinkageMarkerFile_GetFemaleDistanceBetween(simulator->linkage_file, first_test_marker_index, last_seen_marker));
                    } else {
                        dist_male = dist_to_rec(LinkageMarkerFile_GetMaleDistanceBetween(simulator->linkage_file, first_test_marker_index, last_seen_marker));
                        dist_female = dist_male;
                    }
                }
                // Since these two random numbers are drawn only once
                // for all test markers, all test markers will have
                // the same allele origins.
                r1 = genrand_open_open();
                r2 = genrand_open_open();
                // Unless we initialize the allele origins outside of
                // the for-loop we get a "may be used uninitialized"
                // warning because the compiler cannot exclude the
                // possibility that the below for-loop, inside of
                // which the allele origins are set, stops before the
                // first iteration.  We know that the loop will run
                // and set the allele origins to the correct
                // values. Therefore this initialization has no
                // effect except for silencing the warning.
                paternal_origin = maternal_origin = MATERNAL;
                for (int k = 0; k < number_of_test_markers; k++) {
                    int marker_index = Marker_GetIndex(Vector_Get(simulator->test_markers, k));
                    if (r1 < dist_male) {  // recombination happened: origin changes
                        if (Person_GetPaternalAlleleOrigin(person) == PATERNAL) {
                            allele1 = Person_GetMarkerAllele2(father, marker_index);
                            paternal_origin = MATERNAL;
                        } else {
                            allele1 = Person_GetMarkerAllele1(father, marker_index);
                            paternal_origin = PATERNAL;
                        }
                        Person_SetSimulatedMarkerAllele1(person, marker_index, allele1);
                    } else {  // no recombination: same origin as previous allele
                        if (Person_GetPaternalAlleleOrigin(person) == PATERNAL) {
                            allele1 = Person_GetMarkerAllele1(father, marker_index);
                            paternal_origin = PATERNAL;
                        } else {
                            allele1 = Person_GetMarkerAllele2(father, marker_index);
                            paternal_origin = MATERNAL;
                        }
                        Person_SetSimulatedMarkerAllele1(person, marker_index, allele1);
                    }
                    if (r2 < dist_female) {  // recombination happened: origin changes
                        if (Person_GetMaternalAlleleOrigin(person) == PATERNAL) {
                            allele2 = Person_GetMarkerAllele2(mother, marker_index);
                            maternal_origin = MATERNAL;
                        } else {
                            allele2 = Person_GetMarkerAllele1(mother, marker_index);
                            maternal_origin = PATERNAL;
                        }
                        Person_SetSimulatedMarkerAllele2(person, marker_index, allele2);
                    } else {  // no recombination: same origin as previous allele
                        if (Person_GetMaternalAlleleOrigin(person) == PATERNAL) {
                            allele2 = Person_GetMarkerAllele1(mother, marker_index);
                            maternal_origin = PATERNAL;
                        } else {
                            allele2 = Person_GetMarkerAllele2(mother, marker_index);
                            maternal_origin = MATERNAL;
                        }
                        Person_SetSimulatedMarkerAllele2(person, marker_index, allele2);
                    }
                }
                // Since, by assumption, there are no recombinations
                // between test markers we have to wait until the
                // paternal/maternal origin of all test markers was
                // checked before changing the paternal/maternal
                // origin.
                Person_SetPaternalAlleleOrigin(person, paternal_origin);
                Person_SetMaternalAlleleOrigin(person, maternal_origin);
                test_markers_simulated = 1;
                last_seen_marker = Marker_GetIndex(Vector_Get(simulator->test_markers, number_of_test_markers - 1));
            }
        }
        mark_person_as_simulated(person);
        nsimulated++;
    }
}

static void print_family(FILE *fp, PedigreeFileSimulator simulator, Family family) {
    for (int j = 0; j < Family_GetNumberOfMembers(family); j++) {
        Person p = Family_GetMember(family, j);
        fprintf(fp, "%d %s %s %s %d %d",
                Family_GetId(family) + 1,  // +1 to prevent a family with id "0"
                Person_GetId(p),
                Person_GetFatherId(p),
                Person_GetMotherId(p),
                Person_IsSexKnown(p) ? (Person_IsMale(p) ? 1 : 2) : 0,
                Person_IsAffectionStatusKnown(p) ? (Person_IsAffected(p) ? 2 : 1) : 0);
        int number_of_markers = LinkageMarkerFile_GetNumberOfMarkers(simulator->linkage_file);
        for (int k = 0; k < number_of_markers; k++) {
            if (Person_IsGenotypeMissing(p, k)) {
                fprintf(fp, " 0 0");
                continue;
            }
            int a1 = Person_GetMarkerAllele1(p, k);
            int a2 = Person_GetMarkerAllele2(p, k);
            Marker marker = LinkageMarkerFile_GetMarkerByIndex(simulator->linkage_file, k);
            // Note that the distinction between <= 2 and > 2
            // marker alleles is sufficient at this point since
            // any errors related to the number of marker alleles
            // would have been caught when converting to 0/1 in
            // PedigreeFile.c.
            if (Marker_GetNumberOfAlleles(marker) <= 2) {
                fprintf(fp, " %d %d",
                        Marker_ConvertFrom01(marker, a1),
                        Marker_ConvertFrom01(marker, a2));
            } else {
                fprintf(fp, " %d %d", a1, a2);
            }
        }
        fprintf(fp, "\n");
    }
}

IntSet PedigreeFileSimulator_GetMarkerIndexesToKeep(PedigreeFileSimulator simulator)
{
    Vector marker_names = UserLinkageMarkerFile_GetMarkerNames(simulator->user_linkage_file);
    IntSet marker_indexes = IntSet_Create(Vector_GetSize(marker_names));
    for (int i = 0; i < Vector_GetSize(marker_names); i++) {
        const char *marker_name = Vector_Get(marker_names, i);
        int marker_index = LinkageMarkerFile_GetMarkerIndex(simulator->linkage_file, marker_name);
        IntSet_Put(marker_indexes, marker_index);
    }
    for (int i = 0; i < Vector_GetSize(simulator->test_markers); i++) {
        const char *marker_name = Marker_GetName(Vector_Get(simulator->test_markers, i));
        int marker_index = LinkageMarkerFile_GetMarkerIndex(simulator->linkage_file, marker_name);
        IntSet_Put(marker_indexes, marker_index);
    }
    return marker_indexes;
}

IntSet PedigreeFileSimulator_GetTestMarkerIndexes(PedigreeFileSimulator simulator)
{
    IntSet test_marker_indexes = IntSet_Create(Vector_GetSize(simulator->test_markers));
    for (int i = 0; i < Vector_GetSize(simulator->test_markers); i++) {
        const char *marker_name = Marker_GetName(Vector_Get(simulator->test_markers, i));
        int test_marker_index = LinkageMarkerFile_GetMarkerIndex(simulator->linkage_file, marker_name);
        IntSet_Put(test_marker_indexes, test_marker_index);
    }
    return test_marker_indexes;
}
