#include "FailWith.h"
#include "IntSet.h"
#include "Memory.h"
#include "ParseObject.h"
#include "Person.h"
#include <stdlib.h>
#include <stdio.h>

#include "npl.h"

struct PersonStruct {
    const char *id;
    const char *father_id;
    const char *mother_id;
    Person father;
    Person mother;
    int sex;
    int affection_status;
    int *allele1;
    int *allele2;
    int allele_capacity;
    int number_of_markers;
    int paternal_bit_fixed;
    int maternal_bit_fixed;
    int paternal_allele_offset;
    int maternal_allele_offset;
    AlleleOrigin origin_of_maternal_allele;
    AlleleOrigin origin_of_paternal_allele;
    int simulated;
    IntSet missing_genotypes;
};

enum { EMPTY = -2 };
enum { NA = -1 };
enum { MALE, FEMALE };

static int adjust_allele_storage_capacity(Person, int marker);
static void allocate_missing_genotypes(Person, int number_of_markers);
static int double_allele_storage_capacity(Person);
static int get_marker_allele(int which, Person, int marker);
static int initialize_allele_storage(Person);
static int set_allele(int which, Person p, int marker, int allele);

static int set_allele(int which, Person p, int marker, int allele)
{
    if (!adjust_allele_storage_capacity(p, marker))
        return 0;

    switch (which) {
        case 1:
            p->allele1[marker] = allele;
            break;
        case 2:
            p->allele2[marker] = allele;
            break;
        default:
            return 0;
    }
    return 1;
}

static int initialize_allele_storage(Person p)
{
    int *s;

    if ((s = (int *) Memory_Malloc(sizeof(int))) == NULL)
        return 0;
    p->allele1 = s;

    if ((s = (int *) Memory_Malloc(sizeof(int))) == NULL)
        return 0;
    p->allele2 = s;

    p->allele1[0] = p->allele2[0] = EMPTY;
    p->allele_capacity = 1;

    return 1;
}

static int adjust_allele_storage_capacity(Person p, int marker)
{
    if (p->allele_capacity == 0)
        if (!initialize_allele_storage(p))
            return 0;

    while (marker >= p->allele_capacity)
        if (!double_allele_storage_capacity(p))
            return 0;

    return 1;
}

static int double_allele_storage_capacity(Person p)
{
    int *a1, *a2;
    size_t nbytes;
    int i;

    nbytes = 2 * p->allele_capacity * sizeof(int);

    if ((a1 = (int *) Memory_Malloc(nbytes)) == NULL)
        return 0;
    if ((a2 = (int *) Memory_Malloc(nbytes)) == NULL)
        return 0;

    for (i = 0; i < p->allele_capacity; i++) {
        a1[i] = p->allele1[i];
        a2[i] = p->allele2[i];
    }

    for (i = p->allele_capacity; i < 2 * p->allele_capacity; i++) {
        a1[i] = EMPTY;
        a2[i] = EMPTY;
    }

    p->allele1 = a1;
    p->allele2 = a2;
    p->allele_capacity *= 2;

    return 1;
}

static int get_marker_allele(int which, Person person, int marker)
{
    if (marker >= person->number_of_markers) {
        FailWith("you asked for marker %d, but there are only %d markers", marker, person->number_of_markers);
    }
    int result = -999;
    if (which == 1)
        result = person->allele1[marker];
    else if (which == 2)
        result = person->allele2[marker];
    else
        FailWith("get_marker_allele: invalid value for 'which': %d", which);
    return result;
}

Person Person_Create(const char *id)
{
    Person p;

    if ((p = (Person) Memory_Malloc(sizeof *p)) == NULL)
        return NULL;

    p->id = id;
    p->father_id = "unknown";
    p->mother_id = "unknown";
    p->father = NULL;
    p->mother = NULL;
    p->sex = NA;
    p->affection_status = NA;
    p->allele1 = NULL;
    p->allele2 = NULL;
    p->allele_capacity = 0;
    p->number_of_markers = 0;
    p->paternal_bit_fixed = 0;
    p->maternal_bit_fixed = 0;
    p->paternal_allele_offset = -1;
    p->maternal_allele_offset = -1;
    p->simulated = 0;
    p->missing_genotypes = NULL;

    return p;
}

void Person_Destroy(Person p)
{
    Memory_Free(p->allele1);
    Memory_Free(p->allele2);
}

const char *Person_GetId(Person p)
{
    return p->id;
}

void Person_SetFatherId(Person person, const char *id)
{
    person->father_id = id;
}

void Person_SetFather(Person p, Person father)
{
    p->father = father;
}

void Person_SetMother(Person p, Person mother)
{
    p->mother = mother;
}

Person Person_GetFather(Person p)
{
    return p->father;
}

Person Person_GetMother(Person p)
{
    return p->mother;
}

const char *Person_GetFatherId(Person person)
{
    return person->father_id;
}

const char *Person_GetMotherId(Person person)
{
    return person->mother_id;
}

void Person_SetMotherId(Person person, const char *id)
{
    person->mother_id = id;
}

void Person_SetMarkerGenotypes(Person person, int *genotypes, int nmarkers)
{
    person->number_of_markers = nmarkers;
    for (int i = 0; i < nmarkers; i++) {
        set_allele(1, person, i, genotypes[2*i]);
        set_allele(2, person, i, genotypes[2*i + 1]);
    }
}

int Person_GetMarkerAllele1(Person person, int marker_index)
{
    return get_marker_allele(1, person, marker_index);
}

int Person_GetNumberOfMarkers(Person person)
{
    return person->number_of_markers;
}

int Person_GetMarkerAllele2(Person person, int marker_index)
{
    return get_marker_allele(2, person, marker_index);
}

Genotype Person_GetMarkerGenotype(Person p, int marker_index)
{
    int a1 = Person_GetMarkerAllele1(p, marker_index);
    int a2 = Person_GetMarkerAllele2(p, marker_index);
    Genotype result;
    if (a1 == MISSING_ALLELE || a2 == MISSING_ALLELE)
        result = XX;
    else
        result = Genotype_FromAlleles(a1, a2);
    return result;
}

int Person_IsActualGenotypeCompatibleWithGenotype(Person p, int marker_index, Genotype genotype)
{
    return Genotype_AreGenotypesCompatible(Person_GetMarkerGenotype(p, marker_index), genotype);
}

void Person_SetAffected(Person p)
{
    p->affection_status = AFFECTED;
}

void Person_SetUnaffected(Person p)
{
    p->affection_status = UNAFFECTED;
}

int Person_IsAffected(Person p)
{
    return p->affection_status == AFFECTED;
}

int Person_IsAffectionStatusKnown(Person p)
{
    return p->affection_status != NA;
}

void Person_SetMale(Person p)
{
    p->sex = MALE;
}

void Person_SetFemale(Person p)
{
    p->sex = FEMALE;
}

int Person_IsMale(Person p)
{
    return p->sex == MALE;
}

int Person_IsSexKnown(Person p)
{
    return p->sex != NA;
}

void Person_SetPaternalAlleleOffset(Person p, int offset)
{
    p->paternal_allele_offset = offset;
}

int Person_GetPaternalAlleleOffset(Person p)
{
    return p->paternal_allele_offset;
}

int Person_HasPaternalAlleleOffset(Person p)
{
    return p->paternal_allele_offset >= 0;
}

void Person_SetMaternalAlleleOffset(Person p, int offset)
{
    p->maternal_allele_offset = offset;
}

int Person_GetMaternalAlleleOffset(Person p)
{
    return p->maternal_allele_offset;
}

int Person_HasMaternalAlleleOffset(Person p)
{
    return p->maternal_allele_offset >= 0;
}

void Person_SetMaternalAlleleOrigin(Person p, AlleleOrigin origin)
{
    p->origin_of_maternal_allele = origin;
}

void Person_SetPaternalAlleleOrigin(Person p, AlleleOrigin origin)
{
    p->origin_of_paternal_allele = origin;
}

AlleleOrigin Person_GetMaternalAlleleOrigin(Person p)
{
    return p->origin_of_maternal_allele;
}

AlleleOrigin Person_GetPaternalAlleleOrigin(Person p)
{
    return p->origin_of_paternal_allele;
}

const char *Person_GetDescription(Person p)
{
    static char description[120];
    static char genotypes[100];
    int number_of_markers = p->number_of_markers;

    for (int marker_index = 0; marker_index < number_of_markers; marker_index++) {
        int a1 = p->allele1[marker_index];
        int a2 = p->allele2[marker_index];
        int width = 3;
        snprintf(genotypes + marker_index * width, 100 - marker_index * width, " %d%d", a1, a2);
    }

    snprintf(description, 120, "%s f %s m %s %c %c %c%c:%s",
        p->id, p->father_id, p->mother_id,
        (p->sex == MALE) ? 'M' :
        (p->sex == FEMALE) ? 'F' : '?',
        (p->affection_status == AFFECTED) ? 'A' :
        (p->affection_status == UNAFFECTED) ? 'U' : '?',
        (p->paternal_allele_offset != -1) ? '0' + p->paternal_allele_offset : '?',
        (p->maternal_allele_offset != -1) ? '0' + p->maternal_allele_offset : '?',
        genotypes);

    return description;
}

// Note: Unlike Person_SetMarkerAlleles this function does not convert
// between allele codings.  Whatever alleles are passed will be
// assigned.
void Person_SetMarkerAlleles2(Person person, const int *alleles, int number_of_markers)
{
    person->allele1 = Memory_Malloc(number_of_markers * sizeof(int));
    person->allele2 = Memory_Malloc(number_of_markers * sizeof(int));
    for (int i = 0; i < number_of_markers; i++) {
        person->allele1[i] = alleles[2*i];
        person->allele2[i] = alleles[2*i + 1];
    }
    person->allele_capacity = number_of_markers;
    person->number_of_markers = number_of_markers;
    if (g_jla_p_value) {
        allocate_missing_genotypes(person, number_of_markers);
    }
}

int Person_AreActualGenotypesCompatibleWithGenotypes(Vector persons, int marker, PhasedGenotypes genotypes)
{
    int i, number_of_persons = Vector_GetSize(persons);

    for (i = 0; i < number_of_persons; i++) {
        Person person = Vector_Get(persons, i);
        if (!Person_IsActualGenotypeCompatibleWithGenotype(person, marker, PhasedGenotypes_Get(genotypes, i)))
            return 0;
    }
    return 1;
}

void Person_Show(Person person)
{
    fprintf(stderr, "%s\n", Person_GetDescription(person));
}

int Person_HasPaternalBitFixed(Person person)
{
    return person->paternal_bit_fixed;
}

int Person_HasMaternalBitFixed(Person person)
{
    return person->maternal_bit_fixed;
}

void Person_FixPaternalBit(Person person)
{
    person->paternal_bit_fixed = 1;
}

void Person_FixMaternalBit(Person person)
{
    person->maternal_bit_fixed = 1;
}

void Person_SetSimulatedMarkerAllele1(Person person, int marker_index, int allele)
{
    person->allele1[marker_index] = allele;
}

void Person_SetSimulatedMarkerAllele2(Person person, int marker_index, int allele)
{
    person->allele2[marker_index] = allele;
}

int Person_IsSimulated(Person person)
{
    return person->simulated;
}

int Person_SimulatedParents(Person person)
{
    return Person_IsSimulated(person->father) && Person_IsSimulated(person->mother);
}

void Person_SetSimulated(Person person)
{
    person->simulated = 1;
}

void Person_SetNotSimulated(Person person)
{
    person->simulated = 0;
}

void Person_MarkGenotypeAsMissing(Person person, int marker_index)
{
    IntSet_Put(person->missing_genotypes, marker_index);
}

int Person_IsGenotypeMissing(Person person, int marker_index)
{
    return IntSet_Contains(person->missing_genotypes, marker_index);
}

void allocate_missing_genotypes(Person person, int number_of_markers)
{
    if (person->missing_genotypes) {
        return;
    }
    person->missing_genotypes = IntSet_Create(number_of_markers / 100 < 3 ? 3 : number_of_markers / 100);
}
