#include "IO.h"
#include "Marker.h"
#include "Memory.h"
#include "StringHashTable.h"
#include "UserLinkageMarkerFile.h"
#include <ctype.h>
#include <stdlib.h>

struct UserLinkageMarkerFileStruct {
    int number_of_markers;
    Vector marker_names;
    StringHashTable marker_set;
};

static void read_markers(UserLinkageMarkerFile, FILE *);
static void insert_nul_byte_after_first_word(char *s);
static void create_set_of_markers(UserLinkageMarkerFile);

UserLinkageMarkerFile UserLinkageMarkerFile_Create(const char *filename)
{
    UserLinkageMarkerFile user_linkage_file = Memory_Malloc(sizeof *user_linkage_file);
    FILE *fp = fopen(filename, "rb");
    if (fp == NULL) {
        Memory_Free(user_linkage_file);
        return NULL;
    }
    read_markers(user_linkage_file, fp);
    fclose(fp);
    create_set_of_markers(user_linkage_file);
    return user_linkage_file;
}

static void read_markers(UserLinkageMarkerFile user_linkage_file, FILE *fp)
{
    Vector marker_names = Vector_CreateEmpty();
    char *marker_name = NULL;
    while ((marker_name = IO_GetLine(fp)) != NULL) {
        insert_nul_byte_after_first_word(marker_name);
        Vector_PushBack(marker_names, marker_name);
    }
    user_linkage_file->marker_names = marker_names;
}

static void insert_nul_byte_after_first_word(char *s)
{
    while (*s) {
        if (isspace(*s)) {
            *s = '\0';
            return;
        }
        ++s;
    }
}

static void create_set_of_markers(UserLinkageMarkerFile user_linkage_file)
{
    int number_of_markers = Vector_GetSize(user_linkage_file->marker_names);
    int size = (number_of_markers == 1) ? 1 : number_of_markers / 2;
    user_linkage_file->marker_set = StringHashTable_Create(size);
    for (int i = 0; i < number_of_markers; i++) {
        StringHashTable_Put(user_linkage_file->marker_set, Vector_Get(user_linkage_file->marker_names, i), NULL);
    }
}

Vector UserLinkageMarkerFile_GetMarkerNames(UserLinkageMarkerFile user_linkage_file)
{
    return user_linkage_file->marker_names;
}

int UserLinkageMarkerFile_Contains(UserLinkageMarkerFile user_linkage_file, const char *marker_name)
{
    void *p;
    return StringHashTable_Get(user_linkage_file->marker_set, marker_name, &p);
}

int UserLinkageMarkerFile_CountLinkageMarkers(UserLinkageMarkerFile user_linkage_file, Vector markers)
{
    int number_of_linkage_markers = 0;
    for (int i = 0; i < Vector_GetSize(markers); i++) {
        Marker marker = Vector_Get(markers, i);
        if (UserLinkageMarkerFile_Contains(user_linkage_file, Marker_GetName(marker))) {
            number_of_linkage_markers++;
        }
    }
    return number_of_linkage_markers;
}

void UserLinkageMarkerFile_Destroy(UserLinkageMarkerFile user_linkage_file)
{
    for (int i = 0; i < Vector_GetSize(user_linkage_file->marker_names); i++) {
        Memory_Free(Vector_Get(user_linkage_file->marker_names, i));
    }
    Vector_Destroy(user_linkage_file->marker_names);
    StringHashTable_Destroy(user_linkage_file->marker_set);
    Memory_Free(user_linkage_file);
}
