/* GENEHUNTER - Kruglyak, Daly, Reeve-Daly, and Lander
   Whitehead Institute for Biomedical Research, Copyright 1996
   modified by Konstantin Strauch, Johannes Dietter, Manuel Mattheisen,
   and Markus Brugger.    Copyright (c) 1999-2019 */
/******************************************************************************

 #    #    ##       #    #    #           ####
 ##  ##   #  #      #    ##   #          #    #
 # ## #  #    #     #    # #  #          #
 #    #  ######     #    #  # #   ###    #
 #    #  #    #     #    #   ##   ###    #    #
 #    #  #    #     #    #    #   ###     ####

******************************************************************************/
/* pirated from MAPMAKER...call shell for GENEHUNTER code */

#include <getopt.h>  // getopt_long
#include <stdio.h>
#include <stdlib.h>
#ifndef _XOPEN_SOURCE
#define _XOPEN_SOURCE 500   # Feature Test Macro for strdup (see strdup(3))
#endif
#include <string.h>

#define INC_LIB
#define INC_SHELL
#include "npl.h"
void setup_commands(void), init_variables(void), make_help_entries(void);
/* extern char help_filename[];*/

static void usage(void)
{
    fprintf(stderr,
        "Usage: ghm [OPTIONS]\n"
        "\n"
        "-c, --commands FILE\n"
        "    path to commands file\n"
        "\n"
        "-d, --dat FILE\n"
        "    path to dat file\n"
        "\n"
        "-h, --help\n"
        "    show this help message\n"
        "\n"
        "--resume\n"
        "    resume JLA p value simulations\n"
        "\n"
        "--tag LABEL\n"
        "    label to annotate ghm runs\n");
}

static void use_commands_file_as_stdin(const char *filename)
{
    fclose(stdin);
    stdin = fopen(filename, "rb");
    if (stdin == NULL) {
        fprintf(stderr, "Failed to open file: %s\n", filename);
        usage();
        exit(EXIT_FAILURE);
    }
    g_commands_file = strdup(filename);
}

int main(int argc, char **argv)
{
    save_absolute_path_to_ghm_executable(argv[0], &g_ghm_executable);

    int c;
    while (1) {
        int option_index = 0;
        static struct option long_options[] = {
            {"commands", required_argument, 0, 'c'},
            {"dat",      required_argument, 0, 'd'},
            {"help",     no_argument,       0, 'h'},
            {"resume",   no_argument,       0,  0 },
            {"tag",      required_argument, 0, 't'},
            { 0,         0,                 0,  0 }
        };
        c = getopt_long(argc, argv, "c:d:h",
            long_options, &option_index);
        if (c == -1) {
            break;
        }
        switch (c) {
        case 0:
            if (strcmp("resume", long_options[option_index].name) == 0) {
                g_jla_resume = 1;
            }
            break;
        case 'c':  // --commands filename
            use_commands_file_as_stdin(optarg);
            break;
        case 'd':  // --dat filename
            // From here on, g_jla_datfile is used both for its value as a
            // filename and as a flag indicating that g_jla_datfile should
            // be used instead of g_linkage_marker_file as specified in
            // the commands file.  Because of this double purpose of
            // g_jla_datfile we cannot merge g_jla_datfile and
            // g_linkage_marker_file into a single variable.
            g_jla_datfile = strdup(optarg);
            break;
        case 't':  // --tag label
            g_tag = strdup(optarg);
            break;
        case 'h':
        default:
            usage();
            exit(EXIT_SUCCESS);
        }
    }
    if (optind < argc) {
        fprintf(stderr, "Found unused command-line arguments:");
        while (optind < argc) {
            fprintf(stderr, " %s", argv[optind++]);
        }
        fprintf(stderr, "\n");
        usage();
        exit(EXIT_SUCCESS);
    }

    char help_filename[PATH_LENGTH+1];
    custom_lib_init();
    tty_hello();
    seedrand(RANDOM);

    strcpy(help_filename,"ghm.help"); 
    shell_init("GENEHUNTER-MODSCORE - A modified version of GENEHUNTER",
      "4.0 - Joint Linkage and Association Analysis",
      "1996,1998,2000,2001,2002,2003,2004,2005,2006,2007,2008,2009,2010,2015,2019,2023",
	       help_filename);
    banner();

    photo_banner_hook= NULL;
    quit_save_hook= NULL;

    setup_commands();
    init_variables(); /* startup values for global variables - below */

    make_help_entries();

    /***** place any init calls here *****/
    /***** once in command loop, you will be in the shell *****/

    command_loop();
    exit_main();
}


void setup_commands(void)
{

    /*  command-name	             Abbrev	function	   	type  
	1234567890123456789012345    123	12345678901234...  	123
	------------------------- sp ---  tab   --------------	  tab   --- 
    mc("test command",              "tc",	test_command_code,	CMD);*/

    mc("quit",                       "q",	quit,            	CMD);
    mc("photo",                      "",	do_photo,        	CMD);
    mc("system",                     "",	system_command,  	CMD);
    mc("change directory",           "cd",	cd_command,      	CMD);
    mc("run",                        "",	run_from_file,   	CMD);
    mc("time",                       "",	show_time,       	CMD);
    mc("help",                       "?",       help,                   CMD);
    mc("display scores",             "",        set_display_scores,     CMD);
    mc("#",                          "",        comment,                CMD);

    /* put command calls here */

    mc("load markers",               "",      load_markers,             CMD);
    mc("scan pedigrees",             "",      pedigree_prep,            CMD);

    mc("increment",                  "",      set_scan_increment,       CMD);
/*  mc("increment male",             "",      set_scan_increment_m,     CMD);*/
/*  mc("increment female",           "",      set_scan_increment_f,     CMD);*/
    mc("use",                        "",      set_use_map,              CMD);
    mc("units",                      "",      set_units,                CMD);
    mc("map function",               "",      set_map_function,         CMD);
    mc("off end",                    "",      set_off_end_distance,     CMD);
/*  mc("off end male",               "",      set_off_end_distance_m,   CMD);*/
/*  mc("off end female",             "",      set_off_end_distance_f,   CMD);*/

    mc("total stat",                 "",      show_total_stat,          CMD);
    mc("discard",                    "",      set_discard,              CMD);
    mc("score",                      "",      set_score_function,       CMD);
    mc("analysis",                   "",      set_analysis_type,        CMD);
    mc("count recs",                 "",      set_count_recs,           CMD);
    mc("postscript output",          "ps",    set_postscript_output,    CMD);
    mc("haplotype",                  "",      set_haplotype,            CMD);
    mc("haplotype method",           "",      set_haplomethod,          CMD);
    mc("letters",                    "",      set_letters,              CMD);
/*  mc("haplotype ratio",            "hr",    set_haplo_ratio,          CMD);*/
    mc("dump requirements",          "",      set_dump_req,             CMD); /*** new ***/
    mc("pairs used",                 "",      pair_usage,               CMD); /*** new ***/
    mc("dump ibd",                   "",      dump_ibd,                 CMD); /*** new ***/
    mc("exclude",                    "",      exclude_switch,           CMD); /*** new ***/
    mc("estimate",                   "",      estimate_switch,          CMD); /*** new ***/
    mc("haseman elston",             "",      he_switch,                CMD); /*** new ***/
    mc("ml variance",                "",      ml_variance,              CMD); /*** new ***/
    mc("no dom var",                 "",      no_dom_var,               CMD); /*** new ***/
    mc("nonparametric",              "",      nonparametric,            CMD); /*** new ***/
    mc("variance components",        "",      var_com_switch,           CMD); /*** new ***/
    mc("set starting values",        "",      set_start_values,         CMD); /*** new ***/
    mc("means by sex",               "",      set_num_means,            CMD); /*** new ***/

    mc("dump peak",                  "dp",    dump_peak_info,           CMD);
/*  mc("two locus test",             "",      two_locus_test,           CMD);*/

    mc("drawing scale",              "ds",    set_postscript_scaling,   CMD);
    mc("single point",               "",      set_single_point_mode,    CMD);
    mc("max bits",                   "mb",    set_max_bits,             CMD);
    mc("skip large",                 "",      set_skip_large,           CMD);

    mc("simulate",                   "",      simulation,               CMD);

    mc("tdt",                        "",      tdt_prep,                 CMD);
    mc("tdt2",                       "",      tdt2,                     CMD);
    mc("tdt3",                       "",      tdt3,                     CMD);
    mc("tdt4",                       "",      tdt4,                     CMD);
    mc("tdt5",                       "",      tdt5,                     CMD);
    mc("missing tdt",                "",      set_missing_TDT,          CMD);
    mc("one parent tdt",             "",      set_one_parent_TDT,       CMD);
    mc("perm1",                      "",      perm1,                    CMD);
    mc("perm2",                      "",      perm2,                    CMD);
    mc("pexcess1",                   "px1",   pexcess1,                 CMD);
    mc("pexcess2",                   "px2",   pexcess2,                 CMD);
    mc("pexcess3",                   "px3",   pexcess3,                 CMD);
    mc("dhskip",                     "",      set_dhskip,               CMD);
    mc("qtdt",                       "",      set_qtdt,                 CMD);
    mc("qtest",                      "",      qtdt_test,                CMD);

    mc("compute sharing",            "cs",    set_compute_sharing,      CMD);

    mc("log file",                 "",    do_photo,                   CMD);
    mc("output file",              "",    set_output_file,            CMD);
    mc("joint linkage and association","",set_jla,                    CMD);
    mc("read association regions", "",    read_association_regions,   CMD);
    mc("read linkage markers",     "",    read_user_linkage_marker_file, CMD);
    mc("read association models",  "",    read_model_params_file,     CMD);
    mc("markers per set",          "",    read_number_of_markers_per_test_marker_set, CMD);
    mc("maximum test marker distance", "", read_maximum_test_marker_distance, CMD);
    mc("imprinting",                 "",      set_imprinting,           CMD);
    mc("include untyped",            "iu",    set_include_untyped,      CMD);
    mc("modcalc",                    "mc",    set_modcalc,              CMD);
    mc("positions",                  "",      set_positions,            CMD);
    mc("liability class",            "lc",    set_liability_class,      CMD);
    mc("join liability classes",     "jl",    join_liability_classes,   CMD);
    mc("penetrance restriction",     "pr",   set_penetrance_restriction,CMD);
    mc("allfreq restriction",        "ar",    set_allfreq_restriction,  CMD);
    mc("highest allfreq",            "ha",    set_highest_allfreq,      CMD);
    mc("dimensions",                 "",      set_dimensions,           CMD);
    mc("long mod output",            "lm",    set_long_mod_output,      CMD);
    mc("saved models",               "",      set_saved_models,         CMD);
    mc("model",                      "",      model,                    CMD);
    mc("modscore",                   "",      modscore,                 CMD);
    mc("title",                      "",      set_user_title,           CMD);
    mc("read map",                   "",      use_pub_map,              CMD);

    mc("maximization",               "",      set_maximization,         CMD);
    mc("normalize allfreq",          "naf",   set_norm_allfreq,         CMD);
    mc("calculate p value",          "cpv",   pmod_calc,                CMD);
    mc("number of replicates",       "nor",   set_number_of_replicates, CMD);
    mc("sequential simulation",      "seq",   set_sequential_simulation,CMD);
    mc("store replicates",           "str",   set_store_replicates,     CMD);
    mc("untyped founders",           "ufo",   set_untyped_founders,     CMD);
    mc("full information",           "fin",   set_marker_information,   CMD);
    mc("simulate untyped",           "sun",   set_simulate_untyped,     CMD);
    mc("best position",              "bep",   set_best_position,        CMD);
    mc("long p value output",        "lpo",   set_debug_out_pv,         CMD);
    mc("set random seed",            "srs",   set_random_seed,          CMD);
    mc("show distribution",          "sdi",   set_distribution,         CMD);
    mc("algebraic calculation",      "alg",   algebraic_calculation,    CMD);
    mc("mobit simulation",           "mob",   mobit_simulation,         CMD);
    mc("allele frequency estimation", "",     set_allele_frequency_estimation, CMD);
    mc("save allele frequencies",    "",      save_allele_frequencies,  CMD);
    mc("jla test markers",           "",      read_test_markers,        CMD);
    mc("jla delta iterations",       "",      set_jla_delta_iterations, CMD);
    mc("jla hapfreq iterations",     "",      set_jla_hapfreq_iterations, CMD);
    mc("jla nlopt iterations",       "",      set_jla_nlopt_iterations, CMD);
    mc("jla delta modscore",         "",      set_jla_delta_modscore,   CMD);
    // Simulation-specific
    mc("jla p value",                "",      calculate_jla_p_value,    CMD);
    mc("jla replicates",             "",      set_jla_replicates,       CMD);
    mc("jla child processes",        "",      set_jla_child_processes,  CMD);
    mc("jla directory",              "",      set_jla_directory,        CMD);
    mc("jla store replicates",       "",      set_jla_store_replicates, CMD);
    mc("jla debug replicates",       "",      set_jla_debug_replicates, CMD);
    mc("timeout single modscore",    "",      set_timeout_single_modscore,CMD);
}
