#include "ParseObject.h"
#include "Memory.h"
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <errno.h>

struct ParseObjectStruct {
    const char *cur;
    const char *beg;
    const char *end;
};

static const char *skip_space(const char *s);
static const char *skip_nondigit(const char *s);
static char *copy_from_up_to(const char *from, const char *to);

ParseObject ParseObject_Create(const char *s)
{
    ParseObject parse_object;

    if ((parse_object = (ParseObject) Memory_Malloc(sizeof *parse_object)) == NULL)
        return NULL;

    s = skip_space(s);
    parse_object->beg = parse_object->cur = s;
    parse_object->end = s + strlen(s);

    return parse_object;
}

static const char *skip_space(const char *s)
{
    while (*s && isspace(*s))
        ++s;

    return s;
}

int ParseObject_NextInteger(ParseObject parse_object, long *result)
{
    long converted;
    int status = 1;
    const char *rest;

    errno = 0;
    converted = strtol(parse_object->cur, (char **) &rest, 10);

    if (parse_object->cur == rest && converted == 0L) {  // found no digits
        rest = skip_nondigit(skip_space(rest));
        status = 0;
    }
    else if (errno == ERANGE && (converted == LONG_MIN || converted == LONG_MAX))
        status = 0;

    *result = converted;
    parse_object->cur = rest;

    return status;
}

static const char *skip_nondigit(const char *s)
{
    while (*s && !isdigit(*s))
        ++s;

    return s;
}

int ParseObject_HasNextInteger(ParseObject parse_object)
{
    const char *s = parse_object->cur;

    s = skip_space(s);

    return isdigit(*s)
        || (*s == '-' && isdigit(*(s + 1)))
        || (*s == '+' && isdigit(*(s + 1)));
}

void ParseObject_Destroy(ParseObject parse_object)
{
    Memory_Free(parse_object);
}

int ParseObject_NextDouble(ParseObject parse_object, double *result)
{
    int status = 1;
    const char *rest;

    errno = 0;
    double converted = strtod(parse_object->cur, (char **) &rest);

    if (parse_object->cur == rest && converted == 0) {  // found no digits
        status = 0;
    }
    // else if (errno == ERANGE && (converted == double_MIN || converted == double_MAX))
    //     status = 0;

    *result = converted;
    parse_object->cur = rest;

    return status;
}

int ParseObject_NextWord(ParseObject parse_object, char **result)
{
    int status = 1;

    parse_object->cur = skip_space(parse_object->cur);

    const char *s = parse_object->cur;
    while (*s && !isspace(*s))
        ++s;
    if (s == parse_object->cur) {
        status = 0;
        *result = NULL;
    }
    else {
        *result = copy_from_up_to(parse_object->cur, s);
        parse_object->cur = s;
    }

    return status;
}

static char *copy_from_up_to(const char *from, const char *to)
{
    int len = to - from;
    char *buffer = (char *) Memory_Malloc(len + 1);
    strncpy(buffer, from, len);
    buffer[len] = '\0';

    return buffer;
}

Vector ParseObject_SplitIntoWords(ParseObject parse_object)
{
    Vector words = Vector_CreateEmpty();
    char *word;

    while (ParseObject_NextWord(parse_object, &word))
        Vector_PushBack(words, word);

    return words;
}
