#include "Vector.h"
#include "Memory.h"
#include "MathFunctions.h"
#include <stddef.h>
#include <stdarg.h>
#include <string.h>

struct VectorStruct {
    void **store;
    int size;
    int capacity;
};

static Vector collect(Vector vector, Vector_CollectFunction fn, void *additional_arguments, int safe);
void compute_product_of_doubles(void *result_so_far, void *next_element);

Vector Vector_CreateEmpty(void)
{
    Vector vector = (Vector) Memory_Malloc(sizeof(*vector));
    vector->size = vector->capacity = 0;
    vector->store = NULL;

    return vector;
}

Vector Vector_Create(int initial_capacity)
{
    Vector vector = Vector_CreateEmpty();
    if (initial_capacity) {
        vector->store = Memory_Calloc(initial_capacity, sizeof(void *));
        vector->capacity = initial_capacity;
    }
    return vector;
}

int Vector_GetCapacity(Vector vector)
{
    return vector->capacity;
}

void Vector_Destroy(Vector vector)
{
    if (vector->store) {
        Memory_Free(vector->store);
        vector->store = NULL;
    }
    Memory_Free(vector);
}

int Vector_GetSize(Vector vector)
{
    return vector ? vector->size : -1;
}

void Vector_PushBack(Vector vector, void *item)
{
    if (vector->store == NULL) {
        vector->store = (void **) Memory_Malloc(sizeof(void *));
        vector->capacity = 1;
    }
    if (vector->size == vector->capacity) {
        void **p = (void **) Memory_Realloc(vector->store, 2 * vector->capacity * sizeof(void *));
        vector->store = p;
        vector->capacity *= 2;
    }
    vector->store[vector->size++] = item;
}

void Vector_Set(Vector vector, int i, void *item)
{
    if (i >= vector->size) {
        vector->size = i + 1;
    }
    vector->store[i] = item;
}

void *Vector_Get(Vector vector, int i)
{
    return vector->store[i];
}

void **Vector_GetInternalArray(Vector vector)
{
    return vector->store;
}

void Vector_Reverse(Vector vector)
{
    for (int left = 0, right = Vector_GetSize(vector) - 1;
            left < right; left++, right--)
    {
        void *tmp = vector->store[left];
        vector->store[left] = vector->store[right];
        vector->store[right] = tmp;
    }
}

int Vector_Equals(Vector v1, Vector v2, EqualsFunction equals)
{
    if (Vector_GetSize(v1) != Vector_GetSize(v2))
        return 0;
    for (int i = 0; i < Vector_GetSize(v1); i++)
        if (!equals(Vector_Get(v1, i), Vector_Get(v2, i)))
            return 0;
    return 1;
}

Vector Vector_Copy(Vector vector)
{
    int number_of_elements = Vector_GetSize(vector);
    Vector copy = Vector_Create(number_of_elements);
    for (int i = 0; i < number_of_elements; i++)
        Vector_PushBack(copy, Vector_Get(vector, i));
    return copy;
}

Vector Vector_FromArray(void *array, int number_of_elements, int size_of_elements)
{
    Vector vector = Vector_Create(number_of_elements);
    for (int i = 0; i < number_of_elements; i++)
        Vector_PushBack(vector, ((char *) array) + i * size_of_elements);
    return vector;
}

Vector Vector_FromCopyOfArray(void *array, int number_of_elements, int size_of_elements)
{
    void *copy_of_array = Memory_Malloc(number_of_elements * size_of_elements);
    memmove(copy_of_array, array, number_of_elements * size_of_elements);
    return Vector_FromArray(copy_of_array, number_of_elements, size_of_elements);
}

Vector Vector_FindCombinations(Vector vectors)
{
    int number_of_vectors = Vector_GetSize(vectors);
    int *indices = Memory_Malloc(number_of_vectors * sizeof(int));
    for (int i = 0; i < number_of_vectors; i++)
        indices[i] = 0;
    Vector combinations = Vector_CreateEmpty();
    do {
        Vector combination = Vector_Create(number_of_vectors);
        for (int i = 0; i < number_of_vectors; i++)
            Vector_PushBack(combination, Vector_Get(Vector_Get(vectors, i), indices[i]));
        Vector_PushBack(combinations, combination);
    } while (Vector_IncrementIndices(indices, vectors));
    return combinations;
}

int Vector_IncrementIndices(int *indices, Vector vectors)
{
    for (int i = 0; i < Vector_GetSize(vectors); i++) {
        ++indices[i];
        if (indices[i] < Vector_GetSize(Vector_Get(vectors, i)))
            break;
        if (i != Vector_GetSize(vectors) - 1)
            indices[i] = 0;
        else
            return 0;
    }
    return 1;
}

Vector Vector_With(int number_of_elements, ...)
{
    Vector vector = Vector_Create(number_of_elements);
    va_list ap;
    va_start(ap, number_of_elements);
    while (number_of_elements--)
        Vector_PushBack(vector, va_arg(ap, void *));
    va_end(ap);
    return vector;
}

Vector Vector_SafeCollect(Vector vector, Vector_CollectFunction fn, void *additional_arguments)
{
    return collect(vector, fn, additional_arguments, 1);
}

Vector Vector_Collect(Vector vector, Vector_CollectFunction fn, void *additional_arguments)
{
    return collect(vector, fn, additional_arguments, 0);
}

static Vector collect(Vector vector, Vector_CollectFunction fn, void *additional_arguments, int safe)
{
    Vector result = Vector_Create(Vector_GetSize(vector));
    for (int i = 0; i < Vector_GetSize(vector); i++) {
        void *old = Vector_Get(vector, i);
        void *new = fn(old, additional_arguments);
        if (safe && old == new) {
            Vector_Destroy(result);
            return NULL;
        }
        Vector_PushBack(result, new);
    }
    return result;
}

void Vector_Reduce(Vector vector, Vector_ReduceFunction fn, void *result)
{
    for (int i = 0; i < Vector_GetSize(vector); i++)
        fn(result, Vector_Get(vector, i));
}

double Vector_ProductOfDoubles(Vector vector)
{
    double product = 1;
    Vector_Reduce(vector, compute_product_of_doubles, &product);
    return product;
}

void compute_product_of_doubles(void *result_so_far, void *next_element)
{
    double *product = result_so_far;
    double *factor = next_element;
    *product *= *factor;
}
