#include "dSFMT.h"

#include "DoubleVector.h"
#include "LinkageMarkerFile.h"
#include "Marker.h"
#include "Memory.h"
#include "NloptModScoreFunction.h"

#include <float.h>  // DBL_EPSILON
#include <math.h>   // HUGE_VAL
#include <nlopt.h>
#include <stdio.h>
#include <unistd.h>  // usleep

#include <pthread.h>

// Legacy includes.
#define INC_LIB
#define INC_SHELL
#include "npl.h"

#define MIN(x, y) ((x) < (y) ? (x) : (y))

typedef int OptimizeFunction(
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModScore *grid_mod_score,
    ModScore *mod_score,
    double initial_step_size1);

typedef int OptimizeFunctionUserModel(
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModelParams user_model,
    ModScore *grid_mod_score,
    ModScore *mod_score,
    double initial_step_size1);

typedef struct pthread_data_struct {
    OptimizeFunction *optimize;
    Vector annotated_families;
    Vector test_markers;
    DoubleVector em_hapfreqs;
    double position;
    ModScore *grid_mod_score;
    ModScore *nlopt_mod_score;
    double steps;
    int ok;
    int returned;
} *p_data;

typedef struct pthread_user_model_data_struct {
    OptimizeFunctionUserModel *optimize;
    Vector annotated_families;
    Vector test_markers;
    DoubleVector em_hapfreqs;
    double position;
    ModelParams user_model;
    ModScore *grid_mod_score;
    ModScore *nlopt_mod_score;
    double steps;
    int ok;
    int returned;
} *p_data_user_model;

const double FREQ_TOL = 1e-8;
const double PEN_TOL = DBL_EPSILON;

// Number of calls from NLopt to the objective function.
static unsigned long objective_function_calls = 0;

// Due to a bug in the COBYLA algorithm which sometimes results in
// the result of the last evaluation of the target function being
// returned as the "global" minimum, we have to keep track of the
// global minimum ourselves.
static double global_minimum = HUGE_VAL;
static double global_x[20+1] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
// +1 for imprinting
// Without deltas:
// 16 hap_freqs, 1 p, 3+1 pens
// With deltas:
// 8 deltas, 1 p, 3+1 pens

// ===================================================================
// Optimize, target, and constraint functions for N = 1 test markers
// ===================================================================

typedef struct DataStruct {
    Vector annotated_families;
    Vector test_markers;
    DoubleVector em_hapfreqs;
    ModelParams user_model;
} *Data;


// Lower individual constraint
// max(-EMH[i-1] * p, EMH[i-1] * (1-p) - 1) - d[i-1] <= 0
static double d_lower_indiv_constr(int i, unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    double di = x[i-1];
    Data d = data;
    int number_of_ds = DoubleVector_GetSize(d->em_hapfreqs);
    double p = x[number_of_ds];
    double EMH = DoubleVector_Get(d->em_hapfreqs, i-1);
    double term1 = -EMH * p;
    double term2 = EMH * (1 - p) - 1;
    return (term1 > term2 ? term1 : term2) - di;
}

// Lower individual constraint
// max(-EMH[i-1] * p, EMH[i-1] * (1-p) - 1) - d[i-1] <= 0
static double d_user_model_lower_indiv_constr(int i, unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    double di = x[i-1];
    Data d = data;
    double p = d->user_model->disease_allele_frequency;
    double EMH = DoubleVector_Get(d->em_hapfreqs, i-1);
    double term1 = -EMH * p;
    double term2 = EMH * (1 - p) - 1;
    return (term1 > term2 ? term1 : term2) - di;
}

// Lower individual constraint for d2 in 2 test marker case
// max(-EMH[1] * p, EMH[1] * (1-p) - 1) - d2 <= 0
// static double d2_user_model_lower_indiv_constr_2_tm(unsigned n, const double *x, double *grad, void *data) {
//     (void) n;
//     (void) grad;
//     double d1 = x[0];
//     double d2 = -d1;
//     Data d = data;
//     double p = d->user_model->disease_allele_frequency;
//     double EMH = DoubleVector_Get(d->em_hapfreqs, 1);
//     double term1 = -EMH * p;
//     double term2 = EMH * (1 - p) - 1;
//     return (term1 > term2 ? term1 : term2) - d2;
// }

// Upper individual constraint
// d[i-1] - min(1 - EMH[i-1] * p, EMH[i-1] * (1-p)) <= 0
static double d_upper_indiv_constr(int i, unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    double di = x[i-1];
    Data d = data;
    int number_of_ds = DoubleVector_GetSize(d->em_hapfreqs);
    double p = x[number_of_ds];
    double EMH = DoubleVector_Get(d->em_hapfreqs, i-1);
    double term1 = 1 - EMH * p;
    double term2 = EMH * (1 - p);
    return di - (term1 < term2 ? term1 : term2);
}

// Upper individual constraint
// d[i-1] - min(1 - EMH[i-1] * p, EMH[i-1] * (1-p)) <= 0
static double d_user_model_upper_indiv_constr(int i, unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    double di = x[i-1];
    Data d = data;
    double p = d->user_model->disease_allele_frequency;
    double EMH = DoubleVector_Get(d->em_hapfreqs, i-1);
    double term1 = 1 - EMH * p;
    double term2 = EMH * (1 - p);
    return di - (term1 < term2 ? term1 : term2);
}

// Upper individual constraint for d2 in 2 test marker case
// d2 - min(1 - EMH[i-1] * p, EMH[i-1] * (1-p)) <= 0
// static double d2_user_model_upper_indiv_constr_2_tm(unsigned n, const double *x, double *grad, void *data) {
//     (void) n;
//     (void) grad;
//     double d1 = x[0];
//     double d2 = -d1;
//     Data d = data;
//     double p = d->user_model->disease_allele_frequency;
//     double EMH = DoubleVector_Get(d->em_hapfreqs, 1);
//     double term1 = 1 - EMH * p;
//     double term2 = EMH * (1 - p);
//     return d2 - (term1 < term2 ? term1 : term2);
// }

// Penetrance restriction: f0 <= f1_pat/mat <= f2
static double f0_f1_pat_constr_1_tm_pat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[3];
    double f1_pat = x[4];
    return f0 - f1_pat;
}

static double f0_f1_pat_constr_1_tm_mat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[3];
    double f1_pat = x[5];
    return f0 - f1_pat;
}

static double f0_f1_mat_constr_1_tm_pat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[3];
    double f1_mat = x[5];
    return f0 - f1_mat;
}

static double f0_f1_mat_constr_1_tm_mat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[3];
    double f1_mat = x[4];
    return f0 - f1_mat;
}

static double f1_pat_f2_constr_1_tm_pat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1_pat = x[4];
    double f2 = x[6];
    return f1_pat - f2;
}

static double f1_pat_f2_constr_1_tm_mat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1_pat = x[5];
    double f2 = x[6];
    return f1_pat - f2;
}

static double f1_mat_f2_constr_1_tm_pat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1_mat = x[5];
    double f2 = x[6];
    return f1_mat - f2;
}

static double f1_mat_f2_constr_1_tm_mat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1_mat = x[4];
    double f2 = x[6];
    return f1_mat - f2;
}

static double f0_f1_pat_constr_2_tm_pat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[5];
    double f1_pat = x[6];
    return f0 - f1_pat;
}

static double f0_f1_pat_constr_2_tm_mat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[5];
    double f1_pat = x[7];
    return f0 - f1_pat;
}

static double f0_f1_mat_constr_2_tm_pat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[5];
    double f1_mat = x[7];
    return f0 - f1_mat;
}

static double f0_f1_mat_constr_2_tm_mat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[5];
    double f1_mat = x[6];
    return f0 - f1_mat;
}

static double f1_pat_f2_constr_2_tm_pat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1_pat = x[6];
    double f2 = x[8];
    return f1_pat - f2;
}

static double f1_pat_f2_constr_2_tm_mat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1_pat = x[7];
    double f2 = x[8];
    return f1_pat - f2;
}

static double f1_mat_f2_constr_2_tm_pat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1_mat = x[7];
    double f2 = x[8];
    return f1_mat - f2;
}

static double f1_mat_f2_constr_2_tm_mat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1_mat = x[6];
    double f2 = x[8];
    return f1_mat - f2;
}

static double f0_f1_pat_constr_3_tm_pat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[9];
    double f1_pat = x[10];
    return f0 - f1_pat;
}

static double f0_f1_pat_constr_3_tm_mat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[9];
    double f1_pat = x[11];
    return f0 - f1_pat;
}

static double f0_f1_mat_constr_3_tm_pat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[9];
    double f1_mat = x[11];
    return f0 - f1_mat;
}

static double f0_f1_mat_constr_3_tm_mat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[9];
    double f1_mat = x[10];
    return f0 - f1_mat;
}

static double f1_pat_f2_constr_3_tm_pat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1_pat = x[10];
    double f2 = x[12];
    return f1_pat - f2;
}

static double f1_pat_f2_constr_3_tm_mat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1_pat = x[11];
    double f2 = x[12];
    return f1_pat - f2;
}

static double f1_mat_f2_constr_3_tm_pat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1_mat = x[11];
    double f2 = x[12];
    return f1_mat - f2;
}

static double f1_mat_f2_constr_3_tm_mat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1_mat = x[10];
    double f2 = x[12];
    return f1_mat - f2;
}

// Penetrance restriction: f0 <= f1 <= f2
static double f0_f1_constr_1_tm(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[3];
    double f1 = x[4];
    return f0 - f1;
}

static double f1_f2_constr_1_tm(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1 = x[4];
    double f2 = x[5];
    return f1 - f2;
}

static double f0_f1_constr_2_tm(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[5];
    double f1 = x[6];
    return f0 - f1;
}

static double f1_f2_constr_2_tm(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1 = x[6];
    double f2 = x[7];
    return f1 - f2;
}

static double f0_f1_constr_3_tm(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[9];
    double f1 = x[10];
    return f0 - f1;
}

static double f1_f2_constr_3_tm(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1 = x[10];
    double f2 = x[11];
    return f1 - f2;
}

// Penetrance restriction: f0 <= f1_pat/mat <= f2
static double f0_f1_pat_constr_1_tm_pat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[5];
    double f1_pat = x[6];
    return f0 - f1_pat;
}

static double f0_f1_pat_constr_1_tm_mat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[5];
    double f1_pat = x[7];
    return f0 - f1_pat;
}

static double f0_f1_mat_constr_1_tm_pat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[5];
    double f1_mat = x[7];
    return f0 - f1_mat;
}

static double f0_f1_mat_constr_1_tm_mat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[5];
    double f1_mat = x[6];
    return f0 - f1_mat;
}

static double f1_pat_f2_constr_1_tm_pat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1_pat = x[6];
    double f2 = x[8];
    return f1_pat - f2;
}

static double f1_pat_f2_constr_1_tm_mat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1_pat = x[7];
    double f2 = x[8];
    return f1_pat - f2;
}

static double f1_mat_f2_constr_1_tm_pat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1_mat = x[7];
    double f2 = x[8];
    return f1_mat - f2;
}

static double f1_mat_f2_constr_1_tm_mat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1_mat = x[6];
    double f2 = x[8];
    return f1_mat - f2;
}

static double f0_f1_pat_constr_2_tm_pat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[9];
    double f1_pat = x[10];
    return f0 - f1_pat;
}

static double f0_f1_pat_constr_2_tm_mat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[9];
    double f1_pat = x[11];
    return f0 - f1_pat;
}

static double f0_f1_mat_constr_2_tm_pat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[9];
    double f1_mat = x[11];
    return f0 - f1_mat;
}

static double f0_f1_mat_constr_2_tm_mat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[9];
    double f1_mat = x[10];
    return f0 - f1_mat;
}

static double f1_pat_f2_constr_2_tm_pat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1_pat = x[10];
    double f2 = x[12];
    return f1_pat - f2;
}

static double f1_pat_f2_constr_2_tm_mat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1_pat = x[11];
    double f2 = x[12];
    return f1_pat - f2;
}

static double f1_mat_f2_constr_2_tm_pat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1_mat = x[11];
    double f2 = x[12];
    return f1_mat - f2;
}

static double f1_mat_f2_constr_2_tm_mat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1_mat = x[10];
    double f2 = x[12];
    return f1_mat - f2;
}

static double f0_f1_pat_constr_3_tm_pat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[17];
    double f1_pat = x[18];
    return f0 - f1_pat;
}

static double f0_f1_pat_constr_3_tm_mat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[17];
    double f1_pat = x[19];
    return f0 - f1_pat;
}

static double f0_f1_mat_constr_3_tm_pat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[17];
    double f1_mat = x[19];
    return f0 - f1_mat;
}

static double f0_f1_mat_constr_3_tm_mat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[17];
    double f1_mat = x[18];
    return f0 - f1_mat;
}

static double f1_pat_f2_constr_3_tm_pat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1_pat = x[18];
    double f2 = x[20];
    return f1_pat - f2;
}

static double f1_pat_f2_constr_3_tm_mat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1_pat = x[19];
    double f2 = x[20];
    return f1_pat - f2;
}

static double f1_mat_f2_constr_3_tm_pat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1_mat = x[19];
    double f2 = x[20];
    return f1_mat - f2;
}

static double f1_mat_f2_constr_3_tm_mat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1_mat = x[18];
    double f2 = x[20];
    return f1_mat - f2;
}

// Penetrance restriction: f0 <= f1 <= f2
static double f0_f1_constr_1_tm_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[5];
    double f1 = x[6];
    return f0 - f1;
}

static double f1_f2_constr_1_tm_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1 = x[6];
    double f2 = x[7];
    return f1 - f2;
}

static double f0_f1_constr_2_tm_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[9];
    double f1 = x[10];
    return f0 - f1;
}

static double f1_f2_constr_2_tm_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1 = x[10];
    double f2 = x[11];
    return f1 - f2;
}

static double f0_f1_constr_3_tm_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f0 = x[17];
    double f1 = x[18];
    return f0 - f1;
}

static double f1_f2_constr_3_tm_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double f1 = x[18];
    double f2 = x[19];
    return f1 - f2;
}

static double d1_lower_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_lower_indiv_constr(1, n, x, grad, data);
}

static double d2_lower_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_lower_indiv_constr(2, n, x, grad, data);
}

static double d3_lower_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_lower_indiv_constr(3, n, x, grad, data);
}

static double d4_lower_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_lower_indiv_constr(4, n, x, grad, data);
}

static double d5_lower_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_lower_indiv_constr(5, n, x, grad, data);
}

static double d6_lower_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_lower_indiv_constr(6, n, x, grad, data);
}

static double d7_lower_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_lower_indiv_constr(7, n, x, grad, data);
}

static double d8_lower_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_lower_indiv_constr(8, n, x, grad, data);
}

static double d1_upper_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_upper_indiv_constr(1, n, x, grad, data);
}

static double d2_upper_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_upper_indiv_constr(2, n, x, grad, data);
}

static double d3_upper_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_upper_indiv_constr(3, n, x, grad, data);
}

static double d4_upper_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_upper_indiv_constr(4, n, x, grad, data);
}

static double d5_upper_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_upper_indiv_constr(5, n, x, grad, data);
}

static double d6_upper_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_upper_indiv_constr(6, n, x, grad, data);
}

static double d7_upper_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_upper_indiv_constr(7, n, x, grad, data);
}

static double d8_upper_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_upper_indiv_constr(8, n, x, grad, data);
}

static double d1_user_model_lower_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_user_model_lower_indiv_constr(1, n, x, grad, data);
}

static double d2_user_model_lower_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_user_model_lower_indiv_constr(2, n, x, grad, data);
}

static double d3_user_model_lower_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_user_model_lower_indiv_constr(3, n, x, grad, data);
}

static double d4_user_model_lower_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_user_model_lower_indiv_constr(4, n, x, grad, data);
}

static double d5_user_model_lower_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_user_model_lower_indiv_constr(5, n, x, grad, data);
}

static double d6_user_model_lower_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_user_model_lower_indiv_constr(6, n, x, grad, data);
}

static double d7_user_model_lower_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_user_model_lower_indiv_constr(7, n, x, grad, data);
}

static double d8_user_model_lower_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_user_model_lower_indiv_constr(8, n, x, grad, data);
}

static double d1_user_model_upper_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_user_model_upper_indiv_constr(1, n, x, grad, data);
}

static double d2_user_model_upper_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_user_model_upper_indiv_constr(2, n, x, grad, data);
}

static double d3_user_model_upper_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_user_model_upper_indiv_constr(3, n, x, grad, data);
}

static double d4_user_model_upper_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_user_model_upper_indiv_constr(4, n, x, grad, data);
}

static double d5_user_model_upper_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_user_model_upper_indiv_constr(5, n, x, grad, data);
}

static double d6_user_model_upper_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_user_model_upper_indiv_constr(6, n, x, grad, data);
}

static double d7_user_model_upper_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_user_model_upper_indiv_constr(7, n, x, grad, data);
}

static double d8_user_model_upper_indiv_constr(unsigned n, const double *x, double *grad, void *data) {
    return d_user_model_upper_indiv_constr(8, n, x, grad, data);
}

// Group constraint for ds
// sum of ds = 0   (equality constraint)
static double ds_group_constr_1_tm(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double d1 = x[0];
    double d2 = x[1];
    return d1 + d2;
}

// Group constraint for ds
// sum of ds = 0   (equality constraint)
static double ds_group_constr_2_tm(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double d1 = x[0];
    double d2 = x[1];
    double d3 = x[2];
    double d4 = x[3];
    return d1 + d2 + d3 + d4;
}

// Group constraint for ds
// sum of ds = 0   (equality constraint)
static double ds_group_constr_3_tm(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double d1 = x[0];
    double d2 = x[1];
    double d3 = x[2];
    double d4 = x[3];
    double d5 = x[4];
    double d6 = x[5];
    double d7 = x[6];
    double d8 = x[7];
    return d1 + d2 + d3 + d4 + d5 + d6 + d7 + d8;
}

// Group constraints without deltas: sum H(even) = 1-p and sum H(odd)
// = p The constraint that H(x) in [0, 1] is automatically satisfied.

// Group constraints for H_even and H_odd
static double H_even_constr_1_tm_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double H0 = x[0];
    double H2 = x[2];
    double p = x[4];
    return 1 - p - H0 - H2;
}

static double H_odd_constr_1_tm_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double H1 = x[1];
    double H3 = x[3];
    double p = x[4];
    return p - H1 - H3;
}

static double H_even_user_model_constr_1_tm_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    Data d = data;
    double H0 = x[0];
    double H2 = x[2];
    double p = d->user_model->disease_allele_frequency;
    return 1 - p - H0 - H2;
}

static double H_odd_user_model_constr_1_tm_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    Data d = data;
    double H1 = x[1];
    double H3 = x[3];
    double p = d->user_model->disease_allele_frequency;
    return p - H1 - H3;
}

static double H_even_constr_2_tm_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double H0 = x[0];
    double H2 = x[2];
    double H4 = x[4];
    double H6 = x[6];
    double p = x[8];
    return 1 - p - H0 - H2 - H4 - H6;
}

static double H_odd_constr_2_tm_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double H1 = x[1];
    double H3 = x[3];
    double H5 = x[5];
    double H7 = x[7];
    double p = x[8];
    return p - H1 - H3 - H5 - H7;
}

static double H_even_user_model_constr_2_tm_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    Data d = data;
    double H0 = x[0];
    double H2 = x[2];
    double H4 = x[4];
    double H6 = x[6];
    double p = d->user_model->disease_allele_frequency;
    return 1 - p - H0 - H2 - H4 - H6;
}

static double H_odd_user_model_constr_2_tm_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    Data d = data;
    double H1 = x[1];
    double H3 = x[3];
    double H5 = x[5];
    double H7 = x[7];
    double p = d->user_model->disease_allele_frequency;
    return p - H1 - H3 - H5 - H7;
}

static double H_even_constr_3_tm_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double H0 = x[0];
    double H2 = x[2];
    double H4 = x[4];
    double H6 = x[6];
    double H8 = x[8];
    double H10 = x[10];
    double H12 = x[12];
    double H14 = x[14];
    double p = x[16];
    return 1 - p - H0 - H2 - H4 - H6 - H8 - H10 - H12 - H14;
}

static double H_odd_constr_3_tm_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    (void) data;
    double H1 = x[1];
    double H3 = x[3];
    double H5 = x[5];
    double H7 = x[7];
    double H9 = x[9];
    double H11 = x[11];
    double H13 = x[13];
    double H15 = x[15];
    double p = x[16];
    return p - H1 - H3 - H5 - H7 - H9 - H11 - H13 - H15;
}

static double H_even_user_model_constr_3_tm_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    Data d = data;
    double H0 = x[0];
    double H2 = x[2];
    double H4 = x[4];
    double H6 = x[6];
    double H8 = x[8];
    double H10 = x[10];
    double H12 = x[12];
    double H14 = x[14];
    double p = d->user_model->disease_allele_frequency;
    return 1 - p - H0 - H2 - H4 - H6 - H8 - H10 - H12 - H14;
}

static double H_odd_user_model_constr_3_tm_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    Data d = data;
    double H1 = x[1];
    double H3 = x[3];
    double H5 = x[5];
    double H7 = x[7];
    double H9 = x[9];
    double H11 = x[11];
    double H13 = x[13];
    double H15 = x[15];
    double p = d->user_model->disease_allele_frequency;
    return p - H1 - H3 - H5 - H7 - H9 - H11 - H13 - H15;
}

static void print_maximum_1_tm_pat_first(double maximum, const double *x, DoubleVector em_hapfreqs)
{
    double mod = maximum;
    double d1 = x[0], d2 = x[1];
    double p = x[2];
    double f0, f1_pat, f1_mat, f2;
    double EMH0 = DoubleVector_Get(em_hapfreqs, 0);
    double EMH1 = DoubleVector_Get(em_hapfreqs, 1);
    double H0 = EMH0 * (1 - p) - d1;
    double H1 = EMH0 * p + d1;
    double H2 = EMH1 * (1 - p) - d2;
    double H3 = EMH1 * p + d2;
    if (!imprinting) {
        f0 = x[3]; f1_pat = x[4]; f1_mat = x[4]; f2 = x[5];
    } else {
        f0 = x[3]; f1_pat = x[4]; f1_mat = x[5]; f2 = x[6];
    }
    sf(ps, "   mod     p    F0  F1_P  F1_M    F2    H0    H1    H2    H3  EMH0  EMH1    d1    d2\n");
    pr();
    sf(ps, "%6.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f\n", mod, p, f0, f1_pat, f1_mat, f2, H0, H1, H2, H3, EMH0, EMH1, d1, d2);
    pr();
}

static void print_maximum_1_tm_mat_first(double maximum, const double *x, DoubleVector em_hapfreqs)
{
    double mod = maximum;
    double d1 = x[0], d2 = x[1];
    double p = x[2];
    double f0, f1_pat, f1_mat, f2;
    double EMH0 = DoubleVector_Get(em_hapfreqs, 0);
    double EMH1 = DoubleVector_Get(em_hapfreqs, 1);
    double H0 = EMH0 * (1 - p) - d1;
    double H1 = EMH0 * p + d1;
    double H2 = EMH1 * (1 - p) - d2;
    double H3 = EMH1 * p + d2;
    if (!imprinting) {
        f0 = x[3]; f1_pat = x[4]; f1_mat = x[4]; f2 = x[5];
    } else {
        f0 = x[3]; f1_pat = x[5]; f1_mat = x[4]; f2 = x[6];
    }
    sf(ps, "   mod     p    F0  F1_P  F1_M    F2    H0    H1    H2    H3  EMH0  EMH1    d1    d2\n");
    pr();
    sf(ps, "%6.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f\n", mod, p, f0, f1_pat, f1_mat, f2, H0, H1, H2, H3, EMH0, EMH1, d1, d2);
    pr();
}

static void print_maximum_user_model_1_tm(double maximum, const double *x, DoubleVector em_hapfreqs, ModelParams user_model)
{
    double mod = maximum;
    double d1 = x[0], d2 = x[1];
    double p = user_model->disease_allele_frequency;
    double EMH0 = DoubleVector_Get(em_hapfreqs, 0);
    double EMH1 = DoubleVector_Get(em_hapfreqs, 1);
    double H0 = EMH0 * (1 - p) - d1;
    double H1 = EMH0 * p + d1;
    double H2 = EMH1 * (1 - p) - d2;
    double H3 = EMH1 * p + d2;
    double f0 = user_model->f0, f1_pat = user_model->f1_pat, f1_mat = user_model->f1_mat, f2 = user_model->f2;
    sf(ps, "   mod     p    F0  F1_P  F1_M    F2    H0    H1    H2    H3  EMH0  EMH1    d1    d2\n");
    pr();
    sf(ps, "%6.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f\n", mod, p, f0, f1_pat, f1_mat, f2, H0, H1, H2, H3, EMH0, EMH1, d1, d2);
    pr();
}

static void print_maximum_1_tm_pat_first_without_deltas(double maximum, const double *x, DoubleVector em_hapfreqs)
{
    double mod = maximum;
    double H0 = x[0];
    double H1 = x[1];
    double H2 = x[2];
    double H3 = x[3];
    double p = x[4];
    double f0, f1_pat, f1_mat, f2;
    if (!imprinting) {
        f0 = x[5]; f1_pat = x[6]; f1_mat = x[6]; f2 = x[7];
    } else {
        f0 = x[5]; f1_pat = x[6]; f1_mat = x[7]; f2 = x[8];
    }
    double EMH0 = DoubleVector_Get(em_hapfreqs, 0);
    double EMH1 = DoubleVector_Get(em_hapfreqs, 1);
    sf(ps, "   mod     p    F0  F1_P  F1_M    F2    H0    H1    H2    H3  EMH0  EMH1\n");
    pr();
    sf(ps, "%6.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f\n", mod, p, f0, f1_pat, f1_mat, f2, H0, H1, H2, H3, EMH0, EMH1);
    pr();
}

static void print_maximum_1_tm_mat_first_without_deltas(double maximum, const double *x, DoubleVector em_hapfreqs)
{
    double mod = maximum;
    double H0 = x[0];
    double H1 = x[1];
    double H2 = x[2];
    double H3 = x[3];
    double p = x[4];
    double f0, f1_pat, f1_mat, f2;
    if (!imprinting) {
        f0 = x[5]; f1_pat = x[6]; f1_mat = x[6]; f2 = x[7];
    } else {
        f0 = x[5]; f1_pat = x[7]; f1_mat = x[6]; f2 = x[8];
    }
    double EMH0 = DoubleVector_Get(em_hapfreqs, 0);
    double EMH1 = DoubleVector_Get(em_hapfreqs, 1);
    sf(ps, "   mod     p    F0  F1_P  F1_M    F2    H0    H1    H2    H3  EMH0  EMH1\n");
    pr();
    sf(ps, "%6.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f\n", mod, p, f0, f1_pat, f1_mat, f2, H0, H1, H2, H3, EMH0, EMH1);
    pr();
}

static void print_maximum_user_model_1_tm_without_deltas(double maximum, const double *x, DoubleVector em_hapfreqs, ModelParams user_model)
{
    double mod = maximum;
    double H0 = x[0];
    double H1 = x[1];
    double H2 = x[2];
    double H3 = x[3];
    double p = user_model->disease_allele_frequency;
    double f0 = user_model->f0, f1_pat = user_model->f1_pat, f1_mat = user_model->f1_mat, f2 = user_model->f2;
    double EMH0 = DoubleVector_Get(em_hapfreqs, 0);
    double EMH1 = DoubleVector_Get(em_hapfreqs, 1);
    sf(ps, "   mod     p    F0  F1_P  F1_M    F2    H0    H1    H2    H3  EMH0  EMH1\n");
    pr();
    sf(ps, "%6.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f\n", mod, p, f0, f1_pat, f1_mat, f2, H0, H1, H2, H3, EMH0, EMH1);
    pr();
}

static void update_global_minimum_1_tm_pat_first(double value, const double *x, void *data) {
    int n = 0;
    double *grad = 0;
    if (value >= global_minimum) return;
    if (penetrance_restriction) {
        if (imprinting) {
            if (f0_f1_pat_constr_1_tm_pat_first(n, x, grad, data) > PEN_TOL) return;
            if (f0_f1_mat_constr_1_tm_pat_first(n, x, grad, data) > PEN_TOL) return;
            if (f1_pat_f2_constr_1_tm_pat_first(n, x, grad, data) > PEN_TOL) return;
            if (f1_mat_f2_constr_1_tm_pat_first(n, x, grad, data) > PEN_TOL) return;
        } else {
            if (f0_f1_constr_1_tm(n, x, grad, data) > PEN_TOL) return;
            if (f1_f2_constr_1_tm(n, x, grad, data) > PEN_TOL) return;
        }
    }
    if (fabs(ds_group_constr_1_tm(n, x, grad, data)) > FREQ_TOL) return;
    if (d1_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d1_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d2_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d2_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    global_minimum = value;
    for (int i = 0; i < 6 + imprinting; i++) {
        global_x[i] = x[i];
    }
}

static void update_global_minimum_1_tm_mat_first(double value, const double *x, void *data) {
    int n = 0;
    double *grad = 0;
    if (value >= global_minimum) return;
    if (penetrance_restriction) {
        if (imprinting) {
            if (f0_f1_pat_constr_1_tm_mat_first(n, x, grad, data) > PEN_TOL) return;
            if (f0_f1_mat_constr_1_tm_mat_first(n, x, grad, data) > PEN_TOL) return;
            if (f1_pat_f2_constr_1_tm_mat_first(n, x, grad, data) > PEN_TOL) return;
            if (f1_mat_f2_constr_1_tm_mat_first(n, x, grad, data) > PEN_TOL) return;
        } else {
            if (f0_f1_constr_1_tm(n, x, grad, data) > PEN_TOL) return;
            if (f1_f2_constr_1_tm(n, x, grad, data) > PEN_TOL) return;
        }
    }
    if (fabs(ds_group_constr_1_tm(n, x, grad, data)) > FREQ_TOL) return;
    if (d1_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d1_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d2_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d2_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    global_minimum = value;
    for (int i = 0; i < 6 + imprinting; i++) {
        global_x[i] = x[i];
    }
}

static void update_global_minimum_user_model_1_tm(double value, const double *x, void *data) {
    int n = 0;
    double *grad = 0;
    if (value >= global_minimum) return;
    if (fabs(ds_group_constr_1_tm(n, x, grad, data)) > FREQ_TOL) return;
    if (d1_user_model_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d1_user_model_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d2_user_model_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d2_user_model_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    global_minimum = value;
    for (int i = 0; i < 2; i++) {
        global_x[i] = x[i];
    }
}

static void update_global_minimum_1_tm_pat_first_without_deltas(double value, const double *x, void *data) {
    int n = 0;
    double *grad = 0;
    if (value >= global_minimum) return;
    if (penetrance_restriction) {
        if (imprinting) {
            if (f0_f1_pat_constr_1_tm_pat_first_without_deltas(n, x, grad, data) > PEN_TOL) return;
            if (f0_f1_mat_constr_1_tm_pat_first_without_deltas(n, x, grad, data) > PEN_TOL) return;
            if (f1_pat_f2_constr_1_tm_pat_first_without_deltas(n, x, grad, data) > PEN_TOL) return;
            if (f1_mat_f2_constr_1_tm_pat_first_without_deltas(n, x, grad, data) > PEN_TOL) return;
        } else {
            if (f0_f1_constr_1_tm_without_deltas(n, x, grad, data) > PEN_TOL) return;
            if (f1_f2_constr_1_tm_without_deltas(n, x, grad, data) > PEN_TOL) return;
        }
    }
    if (fabs(H_even_constr_1_tm_without_deltas(n, x, grad, data)) > FREQ_TOL) return;
    if (fabs(H_odd_constr_1_tm_without_deltas(n, x, grad, data)) > FREQ_TOL) return;
    global_minimum = value;
    for (int i = 0; i < 8 + imprinting; i++) {
        global_x[i] = x[i];
    }
}

static void update_global_minimum_1_tm_mat_first_without_deltas(double value, const double *x, void *data) {
    int n = 0;
    double *grad = 0;
    if (value >= global_minimum) return;
    if (penetrance_restriction) {
        if (imprinting) {
            if (f0_f1_pat_constr_1_tm_mat_first_without_deltas(n, x, grad, data) > PEN_TOL) return;
            if (f0_f1_mat_constr_1_tm_mat_first_without_deltas(n, x, grad, data) > PEN_TOL) return;
            if (f1_pat_f2_constr_1_tm_mat_first_without_deltas(n, x, grad, data) > PEN_TOL) return;
            if (f1_mat_f2_constr_1_tm_mat_first_without_deltas(n, x, grad, data) > PEN_TOL) return;
        } else {
            if (f0_f1_constr_1_tm_without_deltas(n, x, grad, data) > PEN_TOL) return;
            if (f1_f2_constr_1_tm_without_deltas(n, x, grad, data) > PEN_TOL) return;
        }
    }
    if (fabs(H_even_constr_1_tm_without_deltas(n, x, grad, data)) > FREQ_TOL) return;
    if (fabs(H_odd_constr_1_tm_without_deltas(n, x, grad, data)) > FREQ_TOL) return;
    global_minimum = value;
    for (int i = 0; i < 8 + imprinting; i++) {
        global_x[i] = x[i];
    }
}

static void update_global_minimum_user_model_1_tm_without_deltas(double value, const double *x, void *data) {
    int n = 0;
    double *grad = 0;
    if (value >= global_minimum) return;
    if (fabs(H_even_user_model_constr_1_tm_without_deltas(n, x, grad, data)) > FREQ_TOL) return;
    if (fabs(H_odd_user_model_constr_1_tm_without_deltas(n, x, grad, data)) > FREQ_TOL) return;
    global_minimum = value;
    for (int i = 0; i < 4; i++) {
        global_x[i] = x[i];
    }
}

static double target_1_tm_pat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    objective_function_calls++;
    nlopt_iterations++;
    Data d = data;
    const double *deltas = x;
    double p = x[2];
    double f0, f1_pat, f1_mat, f2;
    DoubleVector hapfreqs = DoubleVector_Create(4);
    for (int i = 0; i < 2; i++) {
        // The order of hapfreqs is always H0, H1, ..., where hapfreqs
        // H0, H2, etc. contain the wildtype allele and hapfreqs H1,
        // H3, etc. contain the mutant allele.  To get this order the
        // even indexes must be wildtype hapfreqs with wildtype allele
        // frequency q and the odd indexes must be mutant hapfreqs
        // with mutant allele frequency p.
        DoubleVector_Set(hapfreqs, 2*i,     DoubleVector_Get(d->em_hapfreqs, i) * (1-p) - deltas[i]);
        DoubleVector_Set(hapfreqs, 2*i + 1, DoubleVector_Get(d->em_hapfreqs, i) * p + deltas[i]);
    }
    if (!imprinting) {
        f0 = x[3]; f1_pat = x[4]; f1_mat = x[4]; f2 = x[5];
    } else {
        f0 = x[3]; f1_pat = x[4]; f1_mat = x[5]; f2 = x[6];
    }
    DoubleVector penetrances = DoubleVector_With(10, f0, f1_pat, f1_mat, f2, 1-f0, 1-f1_pat, 1-f1_mat, 1-f2, .5 * (f1_pat + f1_mat), .5 * ((1-f1_pat) + (1-f1_mat)));
    double score = ModScorer_ComputeModScore(0, d->annotated_families, d->test_markers, p, hapfreqs, penetrances, d->em_hapfreqs);
    DoubleVector_Destroy(hapfreqs);
    DoubleVector_Destroy(penetrances);
    update_global_minimum_1_tm_pat_first(-score, x, data);
    return -score;
}

static double target_1_tm_mat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    objective_function_calls++;
    nlopt_iterations++;
    Data d = data;
    const double *deltas = x;
    double p = x[2];
    double f0, f1_pat, f1_mat, f2;
    DoubleVector hapfreqs = DoubleVector_Create(4);
    for (int i = 0; i < 2; i++) {
        // The order of hapfreqs is always H0, H1, ..., where hapfreqs
        // H0, H2, etc. contain the wildtype allele and hapfreqs H1,
        // H3, etc. contain the mutant allele.  To get this order the
        // even indexes must be wildtype hapfreqs with wildtype allele
        // frequency q and the odd indexes must be mutant hapfreqs
        // with mutant allele frequency p.
        DoubleVector_Set(hapfreqs, 2*i,     DoubleVector_Get(d->em_hapfreqs, i) * (1-p) - deltas[i]);
        DoubleVector_Set(hapfreqs, 2*i + 1, DoubleVector_Get(d->em_hapfreqs, i) * p + deltas[i]);
    }
    if (!imprinting) {
        f0 = x[3]; f1_pat = x[4]; f1_mat = x[4]; f2 = x[5];
    } else {
        f0 = x[3]; f1_pat = x[5]; f1_mat = x[4]; f2 = x[6];
    }
    DoubleVector penetrances = DoubleVector_With(10, f0, f1_pat, f1_mat, f2, 1-f0, 1-f1_pat, 1-f1_mat, 1-f2, .5 * (f1_pat + f1_mat), .5 * ((1-f1_pat) + (1-f1_mat)));
    double score = ModScorer_ComputeModScore(0, d->annotated_families, d->test_markers, p, hapfreqs, penetrances, d->em_hapfreqs);
    DoubleVector_Destroy(hapfreqs);
    DoubleVector_Destroy(penetrances);
    update_global_minimum_1_tm_mat_first(-score, x, data);
    return -score;
}

static double target_user_model_1_tm(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    objective_function_calls++;
    nlopt_iterations++;
    Data d = data;
    const double *deltas = x;
    double p = d->user_model->disease_allele_frequency;
    DoubleVector hapfreqs = DoubleVector_Create(4);
    for (int i = 0; i < 2; i++) {
        // The order of hapfreqs is always H0, H1, ..., where hapfreqs
        // H0, H2, etc. contain the wildtype allele and hapfreqs H1,
        // H3, etc. contain the mutant allele.  To get this order the
        // even indexes must be wildtype hapfreqs with wildtype allele
        // frequency q and the odd indexes must be mutant hapfreqs
        // with mutant allele frequency p.
        DoubleVector_Set(hapfreqs, 2*i,     DoubleVector_Get(d->em_hapfreqs, i) * (1-p) - deltas[i]);
        DoubleVector_Set(hapfreqs, 2*i + 1, DoubleVector_Get(d->em_hapfreqs, i) * p + deltas[i]);
    }
    double f0 = d->user_model->f0, f1_pat = d->user_model->f1_pat, f1_mat = d->user_model->f1_mat, f2 = d->user_model->f2;
    DoubleVector penetrances = DoubleVector_With(10, f0, f1_pat, f1_mat, f2, 1-f0, 1-f1_pat, 1-f1_mat, 1-f2, .5 * (f1_pat + f1_mat), .5 * ((1-f1_pat) + (1-f1_mat)));
    double score = ModScorer_ComputeModScore(0, d->annotated_families, d->test_markers, p, hapfreqs, penetrances, d->em_hapfreqs);
    DoubleVector_Destroy(hapfreqs);
    DoubleVector_Destroy(penetrances);
    update_global_minimum_user_model_1_tm(-score, x, data);
    return -score;
}

static double target_1_tm_pat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    objective_function_calls++;
    nlopt_iterations++;
    Data d = data;
    const double *H = x;
    DoubleVector hapfreqs = DoubleVector_Create(4);
    for (int i = 0; i < 4; i++) {
        DoubleVector_PushBack(hapfreqs, H[i]);
    }
    double p = x[4];
    double f0, f1_pat, f1_mat, f2;
    if (!imprinting) {
        f0 = x[5]; f1_pat = x[6]; f1_mat = x[6]; f2 = x[7];
    } else {
        f0 = x[5]; f1_pat = x[6]; f1_mat = x[7]; f2 = x[8];
    }
    DoubleVector penetrances = DoubleVector_With(10, f0, f1_pat, f1_mat, f2, 1-f0, 1-f1_pat, 1-f1_mat, 1-f2, .5 * (f1_pat + f1_mat), .5 * ((1-f1_pat) + (1-f1_mat)));
    double score = ModScorer_ComputeModScore(0, d->annotated_families, d->test_markers, p, hapfreqs, penetrances, d->em_hapfreqs);
    DoubleVector_Destroy(hapfreqs);
    DoubleVector_Destroy(penetrances);
    update_global_minimum_1_tm_pat_first_without_deltas(-score, x, data);
    return -score;
}

static double target_1_tm_mat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    objective_function_calls++;
    nlopt_iterations++;
    Data d = data;
    const double *H = x;
    DoubleVector hapfreqs = DoubleVector_Create(4);
    for (int i = 0; i < 4; i++) {
        DoubleVector_PushBack(hapfreqs, H[i]);
    }
    double p = x[4];
    double f0, f1_pat, f1_mat, f2;
    if (!imprinting) {
        f0 = x[5]; f1_pat = x[6]; f1_mat = x[6]; f2 = x[7];
    } else {
        f0 = x[5]; f1_pat = x[7]; f1_mat = x[6]; f2 = x[8];
    }
    DoubleVector penetrances = DoubleVector_With(10, f0, f1_pat, f1_mat, f2, 1-f0, 1-f1_pat, 1-f1_mat, 1-f2, .5 * (f1_pat + f1_mat), .5 * ((1-f1_pat) + (1-f1_mat)));
    double score = ModScorer_ComputeModScore(0, d->annotated_families, d->test_markers, p, hapfreqs, penetrances, d->em_hapfreqs);
    DoubleVector_Destroy(hapfreqs);
    DoubleVector_Destroy(penetrances);
    update_global_minimum_1_tm_mat_first_without_deltas(-score, x, data);
    return -score;
}

static double target_user_model_1_tm_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    objective_function_calls++;
    nlopt_iterations++;
    Data d = data;
    const double *H = x;
    DoubleVector hapfreqs = DoubleVector_Create(4);
    for (int i = 0; i < 4; i++) {
        DoubleVector_PushBack(hapfreqs, H[i]);
    }
    double p = d->user_model->disease_allele_frequency;
    double f0 = d->user_model->f0, f1_pat = d->user_model->f1_pat, f1_mat = d->user_model->f1_mat, f2 = d->user_model->f2;
    DoubleVector penetrances = DoubleVector_With(10, f0, f1_pat, f1_mat, f2, 1-f0, 1-f1_pat, 1-f1_mat, 1-f2, .5 * (f1_pat + f1_mat), .5 * ((1-f1_pat) + (1-f1_mat)));
    double score = ModScorer_ComputeModScore(0, d->annotated_families, d->test_markers, p, hapfreqs, penetrances, d->em_hapfreqs);
    DoubleVector_Destroy(hapfreqs);
    DoubleVector_Destroy(penetrances);
    update_global_minimum_user_model_1_tm_without_deltas(-score, x, data);
    return -score;
}

static int optimize_with_1_test_marker_pat_first(
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModScore *gm,
    ModScore *mod_score,
    double initial_step_size1)
{
    //             d1  d2  p       f0 f1_pat f1_mat f2
    double lb[] = {-1, -1, 1e-8,   0, 0,     0,     0};
    double ub[] = { 1,  1, 1-1e-8, 1, 1,     1,     1};
    if (allfreq_restriction) {
        ub[2] = highest_allfreq;
    }
    struct DataStruct data = {
        annotated_families,
        test_markers,
        em_hapfreqs,
        0
    };

    nlopt_opt opt = nlopt_create(NLOPT_LN_COBYLA, 7); // algorithm and dimensionality
    nlopt_set_lower_bounds(opt, lb);
    nlopt_set_upper_bounds(opt, ub);
    nlopt_set_min_objective(opt, target_1_tm_pat_first, &data);
    nlopt_add_equality_constraint(opt, ds_group_constr_1_tm, &data, FREQ_TOL);
    if (penetrance_restriction) {
        nlopt_add_inequality_constraint(opt, f0_f1_pat_constr_1_tm_pat_first, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f0_f1_mat_constr_1_tm_pat_first, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_pat_f2_constr_1_tm_pat_first, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_mat_f2_constr_1_tm_pat_first, &data, DBL_EPSILON);
    }
    nlopt_add_inequality_constraint(opt, d1_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d1_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d2_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d2_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_set_xtol_rel(opt, 1e-4);
    int max_eval = MIN(g_jla_nlopt_iterations_free, 10000);
    nlopt_set_maxeval(opt, max_eval);

    double x[] = {
        DoubleVector_Get(gm->deltas, 0),
        DoubleVector_Get(gm->deltas, 1),
        gm->disease_allele_frequency,
        DoubleVector_Get(gm->penetrances, 0),
        DoubleVector_Get(gm->penetrances, 1),
        DoubleVector_Get(gm->penetrances, 2),
        DoubleVector_Get(gm->penetrances, 3)
    };
    nlopt_iterations = 0;
    nlopt_set_initial_step1(opt, initial_step_size1);

    double negative_mod_score; // minimum value of target function
    int nlopt_status = nlopt_optimize(opt, x, &negative_mod_score);
    int nlopt_success = 1;
    if (nlopt_status < 0) {
        fprintf(stderr, "%s\n", nlopt_result_to_string(nlopt_status));
        nlopt_success = 0;
        sf(ps, "nlopt failed!\n");
        pr();
    } else {
        if (negative_mod_score != global_minimum) {
            sf(ps, "COBYLA = %f vs %f = global minimum\n", negative_mod_score, global_minimum);
            pr();
            negative_mod_score = global_minimum;
            for (int i = 0; i < 6+1; i++) {
                x[i] = global_x[i];
            }
        }
        sf(ps, "No. of nlopt iterations: %d\n", nlopt_iterations);
        pr();
        g_jla_nlopt_iterations_free -= nlopt_iterations;
        DoubleVector hapfreqs = DoubleVector_Create(4);
        DoubleVector ds = DoubleVector_Create(2);
        DoubleVector_PushBack(ds, x[0]);
        DoubleVector_PushBack(ds, x[1]);
        double *deltas = x;
        double p = x[2];
        for (int i = 0; i < 2; i++) {
            DoubleVector_Set(hapfreqs, 2*i,     DoubleVector_Get(em_hapfreqs, i) * (1-p) - deltas[i]);
            DoubleVector_Set(hapfreqs, 2*i + 1, DoubleVector_Get(em_hapfreqs, i) * p + deltas[i]);
        }
        DoubleVector penetrances = DoubleVector_Create(4);
        DoubleVector_PushBack(penetrances, x[3]);
        DoubleVector_PushBack(penetrances, x[4]);
        DoubleVector_PushBack(penetrances, x[5]);
        DoubleVector_PushBack(penetrances, x[6]);
        mod_score->test_markers = test_markers;
        mod_score->position = position;
        mod_score->mod_score = -negative_mod_score;
        mod_score->disease_allele_frequency = p;
        mod_score->penetrances = penetrances;
        mod_score->hapfreqs = hapfreqs;
        mod_score->em_hapfreqs = em_hapfreqs;
        mod_score->deltas = ds;
        print_maximum_1_tm_pat_first(-negative_mod_score, x, em_hapfreqs);
    }
    nlopt_destroy(opt);
    return nlopt_success;
}


static int optimize_with_1_test_marker_mat_first(
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModScore *gm,
    ModScore *mod_score,
    double initial_step_size1)
{
    //             d1  d2  p       f0 f1_mat f1_pat f2
    double lb[] = {-1, -1, 1e-8,   0, 0,     0,     0};
    double ub[] = { 1,  1, 1-1e-8, 1, 1,     1,     1};
    if (allfreq_restriction) {
        ub[2] = highest_allfreq;
    }
    struct DataStruct data = {
        annotated_families,
        test_markers,
        em_hapfreqs,
        0
    };

    nlopt_opt opt = nlopt_create(NLOPT_LN_COBYLA, 7); // algorithm and dimensionality
    nlopt_set_lower_bounds(opt, lb);
    nlopt_set_upper_bounds(opt, ub);
    nlopt_set_min_objective(opt, target_1_tm_mat_first, &data);
    nlopt_add_equality_constraint(opt, ds_group_constr_1_tm, &data, FREQ_TOL);
    if (penetrance_restriction) {
        nlopt_add_inequality_constraint(opt, f0_f1_pat_constr_1_tm_mat_first, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f0_f1_mat_constr_1_tm_mat_first, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_pat_f2_constr_1_tm_mat_first, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_mat_f2_constr_1_tm_mat_first, &data, DBL_EPSILON);
    }
    nlopt_add_inequality_constraint(opt, d1_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d1_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d2_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d2_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_set_xtol_rel(opt, 1e-4);
    int max_eval = MIN(g_jla_nlopt_iterations_free, 10000);
    nlopt_set_maxeval(opt, max_eval);

    double x[] = {
        DoubleVector_Get(gm->deltas, 0),
        DoubleVector_Get(gm->deltas, 1),
        gm->disease_allele_frequency,
        DoubleVector_Get(gm->penetrances, 0),
        DoubleVector_Get(gm->penetrances, 2),
        DoubleVector_Get(gm->penetrances, 1),
        DoubleVector_Get(gm->penetrances, 3)
    };
    nlopt_iterations = 0;
    nlopt_set_initial_step1(opt, initial_step_size1);

    double negative_mod_score; // minimum value of target function
    int nlopt_status = nlopt_optimize(opt, x, &negative_mod_score);
    int nlopt_success = 1;
    if (nlopt_status < 0) {
        fprintf(stderr, "%s\n", nlopt_result_to_string(nlopt_status));
        nlopt_success = 0;
        sf(ps, "nlopt failed!\n");
        pr();
    } else {
        if (negative_mod_score != global_minimum) {
            sf(ps, "COBYLA = %f vs %f = global minimum\n", negative_mod_score, global_minimum);
            pr();
            negative_mod_score = global_minimum;
            for (int i = 0; i < 6+1; i++) {
                x[i] = global_x[i];
            }
        }
        sf(ps, "No. of nlopt iterations: %d\n", nlopt_iterations);
        pr();
        g_jla_nlopt_iterations_free -= nlopt_iterations;
        DoubleVector hapfreqs = DoubleVector_Create(4);
        DoubleVector ds = DoubleVector_Create(2);
        DoubleVector_PushBack(ds, x[0]);
        DoubleVector_PushBack(ds, x[1]);
        double *deltas = x;
        double p = x[2];
        for (int i = 0; i < 2; i++) {
            DoubleVector_Set(hapfreqs, 2*i,     DoubleVector_Get(em_hapfreqs, i) * (1-p) - deltas[i]);
            DoubleVector_Set(hapfreqs, 2*i + 1, DoubleVector_Get(em_hapfreqs, i) * p + deltas[i]);
        }
        DoubleVector penetrances = DoubleVector_Create(4);
        DoubleVector_PushBack(penetrances, x[3]);
        DoubleVector_PushBack(penetrances, x[5]);
        DoubleVector_PushBack(penetrances, x[4]);
        DoubleVector_PushBack(penetrances, x[6]);
        mod_score->test_markers = test_markers;
        mod_score->position = position;
        mod_score->mod_score = -negative_mod_score;
        mod_score->disease_allele_frequency = p;
        mod_score->penetrances = penetrances;
        mod_score->hapfreqs = hapfreqs;
        mod_score->em_hapfreqs = em_hapfreqs;
        mod_score->deltas = ds;
        print_maximum_1_tm_mat_first(-negative_mod_score, x, em_hapfreqs);
    }
    nlopt_destroy(opt);
    return nlopt_success;
}

static int optimize_with_1_test_marker_pat_first_without_deltas(
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModScore *gm,
    ModScore *mod_score,
    double initial_step_size1)
{
    //              H0 H1 H2 H3 p       f0 f1_pat f1_mat f2
    double lb[] = { 0, 0, 0, 0,   1e-8, 0, 0,     0,     0};
    double ub[] = { 1, 1, 1, 1, 1-1e-8, 1, 1,     1,     1};
    if (allfreq_restriction) {
        ub[4] = highest_allfreq;
    }
    struct DataStruct data = {
        annotated_families,
        test_markers,
        em_hapfreqs,
        0
    };

    nlopt_opt opt = nlopt_create(NLOPT_LN_COBYLA, 9); // algorithm and dimensionality
    nlopt_set_lower_bounds(opt, lb);
    nlopt_set_upper_bounds(opt, ub);
    nlopt_set_min_objective(opt, target_1_tm_pat_first_without_deltas, &data);
    nlopt_add_equality_constraint(opt, H_even_constr_1_tm_without_deltas, &data, FREQ_TOL);
    nlopt_add_equality_constraint(opt, H_odd_constr_1_tm_without_deltas, &data, FREQ_TOL);
    if (penetrance_restriction) {
        nlopt_add_inequality_constraint(opt, f0_f1_pat_constr_1_tm_pat_first_without_deltas, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f0_f1_mat_constr_1_tm_pat_first_without_deltas, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_pat_f2_constr_1_tm_pat_first_without_deltas, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_mat_f2_constr_1_tm_pat_first_without_deltas, &data, DBL_EPSILON);
    }
    nlopt_set_xtol_rel(opt, 1e-4);
    int max_eval = MIN(g_jla_nlopt_iterations_free, 10000);
    nlopt_set_maxeval(opt, max_eval);

    double x[] = {
        DoubleVector_Get(gm->hapfreqs, 0),
        DoubleVector_Get(gm->hapfreqs, 1),
        DoubleVector_Get(gm->hapfreqs, 2),
        DoubleVector_Get(gm->hapfreqs, 3),
        gm->disease_allele_frequency,
        DoubleVector_Get(gm->penetrances, 0),
        DoubleVector_Get(gm->penetrances, 1),
        DoubleVector_Get(gm->penetrances, 2),
        DoubleVector_Get(gm->penetrances, 3)
    };
    nlopt_iterations = 0;
    nlopt_set_initial_step1(opt, initial_step_size1);

    double negative_mod_score; // minimum value of target function
    int nlopt_status = nlopt_optimize(opt, x, &negative_mod_score);
    int nlopt_success = 1;
    if (nlopt_status < 0) {
        fprintf(stderr, "%s\n", nlopt_result_to_string(nlopt_status));
        nlopt_success = 0;
        sf(ps, "nlopt failed!\n");
        pr();
    } else {
        if (negative_mod_score != global_minimum) {
            sf(ps, "COBYLA = %f vs %f = global minimum\n", negative_mod_score, global_minimum);
            pr();
            negative_mod_score = global_minimum;
            for (int i = 0; i < 8+1; i++) {
                x[i] = global_x[i];
            }
        }
        sf(ps, "No. of nlopt iterations: %d\n", nlopt_iterations);
        pr();
        g_jla_nlopt_iterations_free -= nlopt_iterations;
        DoubleVector hapfreqs = DoubleVector_Create(4);
        DoubleVector_PushBack(hapfreqs, x[0]);
        DoubleVector_PushBack(hapfreqs, x[1]);
        DoubleVector_PushBack(hapfreqs, x[2]);
        DoubleVector_PushBack(hapfreqs, x[3]);
        double p = x[4];
        DoubleVector penetrances = DoubleVector_Create(4);
        DoubleVector_PushBack(penetrances, x[5]);
        DoubleVector_PushBack(penetrances, x[6]);
        DoubleVector_PushBack(penetrances, x[7]);
        DoubleVector_PushBack(penetrances, x[8]);
        mod_score->test_markers = test_markers;
        mod_score->position = position;
        mod_score->mod_score = -negative_mod_score;
        mod_score->disease_allele_frequency = p;
        mod_score->penetrances = penetrances;
        mod_score->hapfreqs = hapfreqs;
        mod_score->em_hapfreqs = em_hapfreqs;
        print_maximum_1_tm_pat_first_without_deltas(-negative_mod_score, x, em_hapfreqs);
    }
    nlopt_destroy(opt);
    return nlopt_success;
}

static int optimize_with_1_test_marker_mat_first_without_deltas(
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModScore *gm,
    ModScore *mod_score,
    double initial_step_size1)
{
    //              H0 H1 H2 H3 p       f0 f1_mat f1_pat f2
    double lb[] = { 0, 0, 0, 0,   1e-8, 0, 0,     0,     0};
    double ub[] = { 1, 1, 1, 1, 1-1e-8, 1, 1,     1,     1};
    if (allfreq_restriction) {
        ub[4] = highest_allfreq;
    }
    struct DataStruct data = {
        annotated_families,
        test_markers,
        em_hapfreqs,
        0
    };

    nlopt_opt opt = nlopt_create(NLOPT_LN_COBYLA, 9); // algorithm and dimensionality
    nlopt_set_lower_bounds(opt, lb);
    nlopt_set_upper_bounds(opt, ub);
    nlopt_set_min_objective(opt, target_1_tm_mat_first_without_deltas, &data);
    nlopt_add_equality_constraint(opt, H_even_constr_1_tm_without_deltas, &data, FREQ_TOL);
    nlopt_add_equality_constraint(opt, H_odd_constr_1_tm_without_deltas, &data, FREQ_TOL);
    if (penetrance_restriction) {
        nlopt_add_inequality_constraint(opt, f0_f1_pat_constr_1_tm_mat_first_without_deltas, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f0_f1_mat_constr_1_tm_mat_first_without_deltas, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_pat_f2_constr_1_tm_mat_first_without_deltas, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_mat_f2_constr_1_tm_mat_first_without_deltas, &data, DBL_EPSILON);
    }
    nlopt_set_xtol_rel(opt, 1e-4);
    int max_eval = MIN(g_jla_nlopt_iterations_free, 10000);
    nlopt_set_maxeval(opt, max_eval);

    double x[] = {
        DoubleVector_Get(gm->hapfreqs, 0),
        DoubleVector_Get(gm->hapfreqs, 1),
        DoubleVector_Get(gm->hapfreqs, 2),
        DoubleVector_Get(gm->hapfreqs, 3),
        gm->disease_allele_frequency,
        DoubleVector_Get(gm->penetrances, 0),
        DoubleVector_Get(gm->penetrances, 2),
        DoubleVector_Get(gm->penetrances, 1),
        DoubleVector_Get(gm->penetrances, 3)
    };
    nlopt_iterations = 0;
    nlopt_set_initial_step1(opt, initial_step_size1);

    double negative_mod_score; // minimum value of target function
    int nlopt_status = nlopt_optimize(opt, x, &negative_mod_score);
    int nlopt_success = 1;
    if (nlopt_status < 0) {
        fprintf(stderr, "%s\n", nlopt_result_to_string(nlopt_status));
        nlopt_success = 0;
        sf(ps, "nlopt failed!\n");
        pr();
    } else {
        if (negative_mod_score != global_minimum) {
            sf(ps, "COBYLA = %f vs %f = global minimum\n", negative_mod_score, global_minimum);
            pr();
            negative_mod_score = global_minimum;
            for (int i = 0; i < 8+1; i++) {
                x[i] = global_x[i];
            }
        }
        sf(ps, "No. of nlopt iterations: %d\n", nlopt_iterations);
        pr();
        g_jla_nlopt_iterations_free -= nlopt_iterations;
        DoubleVector hapfreqs = DoubleVector_Create(4);
        DoubleVector_PushBack(hapfreqs, x[0]);
        DoubleVector_PushBack(hapfreqs, x[1]);
        DoubleVector_PushBack(hapfreqs, x[2]);
        DoubleVector_PushBack(hapfreqs, x[3]);
        double p = x[4];
        DoubleVector penetrances = DoubleVector_Create(4);
        DoubleVector_PushBack(penetrances, x[5]);
        DoubleVector_PushBack(penetrances, x[7]);
        DoubleVector_PushBack(penetrances, x[6]);
        DoubleVector_PushBack(penetrances, x[8]);
        mod_score->test_markers = test_markers;
        mod_score->position = position;
        mod_score->mod_score = -negative_mod_score;
        mod_score->disease_allele_frequency = p;
        mod_score->penetrances = penetrances;
        mod_score->hapfreqs = hapfreqs;
        mod_score->em_hapfreqs = em_hapfreqs;
        print_maximum_1_tm_mat_first_without_deltas(-negative_mod_score, x, em_hapfreqs);
    }
    nlopt_destroy(opt);
    return nlopt_success;
}

static void *optimize_pthread(void *args)
{
    int old_state;
    pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &old_state);
    p_data p = (p_data) args;
    int ok = p->optimize(p->annotated_families, p->test_markers, p->em_hapfreqs, p->position, p->grid_mod_score, p->nlopt_mod_score, p->steps);
    p->returned = 1;
    p->ok = ok;
    pthread_exit(NULL);
}

static void *optimize_pthread_user_model(void *args)
{
    int old_state;
    pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &old_state);
    p_data_user_model p = (p_data_user_model) args;
    int ok = p->optimize(p->annotated_families, p->test_markers, p->em_hapfreqs, p->position, p->user_model, p->grid_mod_score, p->nlopt_mod_score, p->steps);
    p->returned = 1;
    p->ok = ok;
    pthread_exit(NULL);
}

int run_optimize(
    OptimizeFunction optimize,
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModScore *start_mod_score,
    ModScore *end_mod_score,
    double step)
{
    pthread_t ptid;
    struct pthread_data_struct p = {
        optimize,
        annotated_families,
        test_markers,
        em_hapfreqs,
        position,
        start_mod_score,
        end_mod_score,
        step,
        0,
        0
    };
    global_minimum = HUGE_VAL;  // before every optimize call!
    pthread_create(&ptid, NULL, &optimize_pthread, &p);
    unsigned long prev_calls = objective_function_calls;
    unsigned long delay_in_us = 50000UL;  // 50 ms
    int objective_function_calls_unchanged_max
        = 1000000 * (unsigned) g_timeout_single_modscore / delay_in_us;
    int objective_function_calls_unchanged = 0;
    while (!p.returned) {
        usleep(delay_in_us);
        if (objective_function_calls == prev_calls) {
            objective_function_calls_unchanged++;
            // NLopt didn't call the optimize function for 60 seconds.
            // Seems that the computation is hanging.  Abort it.
            if (objective_function_calls_unchanged == objective_function_calls_unchanged_max) {
                sf(ps, "COBYLA hangs ... skipping step size %f"
                       "  (If the current dataset consists of large pedigrees,"
                       " try to increase \"timeout single modscore\".)\n", step);
                pr();
                break;
            }
        } else {
            objective_function_calls_unchanged = 0;
        }
        prev_calls = objective_function_calls;
    }
    if (!p.returned) {
        pthread_cancel(ptid);
    }
    pthread_join(ptid, NULL);  // ignore return value
    return p.ok;
}

int run_optimize_user_model(
    OptimizeFunctionUserModel optimize,
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModelParams user_model,
    ModScore *start_mod_score,
    ModScore *end_mod_score,
    double step)
{
    pthread_t ptid;
    struct pthread_user_model_data_struct p = {
        optimize,
        annotated_families,
        test_markers,
        em_hapfreqs,
        position,
        user_model,
        start_mod_score,
        end_mod_score,
        step,
        0,
        0
    };
    global_minimum = HUGE_VAL;  // before every optimize call!
    pthread_create(&ptid, NULL, &optimize_pthread_user_model, &p);
    unsigned long prev_calls = objective_function_calls;
    unsigned long delay_in_us = 50000UL;  // 50 ms
    int objective_function_calls_unchanged_max
        = 1000000 * (unsigned) g_timeout_single_modscore / delay_in_us;
    int objective_function_calls_unchanged = 0;
    while (!p.returned) {
        usleep(delay_in_us);
        if (objective_function_calls == prev_calls) {
            objective_function_calls_unchanged++;
            // NLopt didn't call the optimize function for 60 seconds.
            // Seems that the computation is hanging.  Abort it.
            if (objective_function_calls_unchanged == objective_function_calls_unchanged_max) {
                sf(ps, "COBYLA hangs ... skipping step size %f"
                       "  (If the current dataset consists of large pedigrees,"
                       " try to increase \"timeout single modscore\".)\n", step);
                pr();
                break;
            }
        } else {
            objective_function_calls_unchanged = 0;
        }
        prev_calls = objective_function_calls;
    }
    if (!p.returned) {
        pthread_cancel(ptid);
    }
    pthread_join(ptid, NULL);  // ignore return value
    return p.ok;
}

static int optimize_wrapper(
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModScore *grid_mod_score,
    ModScore *synthetic_dummy_mod_score,
    OptimizeFunction optimize,
    ModScore *mod_score)
{
    g_jla_nlopt_iterations_free = g_jla_nlopt_iterations_free_old;
    ModScore *nlopt_mod_score = ModScore_Create();
    ModScore_Update(nlopt_mod_score, grid_mod_score);
    ModScore *tmp = ModScore_Create();
    // Use "big" initial steps until we find a positive mod score.
    double big_steps[] = { 1, .9, .8, .7, .6, .5, .4, .3 };
    const int nbig = sizeof(big_steps) / sizeof(big_steps[0]);
    int mod_score_gt_0 = 0;
    int nfailures = 0;  // # consecutive failing step sizes
    int attempt = 0;
    // Taking the definition of the big_step index i out of the loop
    // allows us to sequentially loop thru big_steps across attempts.
    int i = 0;
    double prev_mod_score = nlopt_mod_score->mod_score;
    while (attempt < 2) {
        while (1) {  // loop through big_steps
            while (1) {  // try step size while mod score improves
                if (g_jla_nlopt_iterations_free <= 0) {
                    break;
                }
                int ok = run_optimize(optimize, annotated_families, test_markers, em_hapfreqs, position,
                    nlopt_mod_score, nlopt_mod_score, big_steps[i]);
                if (!ok) {
                    nfailures++;
                    break;  // try next step size
                } else if (nlopt_mod_score->mod_score > 0) {
                    mod_score_gt_0 = 1;
                    break;
                } else if (nlopt_mod_score->mod_score - prev_mod_score < 1e-4) {
                    break;
                }
                nfailures = 0;
                prev_mod_score = nlopt_mod_score->mod_score;
            }
            if (mod_score_gt_0) {
                break;
            } else if (g_jla_nlopt_iterations_free <= 0) {
                break;
            } else if (nfailures >= nbig) {
                break;
            }
            i = (i + 1) % nbig;
        }
        if (mod_score_gt_0) {
            break;  // success
        } else if (nfailures >= nbig) {
            break;  // no point in retrying step sizes
        } else if (attempt == 0) {  // grant one extra round
            g_jla_nlopt_iterations_free = g_jla_nlopt_iterations_free_old;
        } else {
            break;  // admit failure (to find a mod score > 0)
        }
        attempt++;
    }
    if (mod_score_gt_0) {
        // Use all step sizes but update only if mod score improves.
        double steps[] = { .2, .1, 1, 0 };  // 0 uses in nlopt's default steps
        const int nsteps = sizeof(steps) / sizeof(steps[0]);
        for (int i = 0; i < nsteps; i++) {
            if (g_jla_nlopt_iterations_free <= 0) {
                break;
            }
            sf(ps, "small step size: %f\n", steps[i]);
            pr();
            int ok = run_optimize(optimize, annotated_families, test_markers, em_hapfreqs, position,
                nlopt_mod_score, tmp, steps[i]);
            if (ok && tmp->mod_score > nlopt_mod_score->mod_score) {
                ModScore_Update(nlopt_mod_score, tmp);
            }
        }
    }
    if (attempt > 0) {
        sf(ps, "Total no. of nlopt iterations: %d\n", 2*g_jla_nlopt_iterations_free_old - g_jla_nlopt_iterations_free);
        pr();
    } else {
        sf(ps, "Total no. of nlopt iterations: %d\n", g_jla_nlopt_iterations_free_old - g_jla_nlopt_iterations_free);
        pr();
    }
    if (nlopt_mod_score->mod_score > 0) {
        ModScore_Update(mod_score, nlopt_mod_score);
    } else {
        ModScore_Update(mod_score, synthetic_dummy_mod_score);
    }
    return 1;
}

static int optimize_user_model_wrapper(
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModelParams user_model,
    ModScore *grid_mod_score,
    OptimizeFunctionUserModel optimize,
    ModScore *mod_score)
{
    g_jla_nlopt_iterations_free = g_jla_nlopt_iterations_free_old;
    ModScore *nlopt_mod_score = ModScore_Create();
    ModScore_Update(nlopt_mod_score, grid_mod_score);
    ModScore *tmp = ModScore_Create();
    // Use "big" initial steps until we find a positive mod score.
    double big_steps[] = { 1, .9, .8, .7, .6, .5, .4, .3 };
    const int nbig = sizeof(big_steps) / sizeof(big_steps[0]);
    int mod_score_gt_0 = 0;
    int nfailures = 0;  // # consecutive failing step sizes
    int attempt = 0;
    // Taking the definition of the big_step index i out of the loop
    // allows us to sequentially loop thru big_steps across attempts.
    int i = 0;
    double prev_mod_score = nlopt_mod_score->mod_score;
    while (attempt < 2) {
        while (1) {  // loop through big_steps
            while (1) {  // try step size while mod score improves
                if (g_jla_nlopt_iterations_free <= 0) {
                    break;
                }
                int ok = run_optimize_user_model(optimize, annotated_families, test_markers, em_hapfreqs, position,
                    user_model, nlopt_mod_score, nlopt_mod_score, big_steps[i]);
                if (!ok) {
                    nfailures++;
                    break;  // try next step size
                } else if (nlopt_mod_score->mod_score > 0) {
                    mod_score_gt_0 = 1;
                    break;
                } else if (nlopt_mod_score->mod_score - prev_mod_score < 1e-4) {
                    break;
                }
                nfailures = 0;
                prev_mod_score = nlopt_mod_score->mod_score;
            }
            if (mod_score_gt_0) {
                break;
            } else if (g_jla_nlopt_iterations_free <= 0) {
                break;
            } else if (nfailures >= nbig) {
                break;
            }
            i = (i + 1) % nbig;
        }
        if (mod_score_gt_0) {
            break;  // success
        } else if (nfailures >= nbig) {
            break;  // no point in retrying step sizes
        } else if (attempt == 0) {  // grant one extra round
            g_jla_nlopt_iterations_free = g_jla_nlopt_iterations_free_old;
        } else {
            break;  // admit failure (to find a mod score > 0)
        }
        attempt++;
    }
    if (mod_score_gt_0) {
        // Use all step sizes but update only if mod score improves.
        double steps[] = { .2, .1, 1, 0 };  // 0 uses in nlopt's default steps
        const int nsteps = sizeof(steps) / sizeof(steps[0]);
        for (int i = 0; i < nsteps; i++) {
            if (g_jla_nlopt_iterations_free <= 0) {
                break;
            }
            int ok = run_optimize_user_model(optimize, annotated_families, test_markers, em_hapfreqs, position,
                user_model, nlopt_mod_score, tmp, steps[i]);
            if (ok && tmp->mod_score > nlopt_mod_score->mod_score) {
                ModScore_Update(nlopt_mod_score, tmp);
            }
        }
    }
    if (attempt > 0) {
        sf(ps, "Total no. of nlopt iterations: %d\n", 2*g_jla_nlopt_iterations_free_old - g_jla_nlopt_iterations_free);
        pr();
    } else {
        sf(ps, "Total no. of nlopt iterations: %d\n", g_jla_nlopt_iterations_free_old - g_jla_nlopt_iterations_free);
        pr();
    }
    ModScore_Update(mod_score, nlopt_mod_score);
    return 1;
}

static int optimize_with_1_test_marker_no_imprinting(
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModScore *gm,
    ModScore *mod_score,
    double initial_step_size1)
{
    //             d1  d2  p       f0 f1 f2
    double lb[] = {-1, -1, 1e-8,   0, 0, 0};
    double ub[] = { 1,  1, 1-1e-8, 1, 1, 1};
    if (allfreq_restriction) {
        ub[2] = highest_allfreq;
    }
    struct DataStruct data = {
        annotated_families,
        test_markers,
        em_hapfreqs,
        0
    };

    nlopt_opt opt = nlopt_create(NLOPT_LN_COBYLA, 6); // algorithm and dimensionality
    nlopt_set_lower_bounds(opt, lb);
    nlopt_set_upper_bounds(opt, ub);
    nlopt_set_min_objective(opt, target_1_tm_pat_first, &data);
    nlopt_add_equality_constraint(opt, ds_group_constr_1_tm, &data, FREQ_TOL);
    if (penetrance_restriction) {
        nlopt_add_inequality_constraint(opt, f0_f1_constr_1_tm, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_f2_constr_1_tm, &data, DBL_EPSILON);
    }
    nlopt_add_inequality_constraint(opt, d1_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d1_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d2_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d2_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_set_xtol_rel(opt, 1e-4);
    int max_eval = MIN(g_jla_nlopt_iterations_free, 10000);
    nlopt_set_maxeval(opt, max_eval);

    double x[] = {
        DoubleVector_Get(gm->deltas, 0),
        DoubleVector_Get(gm->deltas, 1),
        gm->disease_allele_frequency,
        DoubleVector_Get(gm->penetrances, 0),
        DoubleVector_Get(gm->penetrances, 1),
        DoubleVector_Get(gm->penetrances, 3)
    };
    nlopt_iterations = 0;
    nlopt_set_initial_step1(opt, initial_step_size1);

    double negative_mod_score; // minimum value of target function
    int nlopt_success = 1;
    if (nlopt_optimize(opt, x, &negative_mod_score) < 0) {
        nlopt_success = 0;
        sf(ps, "nlopt failed!\n");
        pr();
    } else {
        if (negative_mod_score != global_minimum) {
            sf(ps, "COBYLA = %f vs %f = global minimum\n", negative_mod_score, global_minimum);
            pr();
            negative_mod_score = global_minimum;
            for (int i = 0; i < 6; i++) {
                x[i] = global_x[i];
            }
        }
        sf(ps, "No. of nlopt iterations: %d\n", nlopt_iterations);
        pr();
        g_jla_nlopt_iterations_free -= nlopt_iterations;
        DoubleVector hapfreqs = DoubleVector_Create(4);
        DoubleVector ds = DoubleVector_Create(2);
        DoubleVector_PushBack(ds, x[0]);
        DoubleVector_PushBack(ds, x[1]);
        double *deltas = x;
        double p = x[2];
        for (int i = 0; i < 2; i++) {
            DoubleVector_Set(hapfreqs, 2*i,     DoubleVector_Get(em_hapfreqs, i) * (1-p) - deltas[i]);
            DoubleVector_Set(hapfreqs, 2*i + 1, DoubleVector_Get(em_hapfreqs, i) * p + deltas[i]);
        }
        DoubleVector penetrances = DoubleVector_Create(4);
        DoubleVector_PushBack(penetrances, x[3]);
        DoubleVector_PushBack(penetrances, x[4]);
        DoubleVector_PushBack(penetrances, x[4]);
        DoubleVector_PushBack(penetrances, x[5]);
        mod_score->test_markers = test_markers;
        mod_score->position = position;
        mod_score->mod_score = -negative_mod_score;
        mod_score->disease_allele_frequency = p;
        mod_score->penetrances = penetrances;
        mod_score->hapfreqs = hapfreqs;
        mod_score->em_hapfreqs = em_hapfreqs;
        mod_score->deltas = ds;
        print_maximum_1_tm_pat_first(-negative_mod_score, x, em_hapfreqs);
    }
    nlopt_destroy(opt);
    return nlopt_success;
}

static int optimize_user_model_with_1_test_marker(
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModelParams user_model,
    ModScore *gm,
    ModScore *mod_score,
    double initial_step_size1)
{
    //             d1  d2
    double lb[] = {-1, -1};
    double ub[] = { 1,  1};
    struct DataStruct data = {
        annotated_families,
        test_markers,
        em_hapfreqs,
        user_model
    };

    nlopt_opt opt = nlopt_create(NLOPT_LN_COBYLA, 2); // algorithm and dimensionality
    nlopt_set_lower_bounds(opt, lb);
    nlopt_set_upper_bounds(opt, ub);
    nlopt_set_min_objective(opt, target_user_model_1_tm, &data);
    nlopt_add_equality_constraint(opt, ds_group_constr_1_tm, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d1_user_model_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d1_user_model_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d2_user_model_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d2_user_model_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_set_xtol_rel(opt, 1e-4);
    int max_eval = MIN(g_jla_nlopt_iterations_free, 10000);
    nlopt_set_maxeval(opt, max_eval);

    double x[] = {
        DoubleVector_Get(gm->deltas, 0),
        DoubleVector_Get(gm->deltas, 1)
    };
    nlopt_iterations = 0;
    nlopt_set_initial_step1(opt, initial_step_size1);

    double negative_mod_score; // minimum value of target function
    int nlopt_success = 1;
    if (nlopt_optimize(opt, x, &negative_mod_score) < 0) {
        nlopt_success = 0;
        sf(ps, "nlopt failed!\n");
        pr();
    } else {
        if (negative_mod_score != global_minimum) {
            sf(ps, "COBYLA = %f vs %f = global minimum\n", negative_mod_score, global_minimum);
            pr();
            negative_mod_score = global_minimum;
            for (int i = 0; i < 2; i++) {
                x[i] = global_x[i];
            }
        }
        sf(ps, "No. of nlopt iterations: %d\n", nlopt_iterations);
        pr();
        g_jla_nlopt_iterations_free -= nlopt_iterations;
        double p = user_model->disease_allele_frequency;
        DoubleVector hapfreqs = DoubleVector_Create(4);
        DoubleVector ds = DoubleVector_Create(2);
        DoubleVector_PushBack(ds, x[0]);
        DoubleVector_PushBack(ds, x[1]);
        double *deltas = x;
        for (int i = 0; i < 2; i++) {
            DoubleVector_Set(hapfreqs, 2*i,     DoubleVector_Get(em_hapfreqs, i) * (1-p) - deltas[i]);
            DoubleVector_Set(hapfreqs, 2*i + 1, DoubleVector_Get(em_hapfreqs, i) * p + deltas[i]);
        }
        DoubleVector penetrances = DoubleVector_Create(4);
        DoubleVector_PushBack(penetrances, user_model->f0);
        DoubleVector_PushBack(penetrances, user_model->f1_pat);
        DoubleVector_PushBack(penetrances, user_model->f1_mat);
        DoubleVector_PushBack(penetrances, user_model->f2);
        mod_score->test_markers = test_markers;
        mod_score->position = position;
        mod_score->mod_score = -negative_mod_score;
        mod_score->disease_allele_frequency = p;
        mod_score->penetrances = penetrances;
        mod_score->hapfreqs = hapfreqs;
        mod_score->em_hapfreqs = em_hapfreqs;
        mod_score->deltas = ds;
        print_maximum_user_model_1_tm(-negative_mod_score, x, em_hapfreqs, user_model);
    }
    nlopt_destroy(opt);
    return nlopt_success;
}

static int optimize_with_1_test_marker_no_imprinting_without_deltas(
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModScore *gm,
    ModScore *mod_score,
    double initial_step_size1)
{
    //              H0 H1 H2 H3 p       f0 f1 f2
    double lb[] = { 0, 0, 0, 0,   1e-8, 0, 0, 0};
    double ub[] = { 1, 1, 1, 1, 1-1e-8, 1, 1, 1};
    if (allfreq_restriction) {
        ub[4] = highest_allfreq;
    }
    struct DataStruct data = {
        annotated_families,
        test_markers,
        em_hapfreqs,
        0
    };

    nlopt_opt opt = nlopt_create(NLOPT_LN_COBYLA, 8); // algorithm and dimensionality
    nlopt_set_lower_bounds(opt, lb);
    nlopt_set_upper_bounds(opt, ub);
    nlopt_set_min_objective(opt, target_1_tm_pat_first_without_deltas, &data);
    nlopt_add_equality_constraint(opt, H_even_constr_1_tm_without_deltas, &data, FREQ_TOL);
    nlopt_add_equality_constraint(opt, H_odd_constr_1_tm_without_deltas, &data, FREQ_TOL);
    if (penetrance_restriction) {
        nlopt_add_inequality_constraint(opt, f0_f1_constr_1_tm_without_deltas, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_f2_constr_1_tm_without_deltas, &data, DBL_EPSILON);
    }
    nlopt_set_xtol_rel(opt, 1e-4);
    int max_eval = MIN(g_jla_nlopt_iterations_free, 10000);
    nlopt_set_maxeval(opt, max_eval);

    double x[] = {
        DoubleVector_Get(gm->hapfreqs, 0),
        DoubleVector_Get(gm->hapfreqs, 1),
        DoubleVector_Get(gm->hapfreqs, 2),
        DoubleVector_Get(gm->hapfreqs, 3),
        gm->disease_allele_frequency,
        DoubleVector_Get(gm->penetrances, 0),
        DoubleVector_Get(gm->penetrances, 1),
        DoubleVector_Get(gm->penetrances, 3)
    };
    nlopt_iterations = 0;
    nlopt_set_initial_step1(opt, initial_step_size1);

    double negative_mod_score; // minimum value of target function
    int nlopt_success = 1;
    if (nlopt_optimize(opt, x, &negative_mod_score) < 0) {
        nlopt_success = 0;
        sf(ps, "nlopt failed!\n");
        pr();
    } else {
        if (negative_mod_score != global_minimum) {
            sf(ps, "COBYLA = %f vs %f = global minimum\n", negative_mod_score, global_minimum);
            pr();
            negative_mod_score = global_minimum;
            for (int i = 0; i < 8; i++) {
                x[i] = global_x[i];
            }
        }
        sf(ps, "No. of nlopt iterations: %d\n", nlopt_iterations);
        pr();
        g_jla_nlopt_iterations_free -= nlopt_iterations;
        DoubleVector hapfreqs = DoubleVector_Create(4);
        DoubleVector_PushBack(hapfreqs, x[0]);
        DoubleVector_PushBack(hapfreqs, x[1]);
        DoubleVector_PushBack(hapfreqs, x[2]);
        DoubleVector_PushBack(hapfreqs, x[3]);
        double p = x[4];
        DoubleVector penetrances = DoubleVector_Create(4);
        DoubleVector_PushBack(penetrances, x[5]);
        DoubleVector_PushBack(penetrances, x[6]);
        DoubleVector_PushBack(penetrances, x[6]);
        DoubleVector_PushBack(penetrances, x[7]);
        mod_score->test_markers = test_markers;
        mod_score->position = position;
        mod_score->mod_score = -negative_mod_score;
        mod_score->disease_allele_frequency = p;
        mod_score->penetrances = penetrances;
        mod_score->hapfreqs = hapfreqs;
        mod_score->em_hapfreqs = em_hapfreqs;
        print_maximum_1_tm_pat_first_without_deltas(-negative_mod_score, x, em_hapfreqs);
    }
    nlopt_destroy(opt);
    return nlopt_success;
}

static int optimize_user_model_with_1_test_marker_without_deltas(
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModelParams user_model,
    ModScore *gm,
    ModScore *mod_score,
    double initial_step_size1)
{
    //             H0 H1 H2 H3
    double lb[] = {0, 0, 0, 0};
    double ub[] = {1, 1, 1, 1};
    struct DataStruct data = {
        annotated_families,
        test_markers,
        em_hapfreqs,
        user_model
    };

    nlopt_opt opt = nlopt_create(NLOPT_LN_COBYLA, 4); // algorithm and dimensionality
    nlopt_set_lower_bounds(opt, lb);
    nlopt_set_upper_bounds(opt, ub);
    nlopt_set_min_objective(opt, target_user_model_1_tm_without_deltas, &data);
    nlopt_add_equality_constraint(opt, H_even_user_model_constr_1_tm_without_deltas, &data, DBL_EPSILON);
    nlopt_add_equality_constraint(opt, H_odd_user_model_constr_1_tm_without_deltas, &data, DBL_EPSILON);
    nlopt_set_xtol_rel(opt, 1e-4);
    int max_eval = MIN(g_jla_nlopt_iterations_free, 10000);
    nlopt_set_maxeval(opt, max_eval);

    double x[] = {
        DoubleVector_Get(gm->hapfreqs, 0),
        DoubleVector_Get(gm->hapfreqs, 1),
        DoubleVector_Get(gm->hapfreqs, 2),
        DoubleVector_Get(gm->hapfreqs, 3)
    };
    nlopt_iterations = 0;
    nlopt_set_initial_step1(opt, initial_step_size1);

    double negative_mod_score; // minimum value of target function
    int nlopt_success = 1;
    if (nlopt_optimize(opt, x, &negative_mod_score) < 0) {
        nlopt_success = 0;
        sf(ps, "nlopt failed!\n");
        pr();
    } else {
        if (negative_mod_score != global_minimum) {
            sf(ps, "COBYLA = %f vs %f = global minimum\n", negative_mod_score, global_minimum);
            pr();
            negative_mod_score = global_minimum;
            for (int i = 0; i < 4; i++) {
                x[i] = global_x[i];
            }
        }
        sf(ps, "No. of nlopt iterations: %d\n", nlopt_iterations);
        pr();
        g_jla_nlopt_iterations_free -= nlopt_iterations;
        DoubleVector hapfreqs = DoubleVector_Create(4);
        DoubleVector_PushBack(hapfreqs, x[0]);
        DoubleVector_PushBack(hapfreqs, x[1]);
        DoubleVector_PushBack(hapfreqs, x[2]);
        DoubleVector_PushBack(hapfreqs, x[3]);
        double p = user_model->disease_allele_frequency;
        DoubleVector penetrances = DoubleVector_Create(4);
        DoubleVector_PushBack(penetrances, user_model->f0);
        DoubleVector_PushBack(penetrances, user_model->f1_pat);
        DoubleVector_PushBack(penetrances, user_model->f1_mat);
        DoubleVector_PushBack(penetrances, user_model->f2);
        mod_score->test_markers = test_markers;
        mod_score->position = position;
        mod_score->mod_score = -negative_mod_score;
        mod_score->disease_allele_frequency = p;
        mod_score->penetrances = penetrances;
        mod_score->hapfreqs = hapfreqs;
        mod_score->em_hapfreqs = em_hapfreqs;
        print_maximum_user_model_1_tm_without_deltas(-negative_mod_score, x, em_hapfreqs, user_model);
    }
    nlopt_destroy(opt);
    return nlopt_success;
}

// ===================================================================
// Optimize, target, and constraint functions for N = 2 test markers
// ===================================================================

static void print_maximum_2_tm_pat_first(double maximum, const double *x, DoubleVector em_hapfreqs)
{
    double mod = maximum;
    double d1 = x[0], d2 = x[1], d3 = x[2], d4 = x[3];
    double p = x[4];
    double f0, f1_pat, f1_mat, f2;
    double EMH0 = DoubleVector_Get(em_hapfreqs, 0);
    double EMH1 = DoubleVector_Get(em_hapfreqs, 1);
    double EMH2 = DoubleVector_Get(em_hapfreqs, 2);
    double EMH3 = DoubleVector_Get(em_hapfreqs, 3);
    double H0 = EMH0 * (1 - p) - d1;
    double H1 = EMH0 * p + d1;
    double H2 = EMH1 * (1 - p) - d2;
    double H3 = EMH1 * p + d2;
    double H4 = EMH2 * (1 - p) - d3;
    double H5 = EMH2 * p + d3;
    double H6 = EMH3 * (1 - p) - d4;
    double H7 = EMH3 * p + d4;
    if (!imprinting) {
        f0 = x[5]; f1_pat = x[6]; f1_mat = x[6]; f2 = x[7];
    } else {
        f0 = x[5]; f1_pat = x[6]; f1_mat = x[7]; f2 = x[8];
    }
    sf(ps, "   mod     p    F0  F1_P  F1_M    F2    H0    H1    H2    H3    H4    H5    H6    H7  EMH0  EMH1  EMH2  EMH3    d1    d2    d3    d4\n");
    pr();
    sf(ps, "%6.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f\n", mod, p, f0, f1_pat, f1_mat, f2, H0, H1, H2, H3, H4, H5, H6, H7, EMH0, EMH1, EMH2, EMH3, d1, d2, d3, d4);
    pr();
}

static void print_maximum_2_tm_mat_first(double maximum, const double *x, DoubleVector em_hapfreqs)
{
    double mod = maximum;
    double d1 = x[0], d2 = x[1], d3 = x[2], d4 = x[3];
    double p = x[4];
    double f0, f1_pat, f1_mat, f2;
    double EMH0 = DoubleVector_Get(em_hapfreqs, 0);
    double EMH1 = DoubleVector_Get(em_hapfreqs, 1);
    double EMH2 = DoubleVector_Get(em_hapfreqs, 2);
    double EMH3 = DoubleVector_Get(em_hapfreqs, 3);
    double H0 = EMH0 * (1 - p) - d1;
    double H1 = EMH0 * p + d1;
    double H2 = EMH1 * (1 - p) - d2;
    double H3 = EMH1 * p + d2;
    double H4 = EMH2 * (1 - p) - d3;
    double H5 = EMH2 * p + d3;
    double H6 = EMH3 * (1 - p) - d4;
    double H7 = EMH3 * p + d4;
    if (!imprinting) {
        f0 = x[5]; f1_pat = x[6]; f1_mat = x[6]; f2 = x[7];
    } else {
        f0 = x[5]; f1_pat = x[7]; f1_mat = x[6]; f2 = x[8];
    }
    sf(ps, "   mod     p    F0  F1_P  F1_M    F2    H0    H1    H2    H3    H4    H5    H6    H7  EMH0  EMH1  EMH2  EMH3    d1    d2    d3    d4\n");
    pr();
    sf(ps, "%6.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f\n", mod, p, f0, f1_pat, f1_mat, f2, H0, H1, H2, H3, H4, H5, H6, H7, EMH0, EMH1, EMH2, EMH3, d1, d2, d3, d4);
    pr();
}

static void print_maximum_user_model_2_tm(double maximum, const double *x, DoubleVector em_hapfreqs, ModelParams user_model)
{
    double mod = maximum;
    double d1 = x[0], d2 = x[1], d3 = x[2], d4 = x[3];
    double p = user_model->disease_allele_frequency;
    double EMH0 = DoubleVector_Get(em_hapfreqs, 0);
    double EMH1 = DoubleVector_Get(em_hapfreqs, 1);
    double EMH2 = DoubleVector_Get(em_hapfreqs, 2);
    double EMH3 = DoubleVector_Get(em_hapfreqs, 3);
    double H0 = EMH0 * (1 - p) - d1;
    double H1 = EMH0 * p + d1;
    double H2 = EMH1 * (1 - p) - d2;
    double H3 = EMH1 * p + d2;
    double H4 = EMH2 * (1 - p) - d3;
    double H5 = EMH2 * p + d3;
    double H6 = EMH3 * (1 - p) - d4;
    double H7 = EMH3 * p + d4;
    double f0 = user_model->f0, f1_pat = user_model->f1_pat, f1_mat = user_model->f1_mat, f2 = user_model->f2;
    sf(ps, "   mod     p    F0  F1_P  F1_M    F2    H0    H1    H2    H3    H4    H5    H6    H7  EMH0  EMH1  EMH2  EMH3    d1    d2    d3    d4\n");
    pr();
    sf(ps, "%6.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f\n", mod, p, f0, f1_pat, f1_mat, f2, H0, H1, H2, H3, H4, H5, H6, H7, EMH0, EMH1, EMH2, EMH3, d1, d2, d3, d4);
    pr();
}

static void print_maximum_2_tm_pat_first_without_deltas(double maximum, const double *x, DoubleVector em_hapfreqs)
{
    double mod = maximum;
    double H0 = x[0], H1 = x[1], H2 = x[2], H3 = x[3], H4 = x[4], H5 = x[5], H6 = x[6], H7 = x[7];
    double p = x[8];
    double f0, f1_pat, f1_mat, f2;
    if (!imprinting) {
        f0 = x[9]; f1_pat = x[10]; f1_mat = x[10]; f2 = x[11];
    } else {
        f0 = x[9]; f1_pat = x[10]; f1_mat = x[11]; f2 = x[12];
    }
    double EMH0 = DoubleVector_Get(em_hapfreqs, 0);
    double EMH1 = DoubleVector_Get(em_hapfreqs, 1);
    double EMH2 = DoubleVector_Get(em_hapfreqs, 2);
    double EMH3 = DoubleVector_Get(em_hapfreqs, 3);
    sf(ps, "   mod     p    F0  F1_P  F1_M    F2    H0    H1    H2    H3    H4    H5    H6    H7  EMH0  EMH1  EMH2  EMH3\n");
    pr();
    sf(ps, "%6.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f\n", mod, p, f0, f1_pat, f1_mat, f2, H0, H1, H2, H3, H4, H5, H6, H7, EMH0, EMH1, EMH2, EMH3);
    pr();
}

static void print_maximum_2_tm_mat_first_without_deltas(double maximum, const double *x, DoubleVector em_hapfreqs)
{
    double mod = maximum;
    double H0 = x[0], H1 = x[1], H2 = x[2], H3 = x[3], H4 = x[4], H5 = x[5], H6 = x[6], H7 = x[7];
    double p = x[8];
    double f0, f1_pat, f1_mat, f2;
    if (!imprinting) {
        f0 = x[9]; f1_pat = x[10]; f1_mat = x[10]; f2 = x[11];
    } else {
        f0 = x[9]; f1_pat = x[11]; f1_mat = x[10]; f2 = x[12];
    }
    double EMH0 = DoubleVector_Get(em_hapfreqs, 0);
    double EMH1 = DoubleVector_Get(em_hapfreqs, 1);
    double EMH2 = DoubleVector_Get(em_hapfreqs, 2);
    double EMH3 = DoubleVector_Get(em_hapfreqs, 3);
    sf(ps, "   mod     p    F0  F1_P  F1_M    F2    H0    H1    H2    H3    H4    H5    H6    H7  EMH0  EMH1  EMH2  EMH3\n");
    pr();
    sf(ps, "%6.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f\n", mod, p, f0, f1_pat, f1_mat, f2, H0, H1, H2, H3, H4, H5, H6, H7, EMH0, EMH1, EMH2, EMH3);
    pr();
}

static void print_maximum_user_model_2_tm_without_deltas(double maximum, const double *x, DoubleVector em_hapfreqs, ModelParams user_model)
{
    double mod = maximum;
    double H0 = x[0], H1 = x[1], H2 = x[2], H3 = x[3], H4 = x[4], H5 = x[5], H6 = x[6], H7 = x[7];
    double p = user_model->disease_allele_frequency;
    double f0 = user_model->f0, f1_pat = user_model->f1_pat, f1_mat = user_model->f1_mat, f2 = user_model->f2;
    double EMH0 = DoubleVector_Get(em_hapfreqs, 0);
    double EMH1 = DoubleVector_Get(em_hapfreqs, 1);
    double EMH2 = DoubleVector_Get(em_hapfreqs, 2);
    double EMH3 = DoubleVector_Get(em_hapfreqs, 3);
    sf(ps, "   mod     p    F0  F1_P  F1_M    F2    H0    H1    H2    H3    H4    H5    H6    H7  EMH0  EMH1  EMH2  EMH3\n");
    pr();
    sf(ps, "%6.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f\n", mod, p, f0, f1_pat, f1_mat, f2, H0, H1, H2, H3, H4, H5, H6, H7, EMH0, EMH1, EMH2, EMH3);
    pr();
}

static void update_global_minimum_2_tm_pat_first(double value, const double *x, void *data) {
    int n = 0;
    double *grad = 0;
    if (value >= global_minimum) return;
    if (penetrance_restriction) {
        if (imprinting) {
            if (f0_f1_pat_constr_2_tm_pat_first(n, x, grad, data) > PEN_TOL) return;
            if (f0_f1_mat_constr_2_tm_pat_first(n, x, grad, data) > PEN_TOL) return;
            if (f1_pat_f2_constr_2_tm_pat_first(n, x, grad, data) > PEN_TOL) return;
            if (f1_mat_f2_constr_2_tm_pat_first(n, x, grad, data) > PEN_TOL) return;
        } else {
            if (f0_f1_constr_2_tm(n, x, grad, data) > PEN_TOL) return;
            if (f1_f2_constr_2_tm(n, x, grad, data) > PEN_TOL) return;
        }
    }
    if (fabs(ds_group_constr_2_tm(n, x, grad, data)) > FREQ_TOL) return;
    if (d1_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d1_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d2_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d2_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d3_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d3_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d4_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d4_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    global_minimum = value;
    for (int i = 0; i < 8 + imprinting; i++) {
        global_x[i] = x[i];
    }
}

static void update_global_minimum_2_tm_mat_first(double value, const double *x, void *data) {
    int n = 0;
    double *grad = 0;
    if (value >= global_minimum) return;
    if (penetrance_restriction) {
        if (imprinting) {
            if (f0_f1_pat_constr_2_tm_mat_first(n, x, grad, data) > PEN_TOL) return;
            if (f0_f1_mat_constr_2_tm_mat_first(n, x, grad, data) > PEN_TOL) return;
            if (f1_pat_f2_constr_2_tm_mat_first(n, x, grad, data) > PEN_TOL) return;
            if (f1_mat_f2_constr_2_tm_mat_first(n, x, grad, data) > PEN_TOL) return;
        } else {
            if (f0_f1_constr_2_tm(n, x, grad, data) > PEN_TOL) return;
            if (f1_f2_constr_2_tm(n, x, grad, data) > PEN_TOL) return;
        }
    }
    if (fabs(ds_group_constr_2_tm(n, x, grad, data)) > FREQ_TOL) return;
    if (d1_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d1_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d2_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d2_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d3_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d3_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d4_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d4_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    global_minimum = value;
    for (int i = 0; i < 8 + imprinting; i++) {
        global_x[i] = x[i];
    }
}

static void update_global_minimum_user_model_2_tm(double value, const double *x, void *data) {
    int n = 0;
    double *grad = 0;
    if (value >= global_minimum) return;
    if (fabs(ds_group_constr_2_tm(n, x, grad, data)) > FREQ_TOL) return;
    if (d1_user_model_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d1_user_model_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d2_user_model_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d2_user_model_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d3_user_model_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d3_user_model_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d4_user_model_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d4_user_model_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    global_minimum = value;
    for (int i = 0; i < 4; i++) {
        global_x[i] = x[i];
    }
}

static void update_global_minimum_2_tm_pat_first_without_deltas(double value, const double *x, void *data) {
    int n = 0;
    double *grad = 0;
    if (value >= global_minimum) return;
    if (penetrance_restriction) {
        if (imprinting) {
            if (f0_f1_pat_constr_2_tm_pat_first_without_deltas(n, x, grad, data) > PEN_TOL) return;
            if (f0_f1_mat_constr_2_tm_pat_first_without_deltas(n, x, grad, data) > PEN_TOL) return;
            if (f1_pat_f2_constr_2_tm_pat_first_without_deltas(n, x, grad, data) > PEN_TOL) return;
            if (f1_mat_f2_constr_2_tm_pat_first_without_deltas(n, x, grad, data) > PEN_TOL) return;
        } else {
            if (f0_f1_constr_2_tm_without_deltas(n, x, grad, data) > PEN_TOL) return;
            if (f1_f2_constr_2_tm_without_deltas(n, x, grad, data) > PEN_TOL) return;
        }
    }
    if (fabs(H_even_constr_2_tm_without_deltas(n, x, grad, data)) > FREQ_TOL) return;
    if (fabs(H_odd_constr_2_tm_without_deltas(n, x, grad, data)) > FREQ_TOL) return;
    global_minimum = value;
    for (int i = 0; i < 12 + imprinting; i++) {
        global_x[i] = x[i];
    }
}

static void update_global_minimum_2_tm_mat_first_without_deltas(double value, const double *x, void *data) {
    int n = 0;
    double *grad = 0;
    if (value >= global_minimum) return;
    if (penetrance_restriction) {
        if (imprinting) {
            if (f0_f1_pat_constr_2_tm_mat_first_without_deltas(n, x, grad, data) > PEN_TOL) return;
            if (f0_f1_mat_constr_2_tm_mat_first_without_deltas(n, x, grad, data) > PEN_TOL) return;
            if (f1_pat_f2_constr_2_tm_mat_first_without_deltas(n, x, grad, data) > PEN_TOL) return;
            if (f1_mat_f2_constr_2_tm_mat_first_without_deltas(n, x, grad, data) > PEN_TOL) return;
        } else {
            if (f0_f1_constr_2_tm_without_deltas(n, x, grad, data) > PEN_TOL) return;
            if (f1_f2_constr_2_tm_without_deltas(n, x, grad, data) > PEN_TOL) return;
        }
    }
    if (fabs(H_even_constr_2_tm_without_deltas(n, x, grad, data)) > FREQ_TOL) return;
    if (fabs(H_odd_constr_2_tm_without_deltas(n, x, grad, data)) > FREQ_TOL) return;
    global_minimum = value;
    for (int i = 0; i < 12 + imprinting; i++) {
        global_x[i] = x[i];
    }
}

static void update_global_minimum_user_model_2_tm_without_deltas(double value, const double *x, void *data) {
    int n = 0;
    double *grad = 0;
    if (value >= global_minimum) return;
    if (fabs(H_even_user_model_constr_2_tm_without_deltas(n, x, grad, data)) > FREQ_TOL) return;
    if (fabs(H_odd_user_model_constr_2_tm_without_deltas(n, x, grad, data)) > FREQ_TOL) return;
    global_minimum = value;
    for (int i = 0; i < 8; i++) {
        global_x[i] = x[i];
    }
}

static double target_2_tm_pat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    objective_function_calls++;
    nlopt_iterations++;
    Data d = data;
    const double *deltas = x;
    double p = x[4];
    double f0, f1_pat, f1_mat, f2;
    DoubleVector hapfreqs = DoubleVector_Create(8);
    for (int i = 0; i < 4; i++) {
        // The order of hapfreqs is always H0, H1, ..., where hapfreqs
        // H0, H2, etc. contain the wildtype allele and hapfreqs H1,
        // H3, etc. contain the mutant allele.  To get this order the
        // even indexes must be wildtype hapfreqs with wildtype allele
        // frequency q and the odd indexes must be mutant hapfreqs
        // with mutant allele frequency p.
        DoubleVector_Set(hapfreqs, 2*i,     DoubleVector_Get(d->em_hapfreqs, i) * (1-p) - deltas[i]);
        DoubleVector_Set(hapfreqs, 2*i + 1, DoubleVector_Get(d->em_hapfreqs, i) * p + deltas[i]);
    }
    if (!imprinting) {
        f0 = x[5]; f1_pat = x[6]; f1_mat = x[6]; f2 = x[7];
    } else {
        f0 = x[5]; f1_pat = x[6]; f1_mat = x[7]; f2 = x[8];
    }
    DoubleVector penetrances = DoubleVector_With(10, f0, f1_pat, f1_mat, f2, 1-f0, 1-f1_pat, 1-f1_mat, 1-f2, .5 * (f1_pat + f1_mat), .5 * ((1-f1_pat) + (1-f1_mat)));
    double score = ModScorer_ComputeModScore(0, d->annotated_families, d->test_markers, p, hapfreqs, penetrances, d->em_hapfreqs);
    DoubleVector_Destroy(hapfreqs);
    DoubleVector_Destroy(penetrances);
    update_global_minimum_2_tm_pat_first(-score, x, data);
    return -score;
}

static double target_2_tm_mat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    objective_function_calls++;
    nlopt_iterations++;
    Data d = data;
    const double *deltas = x;
    double p = x[4];
    double f0, f1_pat, f1_mat, f2;
    DoubleVector hapfreqs = DoubleVector_Create(8);
    for (int i = 0; i < 4; i++) {
        // The order of hapfreqs is always H0, H1, ..., where hapfreqs
        // H0, H2, etc. contain the wildtype allele and hapfreqs H1,
        // H3, etc. contain the mutant allele.  To get this order the
        // even indexes must be wildtype hapfreqs with wildtype allele
        // frequency q and the odd indexes must be mutant hapfreqs
        // with mutant allele frequency p.
        DoubleVector_Set(hapfreqs, 2*i,     DoubleVector_Get(d->em_hapfreqs, i) * (1-p) - deltas[i]);
        DoubleVector_Set(hapfreqs, 2*i + 1, DoubleVector_Get(d->em_hapfreqs, i) * p + deltas[i]);
    }
    if (!imprinting) {
        f0 = x[5]; f1_pat = x[6]; f1_mat = x[6]; f2 = x[7];
    } else {
        f0 = x[5]; f1_pat = x[7]; f1_mat = x[6]; f2 = x[8];
    }
    DoubleVector penetrances = DoubleVector_With(10, f0, f1_pat, f1_mat, f2, 1-f0, 1-f1_pat, 1-f1_mat, 1-f2, .5 * (f1_pat + f1_mat), .5 * ((1-f1_pat) + (1-f1_mat)));
    double score = ModScorer_ComputeModScore(0, d->annotated_families, d->test_markers, p, hapfreqs, penetrances, d->em_hapfreqs);
    DoubleVector_Destroy(hapfreqs);
    DoubleVector_Destroy(penetrances);
    update_global_minimum_2_tm_mat_first(-score, x, data);
    return -score;
}

static double target_user_model_2_tm(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    objective_function_calls++;
    nlopt_iterations++;
    Data d = data;
    const double *deltas = x;
    double p = d->user_model->disease_allele_frequency;
    DoubleVector hapfreqs = DoubleVector_Create(8);
    for (int i = 0; i < 4; i++) {
        // The order of hapfreqs is always H0, H1, ..., where hapfreqs
        // H0, H2, etc. contain the wildtype allele and hapfreqs H1,
        // H3, etc. contain the mutant allele.  To get this order the
        // even indexes must be wildtype hapfreqs with wildtype allele
        // frequency q and the odd indexes must be mutant hapfreqs
        // with mutant allele frequency p.
        DoubleVector_Set(hapfreqs, 2*i,     DoubleVector_Get(d->em_hapfreqs, i) * (1-p) - deltas[i]);
        DoubleVector_Set(hapfreqs, 2*i + 1, DoubleVector_Get(d->em_hapfreqs, i) * p + deltas[i]);
    }
    double f0 = d->user_model->f0, f1_pat = d->user_model->f1_pat, f1_mat = d->user_model->f1_mat, f2 = d->user_model->f2;
    DoubleVector penetrances = DoubleVector_With(10, f0, f1_pat, f1_mat, f2, 1-f0, 1-f1_pat, 1-f1_mat, 1-f2, .5 * (f1_pat + f1_mat), .5 * ((1-f1_pat) + (1-f1_mat)));
    double score = ModScorer_ComputeModScore(0, d->annotated_families, d->test_markers, p, hapfreqs, penetrances, d->em_hapfreqs);
    DoubleVector_Destroy(hapfreqs);
    DoubleVector_Destroy(penetrances);
    update_global_minimum_user_model_2_tm(-score, x, data);
    return -score;
}

static double target_2_tm_pat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    objective_function_calls++;
    nlopt_iterations++;
    Data d = data;
    const double *H = x;
    DoubleVector hapfreqs = DoubleVector_Create(8);
    for (int i = 0; i < 8; i++) {
        DoubleVector_PushBack(hapfreqs, H[i]);
    }
    double p = x[8];
    double f0, f1_pat, f1_mat, f2;
    if (!imprinting) {
        f0 = x[9]; f1_pat = x[10]; f1_mat = x[10]; f2 = x[11];
    } else {
        f0 = x[9]; f1_pat = x[10]; f1_mat = x[11]; f2 = x[12];
    }
    DoubleVector penetrances = DoubleVector_With(10, f0, f1_pat, f1_mat, f2, 1-f0, 1-f1_pat, 1-f1_mat, 1-f2, .5 * (f1_pat + f1_mat), .5 * ((1-f1_pat) + (1-f1_mat)));
    double score = ModScorer_ComputeModScore(0, d->annotated_families, d->test_markers, p, hapfreqs, penetrances, d->em_hapfreqs);
    DoubleVector_Destroy(hapfreqs);
    DoubleVector_Destroy(penetrances);
    update_global_minimum_2_tm_pat_first_without_deltas(-score, x, data);
    return -score;
}

static double target_2_tm_mat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    objective_function_calls++;
    nlopt_iterations++;
    Data d = data;
    const double *H = x;
    DoubleVector hapfreqs = DoubleVector_Create(8);
    for (int i = 0; i < 8; i++) {
        DoubleVector_PushBack(hapfreqs, H[i]);
    }
    double p = x[8];
    double f0, f1_pat, f1_mat, f2;
    if (!imprinting) {
        f0 = x[9]; f1_pat = x[10]; f1_mat = x[10]; f2 = x[11];
    } else {
        f0 = x[9]; f1_pat = x[11]; f1_mat = x[10]; f2 = x[12];
    }
    DoubleVector penetrances = DoubleVector_With(10, f0, f1_pat, f1_mat, f2, 1-f0, 1-f1_pat, 1-f1_mat, 1-f2, .5 * (f1_pat + f1_mat), .5 * ((1-f1_pat) + (1-f1_mat)));
    double score = ModScorer_ComputeModScore(0, d->annotated_families, d->test_markers, p, hapfreqs, penetrances, d->em_hapfreqs);
    DoubleVector_Destroy(hapfreqs);
    DoubleVector_Destroy(penetrances);
    update_global_minimum_2_tm_mat_first_without_deltas(-score, x, data);
    return -score;
}

static double target_user_model_2_tm_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    objective_function_calls++;
    nlopt_iterations++;
    Data d = data;
    const double *H = x;
    DoubleVector hapfreqs = DoubleVector_Create(8);
    for (int i = 0; i < 8; i++) {
        DoubleVector_PushBack(hapfreqs, H[i]);
    }
    double p = d->user_model->disease_allele_frequency;
    double f0 = d->user_model->f0, f1_pat = d->user_model->f1_pat, f1_mat = d->user_model->f1_mat, f2 = d->user_model->f2;
    DoubleVector penetrances = DoubleVector_With(10, f0, f1_pat, f1_mat, f2, 1-f0, 1-f1_pat, 1-f1_mat, 1-f2, .5 * (f1_pat + f1_mat), .5 * ((1-f1_pat) + (1-f1_mat)));
    double score = ModScorer_ComputeModScore(0, d->annotated_families, d->test_markers, p, hapfreqs, penetrances, d->em_hapfreqs);
    DoubleVector_Destroy(hapfreqs);
    DoubleVector_Destroy(penetrances);
    update_global_minimum_user_model_2_tm_without_deltas(-score, x, data);
    return -score;
}

static int optimize_with_2_test_markers_pat_first(
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModScore *gm,
    ModScore *mod_score,
    double initial_step_size1)
{
    //             d1  d2  d3  d4  p       f0 f1_pat f1_mat f2
    double lb[] = {-1, -1, -1, -1, 1e-8,   0, 0,     0,     0};
    double ub[] = { 1,  1,  1,  1, 1-1e-8, 1, 1,     1,     1};
    if (allfreq_restriction) {
        ub[4] = highest_allfreq;
    }
    struct DataStruct data = {
        annotated_families,
        test_markers,
        em_hapfreqs,
        0
    };

    nlopt_opt opt = nlopt_create(NLOPT_LN_COBYLA, 9); // algorithm and dimensionality
    nlopt_set_lower_bounds(opt, lb);
    nlopt_set_upper_bounds(opt, ub);
    nlopt_set_min_objective(opt, target_2_tm_pat_first, &data);
    nlopt_add_equality_constraint(opt, ds_group_constr_2_tm, &data, FREQ_TOL);
    if (penetrance_restriction) {
        nlopt_add_inequality_constraint(opt, f0_f1_pat_constr_2_tm_pat_first, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f0_f1_mat_constr_2_tm_pat_first, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_pat_f2_constr_2_tm_pat_first, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_mat_f2_constr_2_tm_pat_first, &data, DBL_EPSILON);
    }
    nlopt_add_inequality_constraint(opt, d1_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d1_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d2_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d2_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d3_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d3_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d4_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d4_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_set_xtol_rel(opt, 1e-4);
    int max_eval = MIN(g_jla_nlopt_iterations_free, 10000);
    nlopt_set_maxeval(opt, max_eval);

    double x[] = {
        DoubleVector_Get(gm->deltas, 0),
        DoubleVector_Get(gm->deltas, 1),
        DoubleVector_Get(gm->deltas, 2),
        DoubleVector_Get(gm->deltas, 3),
        gm->disease_allele_frequency,
        DoubleVector_Get(gm->penetrances, 0),
        DoubleVector_Get(gm->penetrances, 1),
        DoubleVector_Get(gm->penetrances, 2),
        DoubleVector_Get(gm->penetrances, 3)
    };
    nlopt_iterations = 0;
    nlopt_set_initial_step1(opt, initial_step_size1);

    double negative_mod_score; // minimum value of target function
    int nlopt_success = 1;
    if (nlopt_optimize(opt, x, &negative_mod_score) < 0) {
        nlopt_success = 0;
        sf(ps, "nlopt failed!\n");
        pr();
    } else {
        if (negative_mod_score != global_minimum) {
            sf(ps, "COBYLA = %f vs %f = global minimum\n", negative_mod_score, global_minimum);
            pr();
            negative_mod_score = global_minimum;
            for (int i = 0; i < 8+1; i++) {
                x[i] = global_x[i];
            }
        }
        sf(ps, "No. of nlopt iterations: %d\n", nlopt_iterations);
        pr();
        g_jla_nlopt_iterations_free -= nlopt_iterations;
        DoubleVector hapfreqs = DoubleVector_Create(8);
        DoubleVector ds = DoubleVector_Create(4);
        DoubleVector_PushBack(ds, x[0]);
        DoubleVector_PushBack(ds, x[1]);
        DoubleVector_PushBack(ds, x[2]);
        DoubleVector_PushBack(ds, x[3]);
        double *deltas = x;
        double p = x[4];
        for (int i = 0; i < 8; i++) {
            DoubleVector_Set(hapfreqs, 2*i,     DoubleVector_Get(em_hapfreqs, i) * (1-p) - deltas[i]);
            DoubleVector_Set(hapfreqs, 2*i + 1, DoubleVector_Get(em_hapfreqs, i) * p + deltas[i]);
        }
        DoubleVector penetrances = DoubleVector_Create(4);
        DoubleVector_PushBack(penetrances, x[5]);
        DoubleVector_PushBack(penetrances, x[6]);
        DoubleVector_PushBack(penetrances, x[7]);
        DoubleVector_PushBack(penetrances, x[8]);
        mod_score->test_markers = test_markers;
        mod_score->position = position;
        mod_score->mod_score = -negative_mod_score;
        mod_score->disease_allele_frequency = p;
        mod_score->penetrances = penetrances;
        mod_score->hapfreqs = hapfreqs;
        mod_score->em_hapfreqs = em_hapfreqs;
        mod_score->deltas = ds;
        print_maximum_2_tm_pat_first(-negative_mod_score, x, em_hapfreqs);
    }
    nlopt_destroy(opt);
    return nlopt_success;
}

static int optimize_with_2_test_markers_mat_first(
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModScore *gm,
    ModScore *mod_score,
    double initial_step_size1)
{
    //             d1  d2  d3  d4  p       f0 f1_mat f1_pat f2
    double lb[] = {-1, -1, -1, -1, 1e-8,   0, 0,     0,     0};
    double ub[] = { 1,  1,  1,  1, 1-1e-8, 1, 1,     1,     1};
    if (allfreq_restriction) {
        ub[4] = highest_allfreq;
    }
    struct DataStruct data = {
        annotated_families,
        test_markers,
        em_hapfreqs,
        0
    };

    nlopt_opt opt = nlopt_create(NLOPT_LN_COBYLA, 9); // algorithm and dimensionality
    nlopt_set_lower_bounds(opt, lb);
    nlopt_set_upper_bounds(opt, ub);
    nlopt_set_min_objective(opt, target_2_tm_mat_first, &data);
    nlopt_add_equality_constraint(opt, ds_group_constr_2_tm, &data, FREQ_TOL);
    if (penetrance_restriction) {
        nlopt_add_inequality_constraint(opt, f0_f1_pat_constr_2_tm_mat_first, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f0_f1_mat_constr_2_tm_mat_first, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_pat_f2_constr_2_tm_mat_first, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_mat_f2_constr_2_tm_mat_first, &data, DBL_EPSILON);
    }
    nlopt_add_inequality_constraint(opt, d1_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d1_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d2_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d2_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d3_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d3_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d4_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d4_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_set_xtol_rel(opt, 1e-4);
    int max_eval = MIN(g_jla_nlopt_iterations_free, 10000);
    nlopt_set_maxeval(opt, max_eval);

    double x[] = {
        DoubleVector_Get(gm->deltas, 0),
        DoubleVector_Get(gm->deltas, 1),
        DoubleVector_Get(gm->deltas, 2),
        DoubleVector_Get(gm->deltas, 3),
        gm->disease_allele_frequency,
        DoubleVector_Get(gm->penetrances, 0),
        DoubleVector_Get(gm->penetrances, 2),
        DoubleVector_Get(gm->penetrances, 1),
        DoubleVector_Get(gm->penetrances, 3)
    };
    nlopt_iterations = 0;
    nlopt_set_initial_step1(opt, initial_step_size1);

    double negative_mod_score; // minimum value of target function
    int nlopt_success = 1;
    if (nlopt_optimize(opt, x, &negative_mod_score) < 0) {
        nlopt_success = 0;
        sf(ps, "nlopt failed!\n");
        pr();
    } else {
        if (negative_mod_score != global_minimum) {
            sf(ps, "COBYLA = %f vs %f = global minimum\n", negative_mod_score, global_minimum);
            pr();
            negative_mod_score = global_minimum;
            for (int i = 0; i < 8+1; i++) {
                x[i] = global_x[i];
            }
        }
        sf(ps, "No. of nlopt iterations: %d\n", nlopt_iterations);
        pr();
        g_jla_nlopt_iterations_free -= nlopt_iterations;
        DoubleVector hapfreqs = DoubleVector_Create(8);
        DoubleVector ds = DoubleVector_Create(4);
        DoubleVector_PushBack(ds, x[0]);
        DoubleVector_PushBack(ds, x[1]);
        DoubleVector_PushBack(ds, x[2]);
        DoubleVector_PushBack(ds, x[3]);
        double *deltas = x;
        double p = x[4];
        for (int i = 0; i < 8; i++) {
            DoubleVector_Set(hapfreqs, 2*i,     DoubleVector_Get(em_hapfreqs, i) * (1-p) - deltas[i]);
            DoubleVector_Set(hapfreqs, 2*i + 1, DoubleVector_Get(em_hapfreqs, i) * p + deltas[i]);
        }
        DoubleVector penetrances = DoubleVector_Create(4);
        DoubleVector_PushBack(penetrances, x[5]);
        DoubleVector_PushBack(penetrances, x[7]);
        DoubleVector_PushBack(penetrances, x[6]);
        DoubleVector_PushBack(penetrances, x[8]);
        mod_score->test_markers = test_markers;
        mod_score->position = position;
        mod_score->mod_score = -negative_mod_score;
        mod_score->disease_allele_frequency = p;
        mod_score->penetrances = penetrances;
        mod_score->hapfreqs = hapfreqs;
        mod_score->em_hapfreqs = em_hapfreqs;
        mod_score->deltas = ds;
        print_maximum_2_tm_mat_first(-negative_mod_score, x, em_hapfreqs);
    }
    nlopt_destroy(opt);
    return nlopt_success;
}

static int optimize_with_2_test_markers_no_imprinting(
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModScore *gm,
    ModScore *mod_score,
    double initial_step_size1)
{
    //             d1  d2  d3  d4  p       f0 f1 f2
    double lb[] = {-1, -1, -1, -1, 1e-8,   0, 0, 0};
    double ub[] = { 1,  1,  1,  1, 1-1e-8, 1, 1, 1};
    if (allfreq_restriction) {
        ub[4] = highest_allfreq;
    }
    struct DataStruct data = {
        annotated_families,
        test_markers,
        em_hapfreqs,
        0
    };

    nlopt_opt opt = nlopt_create(NLOPT_LN_COBYLA, 8); // algorithm and dimensionality
    nlopt_set_lower_bounds(opt, lb);
    nlopt_set_upper_bounds(opt, ub);
    nlopt_set_min_objective(opt, target_2_tm_pat_first, &data);
    nlopt_add_equality_constraint(opt, ds_group_constr_2_tm, &data, FREQ_TOL);
    if (penetrance_restriction) {
        nlopt_add_inequality_constraint(opt, f0_f1_constr_2_tm, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_f2_constr_2_tm, &data, DBL_EPSILON);
    }
    nlopt_add_inequality_constraint(opt, d1_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d1_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d2_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d2_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d3_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d3_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d4_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d4_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_set_xtol_rel(opt, 1e-4);
    int max_eval = MIN(g_jla_nlopt_iterations_free, 10000);
    nlopt_set_maxeval(opt, max_eval);

    double x[] = {
        DoubleVector_Get(gm->deltas, 0),
        DoubleVector_Get(gm->deltas, 1),
        DoubleVector_Get(gm->deltas, 2),
        DoubleVector_Get(gm->deltas, 3),
        gm->disease_allele_frequency,
        DoubleVector_Get(gm->penetrances, 0),
        DoubleVector_Get(gm->penetrances, 1),
        DoubleVector_Get(gm->penetrances, 3)
    };
    nlopt_iterations = 0;
    nlopt_set_initial_step1(opt, initial_step_size1);

    double negative_mod_score; // minimum value of target function
    int nlopt_success = 1;
    if (nlopt_optimize(opt, x, &negative_mod_score) < 0) {
        nlopt_success = 0;
        sf(ps, "nlopt failed!\n");
        pr();
    } else {
        if (negative_mod_score != global_minimum) {
            sf(ps, "COBYLA = %f vs %f = global minimum\n", negative_mod_score, global_minimum);
            pr();
            negative_mod_score = global_minimum;
            for (int i = 0; i < 8; i++) {
                x[i] = global_x[i];
            }
        }
        sf(ps, "No. of nlopt iterations: %d\n", nlopt_iterations);
        pr();
        g_jla_nlopt_iterations_free -= nlopt_iterations;
        DoubleVector hapfreqs = DoubleVector_Create(8);
        DoubleVector ds = DoubleVector_Create(4);
        DoubleVector_PushBack(ds, x[0]);
        DoubleVector_PushBack(ds, x[1]);
        DoubleVector_PushBack(ds, x[2]);
        DoubleVector_PushBack(ds, x[3]);
        double *deltas = x;
        double p = x[4];
        for (int i = 0; i < 8; i++) {
            DoubleVector_Set(hapfreqs, 2*i,     DoubleVector_Get(em_hapfreqs, i) * (1-p) - deltas[i]);
            DoubleVector_Set(hapfreqs, 2*i + 1, DoubleVector_Get(em_hapfreqs, i) * p + deltas[i]);
        }
        DoubleVector penetrances = DoubleVector_Create(4);
        DoubleVector_PushBack(penetrances, x[5]);
        DoubleVector_PushBack(penetrances, x[6]);
        DoubleVector_PushBack(penetrances, x[6]);
        DoubleVector_PushBack(penetrances, x[7]);
        mod_score->test_markers = test_markers;
        mod_score->position = position;
        mod_score->mod_score = -negative_mod_score;
        mod_score->disease_allele_frequency = p;
        mod_score->penetrances = penetrances;
        mod_score->hapfreqs = hapfreqs;
        mod_score->em_hapfreqs = em_hapfreqs;
        mod_score->deltas = ds;
        print_maximum_2_tm_pat_first(-negative_mod_score, x, em_hapfreqs);
    }
    nlopt_destroy(opt);
    return nlopt_success;
}

static int optimize_user_model_with_2_test_markers(
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModelParams user_model,
    ModScore *gm,
    ModScore *mod_score,
    double initial_step_size1)
{
    //             d1  d2  d3  d4
    double lb[] = {-1, -1, -1, -1};
    double ub[] = { 1,  1,  1,  1};
    struct DataStruct data = {
        annotated_families,
        test_markers,
        em_hapfreqs,
        user_model
    };

    nlopt_opt opt = nlopt_create(NLOPT_LN_COBYLA, 4); // algorithm and dimensionality
    nlopt_set_lower_bounds(opt, lb);
    nlopt_set_upper_bounds(opt, ub);
    nlopt_set_min_objective(opt, target_user_model_2_tm, &data);
    nlopt_add_equality_constraint(opt, ds_group_constr_2_tm, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d1_user_model_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d1_user_model_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d2_user_model_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d2_user_model_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d3_user_model_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d3_user_model_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d4_user_model_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d4_user_model_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_set_xtol_rel(opt, 1e-4);
    int max_eval = MIN(g_jla_nlopt_iterations_free, 10000);
    nlopt_set_maxeval(opt, max_eval);

    double x[] = {
        DoubleVector_Get(gm->deltas, 0),
        DoubleVector_Get(gm->deltas, 1),
        DoubleVector_Get(gm->deltas, 2),
        DoubleVector_Get(gm->deltas, 3)
    };
    nlopt_iterations = 0;
    nlopt_set_initial_step1(opt, initial_step_size1);

    double negative_mod_score; // minimum value of target function
    int nlopt_success = 1;
    if (nlopt_optimize(opt, x, &negative_mod_score) < 0) {
        nlopt_success = 0;
        sf(ps, "nlopt failed!\n");
        pr();
    } else {
        if (negative_mod_score != global_minimum) {
            sf(ps, "COBYLA = %f vs %f = global minimum\n", negative_mod_score, global_minimum);
            pr();
            negative_mod_score = global_minimum;
            for (int i = 0; i < 4; i++) {
                x[i] = global_x[i];
            }
        }
        sf(ps, "No. of nlopt iterations: %d\n", nlopt_iterations);
        pr();
        g_jla_nlopt_iterations_free -= nlopt_iterations;
        DoubleVector hapfreqs = DoubleVector_Create(8);
        DoubleVector ds = DoubleVector_Create(4);
        DoubleVector_PushBack(ds, x[0]);
        DoubleVector_PushBack(ds, x[1]);
        DoubleVector_PushBack(ds, x[2]);
        DoubleVector_PushBack(ds, x[3]);
        double *deltas = x;
        double p = user_model->disease_allele_frequency;
        for (int i = 0; i < 8; i++) {
            DoubleVector_Set(hapfreqs, 2*i,     DoubleVector_Get(em_hapfreqs, i) * (1-p) - deltas[i]);
            DoubleVector_Set(hapfreqs, 2*i + 1, DoubleVector_Get(em_hapfreqs, i) * p + deltas[i]);
        }
        DoubleVector penetrances = DoubleVector_Create(4);
        DoubleVector_PushBack(penetrances, user_model->f0);
        DoubleVector_PushBack(penetrances, user_model->f1_pat);
        DoubleVector_PushBack(penetrances, user_model->f1_mat);
        DoubleVector_PushBack(penetrances, user_model->f2);
        mod_score->test_markers = test_markers;
        mod_score->position = position;
        mod_score->mod_score = -negative_mod_score;
        mod_score->disease_allele_frequency = p;
        mod_score->penetrances = penetrances;
        mod_score->hapfreqs = hapfreqs;
        mod_score->em_hapfreqs = em_hapfreqs;
        mod_score->deltas = ds;
        print_maximum_user_model_2_tm(-negative_mod_score, x, em_hapfreqs, user_model);
    }
    nlopt_destroy(opt);
    return nlopt_success;
}

static int optimize_with_2_test_markers_pat_first_without_deltas(
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModScore *gm,
    ModScore *mod_score,
    double initial_step_size1)
{
    //              H0 H1 H2 H3 H4 H5 H6 H7 p       f0 f1_pat f1_mat f2
    double lb[] = { 0, 0, 0, 0, 0, 0, 0, 0,   1e-8, 0, 0,     0,     0};
    double ub[] = { 1, 1, 1, 1, 1, 1, 1, 1, 1-1e-8, 1, 1,     1,     1};
    if (allfreq_restriction) {
        ub[8] = highest_allfreq;
    }
    struct DataStruct data = {
        annotated_families,
        test_markers,
        em_hapfreqs,
        0
    };

    nlopt_opt opt = nlopt_create(NLOPT_LN_COBYLA, 13); // algorithm and dimensionality
    nlopt_set_lower_bounds(opt, lb);
    nlopt_set_upper_bounds(opt, ub);
    nlopt_set_min_objective(opt, target_2_tm_pat_first_without_deltas, &data);
    nlopt_add_equality_constraint(opt, H_even_constr_2_tm_without_deltas, &data, DBL_EPSILON);
    nlopt_add_equality_constraint(opt, H_odd_constr_2_tm_without_deltas, &data, DBL_EPSILON);
    if (penetrance_restriction) {
        nlopt_add_inequality_constraint(opt, f0_f1_pat_constr_2_tm_pat_first_without_deltas, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f0_f1_mat_constr_2_tm_pat_first_without_deltas, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_pat_f2_constr_2_tm_pat_first_without_deltas, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_mat_f2_constr_2_tm_pat_first_without_deltas, &data, DBL_EPSILON);
    }
    nlopt_set_xtol_rel(opt, 1e-4);
    int max_eval = MIN(g_jla_nlopt_iterations_free, 10000);
    nlopt_set_maxeval(opt, max_eval);

    double x[] = {
        DoubleVector_Get(gm->hapfreqs, 0),
        DoubleVector_Get(gm->hapfreqs, 1),
        DoubleVector_Get(gm->hapfreqs, 2),
        DoubleVector_Get(gm->hapfreqs, 3),
        DoubleVector_Get(gm->hapfreqs, 4),
        DoubleVector_Get(gm->hapfreqs, 5),
        DoubleVector_Get(gm->hapfreqs, 6),
        DoubleVector_Get(gm->hapfreqs, 7),
        gm->disease_allele_frequency,
        DoubleVector_Get(gm->penetrances, 0),
        DoubleVector_Get(gm->penetrances, 1),
        DoubleVector_Get(gm->penetrances, 2),
        DoubleVector_Get(gm->penetrances, 3)
    };
    nlopt_iterations = 0;
    nlopt_set_initial_step1(opt, initial_step_size1);

    double negative_mod_score; // minimum value of target function
    int nlopt_success = 1;
    if (nlopt_optimize(opt, x, &negative_mod_score) < 0) {
        nlopt_success = 0;
        sf(ps, "nlopt failed!\n");
        pr();
    } else {
        if (negative_mod_score != global_minimum) {
            sf(ps, "COBYLA = %f vs %f = global minimum\n", negative_mod_score, global_minimum);
            pr();
            negative_mod_score = global_minimum;
            for (int i = 0; i < 12+1; i++) {
                x[i] = global_x[i];
            }
        }
        sf(ps, "No. of nlopt iterations: %d\n", nlopt_iterations);
        pr();
        g_jla_nlopt_iterations_free -= nlopt_iterations;
        DoubleVector hapfreqs = DoubleVector_Create(8);
        DoubleVector_PushBack(hapfreqs, x[0]);
        DoubleVector_PushBack(hapfreqs, x[1]);
        DoubleVector_PushBack(hapfreqs, x[2]);
        DoubleVector_PushBack(hapfreqs, x[3]);
        DoubleVector_PushBack(hapfreqs, x[4]);
        DoubleVector_PushBack(hapfreqs, x[5]);
        DoubleVector_PushBack(hapfreqs, x[6]);
        DoubleVector_PushBack(hapfreqs, x[7]);
        double p = x[8];
        DoubleVector penetrances = DoubleVector_Create(4);
        DoubleVector_PushBack(penetrances, x[9]);
        DoubleVector_PushBack(penetrances, x[10]);
        DoubleVector_PushBack(penetrances, x[11]);
        DoubleVector_PushBack(penetrances, x[12]);
        mod_score->test_markers = test_markers;
        mod_score->position = position;
        mod_score->mod_score = -negative_mod_score;
        mod_score->disease_allele_frequency = p;
        mod_score->penetrances = penetrances;
        mod_score->hapfreqs = hapfreqs;
        mod_score->em_hapfreqs = em_hapfreqs;
        print_maximum_2_tm_pat_first_without_deltas(-negative_mod_score, x, em_hapfreqs);
    }
    nlopt_destroy(opt);
    return nlopt_success;
}

static int optimize_with_2_test_markers_mat_first_without_deltas(
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModScore *gm,
    ModScore *mod_score,
    double initial_step_size1)
{
    //              H0 H1 H2 H3 H4 H5 H6 H7 p       f0 f1_mat f1_pat f2
    double lb[] = { 0, 0, 0, 0, 0, 0, 0, 0,   1e-8, 0, 0,     0,     0};
    double ub[] = { 1, 1, 1, 1, 1, 1, 1, 1, 1-1e-8, 1, 1,     1,     1};
    if (allfreq_restriction) {
        ub[8] = highest_allfreq;
    }
    struct DataStruct data = {
        annotated_families,
        test_markers,
        em_hapfreqs,
        0
    };

    nlopt_opt opt = nlopt_create(NLOPT_LN_COBYLA, 13); // algorithm and dimensionality
    nlopt_set_lower_bounds(opt, lb);
    nlopt_set_upper_bounds(opt, ub);
    nlopt_set_min_objective(opt, target_2_tm_mat_first_without_deltas, &data);
    nlopt_add_equality_constraint(opt, H_even_constr_2_tm_without_deltas, &data, DBL_EPSILON);
    nlopt_add_equality_constraint(opt, H_odd_constr_2_tm_without_deltas, &data, DBL_EPSILON);
    if (penetrance_restriction) {
        nlopt_add_inequality_constraint(opt, f0_f1_pat_constr_2_tm_mat_first_without_deltas, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f0_f1_mat_constr_2_tm_mat_first_without_deltas, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_pat_f2_constr_2_tm_mat_first_without_deltas, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_mat_f2_constr_2_tm_mat_first_without_deltas, &data, DBL_EPSILON);
    }
    nlopt_set_xtol_rel(opt, 1e-4);
    int max_eval = MIN(g_jla_nlopt_iterations_free, 10000);
    nlopt_set_maxeval(opt, max_eval);

    double x[] = {
        DoubleVector_Get(gm->hapfreqs, 0),
        DoubleVector_Get(gm->hapfreqs, 1),
        DoubleVector_Get(gm->hapfreqs, 2),
        DoubleVector_Get(gm->hapfreqs, 3),
        DoubleVector_Get(gm->hapfreqs, 4),
        DoubleVector_Get(gm->hapfreqs, 5),
        DoubleVector_Get(gm->hapfreqs, 6),
        DoubleVector_Get(gm->hapfreqs, 7),
        gm->disease_allele_frequency,
        DoubleVector_Get(gm->penetrances, 0),
        DoubleVector_Get(gm->penetrances, 2),
        DoubleVector_Get(gm->penetrances, 1),
        DoubleVector_Get(gm->penetrances, 3)
    };
    nlopt_iterations = 0;
    nlopt_set_initial_step1(opt, initial_step_size1);

    double negative_mod_score; // minimum value of target function
    int nlopt_success = 1;
    if (nlopt_optimize(opt, x, &negative_mod_score) < 0) {
        nlopt_success = 0;
        sf(ps, "nlopt failed!\n");
        pr();
    } else {
        if (negative_mod_score != global_minimum) {
            sf(ps, "COBYLA = %f vs %f = global minimum\n", negative_mod_score, global_minimum);
            pr();
            negative_mod_score = global_minimum;
            for (int i = 0; i < 12+1; i++) {
                x[i] = global_x[i];
            }
        }
        sf(ps, "No. of nlopt iterations: %d\n", nlopt_iterations);
        pr();
        g_jla_nlopt_iterations_free -= nlopt_iterations;
        DoubleVector hapfreqs = DoubleVector_Create(8);
        DoubleVector_PushBack(hapfreqs, x[0]);
        DoubleVector_PushBack(hapfreqs, x[1]);
        DoubleVector_PushBack(hapfreqs, x[2]);
        DoubleVector_PushBack(hapfreqs, x[3]);
        DoubleVector_PushBack(hapfreqs, x[4]);
        DoubleVector_PushBack(hapfreqs, x[5]);
        DoubleVector_PushBack(hapfreqs, x[6]);
        DoubleVector_PushBack(hapfreqs, x[7]);
        double p = x[8];
        DoubleVector penetrances = DoubleVector_Create(4);
        DoubleVector_PushBack(penetrances, x[9]);
        DoubleVector_PushBack(penetrances, x[11]);
        DoubleVector_PushBack(penetrances, x[10]);
        DoubleVector_PushBack(penetrances, x[12]);
        mod_score->test_markers = test_markers;
        mod_score->position = position;
        mod_score->mod_score = -negative_mod_score;
        mod_score->disease_allele_frequency = p;
        mod_score->penetrances = penetrances;
        mod_score->hapfreqs = hapfreqs;
        mod_score->em_hapfreqs = em_hapfreqs;
        print_maximum_2_tm_mat_first_without_deltas(-negative_mod_score, x, em_hapfreqs);
    }
    nlopt_destroy(opt);
    return nlopt_success;
}

static int optimize_with_2_test_markers_no_imprinting_without_deltas(
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModScore *gm,
    ModScore *mod_score,
    double initial_step_size1)
{
    //              H0 H1 H2 H3 H4 H5 H6 H7 p       f0 f1 f2
    double lb[] = { 0, 0, 0, 0, 0, 0, 0, 0,   1e-8, 0, 0, 0};
    double ub[] = { 1, 1, 1, 1, 1, 1, 1, 1, 1-1e-8, 1, 1, 1};
    if (allfreq_restriction) {
        ub[8] = highest_allfreq;
    }
    struct DataStruct data = {
        annotated_families,
        test_markers,
        em_hapfreqs,
        0
    };

    nlopt_opt opt = nlopt_create(NLOPT_LN_COBYLA, 12); // algorithm and dimensionality
    nlopt_set_lower_bounds(opt, lb);
    nlopt_set_upper_bounds(opt, ub);
    nlopt_set_min_objective(opt, target_2_tm_pat_first_without_deltas, &data);
    nlopt_add_equality_constraint(opt, H_even_constr_2_tm_without_deltas, &data, DBL_EPSILON);
    nlopt_add_equality_constraint(opt, H_odd_constr_2_tm_without_deltas, &data, DBL_EPSILON);
    if (penetrance_restriction) {
        nlopt_add_inequality_constraint(opt, f0_f1_constr_2_tm_without_deltas, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_f2_constr_2_tm_without_deltas, &data, DBL_EPSILON);
    }
    nlopt_set_xtol_rel(opt, 1e-4);
    int max_eval = MIN(g_jla_nlopt_iterations_free, 10000);
    nlopt_set_maxeval(opt, max_eval);

    double x[] = {
        DoubleVector_Get(gm->hapfreqs, 0),
        DoubleVector_Get(gm->hapfreqs, 1),
        DoubleVector_Get(gm->hapfreqs, 2),
        DoubleVector_Get(gm->hapfreqs, 3),
        DoubleVector_Get(gm->hapfreqs, 4),
        DoubleVector_Get(gm->hapfreqs, 5),
        DoubleVector_Get(gm->hapfreqs, 6),
        DoubleVector_Get(gm->hapfreqs, 7),
        gm->disease_allele_frequency,
        DoubleVector_Get(gm->penetrances, 0),
        DoubleVector_Get(gm->penetrances, 1),
        DoubleVector_Get(gm->penetrances, 3)
    };
    nlopt_iterations = 0;
    nlopt_set_initial_step1(opt, initial_step_size1);

    double negative_mod_score; // minimum value of target function
    int nlopt_success = 1;
    if (nlopt_optimize(opt, x, &negative_mod_score) < 0) {
        nlopt_success = 0;
        sf(ps, "nlopt failed!\n");
        pr();
    } else {
        if (negative_mod_score != global_minimum) {
            sf(ps, "COBYLA = %f vs %f = global minimum\n", negative_mod_score, global_minimum);
            pr();
            negative_mod_score = global_minimum;
            for (int i = 0; i < 12; i++) {
                x[i] = global_x[i];
            }
        }
        sf(ps, "No. of nlopt iterations: %d\n", nlopt_iterations);
        pr();
        g_jla_nlopt_iterations_free -= nlopt_iterations;
        DoubleVector hapfreqs = DoubleVector_Create(8);
        DoubleVector_PushBack(hapfreqs, x[0]);
        DoubleVector_PushBack(hapfreqs, x[1]);
        DoubleVector_PushBack(hapfreqs, x[2]);
        DoubleVector_PushBack(hapfreqs, x[3]);
        DoubleVector_PushBack(hapfreqs, x[4]);
        DoubleVector_PushBack(hapfreqs, x[5]);
        DoubleVector_PushBack(hapfreqs, x[6]);
        DoubleVector_PushBack(hapfreqs, x[7]);
        double p = x[8];
        DoubleVector penetrances = DoubleVector_Create(4);
        DoubleVector_PushBack(penetrances, x[9]);
        DoubleVector_PushBack(penetrances, x[10]);
        DoubleVector_PushBack(penetrances, x[10]);
        DoubleVector_PushBack(penetrances, x[11]);
        mod_score->test_markers = test_markers;
        mod_score->position = position;
        mod_score->mod_score = -negative_mod_score;
        mod_score->disease_allele_frequency = p;
        mod_score->penetrances = penetrances;
        mod_score->hapfreqs = hapfreqs;
        mod_score->em_hapfreqs = em_hapfreqs;
        print_maximum_2_tm_pat_first_without_deltas(-negative_mod_score, x, em_hapfreqs);
    }
    nlopt_destroy(opt);
    return nlopt_success;
}

static int optimize_user_model_with_2_test_markers_without_deltas(
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModelParams user_model,
    ModScore *gm,
    ModScore *mod_score,
    double initial_step_size1)
{
    //              H0 H1 H2 H3 H4 H5 H6 H7
    double lb[] = { 0, 0, 0, 0, 0, 0, 0, 0};
    double ub[] = { 1, 1, 1, 1, 1, 1, 1, 1};
    struct DataStruct data = {
        annotated_families,
        test_markers,
        em_hapfreqs,
        user_model
    };

    nlopt_opt opt = nlopt_create(NLOPT_LN_COBYLA, 8); // algorithm and dimensionality
    nlopt_set_lower_bounds(opt, lb);
    nlopt_set_upper_bounds(opt, ub);
    nlopt_set_min_objective(opt, target_user_model_2_tm_without_deltas, &data);
    nlopt_add_equality_constraint(opt, H_even_user_model_constr_2_tm_without_deltas, &data, DBL_EPSILON);
    nlopt_add_equality_constraint(opt, H_odd_user_model_constr_2_tm_without_deltas, &data, DBL_EPSILON);
    nlopt_set_xtol_rel(opt, 1e-4);
    int max_eval = MIN(g_jla_nlopt_iterations_free, 10000);
    nlopt_set_maxeval(opt, max_eval);

    double x[] = {
        DoubleVector_Get(gm->hapfreqs, 0),
        DoubleVector_Get(gm->hapfreqs, 1),
        DoubleVector_Get(gm->hapfreqs, 2),
        DoubleVector_Get(gm->hapfreqs, 3),
        DoubleVector_Get(gm->hapfreqs, 4),
        DoubleVector_Get(gm->hapfreqs, 5),
        DoubleVector_Get(gm->hapfreqs, 6),
        DoubleVector_Get(gm->hapfreqs, 7)
    };
    nlopt_iterations = 0;
    nlopt_set_initial_step1(opt, initial_step_size1);

    double negative_mod_score; // minimum value of target function
    int nlopt_success = 1;
    if (nlopt_optimize(opt, x, &negative_mod_score) < 0) {
        nlopt_success = 0;
        sf(ps, "nlopt failed!\n");
        pr();
    } else {
        if (negative_mod_score != global_minimum) {
            sf(ps, "COBYLA = %f vs %f = global minimum\n", negative_mod_score, global_minimum);
            pr();
            negative_mod_score = global_minimum;
            for (int i = 0; i < 8; i++) {
                x[i] = global_x[i];
            }
        }
        sf(ps, "No. of nlopt iterations: %d\n", nlopt_iterations);
        pr();
        g_jla_nlopt_iterations_free -= nlopt_iterations;
        DoubleVector hapfreqs = DoubleVector_Create(8);
        DoubleVector_PushBack(hapfreqs, x[0]);
        DoubleVector_PushBack(hapfreqs, x[1]);
        DoubleVector_PushBack(hapfreqs, x[2]);
        DoubleVector_PushBack(hapfreqs, x[3]);
        DoubleVector_PushBack(hapfreqs, x[4]);
        DoubleVector_PushBack(hapfreqs, x[5]);
        DoubleVector_PushBack(hapfreqs, x[6]);
        DoubleVector_PushBack(hapfreqs, x[7]);
        double p = user_model->disease_allele_frequency;
        DoubleVector penetrances = DoubleVector_Create(4);
        DoubleVector_PushBack(penetrances, user_model->f0);
        DoubleVector_PushBack(penetrances, user_model->f1_pat);
        DoubleVector_PushBack(penetrances, user_model->f1_mat);
        DoubleVector_PushBack(penetrances, user_model->f2);
        mod_score->test_markers = test_markers;
        mod_score->position = position;
        mod_score->mod_score = -negative_mod_score;
        mod_score->disease_allele_frequency = p;
        mod_score->penetrances = penetrances;
        mod_score->hapfreqs = hapfreqs;
        mod_score->em_hapfreqs = em_hapfreqs;
        print_maximum_user_model_2_tm_without_deltas(-negative_mod_score, x, em_hapfreqs, user_model);
    }
    nlopt_destroy(opt);
    return nlopt_success;
}

// ===================================================================
// Optimize, target, and constraint functions for N = 3 test markers
// ===================================================================

static void print_maximum_3_tm_pat_first(double maximum, const double *x, DoubleVector em_hapfreqs)
{
    double mod = maximum;
    double d1 = x[0], d2 = x[1], d3 = x[2], d4 = x[3],
           d5 = x[4], d6 = x[5], d7 = x[6], d8 = x[7];
    double p = x[8];
    double f0, f1_pat, f1_mat, f2;
    double EMH0 = DoubleVector_Get(em_hapfreqs, 0);
    double EMH1 = DoubleVector_Get(em_hapfreqs, 1);
    double EMH2 = DoubleVector_Get(em_hapfreqs, 2);
    double EMH3 = DoubleVector_Get(em_hapfreqs, 3);
    double EMH4 = DoubleVector_Get(em_hapfreqs, 4);
    double EMH5 = DoubleVector_Get(em_hapfreqs, 5);
    double EMH6 = DoubleVector_Get(em_hapfreqs, 6);
    double EMH7 = DoubleVector_Get(em_hapfreqs, 7);
    double H0 = EMH0 * (1 - p) - d1;
    double H1 = EMH0 * p + d1;
    double H2 = EMH1 * (1 - p) - d2;
    double H3 = EMH1 * p + d2;
    double H4 = EMH2 * (1 - p) - d3;
    double H5 = EMH2 * p + d3;
    double H6 = EMH3 * (1 - p) - d4;
    double H7 = EMH3 * p + d4;
    double H8 = EMH4 * (1 - p) - d5;
    double H9 = EMH4 * p + d5;
    double H10 = EMH5 * (1 - p) - d6;
    double H11 = EMH5 * p + d6;
    double H12 = EMH6 * (1 - p) - d7;
    double H13 = EMH6 * p + d7;
    double H14 = EMH7 * (1 - p) - d8;
    double H15 = EMH7 * p + d8;
    if (!imprinting) {
        f0 = x[9]; f1_pat = x[10]; f1_mat = x[10]; f2 = x[11];
    } else {
        f0 = x[9]; f1_pat = x[10]; f1_mat = x[11]; f2 = x[12];
    }
    sf(ps, "   mod     p    F0  F1_P  F1_M    F2    H0    H1    H2    H3    H4    H5    H6    H7    H8    H9   H10   H11   H12   H13   H14   H15  EMH0  EMH1  EMH2  EMH3  EMH4  EMH5  EMH6  EMH7    d1    d2    d3    d4    d5    d6    d7    d8\n");
    pr();
    sf(ps, "%6.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f\n", mod, p, f0, f1_pat, f1_mat, f2, H0, H1, H2, H3, H4, H5, H6, H7, H8, H9, H10, H11, H12, H13, H14, H15, EMH0, EMH1, EMH2, EMH3, EMH4, EMH5, EMH6, EMH7, d1, d2, d3, d4, d5, d6, d7, d8);
    pr();
}

static void print_maximum_3_tm_mat_first(double maximum, const double *x, DoubleVector em_hapfreqs)
{
    double mod = maximum;
    double d1 = x[0], d2 = x[1], d3 = x[2], d4 = x[3],
           d5 = x[4], d6 = x[5], d7 = x[6], d8 = x[7];
    double p = x[8];
    double f0, f1_pat, f1_mat, f2;
    double EMH0 = DoubleVector_Get(em_hapfreqs, 0);
    double EMH1 = DoubleVector_Get(em_hapfreqs, 1);
    double EMH2 = DoubleVector_Get(em_hapfreqs, 2);
    double EMH3 = DoubleVector_Get(em_hapfreqs, 3);
    double EMH4 = DoubleVector_Get(em_hapfreqs, 4);
    double EMH5 = DoubleVector_Get(em_hapfreqs, 5);
    double EMH6 = DoubleVector_Get(em_hapfreqs, 6);
    double EMH7 = DoubleVector_Get(em_hapfreqs, 7);
    double H0 = EMH0 * (1 - p) - d1;
    double H1 = EMH0 * p + d1;
    double H2 = EMH1 * (1 - p) - d2;
    double H3 = EMH1 * p + d2;
    double H4 = EMH2 * (1 - p) - d3;
    double H5 = EMH2 * p + d3;
    double H6 = EMH3 * (1 - p) - d4;
    double H7 = EMH3 * p + d4;
    double H8 = EMH4 * (1 - p) - d5;
    double H9 = EMH4 * p + d5;
    double H10 = EMH5 * (1 - p) - d6;
    double H11 = EMH5 * p + d6;
    double H12 = EMH6 * (1 - p) - d7;
    double H13 = EMH6 * p + d7;
    double H14 = EMH7 * (1 - p) - d8;
    double H15 = EMH7 * p + d8;
    if (!imprinting) {
        f0 = x[9]; f1_pat = x[10]; f1_mat = x[10]; f2 = x[11];
    } else {
        f0 = x[9]; f1_pat = x[11]; f1_mat = x[10]; f2 = x[12];
    }
    sf(ps, "   mod     p    F0  F1_P  F1_M    F2    H0    H1    H2    H3    H4    H5    H6    H7    H8    H9   H10   H11   H12   H13   H14   H15  EMH0  EMH1  EMH2  EMH3  EMH4  EMH5  EMH6  EMH7    d1    d2    d3    d4    d5    d6    d7    d8\n");
    pr();
    sf(ps, "%6.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f\n", mod, p, f0, f1_pat, f1_mat, f2, H0, H1, H2, H3, H4, H5, H6, H7, H8, H9, H10, H11, H12, H13, H14, H15, EMH0, EMH1, EMH2, EMH3, EMH4, EMH5, EMH6, EMH7, d1, d2, d3, d4, d5, d6, d7, d8);
    pr();
}

static void print_maximum_user_model_3_tm(double maximum, const double *x, DoubleVector em_hapfreqs, ModelParams user_model)
{
    double mod = maximum;
    double d1 = x[0], d2 = x[1], d3 = x[2], d4 = x[3],
           d5 = x[4], d6 = x[5], d7 = x[6], d8 = x[7];
    double p = user_model->disease_allele_frequency;
    double EMH0 = DoubleVector_Get(em_hapfreqs, 0);
    double EMH1 = DoubleVector_Get(em_hapfreqs, 1);
    double EMH2 = DoubleVector_Get(em_hapfreqs, 2);
    double EMH3 = DoubleVector_Get(em_hapfreqs, 3);
    double EMH4 = DoubleVector_Get(em_hapfreqs, 4);
    double EMH5 = DoubleVector_Get(em_hapfreqs, 5);
    double EMH6 = DoubleVector_Get(em_hapfreqs, 6);
    double EMH7 = DoubleVector_Get(em_hapfreqs, 7);
    double H0 = EMH0 * (1 - p) - d1;
    double H1 = EMH0 * p + d1;
    double H2 = EMH1 * (1 - p) - d2;
    double H3 = EMH1 * p + d2;
    double H4 = EMH2 * (1 - p) - d3;
    double H5 = EMH2 * p + d3;
    double H6 = EMH3 * (1 - p) - d4;
    double H7 = EMH3 * p + d4;
    double H8 = EMH4 * (1 - p) - d5;
    double H9 = EMH4 * p + d5;
    double H10 = EMH5 * (1 - p) - d6;
    double H11 = EMH5 * p + d6;
    double H12 = EMH6 * (1 - p) - d7;
    double H13 = EMH6 * p + d7;
    double H14 = EMH7 * (1 - p) - d8;
    double H15 = EMH7 * p + d8;
    double f0 = user_model->f0, f1_pat = user_model->f1_pat, f1_mat = user_model->f1_mat, f2 = user_model->f2;
    sf(ps, "   mod     p    F0  F1_P  F1_M    F2    H0    H1    H2    H3    H4    H5    H6    H7    H8    H9   H10   H11   H12   H13   H14   H15  EMH0  EMH1  EMH2  EMH3  EMH4  EMH5  EMH6  EMH7    d1    d2    d3    d4    d5    d6    d7    d8\n");
    pr();
    sf(ps, "%6.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f\n", mod, p, f0, f1_pat, f1_mat, f2, H0, H1, H2, H3, H4, H5, H6, H7, H8, H9, H10, H11, H12, H13, H14, H15, EMH0, EMH1, EMH2, EMH3, EMH4, EMH5, EMH6, EMH7, d1, d2, d3, d4, d5, d6, d7, d8);
    pr();
}

static void print_maximum_3_tm_pat_first_without_deltas(double maximum, const double *x, DoubleVector em_hapfreqs)
{
    double mod = maximum;
    double H0 = x[0], H1 = x[1], H2 = x[2], H3 = x[3],
           H4 = x[4], H5 = x[5], H6 = x[6], H7 = x[7],
           H8 = x[8], H9 = x[9], H10 = x[10], H11 = x[11],
           H12 = x[12], H13 = x[13], H14 = x[14], H15 = x[15];
    double p = x[16];
    double f0, f1_pat, f1_mat, f2;
    if (!imprinting) {
        f0 = x[17]; f1_pat = x[18]; f1_mat = x[18]; f2 = x[19];
    } else {
        f0 = x[17]; f1_pat = x[18]; f1_mat = x[19]; f2 = x[20];
    }
    double EMH0 = DoubleVector_Get(em_hapfreqs, 0);
    double EMH1 = DoubleVector_Get(em_hapfreqs, 1);
    double EMH2 = DoubleVector_Get(em_hapfreqs, 2);
    double EMH3 = DoubleVector_Get(em_hapfreqs, 3);
    double EMH4 = DoubleVector_Get(em_hapfreqs, 4);
    double EMH5 = DoubleVector_Get(em_hapfreqs, 5);
    double EMH6 = DoubleVector_Get(em_hapfreqs, 6);
    double EMH7 = DoubleVector_Get(em_hapfreqs, 7);
    sf(ps, "   mod     p    F0  F1_P  F1_M    F2    H0    H1    H2    H3    H4    H5    H6    H7    H8    H9   H10   H11   H12   H13   H14   H15  EMH0  EMH1  EMH2  EMH3  EMH4  EMH5  EMH6  EMH7\n");
    pr();
    sf(ps, "%6.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f\n", mod, p, f0, f1_pat, f1_mat, f2, H0, H1, H2, H3, H4, H5, H6, H7, H8, H9, H10, H11, H12, H13, H14, H15, EMH0, EMH1, EMH2, EMH3, EMH4, EMH5, EMH6, EMH7);
    pr();
}

static void print_maximum_3_tm_mat_first_without_deltas(double maximum, const double *x, DoubleVector em_hapfreqs)
{
    double mod = maximum;
    double H0 = x[0], H1 = x[1], H2 = x[2], H3 = x[3],
           H4 = x[4], H5 = x[5], H6 = x[6], H7 = x[7],
           H8 = x[8], H9 = x[9], H10 = x[10], H11 = x[11],
           H12 = x[12], H13 = x[13], H14 = x[14], H15 = x[15];
    double p = x[16];
    double f0, f1_pat, f1_mat, f2;
    if (!imprinting) {
        f0 = x[17]; f1_pat = x[18]; f1_mat = x[18]; f2 = x[19];
    } else {
        f0 = x[17]; f1_pat = x[19]; f1_mat = x[18]; f2 = x[20];
    }
    double EMH0 = DoubleVector_Get(em_hapfreqs, 0);
    double EMH1 = DoubleVector_Get(em_hapfreqs, 1);
    double EMH2 = DoubleVector_Get(em_hapfreqs, 2);
    double EMH3 = DoubleVector_Get(em_hapfreqs, 3);
    double EMH4 = DoubleVector_Get(em_hapfreqs, 4);
    double EMH5 = DoubleVector_Get(em_hapfreqs, 5);
    double EMH6 = DoubleVector_Get(em_hapfreqs, 6);
    double EMH7 = DoubleVector_Get(em_hapfreqs, 7);
    sf(ps, "   mod     p    F0  F1_P  F1_M    F2    H0    H1    H2    H3    H4    H5    H6    H7    H8    H9   H10   H11   H12   H13   H14   H15  EMH0  EMH1  EMH2  EMH3  EMH4  EMH5  EMH6  EMH7\n");
    pr();
    sf(ps, "%6.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f\n", mod, p, f0, f1_pat, f1_mat, f2, H0, H1, H2, H3, H4, H5, H6, H7, H8, H9, H10, H11, H12, H13, H14, H15, EMH0, EMH1, EMH2, EMH3, EMH4, EMH5, EMH6, EMH7);
    pr();
}

static void print_maximum_user_model_3_tm_without_deltas(double maximum, const double *x, DoubleVector em_hapfreqs, ModelParams user_model)
{
    double mod = maximum;
    double H0 = x[0], H1 = x[1], H2 = x[2], H3 = x[3],
           H4 = x[4], H5 = x[5], H6 = x[6], H7 = x[7],
           H8 = x[8], H9 = x[9], H10 = x[10], H11 = x[11],
           H12 = x[12], H13 = x[13], H14 = x[14], H15 = x[15];
    double p = user_model->disease_allele_frequency;
    double f0 = user_model->f0, f1_pat = user_model->f1_pat, f1_mat = user_model->f1_mat, f2 = user_model->f2;
    double EMH0 = DoubleVector_Get(em_hapfreqs, 0);
    double EMH1 = DoubleVector_Get(em_hapfreqs, 1);
    double EMH2 = DoubleVector_Get(em_hapfreqs, 2);
    double EMH3 = DoubleVector_Get(em_hapfreqs, 3);
    double EMH4 = DoubleVector_Get(em_hapfreqs, 4);
    double EMH5 = DoubleVector_Get(em_hapfreqs, 5);
    double EMH6 = DoubleVector_Get(em_hapfreqs, 6);
    double EMH7 = DoubleVector_Get(em_hapfreqs, 7);
    sf(ps, "   mod     p    F0  F1_P  F1_M    F2    H0    H1    H2    H3    H4    H5    H6    H7    H8    H9   H10   H11   H12   H13   H14   H15  EMH0  EMH1  EMH2  EMH3  EMH4  EMH5  EMH6  EMH7\n");
    pr();
    sf(ps, "%6.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f\n", mod, p, f0, f1_pat, f1_mat, f2, H0, H1, H2, H3, H4, H5, H6, H7, H8, H9, H10, H11, H12, H13, H14, H15, EMH0, EMH1, EMH2, EMH3, EMH4, EMH5, EMH6, EMH7);
    pr();
}

static void update_global_minimum_3_tm_pat_first(double value, const double *x, void *data) {
    int n = 0;
    double *grad = 0;
    if (value >= global_minimum) return;
    if (penetrance_restriction) {
        if (imprinting) {
            if (f0_f1_pat_constr_3_tm_pat_first(n, x, grad, data) > PEN_TOL) return;
            if (f0_f1_mat_constr_3_tm_pat_first(n, x, grad, data) > PEN_TOL) return;
            if (f1_pat_f2_constr_3_tm_pat_first(n, x, grad, data) > PEN_TOL) return;
            if (f1_mat_f2_constr_3_tm_pat_first(n, x, grad, data) > PEN_TOL) return;
        } else {
            if (f0_f1_constr_3_tm(n, x, grad, data) > PEN_TOL) return;
            if (f1_f2_constr_3_tm(n, x, grad, data) > PEN_TOL) return;
        }
    }
    if (fabs(ds_group_constr_3_tm(n, x, grad, data)) > FREQ_TOL) return;
    if (d1_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d1_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d2_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d2_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d3_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d3_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d4_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d4_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d5_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d5_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d6_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d6_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d7_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d7_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d8_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d8_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    global_minimum = value;
    for (int i = 0; i < 12 + imprinting; i++) {
        global_x[i] = x[i];
    }
}

static void update_global_minimum_3_tm_mat_first(double value, const double *x, void *data) {
    int n = 0;
    double *grad = 0;
    if (value >= global_minimum) return;
    if (penetrance_restriction) {
        if (imprinting) {
            if (f0_f1_pat_constr_3_tm_mat_first(n, x, grad, data) > PEN_TOL) return;
            if (f0_f1_mat_constr_3_tm_mat_first(n, x, grad, data) > PEN_TOL) return;
            if (f1_pat_f2_constr_3_tm_mat_first(n, x, grad, data) > PEN_TOL) return;
            if (f1_mat_f2_constr_3_tm_mat_first(n, x, grad, data) > PEN_TOL) return;
        } else {
            if (f0_f1_constr_3_tm(n, x, grad, data) > PEN_TOL) return;
            if (f1_f2_constr_3_tm(n, x, grad, data) > PEN_TOL) return;
        }
    }
    if (fabs(ds_group_constr_3_tm(n, x, grad, data)) > FREQ_TOL) return;
    if (d1_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d1_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d2_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d2_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d3_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d3_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d4_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d4_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d5_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d5_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d6_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d6_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d7_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d7_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d8_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d8_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    global_minimum = value;
    for (int i = 0; i < 12 + imprinting; i++) {
        global_x[i] = x[i];
    }
}

static void update_global_minimum_user_model_3_tm(double value, const double *x, void *data) {
    int n = 0;
    double *grad = 0;
    if (value >= global_minimum) return;
    if (fabs(ds_group_constr_3_tm(n, x, grad, data)) > FREQ_TOL) return;
    if (d1_user_model_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d1_user_model_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d2_user_model_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d2_user_model_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d3_user_model_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d3_user_model_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d4_user_model_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d4_user_model_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d5_user_model_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d5_user_model_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d6_user_model_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d6_user_model_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d7_user_model_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d7_user_model_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d8_user_model_lower_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    if (d8_user_model_upper_indiv_constr(n, x, grad, data) > FREQ_TOL) return;
    global_minimum = value;
    for (int i = 0; i < 8; i++) {
        global_x[i] = x[i];
    }
}

static void update_global_minimum_3_tm_pat_first_without_deltas(double value, const double *x, void *data) {
    int n = 0;
    double *grad = 0;
    if (value >= global_minimum) return;
    if (penetrance_restriction) {
        if (imprinting) {
            if (f0_f1_pat_constr_3_tm_pat_first_without_deltas(n, x, grad, data) > PEN_TOL) return;
            if (f0_f1_mat_constr_3_tm_pat_first_without_deltas(n, x, grad, data) > PEN_TOL) return;
            if (f1_pat_f2_constr_3_tm_pat_first_without_deltas(n, x, grad, data) > PEN_TOL) return;
            if (f1_mat_f2_constr_3_tm_pat_first_without_deltas(n, x, grad, data) > PEN_TOL) return;
        } else {
            if (f0_f1_constr_3_tm_without_deltas(n, x, grad, data) > PEN_TOL) return;
            if (f1_f2_constr_3_tm_without_deltas(n, x, grad, data) > PEN_TOL) return;
        }
    }
    if (fabs(H_even_constr_3_tm_without_deltas(n, x, grad, data)) > FREQ_TOL) return;
    if (fabs(H_odd_constr_3_tm_without_deltas(n, x, grad, data)) > FREQ_TOL) return;
    global_minimum = value;
    for (int i = 0; i < 20 + imprinting; i++) {
        global_x[i] = x[i];
    }
}

static void update_global_minimum_3_tm_mat_first_without_deltas(double value, const double *x, void *data) {
    int n = 0;
    double *grad = 0;
    if (value >= global_minimum) return;
    if (penetrance_restriction) {
        if (imprinting) {
            if (f0_f1_pat_constr_3_tm_mat_first_without_deltas(n, x, grad, data) > PEN_TOL) return;
            if (f0_f1_mat_constr_3_tm_mat_first_without_deltas(n, x, grad, data) > PEN_TOL) return;
            if (f1_pat_f2_constr_3_tm_mat_first_without_deltas(n, x, grad, data) > PEN_TOL) return;
            if (f1_mat_f2_constr_3_tm_mat_first_without_deltas(n, x, grad, data) > PEN_TOL) return;
        } else {
            if (f0_f1_constr_3_tm_without_deltas(n, x, grad, data) > PEN_TOL) return;
            if (f1_f2_constr_3_tm_without_deltas(n, x, grad, data) > PEN_TOL) return;
        }
    }
    if (fabs(H_even_constr_3_tm_without_deltas(n, x, grad, data)) > FREQ_TOL) return;
    if (fabs(H_odd_constr_3_tm_without_deltas(n, x, grad, data)) > FREQ_TOL) return;
    global_minimum = value;
    for (int i = 0; i < 20 + imprinting; i++) {
        global_x[i] = x[i];
    }
}

static void update_global_minimum_user_model_3_tm_without_deltas(double value, const double *x, void *data) {
    int n = 0;
    double *grad = 0;
    if (value >= global_minimum) return;
    if (fabs(H_even_user_model_constr_3_tm_without_deltas(n, x, grad, data)) > FREQ_TOL) return;
    if (fabs(H_odd_user_model_constr_3_tm_without_deltas(n, x, grad, data)) > FREQ_TOL) return;
    global_minimum = value;
    for (int i = 0; i < 16; i++) {
        global_x[i] = x[i];
    }
}

static double target_3_tm_pat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    objective_function_calls++;
    nlopt_iterations++;
    Data d = data;
    const double *deltas = x;
    double p = x[8];
    double f0, f1_pat, f1_mat, f2;
    DoubleVector hapfreqs = DoubleVector_Create(16);
    for (int i = 0; i < 8; i++) {
        // The order of hapfreqs is always H0, H1, ..., where hapfreqs
        // H0, H2, etc. contain the wildtype allele and hapfreqs H1,
        // H3, etc. contain the mutant allele.  To get this order the
        // even indexes must be wildtype hapfreqs with wildtype allele
        // frequency q and the odd indexes must be mutant hapfreqs
        // with mutant allele frequency p.
        DoubleVector_Set(hapfreqs, 2*i,     DoubleVector_Get(d->em_hapfreqs, i) * (1-p) - deltas[i]);
        DoubleVector_Set(hapfreqs, 2*i + 1, DoubleVector_Get(d->em_hapfreqs, i) * p + deltas[i]);
    }
    if (!imprinting) {
        f0 = x[9]; f1_pat = x[10]; f1_mat = x[10]; f2 = x[11];
    } else {
        f0 = x[9]; f1_pat = x[10]; f1_mat = x[11]; f2 = x[12];
    }
    DoubleVector penetrances = DoubleVector_With(10, f0, f1_pat, f1_mat, f2, 1-f0, 1-f1_pat, 1-f1_mat, 1-f2, .5 * (f1_pat + f1_mat), .5 * ((1-f1_pat) + (1-f1_mat)));
    double score = ModScorer_ComputeModScore(0, d->annotated_families, d->test_markers, p, hapfreqs, penetrances, d->em_hapfreqs);
    DoubleVector_Destroy(hapfreqs);
    DoubleVector_Destroy(penetrances);
    update_global_minimum_3_tm_pat_first(-score, x, data);
    return -score;
}

static double target_3_tm_mat_first(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    objective_function_calls++;
    nlopt_iterations++;
    Data d = data;
    const double *deltas = x;
    double p = x[8];
    double f0, f1_pat, f1_mat, f2;
    DoubleVector hapfreqs = DoubleVector_Create(16);
    for (int i = 0; i < 8; i++) {
        // The order of hapfreqs is always H0, H1, ..., where hapfreqs
        // H0, H2, etc. contain the wildtype allele and hapfreqs H1,
        // H3, etc. contain the mutant allele.  To get this order the
        // even indexes must be wildtype hapfreqs with wildtype allele
        // frequency q and the odd indexes must be mutant hapfreqs
        // with mutant allele frequency p.
        DoubleVector_Set(hapfreqs, 2*i,     DoubleVector_Get(d->em_hapfreqs, i) * (1-p) - deltas[i]);
        DoubleVector_Set(hapfreqs, 2*i + 1, DoubleVector_Get(d->em_hapfreqs, i) * p + deltas[i]);
    }
    if (!imprinting) {
        f0 = x[9]; f1_pat = x[10]; f1_mat = x[10]; f2 = x[11];
    } else {
        f0 = x[9]; f1_pat = x[11]; f1_mat = x[10]; f2 = x[12];
    }
    DoubleVector penetrances = DoubleVector_With(10, f0, f1_pat, f1_mat, f2, 1-f0, 1-f1_pat, 1-f1_mat, 1-f2, .5 * (f1_pat + f1_mat), .5 * ((1-f1_pat) + (1-f1_mat)));
    double score = ModScorer_ComputeModScore(0, d->annotated_families, d->test_markers, p, hapfreqs, penetrances, d->em_hapfreqs);
    DoubleVector_Destroy(hapfreqs);
    DoubleVector_Destroy(penetrances);
    update_global_minimum_3_tm_mat_first(-score, x, data);
    return -score;
}

static double target_user_model_3_tm(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    objective_function_calls++;
    nlopt_iterations++;
    Data d = data;
    const double *deltas = x;
    double p = d->user_model->disease_allele_frequency;
    DoubleVector hapfreqs = DoubleVector_Create(16);
    for (int i = 0; i < 8; i++) {
        // The order of hapfreqs is always H0, H1, ..., where hapfreqs
        // H0, H2, etc. contain the wildtype allele and hapfreqs H1,
        // H3, etc. contain the mutant allele.  To get this order the
        // even indexes must be wildtype hapfreqs with wildtype allele
        // frequency q and the odd indexes must be mutant hapfreqs
        // with mutant allele frequency p.
        DoubleVector_Set(hapfreqs, 2*i,     DoubleVector_Get(d->em_hapfreqs, i) * (1-p) - deltas[i]);
        DoubleVector_Set(hapfreqs, 2*i + 1, DoubleVector_Get(d->em_hapfreqs, i) * p + deltas[i]);
    }
    double f0 = d->user_model->f0, f1_pat = d->user_model->f1_pat, f1_mat = d->user_model->f1_mat, f2 = d->user_model->f2;
    DoubleVector penetrances = DoubleVector_With(10, f0, f1_pat, f1_mat, f2, 1-f0, 1-f1_pat, 1-f1_mat, 1-f2, .5 * (f1_pat + f1_mat), .5 * ((1-f1_pat) + (1-f1_mat)));
    double score = ModScorer_ComputeModScore(0, d->annotated_families, d->test_markers, p, hapfreqs, penetrances, d->em_hapfreqs);
    DoubleVector_Destroy(hapfreqs);
    DoubleVector_Destroy(penetrances);
    update_global_minimum_user_model_3_tm(-score, x, data);
    return -score;
}

static double target_3_tm_pat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    objective_function_calls++;
    nlopt_iterations++;
    Data d = data;
    const double *H = x;
    DoubleVector hapfreqs = DoubleVector_Create(16);
    for (int i = 0; i < 16; i++) {
        DoubleVector_PushBack(hapfreqs, H[i]);
    }
    double p = x[16];
    double f0, f1_pat, f1_mat, f2;
    if (!imprinting) {
        f0 = x[17]; f1_pat = x[18]; f1_mat = x[18]; f2 = x[19];
    } else {
        f0 = x[17]; f1_pat = x[18]; f1_mat = x[19]; f2 = x[20];
    }
    DoubleVector penetrances = DoubleVector_With(10, f0, f1_pat, f1_mat, f2, 1-f0, 1-f1_pat, 1-f1_mat, 1-f2, .5 * (f1_pat + f1_mat), .5 * ((1-f1_pat) + (1-f1_mat)));
    double score = ModScorer_ComputeModScore(0, d->annotated_families, d->test_markers, p, hapfreqs, penetrances, d->em_hapfreqs);
    DoubleVector_Destroy(hapfreqs);
    DoubleVector_Destroy(penetrances);
    update_global_minimum_3_tm_pat_first_without_deltas(-score, x, data);
    return -score;
}

static double target_3_tm_mat_first_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    objective_function_calls++;
    nlopt_iterations++;
    Data d = data;
    const double *H = x;
    DoubleVector hapfreqs = DoubleVector_Create(16);
    for (int i = 0; i < 16; i++) {
        DoubleVector_PushBack(hapfreqs, H[i]);
    }
    double p = x[16];
    double f0, f1_pat, f1_mat, f2;
    if (!imprinting) {
        f0 = x[17]; f1_pat = x[18]; f1_mat = x[18]; f2 = x[19];
    } else {
        f0 = x[17]; f1_pat = x[19]; f1_mat = x[18]; f2 = x[20];
    }
    DoubleVector penetrances = DoubleVector_With(10, f0, f1_pat, f1_mat, f2, 1-f0, 1-f1_pat, 1-f1_mat, 1-f2, .5 * (f1_pat + f1_mat), .5 * ((1-f1_pat) + (1-f1_mat)));
    double score = ModScorer_ComputeModScore(0, d->annotated_families, d->test_markers, p, hapfreqs, penetrances, d->em_hapfreqs);
    DoubleVector_Destroy(hapfreqs);
    DoubleVector_Destroy(penetrances);
    update_global_minimum_3_tm_mat_first_without_deltas(-score, x, data);
    return -score;
}

static double target_user_model_3_tm_without_deltas(unsigned n, const double *x, double *grad, void *data) {
    (void) n;
    (void) grad;
    objective_function_calls++;
    nlopt_iterations++;
    Data d = data;
    const double *H = x;
    DoubleVector hapfreqs = DoubleVector_Create(16);
    for (int i = 0; i < 16; i++) {
        DoubleVector_PushBack(hapfreqs, H[i]);
    }
    double p = d->user_model->disease_allele_frequency;
    double f0 = d->user_model->f0, f1_pat = d->user_model->f1_pat, f1_mat = d->user_model->f1_mat, f2 = d->user_model->f2;
    DoubleVector penetrances = DoubleVector_With(10, f0, f1_pat, f1_mat, f2, 1-f0, 1-f1_pat, 1-f1_mat, 1-f2, .5 * (f1_pat + f1_mat), .5 * ((1-f1_pat) + (1-f1_mat)));
    double score = ModScorer_ComputeModScore(0, d->annotated_families, d->test_markers, p, hapfreqs, penetrances, d->em_hapfreqs);
    DoubleVector_Destroy(hapfreqs);
    DoubleVector_Destroy(penetrances);
    update_global_minimum_user_model_3_tm_without_deltas(-score, x, data);
    return -score;
}

static int optimize_with_3_test_markers_pat_first(
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModScore *gm,
    ModScore *mod_score,
    double initial_step_size1)
{
    //             d1  d2  d3  d4  d5  d6  d7  d8  p       f0 f1_pat f1_mat f2
    double lb[] = {-1, -1, -1, -1, -1, -1, -1, -1, 1e-8,   0, 0,     0,     0};
    double ub[] = { 1,  1,  1,  1,  1,  1,  1,  1, 1-1e-8, 1, 1,     1,     1};
    if (allfreq_restriction) {
        ub[8] = highest_allfreq;
    }
    struct DataStruct data = {
        annotated_families,
        test_markers,
        em_hapfreqs,
        0
    };

    nlopt_opt opt = nlopt_create(NLOPT_LN_COBYLA, 13); // algorithm and dimensionality
    nlopt_set_lower_bounds(opt, lb);
    nlopt_set_upper_bounds(opt, ub);
    nlopt_set_min_objective(opt, target_3_tm_pat_first, &data);
    nlopt_add_equality_constraint(opt, ds_group_constr_3_tm, &data, FREQ_TOL);
    if (penetrance_restriction) {
        nlopt_add_inequality_constraint(opt, f0_f1_pat_constr_3_tm_pat_first, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f0_f1_mat_constr_3_tm_pat_first, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_pat_f2_constr_3_tm_pat_first, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_mat_f2_constr_3_tm_pat_first, &data, DBL_EPSILON);
    }
    nlopt_add_inequality_constraint(opt, d1_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d1_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d2_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d2_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d3_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d3_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d4_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d4_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d5_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d5_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d6_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d6_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d7_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d7_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d8_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d8_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_set_xtol_rel(opt, 1e-4);
    int max_eval = MIN(g_jla_nlopt_iterations_free, 10000);
    nlopt_set_maxeval(opt, max_eval);

    double x[] = {
        DoubleVector_Get(gm->deltas, 0),
        DoubleVector_Get(gm->deltas, 1),
        DoubleVector_Get(gm->deltas, 2),
        DoubleVector_Get(gm->deltas, 3),
        DoubleVector_Get(gm->deltas, 4),
        DoubleVector_Get(gm->deltas, 5),
        DoubleVector_Get(gm->deltas, 6),
        DoubleVector_Get(gm->deltas, 7),
        gm->disease_allele_frequency,
        DoubleVector_Get(gm->penetrances, 0),
        DoubleVector_Get(gm->penetrances, 1),
        DoubleVector_Get(gm->penetrances, 2),
        DoubleVector_Get(gm->penetrances, 3)
    };
    nlopt_iterations = 0;
    nlopt_set_initial_step1(opt, initial_step_size1);

    double negative_mod_score; // minimum value of target function
    int nlopt_success = 1;
    if (nlopt_optimize(opt, x, &negative_mod_score) < 0) {
        nlopt_success = 0;
        sf(ps, "nlopt failed!\n");
        pr();
    } else {
        if (negative_mod_score != global_minimum) {
            sf(ps, "COBYLA = %f vs %f = global minimum\n", negative_mod_score, global_minimum);
            pr();
            negative_mod_score = global_minimum;
            for (int i = 0; i < 12+1; i++) {
                x[i] = global_x[i];
            }
        }
        sf(ps, "No. of nlopt iterations: %d\n", nlopt_iterations);
        pr();
        g_jla_nlopt_iterations_free -= nlopt_iterations;
        DoubleVector hapfreqs = DoubleVector_Create(16);
        DoubleVector ds = DoubleVector_Create(8);
        DoubleVector_PushBack(ds, x[0]);
        DoubleVector_PushBack(ds, x[1]);
        DoubleVector_PushBack(ds, x[2]);
        DoubleVector_PushBack(ds, x[3]);
        DoubleVector_PushBack(ds, x[4]);
        DoubleVector_PushBack(ds, x[5]);
        DoubleVector_PushBack(ds, x[6]);
        DoubleVector_PushBack(ds, x[7]);
        double *deltas = x;
        double p = x[8];
        for (int i = 0; i < 8; i++) {
            DoubleVector_Set(hapfreqs, 2*i,     DoubleVector_Get(em_hapfreqs, i) * (1-p) - deltas[i]);
            DoubleVector_Set(hapfreqs, 2*i + 1, DoubleVector_Get(em_hapfreqs, i) * p + deltas[i]);
        }
        DoubleVector penetrances = DoubleVector_Create(4);
        DoubleVector_PushBack(penetrances, x[9]);
        DoubleVector_PushBack(penetrances, x[10]);
        DoubleVector_PushBack(penetrances, x[11]);
        DoubleVector_PushBack(penetrances, x[12]);
        mod_score->test_markers = test_markers;
        mod_score->position = position;
        mod_score->mod_score = -negative_mod_score;
        mod_score->disease_allele_frequency = p;
        mod_score->penetrances = penetrances;
        mod_score->hapfreqs = hapfreqs;
        mod_score->em_hapfreqs = em_hapfreqs;
        mod_score->deltas = ds;
        print_maximum_3_tm_pat_first(-negative_mod_score, x, em_hapfreqs);
    }
    nlopt_destroy(opt);
    return nlopt_success;
}

static int optimize_with_3_test_markers_mat_first(
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModScore *gm,
    ModScore *mod_score,
    double initial_step_size1)
{
    //             d1  d2  d3  d4  d5  d6  d7  d8  p       f0 f1_mat f1_pat f2
    double lb[] = {-1, -1, -1, -1, -1, -1, -1, -1, 1e-8,   0, 0,     0,     0};
    double ub[] = { 1,  1,  1,  1,  1,  1,  1,  1, 1-1e-8, 1, 1,     1,     1};
    if (allfreq_restriction) {
        ub[8] = highest_allfreq;
    }
    struct DataStruct data = {
        annotated_families,
        test_markers,
        em_hapfreqs,
        0
    };

    nlopt_opt opt = nlopt_create(NLOPT_LN_COBYLA, 13); // algorithm and dimensionality
    nlopt_set_lower_bounds(opt, lb);
    nlopt_set_upper_bounds(opt, ub);
    nlopt_set_min_objective(opt, target_3_tm_mat_first, &data);
    nlopt_add_equality_constraint(opt, ds_group_constr_3_tm, &data, FREQ_TOL);
    if (penetrance_restriction) {
        nlopt_add_inequality_constraint(opt, f0_f1_pat_constr_3_tm_mat_first, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f0_f1_mat_constr_3_tm_mat_first, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_pat_f2_constr_3_tm_mat_first, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_mat_f2_constr_3_tm_mat_first, &data, DBL_EPSILON);
    }
    nlopt_add_inequality_constraint(opt, d1_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d1_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d2_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d2_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d3_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d3_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d4_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d4_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d5_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d5_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d6_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d6_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d7_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d7_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d8_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d8_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_set_xtol_rel(opt, 1e-4);
    int max_eval = MIN(g_jla_nlopt_iterations_free, 10000);
    nlopt_set_maxeval(opt, max_eval);

    double x[] = {
        DoubleVector_Get(gm->deltas, 0),
        DoubleVector_Get(gm->deltas, 1),
        DoubleVector_Get(gm->deltas, 2),
        DoubleVector_Get(gm->deltas, 3),
        DoubleVector_Get(gm->deltas, 4),
        DoubleVector_Get(gm->deltas, 5),
        DoubleVector_Get(gm->deltas, 6),
        DoubleVector_Get(gm->deltas, 7),
        gm->disease_allele_frequency,
        DoubleVector_Get(gm->penetrances, 0),
        DoubleVector_Get(gm->penetrances, 2),
        DoubleVector_Get(gm->penetrances, 1),
        DoubleVector_Get(gm->penetrances, 3)
    };
    nlopt_iterations = 0;
    nlopt_set_initial_step1(opt, initial_step_size1);

    double negative_mod_score; // minimum value of target function
    int nlopt_success = 1;
    if (nlopt_optimize(opt, x, &negative_mod_score) < 0) {
        nlopt_success = 0;
        sf(ps, "nlopt failed!\n");
        pr();
    } else {
        if (negative_mod_score != global_minimum) {
            sf(ps, "COBYLA = %f vs %f = global minimum\n", negative_mod_score, global_minimum);
            pr();
            negative_mod_score = global_minimum;
            for (int i = 0; i < 12+1; i++) {
                x[i] = global_x[i];
            }
        }
        sf(ps, "No. of nlopt iterations: %d\n", nlopt_iterations);
        pr();
        g_jla_nlopt_iterations_free -= nlopt_iterations;
        DoubleVector hapfreqs = DoubleVector_Create(16);
        DoubleVector ds = DoubleVector_Create(8);
        DoubleVector_PushBack(ds, x[0]);
        DoubleVector_PushBack(ds, x[1]);
        DoubleVector_PushBack(ds, x[2]);
        DoubleVector_PushBack(ds, x[3]);
        DoubleVector_PushBack(ds, x[4]);
        DoubleVector_PushBack(ds, x[5]);
        DoubleVector_PushBack(ds, x[6]);
        DoubleVector_PushBack(ds, x[7]);
        double *deltas = x;
        double p = x[8];
        for (int i = 0; i < 8; i++) {
            DoubleVector_Set(hapfreqs, 2*i,     DoubleVector_Get(em_hapfreqs, i) * (1-p) - deltas[i]);
            DoubleVector_Set(hapfreqs, 2*i + 1, DoubleVector_Get(em_hapfreqs, i) * p + deltas[i]);
        }
        DoubleVector penetrances = DoubleVector_Create(4);
        DoubleVector_PushBack(penetrances, x[9]);
        DoubleVector_PushBack(penetrances, x[11]);
        DoubleVector_PushBack(penetrances, x[10]);
        DoubleVector_PushBack(penetrances, x[12]);
        mod_score->test_markers = test_markers;
        mod_score->position = position;
        mod_score->mod_score = -negative_mod_score;
        mod_score->disease_allele_frequency = p;
        mod_score->penetrances = penetrances;
        mod_score->hapfreqs = hapfreqs;
        mod_score->em_hapfreqs = em_hapfreqs;
        mod_score->deltas = ds;
        print_maximum_3_tm_mat_first(-negative_mod_score, x, em_hapfreqs);
    }
    nlopt_destroy(opt);
    return nlopt_success;
}

static int optimize_with_3_test_markers_no_imprinting(
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModScore *gm,
    ModScore *mod_score,
    double initial_step_size1)
{
    //             d1  d2  d3  d4  d5  d6  d7  d8  p       f0 f1 f2
    double lb[] = {-1, -1, -1, -1, -1, -1, -1, -1, 1e-8,   0, 0, 0};
    double ub[] = { 1,  1,  1,  1,  1,  1,  1,  1, 1-1e-8, 1, 1, 1};
    if (allfreq_restriction) {
        ub[8] = highest_allfreq;
    }
    struct DataStruct data = {
        annotated_families,
        test_markers,
        em_hapfreqs,
        0
    };

    nlopt_opt opt = nlopt_create(NLOPT_LN_COBYLA, 12); // algorithm and dimensionality
    nlopt_set_lower_bounds(opt, lb);
    nlopt_set_upper_bounds(opt, ub);
    nlopt_set_min_objective(opt, target_3_tm_pat_first, &data);
    nlopt_add_equality_constraint(opt, ds_group_constr_3_tm, &data, FREQ_TOL);
    if (penetrance_restriction) {
        nlopt_add_inequality_constraint(opt, f0_f1_constr_3_tm, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_f2_constr_3_tm, &data, DBL_EPSILON);
    }
    nlopt_add_inequality_constraint(opt, d1_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d1_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d2_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d2_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d3_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d3_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d4_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d4_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d5_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d5_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d6_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d6_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d7_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d7_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d8_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d8_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_set_xtol_rel(opt, 1e-4);
    int max_eval = MIN(g_jla_nlopt_iterations_free, 10000);
    nlopt_set_maxeval(opt, max_eval);

    double x[] = {
        DoubleVector_Get(gm->deltas, 0),
        DoubleVector_Get(gm->deltas, 1),
        DoubleVector_Get(gm->deltas, 2),
        DoubleVector_Get(gm->deltas, 3),
        DoubleVector_Get(gm->deltas, 4),
        DoubleVector_Get(gm->deltas, 5),
        DoubleVector_Get(gm->deltas, 6),
        DoubleVector_Get(gm->deltas, 7),
        gm->disease_allele_frequency,
        DoubleVector_Get(gm->penetrances, 0),
        DoubleVector_Get(gm->penetrances, 1),
        DoubleVector_Get(gm->penetrances, 3)
    };
    nlopt_iterations = 0;
    nlopt_set_initial_step1(opt, initial_step_size1);

    double negative_mod_score; // minimum value of target function
    int nlopt_status = nlopt_optimize(opt, x, &negative_mod_score);
    int nlopt_success = 1;
    if (nlopt_status < 0) {
        fprintf(stderr, "%s\n", nlopt_result_to_string(nlopt_status));
        nlopt_success = 0;
        sf(ps, "nlopt failed!\n");
        pr();
    } else {
        if (negative_mod_score != global_minimum) {
            sf(ps, "COBYLA = %f vs %f = global minimum\n", negative_mod_score, global_minimum);
            pr();
            negative_mod_score = global_minimum;
            for (int i = 0; i < 12; i++) {
                x[i] = global_x[i];
            }
        }
        sf(ps, "No. of nlopt iterations: %d\n", nlopt_iterations);
        pr();
        g_jla_nlopt_iterations_free -= nlopt_iterations;
        DoubleVector hapfreqs = DoubleVector_Create(16);
        DoubleVector ds = DoubleVector_Create(8);
        DoubleVector_PushBack(ds, x[0]);
        DoubleVector_PushBack(ds, x[1]);
        DoubleVector_PushBack(ds, x[2]);
        DoubleVector_PushBack(ds, x[3]);
        DoubleVector_PushBack(ds, x[4]);
        DoubleVector_PushBack(ds, x[5]);
        DoubleVector_PushBack(ds, x[6]);
        DoubleVector_PushBack(ds, x[7]);
        double *deltas = x;
        double p = x[8];
        for (int i = 0; i < 8; i++) {
            DoubleVector_Set(hapfreqs, 2*i,     DoubleVector_Get(em_hapfreqs, i) * (1-p) - deltas[i]);
            DoubleVector_Set(hapfreqs, 2*i + 1, DoubleVector_Get(em_hapfreqs, i) * p + deltas[i]);
        }
        DoubleVector penetrances = DoubleVector_Create(4);
        DoubleVector_PushBack(penetrances, x[9]);
        DoubleVector_PushBack(penetrances, x[10]);
        DoubleVector_PushBack(penetrances, x[10]);
        DoubleVector_PushBack(penetrances, x[11]);
        mod_score->test_markers = test_markers;
        mod_score->position = position;
        mod_score->mod_score = -negative_mod_score;
        mod_score->disease_allele_frequency = p;
        mod_score->penetrances = penetrances;
        mod_score->hapfreqs = hapfreqs;
        mod_score->em_hapfreqs = em_hapfreqs;
        mod_score->deltas = ds;
        print_maximum_3_tm_pat_first(-negative_mod_score, x, em_hapfreqs);
    }
    nlopt_destroy(opt);
    return nlopt_success;
}

static int optimize_user_model_with_3_test_markers(
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModelParams user_model,
    ModScore *gm,
    ModScore *mod_score,
    double initial_step_size1)
{
    //             d1  d2  d3  d4  d5  d6  d7  d8
    double lb[] = {-1, -1, -1, -1, -1, -1, -1, -1};
    double ub[] = { 1,  1,  1,  1,  1,  1,  1,  1};
    struct DataStruct data = {
        annotated_families,
        test_markers,
        em_hapfreqs,
        user_model
    };

    nlopt_opt opt = nlopt_create(NLOPT_LN_COBYLA, 8); // algorithm and dimensionality
    nlopt_set_lower_bounds(opt, lb);
    nlopt_set_upper_bounds(opt, ub);
    nlopt_set_min_objective(opt, target_user_model_3_tm, &data);
    nlopt_add_equality_constraint(opt, ds_group_constr_3_tm, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d1_user_model_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d1_user_model_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d2_user_model_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d2_user_model_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d3_user_model_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d3_user_model_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d4_user_model_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d4_user_model_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d5_user_model_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d5_user_model_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d6_user_model_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d6_user_model_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d7_user_model_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d7_user_model_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d8_user_model_lower_indiv_constr, &data, FREQ_TOL);
    nlopt_add_inequality_constraint(opt, d8_user_model_upper_indiv_constr, &data, FREQ_TOL);
    nlopt_set_xtol_rel(opt, 1e-4);
    int max_eval = MIN(g_jla_nlopt_iterations_free, 10000);
    nlopt_set_maxeval(opt, max_eval);

    double x[] = {
        DoubleVector_Get(gm->deltas, 0),
        DoubleVector_Get(gm->deltas, 1),
        DoubleVector_Get(gm->deltas, 2),
        DoubleVector_Get(gm->deltas, 3),
        DoubleVector_Get(gm->deltas, 4),
        DoubleVector_Get(gm->deltas, 5),
        DoubleVector_Get(gm->deltas, 6),
        DoubleVector_Get(gm->deltas, 7)
    };
    nlopt_iterations = 0;
    nlopt_set_initial_step1(opt, initial_step_size1);

    double negative_mod_score; // minimum value of target function
    int nlopt_success = 1;
    if (nlopt_optimize(opt, x, &negative_mod_score) < 0) {
        nlopt_success = 0;
        sf(ps, "nlopt failed!\n");
        pr();
    } else {
        if (negative_mod_score != global_minimum) {
            sf(ps, "COBYLA = %f vs %f = global minimum\n", negative_mod_score, global_minimum);
            pr();
            negative_mod_score = global_minimum;
            for (int i = 0; i < 8; i++) {
                x[i] = global_x[i];
            }
        }
        sf(ps, "No. of nlopt iterations: %d\n", nlopt_iterations);
        pr();
        g_jla_nlopt_iterations_free -= nlopt_iterations;
        DoubleVector hapfreqs = DoubleVector_Create(16);
        DoubleVector ds = DoubleVector_Create(8);
        DoubleVector_PushBack(ds, x[0]);
        DoubleVector_PushBack(ds, x[1]);
        DoubleVector_PushBack(ds, x[2]);
        DoubleVector_PushBack(ds, x[3]);
        DoubleVector_PushBack(ds, x[4]);
        DoubleVector_PushBack(ds, x[5]);
        DoubleVector_PushBack(ds, x[6]);
        DoubleVector_PushBack(ds, x[7]);
        double *deltas = x;
        double p = user_model->disease_allele_frequency;
        for (int i = 0; i < 8; i++) {
            DoubleVector_Set(hapfreqs, 2*i,     DoubleVector_Get(em_hapfreqs, i) * (1-p) - deltas[i]);
            DoubleVector_Set(hapfreqs, 2*i + 1, DoubleVector_Get(em_hapfreqs, i) * p + deltas[i]);
        }
        DoubleVector penetrances = DoubleVector_Create(4);
        DoubleVector_PushBack(penetrances, user_model->f0);
        DoubleVector_PushBack(penetrances, user_model->f1_pat);
        DoubleVector_PushBack(penetrances, user_model->f1_mat);
        DoubleVector_PushBack(penetrances, user_model->f2);
        mod_score->test_markers = test_markers;
        mod_score->position = position;
        mod_score->mod_score = -negative_mod_score;
        mod_score->disease_allele_frequency = p;
        mod_score->penetrances = penetrances;
        mod_score->hapfreqs = hapfreqs;
        mod_score->em_hapfreqs = em_hapfreqs;
        mod_score->deltas = ds;
        print_maximum_user_model_3_tm(-negative_mod_score, x, em_hapfreqs, user_model);
    }
    nlopt_destroy(opt);
    return nlopt_success;
}

static int optimize_with_3_test_markers_pat_first_without_deltas(
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModScore *gm,
    ModScore *mod_score,
    double initial_step_size1)
{
    //              H0  H1  H2  H3  H4  H5  H6  H7  H8  H9  H10 H11 H12 H13 H14 H15 p       f0 f1_pat f1_mat f2
    double lb[] = { 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,    1e-8, 0, 0,     0,     0};
    double ub[] = { 1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1-1e-8, 1, 1,     1,     1};
    if (allfreq_restriction) {
        ub[16] = highest_allfreq;
    }
    struct DataStruct data = {
        annotated_families,
        test_markers,
        em_hapfreqs,
        0
    };

    nlopt_opt opt = nlopt_create(NLOPT_LN_COBYLA, 21); // algorithm and dimensionality
    nlopt_set_lower_bounds(opt, lb);
    nlopt_set_upper_bounds(opt, ub);
    nlopt_set_min_objective(opt, target_3_tm_pat_first_without_deltas, &data);
    nlopt_add_equality_constraint(opt, H_even_constr_3_tm_without_deltas, &data, DBL_EPSILON);
    nlopt_add_equality_constraint(opt, H_odd_constr_3_tm_without_deltas, &data, DBL_EPSILON);
    if (penetrance_restriction) {
        nlopt_add_inequality_constraint(opt, f0_f1_pat_constr_3_tm_pat_first_without_deltas, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f0_f1_mat_constr_3_tm_pat_first_without_deltas, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_pat_f2_constr_3_tm_pat_first_without_deltas, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_mat_f2_constr_3_tm_pat_first_without_deltas, &data, DBL_EPSILON);
    }
    nlopt_set_xtol_rel(opt, 1e-4);
    int max_eval = MIN(g_jla_nlopt_iterations_free, 10000);
    nlopt_set_maxeval(opt, max_eval);

    double x[] = {
        DoubleVector_Get(gm->hapfreqs, 0),
        DoubleVector_Get(gm->hapfreqs, 1),
        DoubleVector_Get(gm->hapfreqs, 2),
        DoubleVector_Get(gm->hapfreqs, 3),
        DoubleVector_Get(gm->hapfreqs, 4),
        DoubleVector_Get(gm->hapfreqs, 5),
        DoubleVector_Get(gm->hapfreqs, 6),
        DoubleVector_Get(gm->hapfreqs, 7),
        DoubleVector_Get(gm->hapfreqs, 8),
        DoubleVector_Get(gm->hapfreqs, 9),
        DoubleVector_Get(gm->hapfreqs, 10),
        DoubleVector_Get(gm->hapfreqs, 11),
        DoubleVector_Get(gm->hapfreqs, 12),
        DoubleVector_Get(gm->hapfreqs, 13),
        DoubleVector_Get(gm->hapfreqs, 14),
        DoubleVector_Get(gm->hapfreqs, 15),
        gm->disease_allele_frequency,
        DoubleVector_Get(gm->penetrances, 0),
        DoubleVector_Get(gm->penetrances, 1),
        DoubleVector_Get(gm->penetrances, 2),
        DoubleVector_Get(gm->penetrances, 3)
    };
    nlopt_iterations = 0;
    nlopt_set_initial_step1(opt, initial_step_size1);

    double dx[21];
    nlopt_get_initial_step(opt, x, dx);
    printf(">>>> initial step size:");
    for (int i = 0; i < 21; i++) {
        printf(" %g", dx[i]);
    }
    printf("\n");

    double negative_mod_score; // minimum value of target function
    int nlopt_success = 1;
    if (nlopt_optimize(opt, x, &negative_mod_score) < 0) {
        nlopt_success = 0;
        sf(ps, "nlopt failed!\n");
        pr();
    } else {
        if (negative_mod_score != global_minimum) {
            sf(ps, "COBYLA = %f vs %f = global minimum\n", negative_mod_score, global_minimum);
            pr();
            negative_mod_score = global_minimum;
            for (int i = 0; i < 20+1; i++) {
                x[i] = global_x[i];
            }
        }
        sf(ps, "No. of nlopt iterations: %d\n", nlopt_iterations);
        pr();
        g_jla_nlopt_iterations_free -= nlopt_iterations;
        DoubleVector hapfreqs = DoubleVector_Create(16);
        DoubleVector_PushBack(hapfreqs, x[0]);
        DoubleVector_PushBack(hapfreqs, x[1]);
        DoubleVector_PushBack(hapfreqs, x[2]);
        DoubleVector_PushBack(hapfreqs, x[3]);
        DoubleVector_PushBack(hapfreqs, x[4]);
        DoubleVector_PushBack(hapfreqs, x[5]);
        DoubleVector_PushBack(hapfreqs, x[6]);
        DoubleVector_PushBack(hapfreqs, x[7]);
        DoubleVector_PushBack(hapfreqs, x[8]);
        DoubleVector_PushBack(hapfreqs, x[9]);
        DoubleVector_PushBack(hapfreqs, x[10]);
        DoubleVector_PushBack(hapfreqs, x[11]);
        DoubleVector_PushBack(hapfreqs, x[12]);
        DoubleVector_PushBack(hapfreqs, x[13]);
        DoubleVector_PushBack(hapfreqs, x[14]);
        DoubleVector_PushBack(hapfreqs, x[15]);
        double p = x[16];
        DoubleVector penetrances = DoubleVector_Create(4);
        DoubleVector_PushBack(penetrances, x[17]);
        DoubleVector_PushBack(penetrances, x[18]);
        DoubleVector_PushBack(penetrances, x[19]);
        DoubleVector_PushBack(penetrances, x[20]);
        mod_score->test_markers = test_markers;
        mod_score->position = position;
        mod_score->mod_score = -negative_mod_score;
        mod_score->disease_allele_frequency = p;
        mod_score->penetrances = penetrances;
        mod_score->hapfreqs = hapfreqs;
        mod_score->em_hapfreqs = em_hapfreqs;
        print_maximum_3_tm_pat_first_without_deltas(-negative_mod_score, x, em_hapfreqs);
    }
    nlopt_destroy(opt);
    return nlopt_success;
}

static int optimize_with_3_test_markers_mat_first_without_deltas(
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModScore *gm,
    ModScore *mod_score,
    double initial_step_size1)
{
    //              H0  H1  H2  H3  H4  H5  H6  H7  H8  H9  H10 H11 H12 H13 H14 H15 p       f0 f1_mat f1_pat f2
    double lb[] = { 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,    1e-8, 0, 0,     0,     0};
    double ub[] = { 1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1-1e-8, 1, 1,     1,     1};
    if (allfreq_restriction) {
        ub[16] = highest_allfreq;
    }
    struct DataStruct data = {
        annotated_families,
        test_markers,
        em_hapfreqs,
        0
    };

    nlopt_opt opt = nlopt_create(NLOPT_LN_COBYLA, 21); // algorithm and dimensionality
    nlopt_set_lower_bounds(opt, lb);
    nlopt_set_upper_bounds(opt, ub);
    nlopt_set_min_objective(opt, target_3_tm_mat_first_without_deltas, &data);
    nlopt_add_equality_constraint(opt, H_even_constr_3_tm_without_deltas, &data, DBL_EPSILON);
    nlopt_add_equality_constraint(opt, H_odd_constr_3_tm_without_deltas, &data, DBL_EPSILON);
    if (penetrance_restriction) {
        nlopt_add_inequality_constraint(opt, f0_f1_pat_constr_3_tm_mat_first_without_deltas, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f0_f1_mat_constr_3_tm_mat_first_without_deltas, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_pat_f2_constr_3_tm_mat_first_without_deltas, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_mat_f2_constr_3_tm_mat_first_without_deltas, &data, DBL_EPSILON);
    }
    nlopt_set_xtol_rel(opt, 1e-4);
    int max_eval = MIN(g_jla_nlopt_iterations_free, 10000);
    nlopt_set_maxeval(opt, max_eval);

    double x[] = {
        DoubleVector_Get(gm->hapfreqs, 0),
        DoubleVector_Get(gm->hapfreqs, 1),
        DoubleVector_Get(gm->hapfreqs, 2),
        DoubleVector_Get(gm->hapfreqs, 3),
        DoubleVector_Get(gm->hapfreqs, 4),
        DoubleVector_Get(gm->hapfreqs, 5),
        DoubleVector_Get(gm->hapfreqs, 6),
        DoubleVector_Get(gm->hapfreqs, 7),
        DoubleVector_Get(gm->hapfreqs, 8),
        DoubleVector_Get(gm->hapfreqs, 9),
        DoubleVector_Get(gm->hapfreqs, 10),
        DoubleVector_Get(gm->hapfreqs, 11),
        DoubleVector_Get(gm->hapfreqs, 12),
        DoubleVector_Get(gm->hapfreqs, 13),
        DoubleVector_Get(gm->hapfreqs, 14),
        DoubleVector_Get(gm->hapfreqs, 15),
        gm->disease_allele_frequency,
        DoubleVector_Get(gm->penetrances, 0),
        DoubleVector_Get(gm->penetrances, 2),
        DoubleVector_Get(gm->penetrances, 1),
        DoubleVector_Get(gm->penetrances, 3)
    };
    nlopt_iterations = 0;
    nlopt_set_initial_step1(opt, initial_step_size1);


    double negative_mod_score; // minimum value of target function
    int nlopt_success = 1;
    if (nlopt_optimize(opt, x, &negative_mod_score) < 0) {
        nlopt_success = 0;
        sf(ps, "nlopt failed!\n");
        pr();
    } else {
        if (negative_mod_score != global_minimum) {
            sf(ps, "COBYLA = %f vs %f = global minimum\n", negative_mod_score, global_minimum);
            pr();
            negative_mod_score = global_minimum;
            for (int i = 0; i < 20+1; i++) {
                x[i] = global_x[i];
            }
        }
        sf(ps, "No. of nlopt iterations: %d\n", nlopt_iterations);
        pr();
        g_jla_nlopt_iterations_free -= nlopt_iterations;
        DoubleVector hapfreqs = DoubleVector_Create(16);
        DoubleVector_PushBack(hapfreqs, x[0]);
        DoubleVector_PushBack(hapfreqs, x[1]);
        DoubleVector_PushBack(hapfreqs, x[2]);
        DoubleVector_PushBack(hapfreqs, x[3]);
        DoubleVector_PushBack(hapfreqs, x[4]);
        DoubleVector_PushBack(hapfreqs, x[5]);
        DoubleVector_PushBack(hapfreqs, x[6]);
        DoubleVector_PushBack(hapfreqs, x[7]);
        DoubleVector_PushBack(hapfreqs, x[8]);
        DoubleVector_PushBack(hapfreqs, x[9]);
        DoubleVector_PushBack(hapfreqs, x[10]);
        DoubleVector_PushBack(hapfreqs, x[11]);
        DoubleVector_PushBack(hapfreqs, x[12]);
        DoubleVector_PushBack(hapfreqs, x[13]);
        DoubleVector_PushBack(hapfreqs, x[14]);
        DoubleVector_PushBack(hapfreqs, x[15]);
        double p = x[16];
        DoubleVector penetrances = DoubleVector_Create(4);
        DoubleVector_PushBack(penetrances, x[17]);
        DoubleVector_PushBack(penetrances, x[19]);
        DoubleVector_PushBack(penetrances, x[18]);
        DoubleVector_PushBack(penetrances, x[20]);
        mod_score->test_markers = test_markers;
        mod_score->position = position;
        mod_score->mod_score = -negative_mod_score;
        mod_score->disease_allele_frequency = p;
        mod_score->penetrances = penetrances;
        mod_score->hapfreqs = hapfreqs;
        mod_score->em_hapfreqs = em_hapfreqs;
        print_maximum_3_tm_mat_first_without_deltas(-negative_mod_score, x, em_hapfreqs);
    }
    nlopt_destroy(opt);
    return nlopt_success;
}

static int optimize_with_3_test_markers_no_imprinting_without_deltas(
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModScore *gm,
    ModScore *mod_score,
    double initial_step_size1)
{
    //              H0  H1  H2  H3  H4  H5  H6  H7  H8  H9  H10 H11 H12 H13 H14 H15 p       f0 f1 f2
    double lb[] = { 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,    1e-8, 0, 0, 0};
    double ub[] = { 1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1-1e-8, 1, 1, 1};
    if (allfreq_restriction) {
        ub[16] = highest_allfreq;
    }
    struct DataStruct data = {
        annotated_families,
        test_markers,
        em_hapfreqs,
        0
    };

    nlopt_opt opt = nlopt_create(NLOPT_LN_COBYLA, 20); // algorithm and dimensionality
    nlopt_set_lower_bounds(opt, lb);
    nlopt_set_upper_bounds(opt, ub);
    nlopt_set_min_objective(opt, target_3_tm_pat_first_without_deltas, &data);
    nlopt_add_equality_constraint(opt, H_even_constr_3_tm_without_deltas, &data, DBL_EPSILON);
    nlopt_add_equality_constraint(opt, H_odd_constr_3_tm_without_deltas, &data, DBL_EPSILON);
    if (penetrance_restriction) {
        nlopt_add_inequality_constraint(opt, f0_f1_constr_3_tm_without_deltas, &data, DBL_EPSILON);
        nlopt_add_inequality_constraint(opt, f1_f2_constr_3_tm_without_deltas, &data, DBL_EPSILON);
    }
    nlopt_set_xtol_rel(opt, 1e-4);
    int max_eval = MIN(g_jla_nlopt_iterations_free, 10000);
    nlopt_set_maxeval(opt, max_eval);

    double x[] = {
        DoubleVector_Get(gm->hapfreqs, 0),
        DoubleVector_Get(gm->hapfreqs, 1),
        DoubleVector_Get(gm->hapfreqs, 2),
        DoubleVector_Get(gm->hapfreqs, 3),
        DoubleVector_Get(gm->hapfreqs, 4),
        DoubleVector_Get(gm->hapfreqs, 5),
        DoubleVector_Get(gm->hapfreqs, 6),
        DoubleVector_Get(gm->hapfreqs, 7),
        DoubleVector_Get(gm->hapfreqs, 8),
        DoubleVector_Get(gm->hapfreqs, 9),
        DoubleVector_Get(gm->hapfreqs, 10),
        DoubleVector_Get(gm->hapfreqs, 11),
        DoubleVector_Get(gm->hapfreqs, 12),
        DoubleVector_Get(gm->hapfreqs, 13),
        DoubleVector_Get(gm->hapfreqs, 14),
        DoubleVector_Get(gm->hapfreqs, 15),
        gm->disease_allele_frequency,
        DoubleVector_Get(gm->penetrances, 0),
        DoubleVector_Get(gm->penetrances, 1),
        DoubleVector_Get(gm->penetrances, 3)
    };
    nlopt_iterations = 0;
    nlopt_set_initial_step1(opt, initial_step_size1);

    double negative_mod_score; // minimum value of target function
    int nlopt_status = nlopt_optimize(opt, x, &negative_mod_score);
    int nlopt_success = 1;
    if (nlopt_status < 0) {
        fprintf(stderr, "%s\n", nlopt_result_to_string(nlopt_status));
        nlopt_success = 0;
        sf(ps, "nlopt failed!\n");
        pr();
    } else {
        if (negative_mod_score != global_minimum) {
            sf(ps, "COBYLA = %f vs %f = global minimum\n", negative_mod_score, global_minimum);
            pr();
            negative_mod_score = global_minimum;
            for (int i = 0; i < 20; i++) {
                x[i] = global_x[i];
            }
        }
        sf(ps, "No. of nlopt iterations: %d\n", nlopt_iterations);
        pr();
        g_jla_nlopt_iterations_free -= nlopt_iterations;
        DoubleVector hapfreqs = DoubleVector_Create(16);
        DoubleVector_PushBack(hapfreqs, x[0]);
        DoubleVector_PushBack(hapfreqs, x[1]);
        DoubleVector_PushBack(hapfreqs, x[2]);
        DoubleVector_PushBack(hapfreqs, x[3]);
        DoubleVector_PushBack(hapfreqs, x[4]);
        DoubleVector_PushBack(hapfreqs, x[5]);
        DoubleVector_PushBack(hapfreqs, x[6]);
        DoubleVector_PushBack(hapfreqs, x[7]);
        DoubleVector_PushBack(hapfreqs, x[8]);
        DoubleVector_PushBack(hapfreqs, x[9]);
        DoubleVector_PushBack(hapfreqs, x[10]);
        DoubleVector_PushBack(hapfreqs, x[11]);
        DoubleVector_PushBack(hapfreqs, x[12]);
        DoubleVector_PushBack(hapfreqs, x[13]);
        DoubleVector_PushBack(hapfreqs, x[14]);
        DoubleVector_PushBack(hapfreqs, x[15]);
        double p = x[16];
        DoubleVector penetrances = DoubleVector_Create(4);
        DoubleVector_PushBack(penetrances, x[17]);
        DoubleVector_PushBack(penetrances, x[18]);
        DoubleVector_PushBack(penetrances, x[18]);
        DoubleVector_PushBack(penetrances, x[19]);
        mod_score->test_markers = test_markers;
        mod_score->position = position;
        mod_score->mod_score = -negative_mod_score;
        mod_score->disease_allele_frequency = p;
        mod_score->penetrances = penetrances;
        mod_score->hapfreqs = hapfreqs;
        mod_score->em_hapfreqs = em_hapfreqs;
        print_maximum_3_tm_pat_first_without_deltas(-negative_mod_score, x, em_hapfreqs);
    }
    nlopt_destroy(opt);
    return nlopt_success;
}

static int optimize_user_model_with_3_test_markers_without_deltas(
    Vector annotated_families,
    Vector test_markers,
    DoubleVector em_hapfreqs,
    double position,
    ModelParams user_model,
    ModScore *gm,
    ModScore *mod_score,
    double initial_step_size1)
{
    //              H0  H1  H2  H3  H4  H5  H6  H7  H8  H9  H10 H11 H12 H13 H14 H15
    double lb[] = { 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0};
    double ub[] = { 1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1};
    struct DataStruct data = {
        annotated_families,
        test_markers,
        em_hapfreqs,
        user_model
    };

    nlopt_opt opt = nlopt_create(NLOPT_LN_COBYLA, 16); // algorithm and dimensionality
    nlopt_set_lower_bounds(opt, lb);
    nlopt_set_upper_bounds(opt, ub);
    nlopt_set_min_objective(opt, target_user_model_3_tm_without_deltas, &data);
    nlopt_add_equality_constraint(opt, H_even_user_model_constr_3_tm_without_deltas, &data, DBL_EPSILON);
    nlopt_add_equality_constraint(opt, H_odd_user_model_constr_3_tm_without_deltas, &data, DBL_EPSILON);
    nlopt_set_xtol_rel(opt, 1e-4);
    int max_eval = MIN(g_jla_nlopt_iterations_free, 10000);
    nlopt_set_maxeval(opt, max_eval);

    double x[] = {
        DoubleVector_Get(gm->hapfreqs, 0),
        DoubleVector_Get(gm->hapfreqs, 1),
        DoubleVector_Get(gm->hapfreqs, 2),
        DoubleVector_Get(gm->hapfreqs, 3),
        DoubleVector_Get(gm->hapfreqs, 4),
        DoubleVector_Get(gm->hapfreqs, 5),
        DoubleVector_Get(gm->hapfreqs, 6),
        DoubleVector_Get(gm->hapfreqs, 7),
        DoubleVector_Get(gm->hapfreqs, 8),
        DoubleVector_Get(gm->hapfreqs, 9),
        DoubleVector_Get(gm->hapfreqs, 10),
        DoubleVector_Get(gm->hapfreqs, 11),
        DoubleVector_Get(gm->hapfreqs, 12),
        DoubleVector_Get(gm->hapfreqs, 13),
        DoubleVector_Get(gm->hapfreqs, 14),
        DoubleVector_Get(gm->hapfreqs, 15)
    };
    nlopt_iterations = 0;
    nlopt_set_initial_step1(opt, initial_step_size1);

    double negative_mod_score; // minimum value of target function
    int nlopt_success = 1;
    if (nlopt_optimize(opt, x, &negative_mod_score) < 0) {
        nlopt_success = 0;
        sf(ps, "nlopt failed!\n");
        pr();
    } else {
        if (negative_mod_score != global_minimum) {
            sf(ps, "COBYLA = %f vs %f = global minimum\n", negative_mod_score, global_minimum);
            pr();
            negative_mod_score = global_minimum;
            for (int i = 0; i < 16; i++) {
                x[i] = global_x[i];
            }
        }
        sf(ps, "No. of nlopt iterations: %d\n", nlopt_iterations);
        pr();
        g_jla_nlopt_iterations_free -= nlopt_iterations;
        DoubleVector hapfreqs = DoubleVector_Create(16);
        DoubleVector_PushBack(hapfreqs, x[0]);
        DoubleVector_PushBack(hapfreqs, x[1]);
        DoubleVector_PushBack(hapfreqs, x[2]);
        DoubleVector_PushBack(hapfreqs, x[3]);
        DoubleVector_PushBack(hapfreqs, x[4]);
        DoubleVector_PushBack(hapfreqs, x[5]);
        DoubleVector_PushBack(hapfreqs, x[6]);
        DoubleVector_PushBack(hapfreqs, x[7]);
        DoubleVector_PushBack(hapfreqs, x[8]);
        DoubleVector_PushBack(hapfreqs, x[9]);
        DoubleVector_PushBack(hapfreqs, x[10]);
        DoubleVector_PushBack(hapfreqs, x[11]);
        DoubleVector_PushBack(hapfreqs, x[12]);
        DoubleVector_PushBack(hapfreqs, x[13]);
        DoubleVector_PushBack(hapfreqs, x[14]);
        DoubleVector_PushBack(hapfreqs, x[15]);
        double p = user_model->disease_allele_frequency;
        DoubleVector penetrances = DoubleVector_Create(4);
        DoubleVector_PushBack(penetrances, user_model->f0);
        DoubleVector_PushBack(penetrances, user_model->f1_pat);
        DoubleVector_PushBack(penetrances, user_model->f1_mat);
        DoubleVector_PushBack(penetrances, user_model->f2);
        mod_score->test_markers = test_markers;
        mod_score->position = position;
        mod_score->mod_score = -negative_mod_score;
        mod_score->disease_allele_frequency = p;
        mod_score->penetrances = penetrances;
        mod_score->hapfreqs = hapfreqs;
        mod_score->em_hapfreqs = em_hapfreqs;
        print_maximum_user_model_3_tm_without_deltas(-negative_mod_score, x, em_hapfreqs, user_model);
    }
    nlopt_destroy(opt);
    return nlopt_success;
}

ModScore *NloptModScoreFunction(
    Vector test_markers,
    Vector annotated_families,
    DoubleVector em_hapfreqs,
    LinkageMarkerFile linkage_file,
    ModelParams user_model,
    int number_of_test_marker_sets,
    ModScore *grid_mod_score)
{
    (void) number_of_test_marker_sets;
    int pat_first = genrand_open_open() < .5;
    global_minimum = HUGE_VAL;
    for (int i = 0; i < 20+1; i++) {
        global_x[i] = 0;
    }
    int nlopt_success = 1;
    ModScore *mod_score = ModScore_Create();
    ModScore *synthetic_dummy_mod_score = ModScore_Create();
    int number_of_test_markers = Vector_GetSize(test_markers);
    double position = LinkageMarkerFile_GetDistanceBetween(linkage_file, Marker_GetIndex(Vector_Get(test_markers, 0)), 0);
    if (!user_model) {
        OptimizeFunction *optimize = 0;
        if (number_of_test_markers == 1) {
            synthetic_dummy_mod_score->penetrances = DoubleVector_With(4, .5, .5, .5, .5);
            synthetic_dummy_mod_score->disease_allele_frequency = highest_allfreq;
            if (g_jla_delta_modscore) {
                synthetic_dummy_mod_score->deltas = DoubleVector_With(2, .0, .0);
            } else {
                double h0 = DoubleVector_Get(em_hapfreqs, 0) * (1 - highest_allfreq);
                double h1 = DoubleVector_Get(em_hapfreqs, 0) * highest_allfreq;
                double h2 = DoubleVector_Get(em_hapfreqs, 1) * (1 - highest_allfreq);
                double h3 = DoubleVector_Get(em_hapfreqs, 1) * highest_allfreq;
                synthetic_dummy_mod_score->hapfreqs = DoubleVector_With(4, h0, h1, h2, h3);
            }
            if (imprinting) {
                if (g_jla_delta_modscore) {
                    optimize = pat_first ? optimize_with_1_test_marker_pat_first
                                         : optimize_with_1_test_marker_mat_first;
                } else {
                    optimize = pat_first ? optimize_with_1_test_marker_pat_first_without_deltas
                                         : optimize_with_1_test_marker_mat_first_without_deltas;
                }
            } else {
                if (g_jla_delta_modscore) {
                    optimize = optimize_with_1_test_marker_no_imprinting;
                } else {
                    optimize = optimize_with_1_test_marker_no_imprinting_without_deltas;
                }
            }
        } else if (number_of_test_markers == 2) {
            synthetic_dummy_mod_score->penetrances = DoubleVector_With(4, .5, .5, .5, .5);
            synthetic_dummy_mod_score->disease_allele_frequency = highest_allfreq;
            if (g_jla_delta_modscore) {
                synthetic_dummy_mod_score->deltas = DoubleVector_With(4, .0, .0, .0, .0);
            } else {
                double h0 = DoubleVector_Get(em_hapfreqs, 0) * (1 - highest_allfreq);
                double h1 = DoubleVector_Get(em_hapfreqs, 0) * highest_allfreq;
                double h2 = DoubleVector_Get(em_hapfreqs, 1) * (1 - highest_allfreq);
                double h3 = DoubleVector_Get(em_hapfreqs, 1) * highest_allfreq;
                double h4 = DoubleVector_Get(em_hapfreqs, 2) * (1 - highest_allfreq);
                double h5 = DoubleVector_Get(em_hapfreqs, 2) * highest_allfreq;
                double h6 = DoubleVector_Get(em_hapfreqs, 3) * (1 - highest_allfreq);
                double h7 = DoubleVector_Get(em_hapfreqs, 3) * highest_allfreq;
                synthetic_dummy_mod_score->hapfreqs = DoubleVector_With(8, h0, h1, h2, h3, h4, h5, h6, h7);
            }
            if (imprinting) {
                if (g_jla_delta_modscore) {
                    optimize = pat_first ? optimize_with_2_test_markers_pat_first
                                         : optimize_with_2_test_markers_mat_first;
                } else {
                    optimize = pat_first ? optimize_with_2_test_markers_pat_first_without_deltas
                                         : optimize_with_2_test_markers_mat_first_without_deltas;
                }
            } else {
                if (g_jla_delta_modscore) {
                    optimize = optimize_with_2_test_markers_no_imprinting;
                } else {
                    optimize = optimize_with_2_test_markers_no_imprinting_without_deltas;
                }
            }
        } else if (number_of_test_markers == 3) {
            synthetic_dummy_mod_score->penetrances = DoubleVector_With(4, .5, .5, .5, .5);
            synthetic_dummy_mod_score->disease_allele_frequency = highest_allfreq;
            if (g_jla_delta_modscore) {
                synthetic_dummy_mod_score->deltas = DoubleVector_With(8, .0, .0, .0, .0, .0, .0, .0, .0);
            } else {
                double h0 = DoubleVector_Get(em_hapfreqs, 0) * (1 - highest_allfreq);
                double h1 = DoubleVector_Get(em_hapfreqs, 0) * highest_allfreq;
                double h2 = DoubleVector_Get(em_hapfreqs, 1) * (1 - highest_allfreq);
                double h3 = DoubleVector_Get(em_hapfreqs, 1) * highest_allfreq;
                double h4 = DoubleVector_Get(em_hapfreqs, 2) * (1 - highest_allfreq);
                double h5 = DoubleVector_Get(em_hapfreqs, 2) * highest_allfreq;
                double h6 = DoubleVector_Get(em_hapfreqs, 3) * (1 - highest_allfreq);
                double h7 = DoubleVector_Get(em_hapfreqs, 3) * highest_allfreq;
                double h8 = DoubleVector_Get(em_hapfreqs, 4) * (1 - highest_allfreq);
                double h9 = DoubleVector_Get(em_hapfreqs, 4) * highest_allfreq;
                double h10 = DoubleVector_Get(em_hapfreqs, 5) * (1 - highest_allfreq);
                double h11 = DoubleVector_Get(em_hapfreqs, 5) * highest_allfreq;
                double h12 = DoubleVector_Get(em_hapfreqs, 6) * (1 - highest_allfreq);
                double h13 = DoubleVector_Get(em_hapfreqs, 6) * highest_allfreq;
                double h14 = DoubleVector_Get(em_hapfreqs, 7) * (1 - highest_allfreq);
                double h15 = DoubleVector_Get(em_hapfreqs, 7) * highest_allfreq;
                synthetic_dummy_mod_score->hapfreqs = DoubleVector_With(16, h0, h1, h2, h3, h4, h5, h6, h7, h8, h9, h10, h11, h12, h13, h14, h15);
            }
            if (imprinting) {
                if (g_jla_delta_modscore) {
                    optimize = pat_first ? optimize_with_3_test_markers_pat_first
                                         : optimize_with_3_test_markers_mat_first;
                } else {
                    optimize = pat_first ? optimize_with_3_test_markers_pat_first_without_deltas
                                         : optimize_with_3_test_markers_mat_first_without_deltas;
                }
            } else {
                if (g_jla_delta_modscore) {
                    optimize = optimize_with_3_test_markers_no_imprinting;
                } else {
                    optimize = optimize_with_3_test_markers_no_imprinting_without_deltas;
                }
            }
        }
        synthetic_dummy_mod_score->position = position;
        synthetic_dummy_mod_score->em_hapfreqs = em_hapfreqs;
        synthetic_dummy_mod_score->test_markers = test_markers;
        synthetic_dummy_mod_score->mod_score = 0.0;
        nlopt_success = optimize_wrapper(
            annotated_families, test_markers, em_hapfreqs,
            position, grid_mod_score, synthetic_dummy_mod_score,
            optimize, mod_score
        );
    } else {
        OptimizeFunctionUserModel *optimize = 0;
        if (number_of_test_markers == 1) {
            if (g_jla_delta_modscore) {
                optimize = optimize_user_model_with_1_test_marker;
            } else {
                optimize = optimize_user_model_with_1_test_marker_without_deltas;
            }
        } else if (number_of_test_markers == 2) {
            if (g_jla_delta_modscore) {
                optimize = optimize_user_model_with_2_test_markers;
            } else {
                optimize = optimize_user_model_with_2_test_markers_without_deltas;
            }
        } else if (number_of_test_markers == 3) {
            if (g_jla_delta_modscore) {
                optimize = optimize_user_model_with_3_test_markers;
            } else {
                optimize = optimize_user_model_with_3_test_markers_without_deltas;
            }
        }
        nlopt_success = optimize_user_model_wrapper(
            annotated_families, test_markers, em_hapfreqs, position,
            user_model, grid_mod_score, optimize,
            mod_score
        );
    }
    Memory_Free(synthetic_dummy_mod_score->deltas);
    Memory_Free(synthetic_dummy_mod_score->hapfreqs);
    Memory_Free(synthetic_dummy_mod_score->penetrances);
    if (nlopt_success == 1)
        return mod_score;
    else
        return NULL;
}
