/* GENEHUNTER-MODSCORE - modscore.c
   module to perform MOD-score analysis
   written by Konstantin Strauch, Johannes Dietter, Manuel Mattheisen,
   and Markus Brugger.    Copyright (c) 2002-2019 */

#define INC_LIB
#define INC_SHELL
#define real_eq(x,y) (fabs((x)-(y))<1e-10)
#include "npl.h"
#include "time.h"
#include "Association.h"
#include "CommandReader.h"
#include "GHMRunner.h"
#include "ModelParamsFile.h"
#include "PedigreeFile.h"
#include "PedigreeFileSimulator.h"

#include <errno.h>       // errno
#include <stdio.h>       // snprintf
#include <sys/stat.h>    // mkdir
#include <sys/types.h>   // mkdir

static void mkdir_or_fail(const char *directory)
{
    errno = 0;
    int ok = mkdir(directory, 0755) == 0 || errno == EEXIST;
    if (!ok) {
        sf(ps, "mkdir failed: %s\n", directory);
        pr();
        exit(1);
    }
}

void init_disease_info (void); /* in locprep.c */
void init_disease_probs (int,PED_MEMBER*);    /* in finalprep.c */
void peel (int,int,PED_MEMBER*,NUC_FAMILY*);  /* in finalprep.c */
double brute_force_analyze (int,int*,int,int*,PED_MEMBER*); /* in finalprep.c*/
void algebra_force(int); /* in finalprep.c, algebraic calculation */
void algebra_cache_opt(double*, int, int*); /* in finalprep.c, algebraic calculation */
void algebra_power_prep(void); /* in finalprep.c, algebraic calculation */
void algebra_calculated(void); /* in finalprep.c, algebraic calculation */
void algebra_force_famnum(int, double*); /* in finalprep.c, algebraic calculation */
void scan_pedigree(int *nz_bits_s, int *nz_bits_s_m, int *nz_bits_s_f, unsigned long *nz_pointer_s, double *pvct_s, double *num_par_score_s, int f_num_in_map_order_s, int *f_uninformative_s, int *f_map_order_s, int real_num_bits, int real_num_bits_m_s, int real_num_bits_f_s, double *score, int score_mask_rdc_s, double exp, double exp2, int pednum, int *best_pv, int num_bits, int num_bits_m_s, int num_bits_f_s, int u_mask, int *f_mask, int **maskT, int this_mod_pos_s);
void fill_placeholder_alleles_1(int, int, PED_MEMBER *);

/* local */
void calc_global_mod (void), calc_single_mod (void);
int optimize_parameter (double*);
int optimize_two_parameters (double*,double*);
int optimize_three_parameters (double*,double*,double*);
int optimize_four_parameters (double*,double*,double*,double*);
int optimize_five_parameters (double*,double*,double*,double*,double*);
double get_next_higher (double), get_next_lower (double);
double get_next_higher_allfreq (double), get_next_lower_allfreq (double);
int penetrance_restriction_violated (void);
int user_penetrance_restriction_violated (double,double,double,double);
int all_penetrances_equal (void);
int all_penetrances_all_liab_equal (void);
int allfreq_restriction_violated (void);
int user_allfreq_restriction_violated (double);
int model_already_checked (void);
void error_too_many_models (void);
int calculate_lod (void);
int calculate_lod_famnum (int);

/*** for maximization options no longer in calc_single or calc_global: ***/
void asymmetric_grid_calc (void);
void fine_maximization (void);

/*** new for PMOD ***/
void symmetric_grid_calc (void); 

static char *lod_str, *pos_str;
static int pos, maxlodpos, modcount, userdefcount=0,
           userdef_pen_viol_count=0, userdef_allfreq_viol_count=0,
           predefined_count, best_model, old_analysis_type,
           allele_frequency_flag, modpos[MAX_MODELS],
           modscore_model_incompatible,
           grand_total_models, total_quick_models, best_model_flag;
static double maxlod;
static double modpen1[MAX_MODELS], modpen2[MAX_MODELS],
              modpen3[MAX_MODELS], modpen4[MAX_MODELS],
              modafreq[MAX_MODELS], mod[MAX_MODELS];

/*** new for maximization options: stores best_model when needed ***/
static int a_gridbest, s_gridbest, s_finebest;

static double userdefpen1[MAX_USERDEF_MODELS],
              userdefpen2[MAX_USERDEF_MODELS],
              userdefpen3[MAX_USERDEF_MODELS],
              userdefpen4[MAX_USERDEF_MODELS],
              userdefafreq[MAX_USERDEF_MODELS];
static double **pospen1, **pospen2, **pospen3, **pospen4, *posafreq;
static double **saved_model_pen1, **saved_model_pen2, **saved_model_pen3,
              **saved_model_pen4, *saved_model_afreq;
double ***saved_model_vectors;  /* now extern for PMOD */

static int current_saved_model, initial_count,
           initial_fill_storage, model_storage_full;


command model(void)   /* add a user-defined model to the MOD-score analysis */
{
  double pen1, pen2, pen3, pen4, afreq;
  int count, model_added;

  model_added = FALSE;
  if (scan_done && modcalc==FALSE) {
    print("Pedigrees have already been analyzed with modcalc turned off.\n");
    print("This command will therefore be ignored.\n");
  } else if (imprinting) {

    get_arg(rtoken,-999.0,&afreq);
    get_arg(rtoken,-999.0,&pen1);
    get_arg(rtoken,-999.0,&pen2);
    get_arg(rtoken,-999.0,&pen3);
    get_arg(rtoken,-999.0,&pen4);
    if (afreq < 0.0 || afreq > 1.0 ||
	pen1 < 0.0 || pen1 > 1.0 ||
	pen2 < 0.0 || pen2 > 1.0 ||
	pen3 < 0.0 || pen3 > 1.0 ||
	pen4 < 0.0 || pen4 > 1.0)
      error("parameters must be given in the order <p> <f+/+> <fm/+> <f+/m> <fm/m>");
    if (real_eq(afreq, 0.0) || real_eq(afreq, 1.0))
      error("disease allele frequency (p) must be > 0.0 and < 1.0");
    if (real_eq(pen1, pen2) && real_eq(pen1, pen3) && real_eq(pen1, pen4))
      error("all penetrances are equal");
    for (count=0; count<userdefcount; count++)
      if (real_eq(userdefpen1[count], pen1) &&
	  real_eq(userdefpen2[count], pen2) &&
	  real_eq(userdefpen3[count], pen3) &&
	  real_eq(userdefpen4[count], pen4) &&
	  real_eq(userdefafreq[count], afreq))
	error("this trait model has already been entered before");
    if (userdefcount >= MAX_USERDEF_MODELS) {
      sf(ps, "number of added models > MAX_USERDEF_MODELS (%d) - please edit npl.h", MAX_USERDEF_MODELS);
      error(ps);
    }
    userdefpen1[userdefcount] = pen1;
    userdefpen2[userdefcount] = pen2;
    userdefpen3[userdefcount] = pen3;
    userdefpen4[userdefcount] = pen4;
    userdefafreq[userdefcount] = afreq;
    userdefcount++;
    model_added = TRUE;
    print("The following trait model will be added to the MOD-score analysis:\n");
    sf(ps, "p = %8.6lf,  (f+/+, fm/+, f+/m, fm/m) = (%6.4lf, %6.4lf, %6.4lf, %6.4lf)\n", afreq, pen1, pen2, pen3, pen4); pr();
    if (user_penetrance_restriction_violated(pen1, pen2, pen3, pen4)) {
      print("Warning: This model will only be used with 'penetrance restriction off'.\n");
      userdef_pen_viol_count++;
    }
    if (user_allfreq_restriction_violated(afreq)) {
      sf(ps, "Warning: With the current highest allfreq of %.2lf, this model will only\n", highest_allfreq); pr();
      print("         be used if 'allfreq restriction' is 'off'.\n");
      userdef_allfreq_viol_count++;
    }

  } else {  /* no imprinting */

    get_arg(rtoken,-999.0,&afreq);
    get_arg(rtoken,-999.0,&pen1);
    get_arg(rtoken,-999.0,&pen2);
    get_arg(rtoken,-999.0,&pen3);
    get_arg(rtoken,-999.0,&pen4);
    if (afreq < 0.0 || afreq > 1.0 ||
	pen1 < 0.0 || pen1 > 1.0 ||
	pen2 < 0.0 || pen2 > 1.0 ||
	pen3 < 0.0 || pen3 > 1.0 ||
	pen4 > -990.0)
      error("parameters must be given in the order <p> <f+/+> <fHet> <fm/m>");
    if (real_eq(afreq, 0.0) || real_eq(afreq, 1.0))
      error("disease allele frequency (p) must be > 0.0 and < 1.0");
    if (real_eq(pen1, pen2) && real_eq(pen1, pen3))
      error("all penetrances are equal");
    for (count=0; count<userdefcount; count++)
      if (real_eq(userdefpen1[count], pen1) &&
	  real_eq(userdefpen2[count], pen2) &&
	  real_eq(userdefpen3[count], pen2) &&
	  real_eq(userdefpen4[count], pen3) &&
	  real_eq(userdefafreq[count], afreq))
	error("this trait model has already been entered before");
    if (userdefcount >= MAX_USERDEF_MODELS) {
      sf(ps, "number of added models > MAX_USERDEF_MODELS (%d) - please edit npl.h", MAX_USERDEF_MODELS);
      error(ps);
    }
    userdefpen1[userdefcount] = pen1;
    userdefpen2[userdefcount] = pen2;
    userdefpen3[userdefcount] = pen2;
    userdefpen4[userdefcount] = pen3;
    userdefafreq[userdefcount] = afreq;
    userdefcount++;
    model_added = TRUE;
    print("The following trait model will be added to the MOD-score analysis:\n");
    sf(ps, "p = %8.6lf,  (f+/+, fHet, fm/m) = (%6.4lf, %6.4lf, %6.4lf)\n",
       afreq, pen1, pen2, pen3); pr();
    if (user_penetrance_restriction_violated(pen1, pen2, pen2, pen3)) {
      print("Warning: This model will only be used with 'penetrance restriction off'.\n");
      userdef_pen_viol_count++;
    }
    if (user_allfreq_restriction_violated(afreq)) {
      sf(ps, "Warning: With the current highest allfreq of %.2lf, this model will only\n", highest_allfreq); pr();
      print("         be used if 'allfreq restriction' is 'off'.\n");
      userdef_allfreq_viol_count++;
    }
  }
  if (modscore_done && model_added) {
      /* do not allow p-value calculation if a new trait model
	 is added after executing 'modscore' */
      changed_modscore_options = TRUE;
  }
}


command modscore(void)
{
  int i;

  if (g_association && modscore_done) {
    sf(ps, "Calling modscore command more than once is not allowed in JLA analysis mode.\n");
    pr();
    return;
  }

  replicates_error_flag = TRUE; /* will be set to 'FALSE' at the end and therefore only 
                                be true when the 'command modscore' is not finished accurately */

  if (single_point_mode)
    error("MOD-score analysis does not work with single point mode");

  if (analysis_type == NPL_ANALYSIS)
    error("MOD-score analysis only works if LOD calculation is activated");

  if (!modcalc)
    error("'modcalc' must be set to 'global' or 'single' before scan pedigrees");

  if (!scan_done)
    error("no scan done; need to scan pedigrees first");

  if (pedigrees_calculated == 0)
    error("no pedigree analyzed so far, please check pedigree file for 'informative' pedigrees");

  if (scan_error_flag) error("a MOD-score analysis cannot be performed since an error occurred during scan pedigrees");

  if (datafile_model_incompatible) error("the datafile disease model is incompatible with the trait phenotypes\nPlease specify a different trait model in the locus datafile\n(try a small phenocopy rate and incomplete penetrance).");

  if (modliab > num_liability_classes) {
    sf(ps,"liability class to be optimized (%d) does not exist", modliab);
    error(ps);
  }

  if (!pmod_calc_flag && replicates_count > 0) {
    /* modscore() is called after pmod_calc has finished */
    print("Since p-values have already been calculated, this command will be ignored.\n");
    return;
  }

  lod_str = get_temp_string();
  pos_str = get_temp_string();
  old_analysis_type = analysis_type;
  analysis_type = LOD_ANALYSIS;

  if(!modscore_done){
	  initial_count = 0;  /** need this to keep track of saved models  **/
	  current_saved_model = 0;
	  model_storage_full = FALSE;
	  initial_fill_storage = FALSE;
  }
  best_model_flag = FALSE;

  /* The following was in calc_single_mod before. Now in case of p-value
     calculation also needed for 'modcalc global'*/

  if (saved_models == -1) {  /* that is, save all initial models */
    if (maxim_type == 1 || maxim_type == 5) {
      if ((!imprinting) && penetrance_restriction) saved_models = 81;
      else if ((!imprinting) && (!penetrance_restriction)) saved_models = 135;
      else if (imprinting && penetrance_restriction) saved_models = 261;
      else if (imprinting && (!penetrance_restriction)) saved_models = 567;
    } else if (maxim_type == 2 || maxim_type == 4) { 
      if ((!imprinting) && penetrance_restriction) saved_models = 236;
      else if ((!imprinting) && (!penetrance_restriction)) saved_models = 848;
      else if (imprinting && penetrance_restriction) saved_models = 988;
      else if (imprinting && (!penetrance_restriction)) saved_models = 4928;
    } else if (maxim_type == 6) {
      saved_models = 0;
    }
    saved_models += userdefcount;
    if (penetrance_restriction) saved_models -= userdef_pen_viol_count;
    /* don't count userdef'd models that violate penetrance restriction */
    if (allfreq_restriction) saved_models -= userdef_allfreq_viol_count;
    /* don't count userdef'd models with disease allfreq > highest_allfreq */
    if (num_liability_classes > 1) saved_models++;
    /* if more than one liability class, also save model w/all pen equal */
  }

  /* In case a MOD-score analysis with 'modcalc global' is performed
     without p-value calculation, saving scoring functions is without any
     gain in terms of computation time, therefore saved_models will be
     set to '0' and hence no scoring functions will be stored */
  if (num_replicates < 1 && modcalc == 1)
    saved_models = 0;

  /* The following was in calc_single_mod before. Now for PMOD also 
     needed for calc_global_mod: */
  if (!modscore_done && saved_models) {
    /*** allocate memory for saved model vectors ***/
    /* Modified: In case p-values will be calculated for MOD scores 
      be sure to allocate all memory and not only for those families
      which are analyzed in the real data set. Address variable not with
      'famnum' but with 'pointer' on famnum = 'famnum_pointer[famnum]'.
      Since real_num_vecs is not calculated for pedigrees which have been
      skipped during the analysis of the real data set, only allocate here
      saved_model_vectors for those pedigrees which have not been skipped
      in finalprep.c and allocate memory for the other pedigrees later
      (see finalprep.c) */
    array(saved_model_vectors, num_pedigrees_allocated, double**);
    array(saved_model_allocated, num_pedigrees_allocated, bool);
    matrix (quick_model_famnum, num_pedigrees_allocated, saved_models, int);

    for (famnum=0; famnum<num_pedigrees_allocated; famnum++) {
      saved_model_allocated[famnum] = FALSE;
      for (i=0; i<saved_models; i++)
        quick_model_famnum[famnum][i] = 0;
    }

    for (famnum=0; famnum<num_pedigrees; famnum++) {
      if (real_num_vecs_calculated[famnum_pointer[famnum]] == TRUE) {
        matrix(saved_model_vectors[famnum_pointer[famnum]], saved_models,
               save_real_num_vecs[famnum_pointer[famnum]], double);
        saved_model_allocated[famnum_pointer[famnum]] = TRUE;
      }
    }
    /* We need to store the penetrances of each saved model for all       *
     * liability classes to cover the case that a MOD-score analysis has  *
     * already been done for another liability class, since this leads to *
     * different penetrances (of other liability classes optimized in     *
     * the previous MOD-score analyses) at distinct loci.                 */
    matrix(saved_model_pen1, saved_models, num_liability_classes, double);
    matrix(saved_model_pen2, saved_models, num_liability_classes, double);
    matrix(saved_model_pen3, saved_models, num_liability_classes, double);
    matrix(saved_model_pen4, saved_models, num_liability_classes, double);
    array(saved_model_afreq, saved_models, double);
  }

  if (g_association) {
      if (g_jla_p_value) {
          PedigreeFile ped_file = PedigreeFile_Create(g_pedigree_file);
          Vector test_markers = Vector_Create(Vector_GetSize(g_test_marker_names));
          PedigreeFileSimulator simulator = NULL;
          int number_of_test_markers = Vector_GetSize(g_test_marker_names);
          for (int i = 0; i < number_of_test_markers; i++) {
              const char *marker_name = Vector_Get(g_test_marker_names, i);
              int marker_index = LinkageMarkerFile_GetMarkerIndex(g_linkage_file, marker_name);
              Marker marker = LinkageMarkerFile_GetMarkerByIndex(g_linkage_file, marker_index);
              Vector_PushBack(test_markers, marker);
          }
          CommandReader reader = CommandReader_Create(g_commands_file);
          GHMRunner runner = GHMRunner_Create(reader, number_of_test_markers);
          char directory[200];
          int ok = snprintf(directory, sizeof directory, "%s/real", g_jla_directory) < (int) sizeof directory;
          if (!ok) {
              sf(ps, "Buffer too short for directory name: \"%s/real\"\n", g_jla_directory);
              pr();
              exit(1);
          }
          mkdir_or_fail(directory);
          struct GHMResults results = GHMRunner_Run(runner, directory, g_random_seed);
          if (g_model_params_file) {
              if (ModelParamsFile_GetNumberOfModels(g_model_params_file) > 1) {
                  sf(ps, "In the case of p value simulation, only a single user-defined model can be analyzed!\n");
                  pr();
                  exit(1);
              }
              simulator = PedigreeFileSimulator_Create(
                      g_families, ped_file, g_linkage_file, g_user_linkage_file, test_markers, g_model_params_file, results.hapfreqs);
          }
          else {
              simulator = PedigreeFileSimulator_Create(
                      g_families, ped_file, g_linkage_file, g_user_linkage_file, test_markers, NULL, results.hapfreqs);
          }
          if (simulator) {
              simulate_jla_p_value(simulator, reader, results.mod_score);
          }
      } else {
          compute_mod_scores();
      }
  } else if (modcalc == 1) {
      calc_global_mod();
  } else if (modcalc == 2) {
      calc_single_mod();
  }

  analysis_type = old_analysis_type;
  if (modscore_done)
    changed_modscore_options = TRUE; /* this is checked in pmod.c;
				        do not allow p-value calculation if
				        modscore is called more than once */
  modscore_done = TRUE;
  replicates_error_flag = FALSE;
}


/************************* modcalc global *************************/

void calc_global_mod (void)
{
  int i, real_userdefcount;   /* changed_parm now in fine_maximization() */
  double one_dominance_index, one_imprinting_index;
  unsigned long int mod_memcount=0; /* only for p-value calculation used in calc_global_mod */

  /* print statements for real data set */
  sf(ps,"\nMOD-score calculation with option 'global'.\n");
  if (!pmod_calc_flag)
    pr();
  else if (replicates_store_result && pmod_calc_flag)
    fpr(replifp); 

  if ((lowpos<-990.0) && (highpos>990.0))
    sf(ps,"Maximization will be done globally over all genetic positions,\n"); 
  else 
    sf(ps,"Maximization will be done globally over all pos between %.2lf and %.2lf cM,\n", lowpos, highpos);
  if (!pmod_calc_flag)
    pr();
  else if (replicates_store_result && pmod_calc_flag)
    fpr(replifp); 
  sf(ps,"by varying penetrances of liability class %d and disease allele frequency.\n", modliab);
  if (!pmod_calc_flag)
    pr();
  else if (replicates_store_result && pmod_calc_flag)
    fpr(replifp); 

  if (dimensions == 1)
    sf(ps,"Only 1 parameter is changed at a time for fine maximization.\n");
  else 
    sf(ps,"Up to %d parameters are jointly changed for fine maximization.\n",
          dimensions);  
  if (!pmod_calc_flag)
    pr();
  else if (replicates_store_result && pmod_calc_flag)
    fpr(replifp); 

  sf(ps,"Imprinting mode is '%s'.\n", imprinting?"on":"off");
  if (!pmod_calc_flag)
    pr();
  else if (replicates_store_result && pmod_calc_flag)
    fpr(replifp); 

  sf(ps,"Penetrance restriction \"f+/+ <= (fm/+,f+/m) <= fm/m\" is '%s'.\n",
       penetrance_restriction?"on":"off");

  if (!pmod_calc_flag)
    pr();
  else if (replicates_store_result && pmod_calc_flag)
    fpr(replifp); 

  if (allfreq_restriction)
    sf(ps,"Disease allele frequency restriction \"p <= %.2lf\" is 'on'.\n",
         highest_allfreq);
  else
    sf(ps,"Disease allele frequency restriction is 'off'.\n");
  if (!pmod_calc_flag)
    pr();
  else if (replicates_store_result && pmod_calc_flag)
    fpr(replifp);

  if (maxim_type == 1) {
    sf(ps,"Maximization will be performed with the standard set of predefined\ntrait models, followed by a fine maximization.\n");
  } else if (maxim_type == 2) {
    sf(ps,"Maximization will be performed with the larger set of predefined\ntrait models, followed by a fine maximization.\n");
  } else if (maxim_type == 4) {
    sf(ps,"Maximization will be done only with the larger set of predefined\ntrait models, without a fine maximization.\n");
  } else if (maxim_type == 5) {
    sf(ps,"Maximization will be done only with the standard set of predefined\ntrait models, without a fine maximization.\n");
  } else if (maxim_type == 6) {
    sf(ps,"Maximization will be done only with the set of user-defined trait models\n(those entered by the 'model' command), without a fine maximization.\n");
  }
  if (!pmod_calc_flag)
    pr();
  else if (replicates_store_result && pmod_calc_flag)
    fpr(replifp);

  if (saved_models == 1)
    sf(ps,"In case of p-value calculation, 1 model is saved in memory for MOD\n");
  else if (num_replicates > 0)
    sf(ps,"In case of p-value calculation, %d models are saved in memory for MOD\n",
       saved_models);
  if (!pmod_calc_flag)
    pr();
  else if (replicates_store_result && pmod_calc_flag)
    fpr(replifp); 
 
  if (saved_models && num_replicates > 0) {
    for (famnum=0; famnum<num_pedigrees; famnum++)
      mod_memcount += save_real_num_vecs[famnum_pointer[famnum]] * sizeof(double);
    mod_memcount *= saved_models;
    sf(ps,"calculation. With the current dataset, this will need %lu bytes of\nmemory.\n",
       mod_memcount);
    if (!pmod_calc_flag)
      pr();
    else if (replicates_store_result && pmod_calc_flag)
      fpr(replifp); 
  }

  sf(ps,"\nChecking models, looking at all positions for each model...\n\n");
  if (!pmod_calc_flag)
    pr();
  else if (replicates_store_result && pmod_calc_flag)
    fpr(replifp); 

  if (ind_sp==0) {
    if (imprinting) 
      sf(ps,"    p     ( f+/+ ,  fm/+ ,  f+/m ,  fm/m )  position     MOD\n");
     else 
      sf(ps,"    p     ( f+/+ ,  fHet ,  fm/m )  position     MOD\n");
  } else if(ind_sp==1) {
      if (imprinting) 
        sf(ps,"    p     ( f+/+ ,  fm/+ ,  f+/m ,  fm/m )  pos.(avg. female  male)     MOD\n");
      else 
        sf(ps,"    p     ( f+/+ ,  fHet ,  fm/m )  pos.(avg. female  male)     MOD\n");
  }
  if (!pmod_calc_flag)
    pr();
  else if (replicates_store_result && pmod_calc_flag)
    fpr(replifp);

  /*** now delete family LODs and total LODs for pos. not considered
       (get rid of possibly remaining junk in arrays) ***/
  for (i=0; i<first_mod_pos; i++) {
    for (famnum=0; famnum<num_pedigrees; famnum++) sub_lods[famnum][i] = 0.0;
    total_lod[i] = 0.0;
  }
  for (i=last_mod_pos+1; i<num_w_pos; i++) {
    for (famnum=0; famnum<num_pedigrees; famnum++) sub_lods[famnum][i] = 0.0;
    total_lod[i] = 0.0;
  }

  modcount = 0;
  total_quick_models = 0;

  /*** get maximum LOD of the base model done with scan pedigrees
       - or, if a modscore round has already been performed for
       another liability class, of the best model obtained previously ***/
  maxlodpos = -1;
  maxlod = -10000.0;
  for (i=first_mod_pos; i<=last_mod_pos; i++) 
     if (total_lod[i] > maxlod) {
       maxlod = total_lod[i];
       maxlodpos = i;
     }
  if (ind_sp==1) {
    sf(pos_str,"%7.2lf %7.2lf %7.2lf",0.5*(wpos_m[maxlodpos] + wpos_f[maxlodpos]), wpos_f[maxlodpos], wpos_m[maxlodpos]);
  } else if (ind_sp==0) {
    sf(pos_str,"%7.2lf",wpos[maxlodpos]);
  }  
  if (maxlod < -9000.0) strcpy(lod_str," -INFINITY");
  else sf(lod_str,"%10.6lf", maxlod);

  if (imprinting)
    sf(ps,"%8.6lf  (%6.4lf, %6.4lf, %6.4lf, %6.4lf)  %s %s   initial MOD\n",
       affected_freq,
       penetrance[modliab][HOM_UNAFF], penetrance[modliab][HET_AU],
       penetrance[modliab][HET_UA], penetrance[modliab][HOM_AFF],
       pos_str, lod_str);
  else
    sf(ps,"%8.6lf  (%6.4lf, %6.4lf, %6.4lf)  %s %s   initial MOD\n",
       affected_freq,
       penetrance[modliab][HOM_UNAFF], penetrance[modliab][HET_AU],
       penetrance[modliab][HOM_AFF], pos_str, lod_str);
  if (!pmod_calc_flag)
    pr();
  else if (replicates_store_result && pmod_calc_flag)
    fpr(replifp);

  modpen1[0] = penetrance[modliab][HOM_UNAFF];
  modpen2[0] = penetrance[modliab][HET_AU];
  modpen3[0] = penetrance[modliab][HET_UA];
  modpen4[0] = penetrance[modliab][HOM_AFF];
  modafreq[0] = affected_freq;
  modpos[0] = maxlodpos;
  mod[0] = maxlod;
  best_model = 0;

  modcount++;  /*** should be '1' now ***/

  /*** now start maximization ***/
  /*** initially, check LOD for user-defined models (if there are any) ***/

  for (i=0; i<userdefcount; i++) {
    penetrance[modliab][HOM_UNAFF] = userdefpen1[i];
    penetrance[modliab][HET_AU]    = userdefpen2[i];
    penetrance[modliab][HET_UA]    = userdefpen3[i];
    penetrance[modliab][HOM_AFF]   = userdefpen4[i];
    affected_freq = userdefafreq[i];
    if (penetrance_restriction_violated()
	|| allfreq_restriction_violated() || all_penetrances_equal()
	|| model_already_checked()) continue;
    modscore_model_incompatible = calculate_lod();
    if (modcount >= MAX_MODELS) error_too_many_models();
    modpen1[modcount] = penetrance[modliab][HOM_UNAFF];
    modpen2[modcount] = penetrance[modliab][HET_AU];
    modpen3[modcount] = penetrance[modliab][HET_UA];
    modpen4[modcount] = penetrance[modliab][HOM_AFF];
    modafreq[modcount] = affected_freq;
    if (modscore_model_incompatible) {
      modpos[modcount] = -1;
      mod[modcount] = -10000.0;
    } else {
      modpos[modcount] = maxlodpos;
      mod[modcount] = maxlod;
    }
    if ((!modscore_model_incompatible) && (mod[modcount] > mod[best_model])) {
      best_model = modcount;
      sf(ps,"   new max MOD\n"); 
      if (!pmod_calc_flag)
        pr();
      else if (replicates_store_result && pmod_calc_flag)
        fpr(replifp);
    } else {
      if (!pmod_calc_flag)
        nl();
      else if (replicates_store_result && pmod_calc_flag)
        fnl(replifp);
    }
    modcount++;
  }
  real_userdefcount = modcount;

  /*** then look at fixed grid of predefined models,
       only do this if maximization option is != 'user',
       with 'dense' and 'grid' option, 'standard' grid is
       computed first, new (symmetric) grid afterwards.
       With option 'dense', fine maximization will then be
       done with both best models (standard and new grid)
       in order to make sure that best score with 'dense'
       is always at least equal to 'standard' option score ***/

  if ((maxim_type == 1) || (maxim_type == 2) || 
      (maxim_type == 4) || (maxim_type == 5)) {
    asymmetric_grid_calc();
    a_gridbest = best_model;
    if ((maxim_type == 2) || (maxim_type == 4)) {
      symmetric_grid_calc();
      s_gridbest = best_model;
    } else {
      s_gridbest = a_gridbest;
    }
  }


  if ((mod[best_model] < 0.0) || (num_liability_classes > 1)) {
    /* add a model w/all penetrances equal; in case of only one
       liability class, this will lead to a MOD score of zero */
    penetrance[modliab][HOM_UNAFF] = 0.01;
    penetrance[modliab][HET_AU]    = 0.01;
    penetrance[modliab][HET_UA]    = 0.01;
    penetrance[modliab][HOM_AFF]   = 0.01;
    affected_freq = 0.01;
    modscore_model_incompatible = calculate_lod();
    if (modcount >= MAX_MODELS) error_too_many_models();
    modpen1[modcount] = penetrance[modliab][HOM_UNAFF];
    modpen2[modcount] = penetrance[modliab][HET_AU];
    modpen3[modcount] = penetrance[modliab][HET_UA];
    modpen4[modcount] = penetrance[modliab][HOM_AFF];
    modafreq[modcount] = affected_freq;
    if (modscore_model_incompatible) {
      modpos[modcount] = -1;
      mod[modcount] = -10000.0;
    } else {
      modpos[modcount] = maxlodpos;
      mod[modcount] = maxlod; /* will be zero with only one liab.class */
    }
    if ((!modscore_model_incompatible) &&
        (mod[modcount] > mod[best_model])) {
      a_gridbest = s_gridbest = best_model = modcount;
      sf(ps,"   new max MOD\n");
      if (!pmod_calc_flag)
        pr();
      else if (replicates_store_result && pmod_calc_flag)
        fpr(replifp);
    } else {
      if (!pmod_calc_flag)
        nl();
      else if (replicates_store_result && pmod_calc_flag)
        fnl(replifp);
    }
    modcount++;
  }

  if (initial_count == 0) {
    /* This is only true for pos==first_mod_pos and if we are doing
       the first modscore round in case of more than one liability
       class (i.e., no maximization has been done yet for another
       liability class; thus, the penetrances of all other liability
       classes do not differ between genetic positions).  In this
       case, we also save the model calculated above with all
       penetrances equal for the currently optimized liability
       class, and therefore, we count it in initial_count. -
       In all other cases (either only one liability class or more than
       one liability class but modscore done before), the variable
       initial_count has already been set to a nonzero value before. */
    initial_count = modcount;
    if (all_penetrances_all_liab_equal()) initial_count--;
    /* In case that the datafile penetrances are equal for all other
       liability classes as well, the model calculated above with all
       penetrances equal for the currently optimized liability class
       was not saved, and so it is not counted here. */
  }

  predefined_count = modcount - real_userdefcount;
  
  if ((maxim_type == 1) || (maxim_type == 2)) {
    sf(ps, "\n%d models tested initially (%d user-defined and %d predefined).\n",
       modcount, real_userdefcount, predefined_count); 
    if (!pmod_calc_flag)
       pr();
    else if (replicates_store_result && pmod_calc_flag)
       fpr(replifp);
    sf(ps,"Now starting fine maximization.\n\n");
    if (!pmod_calc_flag)
       pr();
    else if (replicates_store_result && pmod_calc_flag)
       fpr(replifp);
  }

  /*** now take maximum of these models and initialize 
       penetrances and affected_freq ***/

  penetrance[modliab][HOM_UNAFF] = modpen1[best_model];
  penetrance[modliab][HET_AU]    = modpen2[best_model];
  penetrance[modliab][HET_UA]    = modpen3[best_model];
  penetrance[modliab][HOM_AFF]   = modpen4[best_model];
  affected_freq = modafreq[best_model];

  /*** start fine maximization with options 'standard' and 
       'dense'; if 'dense' maximization is desired and if best
       models after asymmetric_grid_calc and symmetric_grid_calc
       differ, do fine maximization with both best models in order
       to make sure that 'dense' result is always at least equal
       to 'standard' result ***/

  if ((maxim_type == 1) || (maxim_type == 2)) {
    best_model = s_gridbest;
    fine_maximization(); /*** for best model after symmetric_grid_calc ***/
    if ((maxim_type ==2) && (s_gridbest != a_gridbest)) {
      s_finebest = best_model;
      penetrance[modliab][HOM_UNAFF] = modpen1[a_gridbest];
      penetrance[modliab][HET_AU]    = modpen2[a_gridbest];
      penetrance[modliab][HET_UA]    = modpen3[a_gridbest];
      penetrance[modliab][HOM_AFF]   = modpen4[a_gridbest];
      affected_freq = modafreq[a_gridbest];
      best_model = a_gridbest;
      fine_maximization(); /*** for best model after asymmetric_grid_calc ***/
      if (mod[s_finebest] > mod[best_model])
        best_model = s_finebest;
    }
  }

  /*** calculate again the LOD under the best model in order to
       fill the total_lod and sub_lods arrays so that the user
       can execute the total stat (het) command afterwards,
       total_lod array is also needed for replicates (pmod.c) ***/

  penetrance[modliab][HOM_UNAFF] = modpen1[best_model];
  penetrance[modliab][HET_AU]    = modpen2[best_model];
  penetrance[modliab][HET_UA]    = modpen3[best_model];
  penetrance[modliab][HOM_AFF]   = modpen4[best_model];
  affected_freq = modafreq[best_model];

  if (!pmod_calc_flag)  
    nl();
  else if (replicates_store_result)
    fnl(replifp);
  best_model_flag = TRUE;
  calculate_lod();
  best_model_flag = FALSE;
  if (!pmod_calc_flag) print("   Best model\n");
  else if (replicates_store_result) fprint(replifp, "   Best model\n");
  
  if (display_scores && !pmod_calc_flag) { 
    print("\nResults by pedigree and disease-locus position...\n");
    for (famnum=0; famnum<num_pedigrees; famnum++) {
      sf(ps,"\nScores for pedigree %s:\n", pedigree_names[famnum]); pr();
      if (old_analysis_type == NPL_ANALYSIS)
	error("inconsistency in MOD routine (old_analysis_type=NPL_ANALYSIS)");
      else if (old_analysis_type == LOD_ANALYSIS) {
        if (ind_sp==1) print("pos.(avg. female  male)   LOD score    information\n");
        if (ind_sp==0) print("position  LOD score    information\n");
      } else { /* BOTH */        
        if (ind_sp==1) print("pos.(avg. female  male)   LOD score    NPL score p-value    information\n");
        if (ind_sp==0) print("position  LOD score    NPL score  p-value    information\n");
      }  

      if (ind_sp==1){
        for (i=0; i<num_w_pos; i++) {
	  if (old_analysis_type == LOD_ANALYSIS) {
	    sf(ps,"%7.2lf %7.2lf %7.2lf %10.6lf    %9.6lf\n",
	       0.5*(wpos_m[i]+wpos_f[i]), wpos_f[i], wpos_m[i],
	       sub_lods[famnum][i], pvec_entropy[famnum][i]);
	    pr();
	  } else { /* BOTH */
	    sf(ps,"%7.2lf %7.2lf %7.2lf %10.6lf    %9.6lf %9.6lf   %9.6lf\n",
	       0.5*(wpos_m[i]+wpos_f[i]), wpos_f[i], wpos_m[i],
	       sub_lods[famnum][i],  apm_stat[famnum][i],
	       p_value[famnum][i], pvec_entropy[famnum][i]);
	    pr();
	  }
        }
      } else if (ind_sp==0) {
        for (i=0; i<num_w_pos; i++) {
	  if (old_analysis_type == LOD_ANALYSIS) {
	    sf(ps,"%7.2lf  %10.6lf    %9.6lf\n",
	       wpos[i] , sub_lods[famnum][i], pvec_entropy[famnum][i]);
	    pr();
	  } else { /* BOTH */
	    sf(ps,"%7.2lf %10.6lf    %9.6lf %9.6lf   %9.6lf\n",
	       wpos[i], sub_lods[famnum][i],  apm_stat[famnum][i],
	       p_value[famnum][i], pvec_entropy[famnum][i]);
	    pr();
	  }
	}
      }	      
    } /* for famnum */

    if (ind_sp==1) {
      sf(pos_str,"%.2lf %.2lf %.2lf",
	 0.5*(wpos_m[modpos[best_model]] + wpos_f[modpos[best_model]]),
	 wpos_f[modpos[best_model]], wpos_m[modpos[best_model]]);
    } else if (ind_sp==0) {
      sf(pos_str,"%.2lf", wpos[modpos[best_model]]);
    }

    if (mod[best_model] < -9000.0) strcpy(lod_str,"-INFINITY");
    else sf(lod_str,"%.6lf", mod[best_model]);
    if (ind_sp==0) {
      sf(ps,"\nMOD score for the complete dataset:  %s  (obtained at %s cM)\n",
	 lod_str, pos_str);
      pr();
    } else if (ind_sp==1) {
      sf(ps,"\nMOD score for the complete dataset:  %s  \n", lod_str);
      pr();
      sf(ps,"obtained at position (avg., female, male):  %s cM\n", pos_str);
      pr();
    } 

    print("\nBest-fitting model:\n");
    if (imprinting)
      sf(ps,"p = %8.6lf   (f+/+, fm/+, f+/m, fm/m) = (%6.4lf, %6.4lf, %6.4lf, %6.4lf)\n",
	 affected_freq,
	 penetrance[modliab][HOM_UNAFF], penetrance[modliab][HET_AU],
	 penetrance[modliab][HET_UA], penetrance[modliab][HOM_AFF]);
    else
      sf(ps,"p = %8.6lf   (f+/+, fHet, fm/m) = (%6.4lf, %6.4lf, %6.4lf)\n",
	 affected_freq,
	 penetrance[modliab][HOM_UNAFF], penetrance[modliab][HET_AU],
	 penetrance[modliab][HOM_AFF]);
    pr();
  } /* if (display_scores && !pmod_calc_flag) */

  if (real_eq(modpen1[best_model], modpen4[best_model]))
    one_dominance_index = 0.0;
  else
    one_dominance_index =
      (modpen2[best_model] + modpen3[best_model]
       - modpen1[best_model] - modpen4[best_model]) /
      (modpen4[best_model] - modpen1[best_model]);
  sf(ps, "\nDominance index (D): %5.2lf", one_dominance_index);
  if (!pmod_calc_flag)
    pr();
  else if (replicates_store_result && pmod_calc_flag)
    fpr(replifp);
  if (imprinting) {
    if (real_eq(modpen1[best_model], modpen4[best_model]))
      one_imprinting_index = 0.0;
    else
      one_imprinting_index =
      (modpen2[best_model] - modpen3[best_model]) /
      (modpen4[best_model] - modpen1[best_model]);
    sf(ps, "     Imprinting index (I): %5.2lf\n", one_imprinting_index);
    if (!pmod_calc_flag)
      pr();
    else if (replicates_store_result && pmod_calc_flag)
      fpr(replifp);
  } else {
    if (!pmod_calc_flag)
      nl();
    else if (replicates_store_result && pmod_calc_flag)
      fnl(replifp);
  }
  
  sf(ps,"\nTotal no. of models tested: %d  (%d user-def'd, %d predef'd, %d fine-max.)\n",
     modcount, real_userdefcount, predefined_count,
     modcount-real_userdefcount-predefined_count);
  if (!pmod_calc_flag)
    pr();
  else if (replicates_store_result && pmod_calc_flag)
    fpr(replifp);

  if (!pmod_calc_flag) {
    print("\nPlease use the 'total stat (het)' command to display the total\n");
    print("LOD scores under the best-fitting model for all genetic positions.\n");
    if (!postscript_output)
      print("In order to obtain a plot as well, 'ps' must be turned 'on'.\n");
  }
}


/************************* modcalc single *************************/

void calc_single_mod (void)
{
  int i, real_userdefcount;
  int *best_pvector=NULL;
  unsigned long int mod_memcount=0;
  double **single_results;
  double *num_par_score=NULL, *score=NULL;
  char *filename, **ylabel;
  FILE *fp;

  sf(ps,"\nMOD-score calculation with option 'single'.\n");
  if (!pmod_calc_flag)
    pr();
  else if (replicates_store_result && pmod_calc_flag)
    fpr(replifp);
  if ((lowpos<-990.0) && (highpos>990.0)) 
    sf(ps,"Maximization will be done separately for each genetic position,\n");
  else 
    sf(ps,"Maximization will be done separately for each pos between %.2lf and %.2lf cM,\n", lowpos, highpos);
  if (!pmod_calc_flag)
    pr();
  else if (replicates_store_result && pmod_calc_flag)
    fpr(replifp); 
  sf(ps,"by varying penetrances of liability class %d and disease allele frequency.\n", modliab);
  if (!pmod_calc_flag)
    pr();
  else if (replicates_store_result && pmod_calc_flag)
    fpr(replifp); 
  if (dimensions == 1)
    sf(ps,"Only 1 parameter is changed at a time for fine maximization.\n");
  else
    sf(ps,"Up to %d parameters are jointly changed for fine maximization.\n", dimensions);
  if (!pmod_calc_flag)
    pr();
  else if (replicates_store_result && pmod_calc_flag)
    fpr(replifp); 

  sf(ps,"Imprinting mode is '%s'.\n", imprinting?"on":"off");
  if (!pmod_calc_flag)
    pr();
  else if (replicates_store_result && pmod_calc_flag)
    fpr(replifp); 

  sf(ps,"Penetrance restriction \"f+/+ <= (fm/+,f+/m) <= fm/m\" is '%s'.\n",
     penetrance_restriction?"on":"off");
  if (!pmod_calc_flag)
    pr();
  else if (replicates_store_result && pmod_calc_flag)
    fpr(replifp);
  
  if (maxim_type == 1) {
    sf(ps,"Maximization will be performed with the standard set of predefined\ntrait models, followed by a fine maximization.\n");
  } else if (maxim_type == 2) {
    sf(ps,"Maximization will be performed with the larger set of predefined\ntrait models, followed by a fine maximization.\n");
  } else if (maxim_type == 4) {
    sf(ps,"Maximization will be done only with the larger set of predefined\ntrait models, without a fine maximization.\n");
  } else if (maxim_type == 5) {
    sf(ps,"Maximization will be done only with the standard set of predefined\ntrait models, without a fine maximization.\n");
  } else if (maxim_type == 6) {
    sf(ps,"Maximization will be done only with the set of user-defined trait models\n(those entered by the 'model' command), without a fine maximization.\n");
  }
  if (!pmod_calc_flag)
    pr();
  else if (replicates_store_result && pmod_calc_flag)
    fpr(replifp); 
  
  if (allfreq_restriction)
    sf(ps,"Disease allele frequency restriction \"p <= %.2lf\" is 'on'.\n",
       highest_allfreq);
  else
    sf(ps,"Disease allele frequency restriction is 'off'.\n");
  if (!pmod_calc_flag)
    pr();
  else if (replicates_store_result && pmod_calc_flag)
    fpr(replifp);

  /* giving the numbers of saved_models == -1 now in modscore() */

  if (saved_models == 1)
    sf(ps,"1 model is saved in memory for MOD calculation.\n");
  else
    sf(ps,"%d models are saved in memory for MOD calculation.\n",
       saved_models);
  if (!pmod_calc_flag)
    pr();
  else if (replicates_store_result && pmod_calc_flag)
    fpr(replifp); 

  if (saved_models) {
    for (famnum=0; famnum<num_pedigrees; famnum++)
      mod_memcount += save_real_num_vecs[famnum_pointer[famnum]] * sizeof(double);
    mod_memcount *= saved_models;
    sf(ps,"With the current dataset, this will need %lu bytes of memory.\n",
       mod_memcount);
    if (!pmod_calc_flag)
      pr();
    else if (replicates_store_result && pmod_calc_flag)
      fpr(replifp); 
  }
  sf(ps,"Long output with MOD-score analysis is '%s'.\n",
     long_mod_output?"on":"off");
  if (!pmod_calc_flag)
    pr();
  else if (replicates_store_result && pmod_calc_flag)
    fpr(replifp); 

  sf(ps,"\nChecking models, separately for each position...\n");
  if (!pmod_calc_flag)
    pr();
  else if (replicates_store_result && pmod_calc_flag)
    fpr(replifp); 

  if (ind_sp==0 && !long_mod_output) {
    if (imprinting)
      sf(ps,"\nposition     p     ( f+/+ ,  fm/+ ,  f+/m ,  fm/m )   D     I        MOD\n");
    else
      sf(ps,"\nposition     p     ( f+/+ ,  fHet ,  fm/m )   D        MOD\n");
    if (!pmod_calc_flag)
      pr();
    else if (replicates_store_result && pmod_calc_flag)
      fpr(replifp); 
  } else if (ind_sp==1 && !long_mod_output) {
    if (imprinting)
      sf(ps,"\npos.(avg. female  male)      p     ( f+/+ ,  fm/+ ,  f+/m ,  fm/m )   D     I        MOD\n");
    else
      sf(ps,"\npos.(avg. female  male)      p     ( f+/+ ,  fHet ,  fm/m )   D        MOD\n");
    if (!pmod_calc_flag)
      pr();
    else if (replicates_store_result && pmod_calc_flag)
      fpr(replifp); 
  }

  if (!modscore_done) {
    /* allocate storage first time --------------------------------- *
     * allocate storage for penetrances of all liability classes and *
     * disease allele frequency for each position so that several    *
     * rounds of MOD-score analysis can be done for different        *
     * liability classes, with different models for each position    */
    matrix(pospen1, num_mod_pos, num_liability_classes, double);
    matrix(pospen2, num_mod_pos, num_liability_classes, double);
    matrix(pospen3, num_mod_pos, num_liability_classes, double);
    matrix(pospen4, num_mod_pos, num_liability_classes, double);
    array(posafreq, num_mod_pos, double);
    for (pos=0; pos<num_mod_pos; pos++) {
      for (i=0; i<num_liability_classes; i++) {
	/* WARNING: index of penetrance array for the liability class    *
	 * ranges from 1 to n, and NOT from 0 to n-1.  In fact, the      *
	 * function init_disease_info() in locprep.c writes the average  *
	 * over all liability classes from 1 to n into penetrance[0] -   *
	 * which is never used.  Therefore, we do not need element 0     *
	 * for the pospen matrices, and accordingly, the index ranges    *
	 * from 0 to n-1 for pospen1..4.  The appropriate way to address *
	 * the matrix for a given liability class to be optimized is     *
	 * therefore pospen..[modliab-1][..].                            */
	pospen1[pos][i] = penetrance[i+1][HOM_UNAFF];
	pospen2[pos][i] = penetrance[i+1][HET_AU];
	pospen3[pos][i] = penetrance[i+1][HET_UA];
	pospen4[pos][i] = penetrance[i+1][HOM_AFF];
      }
      posafreq[pos] = affected_freq;
    }
    /* matrix and array saved_model_pen and afreq now in modscore()
       since models will also be stored for p-value calculation of
       modcalc global MOD scores */
  } else if (modscore_done && pmod_calc_flag) {
    /* PMOD: for each replicate new allocation --------------------- */
    /* with PMOD mod-score analysis of only ONE liab class possible  */
    unmatrix(pospen1, num_mod_pos, double);
    unmatrix(pospen2, num_mod_pos, double);
    unmatrix(pospen3, num_mod_pos, double);
    unmatrix(pospen4, num_mod_pos, double);
    unarray(posafreq, double);

    matrix(pospen1, num_mod_pos, num_liability_classes, double);
    matrix(pospen2, num_mod_pos, num_liability_classes, double);
    matrix(pospen3, num_mod_pos, num_liability_classes, double);
    matrix(pospen4, num_mod_pos, num_liability_classes, double);
    array(posafreq, num_mod_pos, double);
    for (pos=0; pos<num_mod_pos; pos++) {
      for (i=0; i<num_liability_classes; i++) {
	/* WARNING: index of penetrance array for the liability class    *
	 * ranges from 1 to n, and NOT from 0 to n-1.  In fact, the      *
	 * function init_disease_info() in locprep.c writes the average  *
	 * over all liability classes from 1 to n into penetrance[0] -   *
	 * which is never used.  Therefore, we do not need element 0     *
	 * for the pospen matrices, and accordingly, the index ranges    *
	 * from 0 to n-1 for pospen1..4.  The appropriate way to address *
	 * the matrix for a given liability class to be optimized is     *
	 * therefore pospen..[modliab-1][..].                            */
	    pospen1[pos][i] = penetrance[i+1][HOM_UNAFF];
	    pospen2[pos][i] = penetrance[i+1][HET_AU];
	    pospen3[pos][i] = penetrance[i+1][HET_UA];
	    pospen4[pos][i] = penetrance[i+1][HOM_AFF];
      }
      posafreq[pos] = affected_freq;
    }
  }

  /*** single_results[0] contains the MOD for each position (later),
       single_results[1] contains the dominance index (D) for each pos.,
       single_results[2] contains the imprinting index (I) for each pos. ***/
  if (imprinting) {
    matrix (single_results, 3, num_w_pos, double);
    matrix (ylabel, 3, 20, char);
  }
  else {
    matrix (single_results, 2, num_w_pos, double);
    matrix (ylabel, 2, 20, char);
  }

  /*** now delete family LODs, total LODs, dominance/imprinting index for
       pos. not considered (get rid of possibly remaining junk in arrays) ***/
  for (pos=0; pos<first_mod_pos; pos++) {
    for (famnum=0; famnum<num_pedigrees; famnum++) sub_lods[famnum][pos] = 0.0;
    total_lod[pos] = 0.0;
    single_results[1][pos] = 0.0;  /* dominance index (D) */
    if (imprinting) single_results[2][pos] = 0.0;  /* imprinting index (I) */
  }
  for (pos=last_mod_pos+1; pos<num_w_pos; pos++) {
    for (famnum=0; famnum<num_pedigrees; famnum++) sub_lods[famnum][pos] = 0.0;
    total_lod[pos] = 0.0;
    single_results[1][pos] = 0.0;  /* dominance index (D) */
    if (imprinting) single_results[2][pos] = 0.0;  /* imprinting index (I) */
  }

  grand_total_models = 0;  total_quick_models = 0;

  /* now in modscore():
  current_saved_model = 0;
  initial_count = 0;
  initial_fill_storage = FALSE;
  model_storage_full = FALSE;   */

  if ((num_liability_classes > 1) && (modscore_done)) initial_count = 1;
  /* In case that we have more than one liability class, it only makes
     sense to fix the initial models in storage if this is the first
     modscore round.  If modscore has already been executed for another
     liability class, the penetrances of that liability class will most
     likely not be the same for different genetic positions.  Therefore,
     in that case, even the initial models will differ, and accordingly
     they should not be stored in memory with priority.  With the above
     statement, they will be overwritten when the model storage is full,
     in the same way as the fine-maximization models.
     'initial_count = 1' because 1 will be subtracted in calculate_lod
     to otherwise account for the datafile model, and further, with this
     definition, the condition 'initial_count' (!= 0) is also satisfied. */

  for (pos=first_mod_pos; pos<=last_mod_pos; pos++) { /* overall pos. loop */

    /* do not calculate this position when seq_sim is on and threshold reached */
    if (pmod_calc_flag && ((sequential_sim != 0) && (seq_sim_single[pos] == 1)))
      continue;

    if (ind_sp==1){

      sf(pos_str,"%7.2lf %7.2lf %7.2lf", 0.5*(wpos_m[pos] + wpos_f[pos]), wpos_f[pos], wpos_m[pos]);
      if (long_mod_output) {
        sf(ps,"\nPosition: %s cM\n\n", pos_str);
        if (!pmod_calc_flag)
          pr();
        else if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
        if (imprinting)
          sf(ps,"    p     ( f+/+ ,  fm/+ ,  f+/m ,  fm/m )  pos.(avg. female  male)     MOD\n");
        else
          sf(ps,"    p     ( f+/+ ,  fHet ,  fm/m )  pos.(avg. female  male)     MOD\n");
        if (!pmod_calc_flag)
          pr();
        else if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      } else {
        sf(ps,"%s ", pos_str);
        if (!pmod_calc_flag) {
          pr();
          flush();
	} else if (replicates_store_result && pmod_calc_flag) {
          fpr(replifp);
          fflush(replifp);
	}
      }

    } else if (ind_sp==0) {
      sf(pos_str,"%7.2lf", wpos[pos]);
      if (long_mod_output) {
        sf(ps,"\nPosition: %s cM\n\n", pos_str);
        if (!pmod_calc_flag)
          pr();
        else if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
        if (imprinting)
	  sf(ps,"    p     ( f+/+ ,  fm/+ ,  f+/m ,  fm/m )  position     MOD\n");
        else
	  sf(ps,"    p     ( f+/+ ,  fHet ,  fm/m )  position     MOD\n");
        if (!pmod_calc_flag)
          pr();
        else if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      } else {
        sf(ps,"%s ", pos_str);
        if (!pmod_calc_flag) {
          pr();
          flush();
	} else if (replicates_store_result && pmod_calc_flag) {
          fpr(replifp);
          fflush(replifp);
	}
      }
    }


    /* first initialize save_p_one_pos for all families */
    for (famnum=0; famnum<num_pedigrees; famnum++) {
      if(!algebra)
	    array(num_par_score, save_real_num_vecs[famnum_pointer[famnum]], double); /* dummy, not used */
      array(score, save_real_num_vecs[famnum_pointer[famnum]], double);  /* dummy, not used */
      array(best_pvector, num_in_map_order+1, int);      /* dummy, not used */
      scan_pedigree(save_nz_bits[famnum], save_nz_bits_m[famnum], save_nz_bits_f[famnum], save_nz_pointer[famnum],
		    save_pvct[famnum], num_par_score,
		    save_f_num_in_map_order[famnum],
		    save_f_uninformative[famnum], save_f_map_order[famnum],
		    save_real_num_bits[famnum], save_real_num_bits_m[famnum], save_real_num_bits_f[famnum],
		    score, 0, 0.0, 0.0,
		    famnum, best_pvector,
		    save_num_bits[famnum], save_num_bits_m[famnum], save_num_bits_f[famnum], save_uninformative_mask[famnum],
		    save_flip_mask[famnum], save_maskTest[famnum], pos);
		    	    

      unarray(best_pvector, int);
      unarray(score, double);
      if(!algebra)
	      unarray(num_par_score, double); 
    }  /* for famnum */

    /*** initialize disease allele frequency and penetrances -
	 do this for all liability classes in case that a MOD-score
	 analysis has already been done for another liability class ***/

    for (i=0; i<num_liability_classes; i++) {
      penetrance[i+1][HOM_UNAFF] = pospen1[pos-first_mod_pos][i];
      penetrance[i+1][HET_AU] = pospen2[pos-first_mod_pos][i];
      penetrance[i+1][HET_UA] = pospen3[pos-first_mod_pos][i];
      penetrance[i+1][HOM_AFF] = pospen4[pos-first_mod_pos][i];
    }
    affected_freq = posafreq[pos-first_mod_pos];

    modcount = 0;

    /*** get LOD of the base model done with scan pedigrees
	 - or, if a modscore round has already been performed for
	 another liability class, of the best model obtained previously;
	 variable is called 'maxlod' to make it consistent with the
	 modcalc global option, although we're just looking at the
	 LOD at a single position here with the modcalc single option  ***/

    maxlod = total_lod[pos];

    if (long_mod_output) {
      if (maxlod < -9000.0) strcpy(lod_str," -INFINITY");
      else sf(lod_str,"%10.6lf", maxlod);
      if (imprinting)
	sf(ps,
	   "%8.6lf  (%6.4lf, %6.4lf, %6.4lf, %6.4lf)  %s %s   initial MOD\n",
	   affected_freq,
	   penetrance[modliab][HOM_UNAFF], penetrance[modliab][HET_AU],
	   penetrance[modliab][HET_UA], penetrance[modliab][HOM_AFF],
	   pos_str, lod_str);
      else
	sf(ps,"%8.6lf  (%6.4lf, %6.4lf, %6.4lf)  %s %s   initial MOD\n",
	   affected_freq,
	   penetrance[modliab][HOM_UNAFF], penetrance[modliab][HET_AU],
	   penetrance[modliab][HOM_AFF], pos_str, lod_str);
      if (!pmod_calc_flag)
        pr();
      else if (replicates_store_result && pmod_calc_flag)
        fpr(replifp);
    }
    modpen1[0] = penetrance[modliab][HOM_UNAFF];
    modpen2[0] = penetrance[modliab][HET_AU];
    modpen3[0] = penetrance[modliab][HET_UA];
    modpen4[0] = penetrance[modliab][HOM_AFF];
    modafreq[0] = affected_freq;
    modpos[0] = pos; /* for completeness; not needed in modcalc single mode */
    mod[0] = maxlod;
    best_model = 0;

    modcount++;  /*** should be '1' now ***/

    /*** now start maximization ***/
    /*** initially, check user-defined models (if there are any) ***/

    for (i=0; i<userdefcount; i++) {
      penetrance[modliab][HOM_UNAFF] = userdefpen1[i];
      penetrance[modliab][HET_AU]    = userdefpen2[i];
      penetrance[modliab][HET_UA]    = userdefpen3[i];
      penetrance[modliab][HOM_AFF]   = userdefpen4[i];
      affected_freq = userdefafreq[i];
      if (penetrance_restriction_violated()
	  || allfreq_restriction_violated() || all_penetrances_equal()
	  || model_already_checked()) continue;
      modscore_model_incompatible = calculate_lod();
      if (modcount >= MAX_MODELS) error_too_many_models();
      modpen1[modcount] = penetrance[modliab][HOM_UNAFF];
      modpen2[modcount] = penetrance[modliab][HET_AU];
      modpen3[modcount] = penetrance[modliab][HET_UA];
      modpen4[modcount] = penetrance[modliab][HOM_AFF];
      modafreq[modcount] = affected_freq;
      if (modscore_model_incompatible) {
	modpos[modcount] = -1;
	mod[modcount] = -10000.0;
      } else {
	modpos[modcount] = maxlodpos; /* for completeness; not needed (=pos) */
	mod[modcount] = maxlod;
      }
      if ((!modscore_model_incompatible) &&
	  (mod[modcount] > mod[best_model])) {
	best_model = modcount;
	if (long_mod_output) { 
	  sf(ps, "   new max MOD\n");
	  if (!pmod_calc_flag)
	    pr();
	  else if (replicates_store_result && pmod_calc_flag)
	    fpr(replifp);
	}
      } else if (long_mod_output) {
        if (!pmod_calc_flag)
          nl();
        else if (replicates_store_result && pmod_calc_flag)
          fnl(replifp);
      }
      modcount++;
    }
    real_userdefcount = modcount;

    /*** then look at fixed grid of predefined models,
	 only do this if maximization option is != 'user',
	 with 'dense' and 'grid' option, 'standard' grid is
	 computed first, new (symmetric) grid afterwards.
	 With option 'dense', fine maximization will then be
	 done with both best models (standard and new grid)
	 in order to make sure that best score with 'dense'
	 is always at least equal to 'standard' option score ***/

    if ((maxim_type == 1) || (maxim_type == 2) || 
        (maxim_type == 4) || (maxim_type == 5)) { 
      asymmetric_grid_calc();
      a_gridbest = best_model;
      if ((maxim_type == 2) || (maxim_type == 4)) { 
        symmetric_grid_calc();
        s_gridbest = best_model;
      } else {
        s_gridbest = a_gridbest;
      }
    }

    if ((initial_count == 0) && (num_liability_classes == 1))
      initial_count = modcount;
      /* initial_count equals zero only for pos==first_mod_pos,
	 but not if modscore has been executed before in case that
	 there is more than one liability class -
	 then, initial_count has been set to 1 just before the pos loop. -
	 With only one liability class, even if the best mod obtained so
	 far is < 0 and thus a trivial model with all penetrances equal
	 is calculated below, this trivial model will not be saved in
	 calculate_lod.  Therefore, we don't count it in initial_count. */
    if ((mod[best_model] < 0.0) || (num_liability_classes > 1)) {
      /* add a model w/all penetrances equal; in case of only one
	 liability class, this will lead to a MOD score of zero */
      penetrance[modliab][HOM_UNAFF] = 0.01;
      penetrance[modliab][HET_AU]    = 0.01;
      penetrance[modliab][HET_UA]    = 0.01;
      penetrance[modliab][HOM_AFF]   = 0.01;
      affected_freq = 0.01;
      modscore_model_incompatible = calculate_lod();
      if (modcount >= MAX_MODELS) error_too_many_models();
      modpen1[modcount] = penetrance[modliab][HOM_UNAFF];
      modpen2[modcount] = penetrance[modliab][HET_AU];
      modpen3[modcount] = penetrance[modliab][HET_UA];
      modpen4[modcount] = penetrance[modliab][HOM_AFF];
      modafreq[modcount] = affected_freq;
      if (modscore_model_incompatible) {
	modpos[modcount] = -1;
	mod[modcount] = -10000.0;
      } else {
	modpos[modcount] = maxlodpos; /* for completeness; not needed */
	mod[modcount] = maxlod; /* will be zero with only one liab.class */
      }
      if ((!modscore_model_incompatible) &&
	  (mod[modcount] > mod[best_model])) {
	a_gridbest = s_gridbest = best_model = modcount;
	if (long_mod_output) {
	  sf(ps, "   new max MOD\n");
	  if (!pmod_calc_flag)
	    pr();
	  else if (replicates_store_result && pmod_calc_flag)
	    fpr(replifp);
	}
      } else if (long_mod_output) {
	if (!pmod_calc_flag)
	  nl();
	else if (replicates_store_result && pmod_calc_flag)
	  fnl(replifp);
      }
      modcount++;
    }

    if (initial_count == 0) {
      /* This is only true for pos==first_mod_pos and if we are doing
	 the first modscore round in case of more than one liability
	 class (i.e., no maximization has been done yet for another
	 liability class; thus, the penetrances of all other liability
	 classes do not differ between genetic positions).  In this
	 case, we also save the model calculated above with all
	 penetrances equal for the currently optimized liability
	 class, and therefore, we count it in initial_count. -
	 In all other cases (either only one liability class or more than
	 one liability class but modscore done before), the variable
	 initial_count has already been set to a nonzero value before. */
      initial_count = modcount;
      if (all_penetrances_all_liab_equal()) initial_count--;
      /* In case that the datafile penetrances are equal for all other
	 liability classes as well, the model calculated above with all
	 penetrances equal for the currently optimized liability class
	 was not saved, and so it is not counted here. */
    }

    predefined_count = modcount - real_userdefcount;
    if (long_mod_output && ((maxim_type ==1) || (maxim_type == 2))) {
      sf(ps,
	 "%d models tested initially (%d user-defined and %d predefined).\n",
	 modcount, real_userdefcount, predefined_count); 
      if (!pmod_calc_flag)
	pr();
      else if (replicates_store_result && pmod_calc_flag)
	fpr(replifp);
      sf(ps,"Now starting fine maximization.\n");
      if (!pmod_calc_flag)
        pr();
      else if (replicates_store_result && pmod_calc_flag)
        fpr(replifp);
    }

    /*** now take maximum of these models and initialize
	 penetrances and affected_freq ***/

    penetrance[modliab][HOM_UNAFF] = modpen1[best_model];
    penetrance[modliab][HET_AU]    = modpen2[best_model];
    penetrance[modliab][HET_UA]    = modpen3[best_model];
    penetrance[modliab][HOM_AFF]   = modpen4[best_model];
    affected_freq = modafreq[best_model];

    /*** start fine maximization with options 'standard' and 
	 'dense'; if 'dense' maximization is desired and if best
	 models after asymmetric_grid_calc and symmetric_grid_calc
	 differ, do fine maximization with both best models in order
	 to make sure that 'dense' result is always at least equal
	 to 'standard' result ***/

    if ((maxim_type == 1) || (maxim_type == 2)) { 
      best_model = s_gridbest;
      fine_maximization(); /*** for best model after symmetric grid ***/
      if ((maxim_type ==2) && (s_gridbest != a_gridbest)) {
        s_finebest = best_model;
        penetrance[modliab][HOM_UNAFF] = modpen1[a_gridbest];
        penetrance[modliab][HET_AU]    = modpen2[a_gridbest];
        penetrance[modliab][HET_UA]    = modpen3[a_gridbest];
        penetrance[modliab][HOM_AFF]   = modpen4[a_gridbest];
        affected_freq = modafreq[a_gridbest];
        best_model = a_gridbest;
        fine_maximization(); /*** for best model after asymmetric grid ***/
        if (mod[s_finebest] > mod[best_model])
          best_model = s_finebest;
      }
    }

    /*** calculate again the LOD under the best model in order to
	 fill the total_lod and sub_lods arrays (for pos only) so that
	 the scores can be printed afterwards,
	 total_lod array is also needed for replicates (pmod.c) ***/

    penetrance[modliab][HOM_UNAFF] = modpen1[best_model];
    penetrance[modliab][HET_AU]    = modpen2[best_model];
    penetrance[modliab][HET_UA]    = modpen3[best_model];
    penetrance[modliab][HOM_AFF]   = modpen4[best_model];
    affected_freq = modafreq[best_model];

    best_model_flag = TRUE;
    calculate_lod();
    best_model_flag = FALSE;
    if (long_mod_output) {
      if (!pmod_calc_flag) print("   Best model\n");
      else if (replicates_store_result) fprint(replifp, "   Best model\n");
    }

    /* dominance index (D): */
    if (real_eq(modpen1[best_model], modpen4[best_model]))
      single_results[1][pos] = 0.0;
    else
      single_results[1][pos] =  
	  (modpen2[best_model] + modpen3[best_model]
	   - modpen1[best_model] - modpen4[best_model]) /
	  (modpen4[best_model] - modpen1[best_model]);

    /* imprinting index (I): */
    if (imprinting) {
      if (real_eq(modpen1[best_model], modpen4[best_model]))
	single_results[2][pos] = 0.0;
      else
	single_results[2][pos] =  
	                (modpen2[best_model] - modpen3[best_model]) /
			(modpen4[best_model] - modpen1[best_model]);
    }

    pospen1[pos-first_mod_pos][modliab-1] = modpen1[best_model];
    pospen2[pos-first_mod_pos][modliab-1] = modpen2[best_model];
    pospen3[pos-first_mod_pos][modliab-1] = modpen3[best_model];
    pospen4[pos-first_mod_pos][modliab-1] = modpen4[best_model];
    posafreq[pos-first_mod_pos] = modafreq[best_model];

    if (long_mod_output) {
      sf(ps,"Total no. of models tested: %d  (%d user-def'd, %d predef'd, %d fine-max.)\n",
	 modcount, real_userdefcount, predefined_count,
	 modcount-real_userdefcount-predefined_count);
      if (!pmod_calc_flag)
        pr();
      else if (replicates_store_result && pmod_calc_flag)
        fpr(replifp);
    } else {  /* long_mod_output==FALSE */
      if (maxlod < -9000.0) strcpy(lod_str," -INFINITY");
      else sf(lod_str,"%10.6lf", maxlod);
      if (imprinting)
	sf(ps," %8.6lf  (%6.4lf, %6.4lf, %6.4lf, %6.4lf) %5.2lf %5.2lf %s\n",
	   affected_freq,
	   penetrance[modliab][HOM_UNAFF], penetrance[modliab][HET_AU],
	   penetrance[modliab][HET_UA], penetrance[modliab][HOM_AFF],
	   single_results[1][pos], single_results[2][pos], lod_str);
      else
	sf(ps," %8.6lf  (%6.4lf, %6.4lf, %6.4lf) %5.2lf %s\n",
	   affected_freq,
	   penetrance[modliab][HOM_UNAFF], penetrance[modliab][HET_AU],
	   penetrance[modliab][HOM_AFF], single_results[1][pos], lod_str);
      if (!pmod_calc_flag)
        pr();
      else if (replicates_store_result && pmod_calc_flag)
        fpr(replifp);
    }
    grand_total_models += modcount;

  } /* for pos */

    if (display_scores && !pmod_calc_flag) {
    print("\nResults by pedigree and disease-locus position...\n");
    for (famnum=0; famnum<num_pedigrees; famnum++) {
      sf(ps,"\nScores for pedigree %s:\n", pedigree_names[famnum]); pr();
      if (old_analysis_type == NPL_ANALYSIS) {
	error("inconsistency in MOD routine (old_analysis_type=NPL_ANALYSIS)");
      } else if (old_analysis_type == LOD_ANALYSIS) {
        if (ind_sp==1) print("pos.(avg. female  male)  MOD score    information\n");
        if (ind_sp==0) print("position  MOD score    information\n");
      } else { /* BOTH */
        if (ind_sp==1) print("pos.(avg. female  male)    MOD score    NPL score p-value     information\n");
        if (ind_sp==0) print("position  MOD score    NPL score  p-value    information\n");
      }

      for (pos=0; pos<num_w_pos; pos++) {
	 if (ind_sp==1){
	  if (old_analysis_type == LOD_ANALYSIS) {
	    sf(ps,"%7.2lf %7.2lf %7.2lf %10.6lf    %9.6lf\n",0.5*(wpos_m[pos] + wpos_f[pos]),
	       wpos_f[pos], wpos_m[pos], sub_lods[famnum][pos], pvec_entropy[famnum][pos]);
	    pr();
	  } else { /* BOTH */
	  sf(ps,"%7.2lf %7.2lf %7.2lf  %10.6lf    %9.6lf %9.6lf   %9.6lf\n",
	       0.5*(wpos_f[pos] + wpos_m[pos]), wpos_f[pos], wpos_m[pos], sub_lods[famnum][pos], apm_stat[famnum][pos],
	    p_value[famnum][pos], pvec_entropy[famnum][pos]);
	    pr();
	  }  
	    
	} else if (ind_sp==0){
	  
	    if (old_analysis_type == LOD_ANALYSIS) {
	      sf(ps,"%7.2lf  %10.6lf    %9.6lf\n",wpos[pos], sub_lods[famnum][pos], pvec_entropy[famnum][pos]);
	      pr();
	    } else { /* BOTH */
	     sf(ps,"%7.2lf  %10.6lf    %9.6lf %9.6lf   %9.6lf\n",
	        wpos[pos], sub_lods[famnum][pos], apm_stat[famnum][pos],
	      p_value[famnum][pos], pvec_entropy[famnum][pos]);
	      pr();
	    }  	 	  
	 } 
      }      
    } /* for famnum */
  } /* if (display_scores && !pmod_calc_flag) */

  /* if (replicates_store_result && modscore_done) { * skipped !pmod_calc_flag 
    sf(ps,"\nResults by pedigree and disease-locus position...\n");
    fpr(replifp);
    for (famnum=0; famnum<num_pedigrees; famnum++) {
      sf(ps,"\nScores for pedigree %s:\n", pedigree_names[famnum]);
      fpr(replifp);
      if (old_analysis_type == NPL_ANALYSIS) {
	error("inconsistency in MOD routine (old_analysis_type=NPL_ANALYSIS)");
      } else if (old_analysis_type == LOD_ANALYSIS) {
        if (ind_sp==1) sf(ps,"pos.(avg. female  male)  MOD score    information\n");
        if (ind_sp==0) sf(ps,"position  MOD score    information\n");
	fpr(replifp);
      } else { * BOTH *
        if (ind_sp==1) print("pos.(avg. female  male)    MOD score    NPL score p-value     information\n");
        if (ind_sp==0) sf(ps,"position  MOD score    NPL score  p-value    information\n");
	fpr(replifp);
      }

      for (pos=0; pos<num_w_pos; pos++) {
	 if (ind_sp==1){
	  if (old_analysis_type == LOD_ANALYSIS) {
	    sf(ps,"%7.2lf %7.2lf %7.2lf %10.6lf    %9.6lf\n",0.5*(wpos_m[pos] + wpos_f[pos]),
	       wpos_f[pos], wpos_m[pos], sub_lods[famnum][pos], pvec_entropy[famnum][pos]);
	    fpr(replifp);
	  } else { * BOTH *
	    sf(ps,"%7.2lf %7.2lf %7.2lf  %10.6lf    %9.6lf %9.6lf   %9.6lf\n",
	       0.5*(wpos_f[pos] + wpos_m[pos]), wpos_f[pos], wpos_m[pos], sub_lods[famnum][pos], apm_stat[famnum][pos],
	    p_value[famnum][pos], pvec_entropy[famnum][pos]);
	    fpr(replifp);
	  }  
	    
	} else if (ind_sp==0){
	  
	    if (old_analysis_type == LOD_ANALYSIS) {
	      sf(ps,"%7.2lf  %10.6lf    %9.6lf\n",wpos[pos], sub_lods[famnum][pos], pvec_entropy[famnum][pos]);
	      fpr(replifp);
	    } else { * BOTH *
	      sf(ps,"%7.2lf  %10.6lf    %9.6lf %9.6lf   %9.6lf\n",
	        wpos[pos], sub_lods[famnum][pos], apm_stat[famnum][pos],
	      p_value[famnum][pos], pvec_entropy[famnum][pos]);
	      fpr(replifp);
	    }  	 	  
	 } 
      }      
    } * for famnum *
  } * if (display_scores && !pmod_calc_flag) */

  if (long_mod_output || (display_scores && !pmod_calc_flag)) {
  
    if (ind_sp==1) {
  
      sf(ps,"\nSummary of the MOD-score results for the complete dataset:\n\n");
      if (!pmod_calc_flag)
        pr();
      else if (replicates_store_result && pmod_calc_flag)
        fpr(replifp);
      if (imprinting)
        sf(ps,"pos.(avg. female  male)     p      ( f+/+ ,  fm/+ ,  f+/m ,  fm/m )   D     I        MOD\n");
      else
        sf(ps,"pos.(avg. female  male)     p      ( f+/+ ,  fHet ,  fm/m )   D        MOD\n");
      if (!pmod_calc_flag)
        pr();
      else if (replicates_store_result && pmod_calc_flag)
        fpr(replifp);

      for (pos=first_mod_pos; pos<=last_mod_pos; pos++) {

        sf(pos_str,"%7.2lf %7.2lf %7.2lf",0.5*(wpos_m[pos] + wpos_f[pos]), wpos_f[pos], wpos_m[pos]);
        if (total_lod[pos] < -9000.0) strcpy(lod_str," -INFINITY");
        else sf(lod_str,"%10.6lf", total_lod[pos]);
        if (imprinting)
	  sf(ps,
	     "%s  %8.6lf  (%6.4lf, %6.4lf, %6.4lf, %6.4lf) %5.2lf %5.2lf %s\n",
	     pos_str, posafreq[pos-first_mod_pos],
	     pospen1[pos-first_mod_pos][modliab-1],
	     pospen2[pos-first_mod_pos][modliab-1],
	     pospen3[pos-first_mod_pos][modliab-1],
	     pospen4[pos-first_mod_pos][modliab-1],
	     single_results[1][pos], single_results[2][pos], lod_str);
        else
	  sf(ps,"%s  %8.6lf  (%6.4lf, %6.4lf, %6.4lf) %5.2lf %s\n",
	     pos_str, posafreq[pos-first_mod_pos],
	     pospen1[pos-first_mod_pos][modliab-1],
	     pospen2[pos-first_mod_pos][modliab-1],
	     pospen4[pos-first_mod_pos][modliab-1],
	     single_results[1][pos], lod_str);
	if (!pmod_calc_flag)
          pr();
        else if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
    
    } else if (ind_sp==0) {
    
      sf(ps,"\nSummary of the MOD-score results for the complete dataset:\n\n");
      if (!pmod_calc_flag)
        pr();
      else if (replicates_store_result && pmod_calc_flag)
        fpr(replifp);
      if (imprinting)
        sf(ps,"position     p     ( f+/+ ,  fm/+ ,  f+/m ,  fm/m )   D     I        MOD\n");
      else
        sf(ps,"position     p     ( f+/+ ,  fHet ,  fm/m )   D        MOD\n");
      if (!pmod_calc_flag)
        pr();
      else if (replicates_store_result && pmod_calc_flag)
        fpr(replifp);
    
      for (pos=first_mod_pos; pos<=last_mod_pos; pos++) {
    
        sf(pos_str,"%7.2lf",wpos[pos]);
        if (total_lod[pos] < -9000.0) strcpy(lod_str," -INFINITY");
        else sf(lod_str,"%10.6lf", total_lod[pos]);
        if (imprinting)
	  sf(ps,
	     "%s  %8.6lf  (%6.4lf, %6.4lf, %6.4lf, %6.4lf) %5.2lf %5.2lf %s\n",
	     pos_str, posafreq[pos-first_mod_pos],
	     pospen1[pos-first_mod_pos][modliab-1],
	     pospen2[pos-first_mod_pos][modliab-1],
	     pospen3[pos-first_mod_pos][modliab-1],
	     pospen4[pos-first_mod_pos][modliab-1],
	     single_results[1][pos], single_results[2][pos], lod_str);
        else
	  sf(ps,"%s  %8.6lf  (%6.4lf, %6.4lf, %6.4lf) %5.2lf %s\n",
	     pos_str, posafreq[pos-first_mod_pos],
	     pospen1[pos-first_mod_pos][modliab-1],
	     pospen2[pos-first_mod_pos][modliab-1],
	     pospen4[pos-first_mod_pos][modliab-1],
	     single_results[1][pos], lod_str);
        if (!pmod_calc_flag)
          pr();
        else if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
    }
  }
  sf(ps,"\nGrand total number of LOD evaluations for this MOD-score analysis: %d\n", grand_total_models);
  if (!pmod_calc_flag)
    pr();
  else if (replicates_store_result && pmod_calc_flag)
    fpr(replifp);

  sf(ps,"This number includes %d LOD scores yielded by the original model.\n", num_mod_pos);
  if (!pmod_calc_flag)
    pr();
  else if (replicates_store_result && pmod_calc_flag)
    fpr(replifp);

  sf(ps,"For %d evaluations, recomputation was avoided by model-vector saving.\n", total_quick_models);
  if (!pmod_calc_flag)
    pr();
  else if (replicates_store_result && pmod_calc_flag)
    fpr(replifp);

  /* give postscript-output only for real data set: */
  if (num_mod_pos > 1 && postscript_output && !pmod_calc_flag) {
  
    for (i=0; i<num_w_pos; i++) 
       wpos[i]=0.5*(wpos_m[i] + wpos_f[i]);
  
    nl();

    for (pos=0; pos<num_w_pos; pos++) single_results[0][pos] = total_lod[pos];
    strcpy (ylabel[0], "MOD score");
    strcpy (ylabel[1], "Dom. index (D)");
    if (imprinting) strcpy (ylabel[2], "Impr. index (I)");

    filename = get_temp_string();
    strcpy(filename,"modscore_results.ps");
    fp = open_out_file(filename,"postscript plot");
    
    if (imprinting) draw_mult_y_plots(fp, wpos, single_results, num_w_pos,
				      3, "Results of MOD-score analysis",
				      ylabel, "");
    else draw_mult_y_plots(fp, wpos, single_results, num_w_pos,
			   2, "Results of MOD-score analysis",
			   ylabel, "");
    close_file (fp);
  }

  if (imprinting) {
    unmatrix (single_results, 3, double); 
    unmatrix (ylabel, 3, char);
  } else {
    unmatrix (single_results, 2, double);
    unmatrix (ylabel, 2, char);
  }
}


int optimize_parameter(double *parm_pointer)
{
  int i, done, improved=0;
  double bestparm, newparm;

  bestparm = *parm_pointer;
  for (i=0; i<=1; i++) do { /* minimum two runs */
    done = 0;
    if (allele_frequency_flag) { /* change allfreq (after changing penetrances) */
      if (i) newparm = get_next_higher_allfreq(bestparm);
      else newparm = get_next_lower_allfreq(bestparm);
      if (real_eq(newparm, 0.0) || real_eq(newparm, 1.0)) break;
    } else { /* change penetrances */
      if (i) newparm = get_next_higher(bestparm);
      else newparm = get_next_lower(bestparm); /* first call choice */
    }
    if (real_eq(newparm, bestparm)) break;
    *parm_pointer = newparm;
    if (!imprinting) penetrance[modliab][HET_UA] = penetrance[modliab][HET_AU];
    if (penetrance_restriction_violated()
	|| allfreq_restriction_violated() || all_penetrances_equal()
	|| model_already_checked()) break;

    /* calculate_lod() returns '-1' if disease pvectors sum to 0.0, else '0'*/
    modscore_model_incompatible = calculate_lod();
    if (modcount >= MAX_MODELS) error_too_many_models();
    modpen1[modcount] = penetrance[modliab][HOM_UNAFF];
    modpen2[modcount] = penetrance[modliab][HET_AU];
    modpen3[modcount] = penetrance[modliab][HET_UA];
    modpen4[modcount] = penetrance[modliab][HOM_AFF];
    modafreq[modcount] = affected_freq;
    if (modscore_model_incompatible) {
      modpos[modcount] = -1;
      mod[modcount] = -10000.0;
    } else {
      modpos[modcount] = maxlodpos;
      mod[modcount] = maxlod;
    }
    if ((!modscore_model_incompatible) && (mod[modcount] > mod[best_model])) {
      /* if new max MOD and model not skipped (disease pvectors sum != 0.0) */
      bestparm = *parm_pointer;
      best_model = modcount;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"   new max MOD\n");
        if (!pmod_calc_flag)
          pr();
        else if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      improved = 1;
    } else { 
      /* if not new max MOD or model skipped (disease pvectors sum to 0.0) */
      if (modcalc==1 || long_mod_output) {
	if (!pmod_calc_flag)
          nl();
        else if (replicates_store_result && pmod_calc_flag)
          fnl(replifp);
      }
      done = 1;
    }
    modcount++;
  } while (!done);
  *parm_pointer = bestparm;
  if (!imprinting) penetrance[modliab][HET_UA] = penetrance[modliab][HET_AU];
  return(improved);
}


int optimize_two_parameters(double *parm_pointer1, double *parm_pointer2)
{
  int i, j, done, improved, overall_improved=0;
  double bestparm1, bestparm2, newparm1, newparm2;

  bestparm1 = *parm_pointer1;
  bestparm2 = *parm_pointer2;

  do {
    improved = 0;
    for (i=0; i<=1; i++) for (j=0; j<=1; j++) do {
      done = 0;
      if (allele_frequency_flag) {
	if (i) newparm1 = get_next_higher_allfreq(bestparm1);
	else newparm1 = get_next_lower_allfreq(bestparm1);
	if (real_eq(newparm1, 0.0) || real_eq(newparm1, 1.0)) break;
      } else {
	if (i) newparm1 = get_next_higher(bestparm1);
	else newparm1 = get_next_lower(bestparm1);
      }
      if (real_eq(newparm1, bestparm1)) break;
      if (j) newparm2 = get_next_higher(bestparm2);
      else newparm2 = get_next_lower(bestparm2);
      if (real_eq(newparm2, bestparm2)) break;
      *parm_pointer1 = newparm1;
      *parm_pointer2 = newparm2;
      if (!imprinting) penetrance[modliab][HET_UA] = penetrance[modliab][HET_AU];
      if (penetrance_restriction_violated()
	  || allfreq_restriction_violated() || all_penetrances_equal()
	  || model_already_checked()) break;
      modscore_model_incompatible = calculate_lod();
      if (modcount >= MAX_MODELS) error_too_many_models();
      modpen1[modcount] = penetrance[modliab][HOM_UNAFF];
      modpen2[modcount] = penetrance[modliab][HET_AU];
      modpen3[modcount] = penetrance[modliab][HET_UA];
      modpen4[modcount] = penetrance[modliab][HOM_AFF];
      modafreq[modcount] = affected_freq;
      if (modscore_model_incompatible) {
	modpos[modcount] = -1;
	mod[modcount] = -10000.0;
      } else {
	modpos[modcount] = maxlodpos;
	mod[modcount] = maxlod;
      }
      if ((!modscore_model_incompatible) &&
	  (mod[modcount] > mod[best_model])) {
	bestparm1 = *parm_pointer1;
	bestparm2 = *parm_pointer2;
	best_model = modcount;
	if (modcalc==1 || long_mod_output) {
	  sf(ps,"   new max MOD\n");
	  if (!pmod_calc_flag)
	    pr();
	  else if (replicates_store_result && pmod_calc_flag)
	    fpr(replifp);
	}
	improved = 1;
      } else {
	if (modcalc==1 || long_mod_output) {
	  if (!pmod_calc_flag)
	    nl();
          else if (replicates_store_result && pmod_calc_flag)
            fnl(replifp);
	}
    	done = 1;
      }
      modcount++;
    } while (!done);
    if (improved) overall_improved = 1;
  } while (improved);
  *parm_pointer1 = bestparm1;
  *parm_pointer2 = bestparm2;
  if (!imprinting) penetrance[modliab][HET_UA] = penetrance[modliab][HET_AU];
  return(overall_improved);
}


int optimize_three_parameters(double *parm_pointer1, double *parm_pointer2,
			      double *parm_pointer3)
{
  int i, j, k, done, improved, overall_improved=0;
  double bestparm1, bestparm2, bestparm3, newparm1, newparm2, newparm3;

  bestparm1 = *parm_pointer1;
  bestparm2 = *parm_pointer2;
  bestparm3 = *parm_pointer3;

  do {
    improved = 0;
    for (i=0; i<=1; i++) for (j=0; j<=1; j++) for (k=0; k<=1; k++) do {
      done = 0;
      if (allele_frequency_flag) {
	if (i) newparm1 = get_next_higher_allfreq(bestparm1);
	else newparm1 = get_next_lower_allfreq(bestparm1);
	if (real_eq(newparm1, 0.0) || real_eq(newparm1, 1.0)) break;
      } else {
	if (i) newparm1 = get_next_higher(bestparm1);
	else newparm1 = get_next_lower(bestparm1);
      }
      if (real_eq(newparm1, bestparm1)) break;
      if (j) newparm2 = get_next_higher(bestparm2);
      else newparm2 = get_next_lower(bestparm2);
      if (real_eq(newparm2, bestparm2)) break;
      if (k) newparm3 = get_next_higher(bestparm3);
      else newparm3 = get_next_lower(bestparm3);
      if (real_eq(newparm3, bestparm3)) break;
      *parm_pointer1 = newparm1;
      *parm_pointer2 = newparm2;
      *parm_pointer3 = newparm3;
      if (!imprinting) penetrance[modliab][HET_UA] = penetrance[modliab][HET_AU];
      if (penetrance_restriction_violated()
	  || allfreq_restriction_violated() || all_penetrances_equal()
	  || model_already_checked()) break;
      modscore_model_incompatible = calculate_lod();
      if (modcount >= MAX_MODELS) error_too_many_models();
      modpen1[modcount] = penetrance[modliab][HOM_UNAFF];
      modpen2[modcount] = penetrance[modliab][HET_AU];
      modpen3[modcount] = penetrance[modliab][HET_UA];
      modpen4[modcount] = penetrance[modliab][HOM_AFF];
      modafreq[modcount] = affected_freq;
      if (modscore_model_incompatible) {
	modpos[modcount] = -1;
	mod[modcount] = -10000.0;
      } else {
	modpos[modcount] = maxlodpos;
	mod[modcount] = maxlod;
      }
      if ((!modscore_model_incompatible) &&
	  (mod[modcount] > mod[best_model])) {
	bestparm1 = *parm_pointer1;
	bestparm2 = *parm_pointer2;
	bestparm3 = *parm_pointer3;
	best_model = modcount;
	if (modcalc==1 || long_mod_output) {
	  sf(ps, "   new max MOD\n");
	  if (!pmod_calc_flag)
	    pr();
	  else if (replicates_store_result && pmod_calc_flag)
	    fpr(replifp);
	}
	improved = 1;
      } else {
	if (modcalc==1 || long_mod_output) {
	  if (!pmod_calc_flag)
	    nl();
	  else if (replicates_store_result && pmod_calc_flag)
	    fnl(replifp);
	}
	done = 1;
      }
      modcount++;
    } while (!done);
    if (improved) overall_improved = 1;
  } while (improved);
  *parm_pointer1 = bestparm1;
  *parm_pointer2 = bestparm2;
  *parm_pointer3 = bestparm3;
  if (!imprinting) penetrance[modliab][HET_UA] = penetrance[modliab][HET_AU];
  return(overall_improved);
}


int optimize_four_parameters(double *parm_pointer1, double *parm_pointer2,
			     double *parm_pointer3, double *parm_pointer4)
{
  int i, j, k, l, done, improved, overall_improved=0;
  double bestparm1, bestparm2, bestparm3, bestparm4,
      newparm1, newparm2, newparm3, newparm4;

  bestparm1 = *parm_pointer1;
  bestparm2 = *parm_pointer2;
  bestparm3 = *parm_pointer3;
  bestparm4 = *parm_pointer4;

  do {
    improved = 0;
    for (i=0; i<=1; i++) for (j=0; j<=1; j++) for (k=0; k<=1; k++) for (l=0; l<=1; l++) do {
      done = 0;
      if (allele_frequency_flag) {
	if (i) newparm1 = get_next_higher_allfreq(bestparm1);
	else newparm1 = get_next_lower_allfreq(bestparm1);
	if (real_eq(newparm1, 0.0) || real_eq(newparm1, 1.0)) break;
      } else {
	if (i) newparm1 = get_next_higher(bestparm1);
	else newparm1 = get_next_lower(bestparm1);
      }
      if (real_eq(newparm1, bestparm1)) break;
      if (j) newparm2 = get_next_higher(bestparm2);
      else newparm2 = get_next_lower(bestparm2);
      if (real_eq(newparm2, bestparm2)) break;
      if (k) newparm3 = get_next_higher(bestparm3);
      else newparm3 = get_next_lower(bestparm3);
      if (real_eq(newparm3, bestparm3)) break;
      if (l) newparm4 = get_next_higher(bestparm4);
      else newparm4 = get_next_lower(bestparm4);
      if (real_eq(newparm4, bestparm4)) break;
      *parm_pointer1 = newparm1;
      *parm_pointer2 = newparm2;
      *parm_pointer3 = newparm3;
      *parm_pointer4 = newparm4;
      if (!imprinting) penetrance[modliab][HET_UA] = penetrance[modliab][HET_AU];
      if (penetrance_restriction_violated()
	  || allfreq_restriction_violated() || all_penetrances_equal()
	  || model_already_checked()) break;
      modscore_model_incompatible = calculate_lod();
      if (modcount >= MAX_MODELS) error_too_many_models();
      modpen1[modcount] = penetrance[modliab][HOM_UNAFF];
      modpen2[modcount] = penetrance[modliab][HET_AU];
      modpen3[modcount] = penetrance[modliab][HET_UA];
      modpen4[modcount] = penetrance[modliab][HOM_AFF];
      modafreq[modcount] = affected_freq;
      if (modscore_model_incompatible) {
	modpos[modcount] = -1;
	mod[modcount] = -10000.0;
      } else {
	modpos[modcount] = maxlodpos;
	mod[modcount] = maxlod;
      }
      if ((!modscore_model_incompatible) &&
	  (mod[modcount] > mod[best_model])) {
	bestparm1 = *parm_pointer1;
	bestparm2 = *parm_pointer2;
	bestparm3 = *parm_pointer3;
	bestparm4 = *parm_pointer4;
	best_model = modcount;
	if (modcalc==1 || long_mod_output) { 
	  sf(ps,"   new max MOD\n");
 	  if (!pmod_calc_flag)
	    pr();
	  else if (replicates_store_result && pmod_calc_flag)
	    fpr(replifp);
	}
	improved = 1;
      } else {
	if (modcalc==1 || long_mod_output) {
	  if (!pmod_calc_flag)
	    nl();
	  else if (replicates_store_result && pmod_calc_flag)
	    fnl(replifp);
	}
	done = 1;
      }
      modcount++;
    } while (!done);
    if (improved) overall_improved = 1;
  } while (improved);
  *parm_pointer1 = bestparm1;
  *parm_pointer2 = bestparm2;
  *parm_pointer3 = bestparm3;
  *parm_pointer4 = bestparm4;
  if (!imprinting) penetrance[modliab][HET_UA] = penetrance[modliab][HET_AU];
  return(overall_improved);
}


int optimize_five_parameters(double *parm_pointer1, double *parm_pointer2,
			     double *parm_pointer3, double *parm_pointer4,
			     double *parm_pointer5)
{
  int i, j, k, l, m, done, improved, overall_improved=0;
  double bestparm1, bestparm2, bestparm3, bestparm4, bestparm5,
      newparm1, newparm2, newparm3, newparm4, newparm5;

  bestparm1 = *parm_pointer1;
  bestparm2 = *parm_pointer2;
  bestparm3 = *parm_pointer3;
  bestparm4 = *parm_pointer4;
  bestparm5 = *parm_pointer5;

  do {
    improved = 0;
    for (i=0; i<=1; i++) for (j=0; j<=1; j++) for (k=0; k<=1; k++) for (l=0; l<=1; l++) for (m=0; m<=1; m++) do {
      done = 0;
      if (allele_frequency_flag) {
	if (i) newparm1 = get_next_higher_allfreq(bestparm1);
	else newparm1 = get_next_lower_allfreq(bestparm1);
	if (real_eq(newparm1, 0.0) || real_eq(newparm1, 1.0)) break;
      } else {
	if (i) newparm1 = get_next_higher(bestparm1);
	else newparm1 = get_next_lower(bestparm1);
      }
      if (real_eq(newparm1, bestparm1)) break;
      if (j) newparm2 = get_next_higher(bestparm2);
      else newparm2 = get_next_lower(bestparm2);
      if (real_eq(newparm2, bestparm2)) break;
      if (k) newparm3 = get_next_higher(bestparm3);
      else newparm3 = get_next_lower(bestparm3);
      if (real_eq(newparm3, bestparm3)) break;
      if (l) newparm4 = get_next_higher(bestparm4);
      else newparm4 = get_next_lower(bestparm4);
      if (real_eq(newparm4, bestparm4)) break;
      if (m) newparm5 = get_next_higher(bestparm5);
      else newparm5 = get_next_lower(bestparm5);
      if (real_eq(newparm5, bestparm5)) break;
      *parm_pointer1 = newparm1;
      *parm_pointer2 = newparm2;
      *parm_pointer3 = newparm3;
      *parm_pointer4 = newparm4;
      *parm_pointer5 = newparm5;
      if (!imprinting) penetrance[modliab][HET_UA] = penetrance[modliab][HET_AU];
      if (penetrance_restriction_violated()
	  || allfreq_restriction_violated() || all_penetrances_equal()
	  || model_already_checked()) break;
      modscore_model_incompatible = calculate_lod();
      if (modcount >= MAX_MODELS) error_too_many_models();
      modpen1[modcount] = penetrance[modliab][HOM_UNAFF];
      modpen2[modcount] = penetrance[modliab][HET_AU];
      modpen3[modcount] = penetrance[modliab][HET_UA];
      modpen4[modcount] = penetrance[modliab][HOM_AFF];
      modafreq[modcount] = affected_freq;
      if (modscore_model_incompatible) {
	modpos[modcount] = -1;
	mod[modcount] = -10000.0;
      } else {
	modpos[modcount] = maxlodpos;
	mod[modcount] = maxlod;
      }
      if ((!modscore_model_incompatible) &&
	  (mod[modcount] > mod[best_model])) {
	bestparm1 = *parm_pointer1;
	bestparm2 = *parm_pointer2;
	bestparm3 = *parm_pointer3;
	bestparm4 = *parm_pointer4;
	bestparm5 = *parm_pointer5;
	best_model = modcount;
	if (modcalc==1 || long_mod_output) {
	  sf(ps,"   new max MOD\n");
	  if (!pmod_calc_flag)
	    pr();
	  else if (replicates_store_result && pmod_calc_flag)
	    fpr(replifp);
	  }
	improved = 1;
      } else {
	if (modcalc==1 || long_mod_output) {
	  if (!pmod_calc_flag)
	    nl();
	  else if (replicates_store_result && pmod_calc_flag)
	    fnl(replifp);
	}
	done = 1;
      }
      modcount++;
    } while (!done);
    if (improved) overall_improved = 1;
  } while (improved);
  *parm_pointer1 = bestparm1;
  *parm_pointer2 = bestparm2;
  *parm_pointer3 = bestparm3;
  *parm_pointer4 = bestparm4;
  *parm_pointer5 = bestparm5;
  if (!imprinting) penetrance[modliab][HET_UA] = penetrance[modliab][HET_AU];
  return(overall_improved);
}


double get_next_higher(double bestparm)
{
  int i;

  if (maxim_type == 2) {
    if (bestparm >= 0.9999 || real_eq(bestparm, 0.9999)) return(1.0);
    if (bestparm >= 0.9998 || real_eq(bestparm, 0.9998)) return(0.9999);
    if (bestparm >= 0.9997 || real_eq(bestparm, 0.9997)) return(0.9998);
    if (bestparm >= 0.9996 || real_eq(bestparm, 0.9996)) return(0.9997);
    if (bestparm >= 0.9995 || real_eq(bestparm, 0.9995)) return(0.9996);
    if (bestparm >= 0.9994 || real_eq(bestparm, 0.9994)) return(0.9995);
    if (bestparm >= 0.9992 || real_eq(bestparm, 0.9992)) return(0.9994);
    if (bestparm >= 0.999  || real_eq(bestparm, 0.999))  return(0.9992);
    if (bestparm >= 0.998  || real_eq(bestparm, 0.998))  return(0.999);
    if (bestparm >= 0.997  || real_eq(bestparm, 0.997))  return(0.998);
    if (bestparm >= 0.996  || real_eq(bestparm, 0.996))  return(0.997);
    if (bestparm >= 0.995  || real_eq(bestparm, 0.995))  return(0.996);
    if (bestparm >= 0.994  || real_eq(bestparm, 0.994))  return(0.995);
    if (bestparm >= 0.992  || real_eq(bestparm, 0.992))  return(0.994);
    if (bestparm >= 0.99   || real_eq(bestparm, 0.99))   return(0.992);
    if (bestparm >= 0.985  || real_eq(bestparm, 0.985))  return(0.99);
    if (bestparm >= 0.98   || real_eq(bestparm, 0.98 ))  return(0.985);
    if (bestparm >= 0.975  || real_eq(bestparm, 0.975))  return(0.98);
    if (bestparm >= 0.97   || real_eq(bestparm, 0.97))   return(0.975);
    if (bestparm >= 0.965  || real_eq(bestparm, 0.965))  return(0.97);
    if (bestparm >= 0.96   || real_eq(bestparm, 0.96))   return(0.965);
    if (bestparm >= 0.955  || real_eq(bestparm, 0.955))  return(0.96);
    if (bestparm >= 0.95   || real_eq(bestparm, 0.95))   return(0.955);
    for (i=94; i>=5; i--) {
      if (bestparm >= ((double)i)/100.0 || real_eq(bestparm, ((double)i)/100.0))
        return(((double)(i+1))/100.0);
    }
  } else {
    if (bestparm >= 0.99 || real_eq(bestparm, 0.99)) return(1.0);
    for (i=98; i>=5; i--) {
      if (bestparm >= ((double)i)/100.0 || real_eq(bestparm, ((double)i)/100.0))
        return(((double)(i+1))/100.0);
    }
  }
  if (bestparm >= 0.045  || real_eq(bestparm, 0.045))   return(0.05);
  if (bestparm >= 0.04   || real_eq(bestparm, 0.04))    return(0.045);
  if (bestparm >= 0.035  || real_eq(bestparm, 0.035))   return(0.04);
  if (bestparm >= 0.03   || real_eq(bestparm, 0.03))    return(0.035);
  if (bestparm >= 0.025  || real_eq(bestparm, 0.025))   return(0.03);
  if (bestparm >= 0.02   || real_eq(bestparm, 0.02))    return(0.025);
  if (bestparm >= 0.015  || real_eq(bestparm, 0.015))   return(0.02);
  if (bestparm >= 0.01   || real_eq(bestparm, 0.01))    return(0.015);
  if (bestparm >= 0.008  || real_eq(bestparm, 0.008))   return(0.01);
  if (bestparm >= 0.006  || real_eq(bestparm, 0.006))   return(0.008);
  if (bestparm >= 0.005  || real_eq(bestparm, 0.005))   return(0.006);
  if (bestparm >= 0.004  || real_eq(bestparm, 0.004))   return(0.005);
  if (bestparm >= 0.003  || real_eq(bestparm, 0.003))   return(0.004);
  if (bestparm >= 0.002  || real_eq(bestparm, 0.002))   return(0.003);
  if (bestparm >= 0.001  || real_eq(bestparm, 0.001))   return(0.002);
  if (bestparm >= 0.0008 || real_eq(bestparm, 0.0008))  return(0.001);
  if (bestparm >= 0.0006 || real_eq(bestparm, 0.0006))  return(0.0008);
  if (bestparm >= 0.0005 || real_eq(bestparm, 0.0005))  return(0.0006);
  if (bestparm >= 0.0004 || real_eq(bestparm, 0.0004))  return(0.0005);
  if (bestparm >= 0.0003 || real_eq(bestparm, 0.0003))  return(0.0004);
  if (bestparm >= 0.0002 || real_eq(bestparm, 0.0002))  return(0.0003);
  if (bestparm >= 0.0001 || real_eq(bestparm, 0.0001))  return(0.0002);
  return(0.0001);
}


double get_next_lower(double bestparm)
{
  int i;

  if (bestparm <= 0.0001 || real_eq(bestparm, 0.0001))  return(0.0);
  if (bestparm <= 0.0002 || real_eq(bestparm, 0.0002))  return(0.0001);
  if (bestparm <= 0.0003 || real_eq(bestparm, 0.0003))  return(0.0002);
  if (bestparm <= 0.0004 || real_eq(bestparm, 0.0004))  return(0.0003);
  if (bestparm <= 0.0005 || real_eq(bestparm, 0.0005))  return(0.0004);
  if (bestparm <= 0.0006 || real_eq(bestparm, 0.0006))  return(0.0005);
  if (bestparm <= 0.0008 || real_eq(bestparm, 0.0008))  return(0.0006);
  if (bestparm <= 0.001  || real_eq(bestparm, 0.001))   return(0.0008);
  if (bestparm <= 0.002  || real_eq(bestparm, 0.002))   return(0.001);
  if (bestparm <= 0.003  || real_eq(bestparm, 0.003))   return(0.002);
  if (bestparm <= 0.004  || real_eq(bestparm, 0.004))   return(0.003);
  if (bestparm <= 0.005  || real_eq(bestparm, 0.005))   return(0.004);
  if (bestparm <= 0.006  || real_eq(bestparm, 0.006))   return(0.005);
  if (bestparm <= 0.008  || real_eq(bestparm, 0.008))   return(0.006);
  if (bestparm <= 0.01   || real_eq(bestparm, 0.01))    return(0.008);
  if (bestparm <= 0.015  || real_eq(bestparm, 0.015))   return(0.01);
  if (bestparm <= 0.02   || real_eq(bestparm, 0.02))    return(0.015);
  if (bestparm <= 0.025  || real_eq(bestparm, 0.025))   return(0.02);
  if (bestparm <= 0.03   || real_eq(bestparm, 0.03))    return(0.025);
  if (bestparm <= 0.035  || real_eq(bestparm, 0.035))   return(0.03);
  if (bestparm <= 0.04   || real_eq(bestparm, 0.04))    return(0.035);
  if (bestparm <= 0.045  || real_eq(bestparm, 0.045))   return(0.04);
  if (bestparm <= 0.05   || real_eq(bestparm, 0.05))    return(0.045);
  if (maxim_type ==2) {
    for (i=6; i<=95; i++) {
      if (bestparm <= ((double)i)/100.0 || real_eq(bestparm, ((double)i)/100.0))
        return(((double)(i-1))/100.0);
    }
    if (bestparm <= 0.955  || real_eq(bestparm, 0.955))    return(0.95);
    if (bestparm <= 0.96   || real_eq(bestparm, 0.96))     return(0.955);
    if (bestparm <= 0.965  || real_eq(bestparm, 0.965))    return(0.96);
    if (bestparm <= 0.97   || real_eq(bestparm, 0.97))     return(0.965);
    if (bestparm <= 0.975  || real_eq(bestparm, 0.975))    return(0.97);
    if (bestparm <= 0.98   || real_eq(bestparm, 0.98))     return(0.975);
    if (bestparm <= 0.985  || real_eq(bestparm, 0.985))    return(0.98);
    if (bestparm <= 0.99   || real_eq(bestparm, 0.99))     return(0.985);
    if (bestparm <= 0.992  || real_eq(bestparm, 0.992))    return(0.99);
    if (bestparm <= 0.994  || real_eq(bestparm, 0.994))    return(0.992);
    if (bestparm <= 0.995  || real_eq(bestparm, 0.995))    return(0.994);
    if (bestparm <= 0.996  || real_eq(bestparm, 0.996))    return(0.995);
    if (bestparm <= 0.997  || real_eq(bestparm, 0.997))    return(0.996);
    if (bestparm <= 0.998  || real_eq(bestparm, 0.998))    return(0.997);
    if (bestparm <= 0.999  || real_eq(bestparm, 0.999))    return(0.998);
    if (bestparm <= 0.9992 || real_eq(bestparm, 0.9992))   return(0.999);
    if (bestparm <= 0.9994 || real_eq(bestparm, 0.9994))   return(0.9992);
    if (bestparm <= 0.9995 || real_eq(bestparm, 0.9995))   return(0.9994);
    if (bestparm <= 0.9996 || real_eq(bestparm, 0.9996))   return(0.9995);
    if (bestparm <= 0.9997 || real_eq(bestparm, 0.9997))   return(0.9996);
    if (bestparm <= 0.9998 || real_eq(bestparm, 0.9998))   return(0.9997);
    if (bestparm <= 0.9999 || real_eq(bestparm, 0.9999))   return(0.9998);
    return(0.9999);
  } else {  
    for (i=6; i<=99; i++) {
      if (bestparm <= ((double)i)/100.0 || real_eq(bestparm, ((double)i)/100.0))
        return(((double)(i-1))/100.0);
    }
    return(0.99);
  }
}


double get_next_higher_allfreq(double bestparm)
{
  int i;

  if (maxim_type == 2) {
    if (bestparm >= 0.999999 || real_eq(bestparm, 0.999999))  return(1.0);
    if (bestparm >= 0.999998 || real_eq(bestparm, 0.999998))  return(0.999999);
    if (bestparm >= 0.999997 || real_eq(bestparm, 0.999997))  return(0.999998);
    if (bestparm >= 0.999996 || real_eq(bestparm, 0.999996))  return(0.999997);
    if (bestparm >= 0.999995 || real_eq(bestparm, 0.999995))  return(0.999996);
    if (bestparm >= 0.999994 || real_eq(bestparm, 0.999994))  return(0.999995);
    if (bestparm >= 0.999992 || real_eq(bestparm, 0.999992))  return(0.999994);
    if (bestparm >= 0.99999  || real_eq(bestparm, 0.99999))   return(0.999992);
    if (bestparm >= 0.99998  || real_eq(bestparm, 0.99998))   return(0.99999);
    if (bestparm >= 0.99997  || real_eq(bestparm, 0.99997))   return(0.99998);
    if (bestparm >= 0.99996  || real_eq(bestparm, 0.99996))   return(0.99997);
    if (bestparm >= 0.99995  || real_eq(bestparm, 0.99995))   return(0.99996);
    if (bestparm >= 0.99994  || real_eq(bestparm, 0.99994))   return(0.99995);
    if (bestparm >= 0.99992  || real_eq(bestparm, 0.99992))   return(0.99994);
    if (bestparm >= 0.9999   || real_eq(bestparm, 0.9999))    return(0.99992);
    if (bestparm >= 0.9998   || real_eq(bestparm, 0.9998))    return(0.9999);
    if (bestparm >= 0.9997   || real_eq(bestparm, 0.9997))    return(0.9998);
    if (bestparm >= 0.9996   || real_eq(bestparm, 0.9996))    return(0.9997);
    if (bestparm >= 0.9995   || real_eq(bestparm, 0.9995))    return(0.9996);
    if (bestparm >= 0.9994   || real_eq(bestparm, 0.9994))    return(0.9995);
    if (bestparm >= 0.9992   || real_eq(bestparm, 0.9992))    return(0.9994);
    if (bestparm >= 0.999    || real_eq(bestparm, 0.999))     return(0.9992);
    if (bestparm >= 0.998    || real_eq(bestparm, 0.998))     return(0.999);
    if (bestparm >= 0.997    || real_eq(bestparm, 0.997))     return(0.998);
    if (bestparm >= 0.996    || real_eq(bestparm, 0.996))     return(0.997);
    if (bestparm >= 0.995    || real_eq(bestparm, 0.995))     return(0.996);
    if (bestparm >= 0.994    || real_eq(bestparm, 0.994))     return(0.995);
    if (bestparm >= 0.992    || real_eq(bestparm, 0.992))     return(0.994);
    if (bestparm >= 0.99     || real_eq(bestparm, 0.99))      return(0.992);
    if (bestparm >= 0.985    || real_eq(bestparm, 0.985))     return(0.99);
    if (bestparm >= 0.98     || real_eq(bestparm, 0.98))      return(0.985);
    if (bestparm >= 0.975    || real_eq(bestparm, 0.975))     return(0.98);
    if (bestparm >= 0.97     || real_eq(bestparm, 0.97))      return(0.975);
    if (bestparm >= 0.965    || real_eq(bestparm, 0.965))     return(0.97);
    if (bestparm >= 0.96     || real_eq(bestparm, 0.96))      return(0.965);
    if (bestparm >= 0.955    || real_eq(bestparm, 0.955))     return(0.96);
    if (bestparm >= 0.95     || real_eq(bestparm, 0.95))      return(0.955);
    for (i=94; i>=5; i--) {
      if (bestparm >= ((double)i)/100.0 || real_eq(bestparm, ((double)i)/100.0))
        return(((double)(i+1))/100.0);
    }
  } else {
    if (bestparm >= 0.99 || real_eq(bestparm, 0.99))  return(1.0);
    for (i=98; i>=5; i--) {
      if (bestparm >= ((double)i)/100.0 || real_eq(bestparm, ((double)i)/100.0))
        return(((double)(i+1))/100.0);
    }
  }
  if (bestparm >= 0.045  || real_eq(bestparm, 0.045))   return(0.05);
  if (bestparm >= 0.04   || real_eq(bestparm, 0.04))    return(0.045);
  if (bestparm >= 0.035  || real_eq(bestparm, 0.035))   return(0.04);
  if (bestparm >= 0.03   || real_eq(bestparm, 0.03))    return(0.035);
  if (bestparm >= 0.025  || real_eq(bestparm, 0.025))   return(0.03);
  if (bestparm >= 0.02   || real_eq(bestparm, 0.02))    return(0.025);
  if (bestparm >= 0.015  || real_eq(bestparm, 0.015))   return(0.02);
  if (bestparm >= 0.01   || real_eq(bestparm, 0.01))    return(0.015);
  if (bestparm >= 0.008  || real_eq(bestparm, 0.008))   return(0.01);
  if (bestparm >= 0.006  || real_eq(bestparm, 0.006))   return(0.008);
  if (bestparm >= 0.005  || real_eq(bestparm, 0.005))   return(0.006);
  if (bestparm >= 0.004  || real_eq(bestparm, 0.004))   return(0.005);
  if (bestparm >= 0.003  || real_eq(bestparm, 0.003))   return(0.004);
  if (bestparm >= 0.002  || real_eq(bestparm, 0.002))   return(0.003);
  if (bestparm >= 0.001  || real_eq(bestparm, 0.001))   return(0.002);
  if (bestparm >= 0.0008 || real_eq(bestparm, 0.0008))  return(0.001);
  if (bestparm >= 0.0006 || real_eq(bestparm, 0.0006))  return(0.0008);
  if (bestparm >= 0.0005 || real_eq(bestparm, 0.0005))  return(0.0006);
  if (bestparm >= 0.0004 || real_eq(bestparm, 0.0004))  return(0.0005);
  if (bestparm >= 0.0003 || real_eq(bestparm, 0.0003))  return(0.0004);
  if (bestparm >= 0.0002 || real_eq(bestparm, 0.0002))  return(0.0003);
  if (bestparm >= 0.0001 || real_eq(bestparm, 0.0001))  return(0.0002);
  if (bestparm >= 0.00008 || real_eq(bestparm, 0.00008))  return(0.0001);
  if (bestparm >= 0.00006 || real_eq(bestparm, 0.00006))  return(0.00008);
  if (bestparm >= 0.00005 || real_eq(bestparm, 0.00005))  return(0.00006);
  if (bestparm >= 0.00004 || real_eq(bestparm, 0.00004))  return(0.00005);
  if (bestparm >= 0.00003 || real_eq(bestparm, 0.00003))  return(0.00004);
  if (bestparm >= 0.00002 || real_eq(bestparm, 0.00002))  return(0.00003);
  if (bestparm >= 0.00001 || real_eq(bestparm, 0.00001))  return(0.00002);
  if (bestparm >= 0.000008 || real_eq(bestparm, 0.000008))  return(0.00001);
  if (bestparm >= 0.000006 || real_eq(bestparm, 0.000006))  return(0.000008);
  if (bestparm >= 0.000005 || real_eq(bestparm, 0.000005))  return(0.000006);
  if (bestparm >= 0.000004 || real_eq(bestparm, 0.000004))  return(0.000005);
  if (bestparm >= 0.000003 || real_eq(bestparm, 0.000003))  return(0.000004);
  if (bestparm >= 0.000002 || real_eq(bestparm, 0.000002))  return(0.000003);
  if (bestparm >= 0.000001 || real_eq(bestparm, 0.000001))  return(0.000002);
  return(0.000001);
}


double get_next_lower_allfreq(double bestparm)
{
  int i;

  if (bestparm <= 0.000001 || real_eq(bestparm, 0.000001))  return(0.0);
  if (bestparm <= 0.000002 || real_eq(bestparm, 0.000002))  return(0.000001);
  if (bestparm <= 0.000003 || real_eq(bestparm, 0.000003))  return(0.000002);
  if (bestparm <= 0.000004 || real_eq(bestparm, 0.000004))  return(0.000003);
  if (bestparm <= 0.000005 || real_eq(bestparm, 0.000005))  return(0.000004);
  if (bestparm <= 0.000006 || real_eq(bestparm, 0.000006))  return(0.000005);
  if (bestparm <= 0.000008 || real_eq(bestparm, 0.000008))  return(0.000006);
  if (bestparm <= 0.00001 || real_eq(bestparm, 0.00001))  return(0.000008);
  if (bestparm <= 0.00002 || real_eq(bestparm, 0.00002))  return(0.00001);
  if (bestparm <= 0.00003 || real_eq(bestparm, 0.00003))  return(0.00002);
  if (bestparm <= 0.00004 || real_eq(bestparm, 0.00004))  return(0.00003);
  if (bestparm <= 0.00005 || real_eq(bestparm, 0.00005))  return(0.00004);
  if (bestparm <= 0.00006 || real_eq(bestparm, 0.00006))  return(0.00005);
  if (bestparm <= 0.00008 || real_eq(bestparm, 0.00008))  return(0.00006);
  if (bestparm <= 0.0001 || real_eq(bestparm, 0.0001))  return(0.00008);
  if (bestparm <= 0.0002 || real_eq(bestparm, 0.0002))  return(0.0001);
  if (bestparm <= 0.0003 || real_eq(bestparm, 0.0003))  return(0.0002);
  if (bestparm <= 0.0004 || real_eq(bestparm, 0.0004))  return(0.0003);
  if (bestparm <= 0.0005 || real_eq(bestparm, 0.0005))  return(0.0004);
  if (bestparm <= 0.0006 || real_eq(bestparm, 0.0006))  return(0.0005);
  if (bestparm <= 0.0008 || real_eq(bestparm, 0.0008))  return(0.0006);
  if (bestparm <= 0.001  || real_eq(bestparm, 0.001))   return(0.0008);
  if (bestparm <= 0.002  || real_eq(bestparm, 0.002))   return(0.001);
  if (bestparm <= 0.003  || real_eq(bestparm, 0.003))   return(0.002);
  if (bestparm <= 0.004  || real_eq(bestparm, 0.004))   return(0.003);
  if (bestparm <= 0.005  || real_eq(bestparm, 0.005))   return(0.004);
  if (bestparm <= 0.006  || real_eq(bestparm, 0.006))   return(0.005);
  if (bestparm <= 0.008  || real_eq(bestparm, 0.008))   return(0.006);
  if (bestparm <= 0.01   || real_eq(bestparm, 0.01))    return(0.008);
  if (bestparm <= 0.015  || real_eq(bestparm, 0.015))   return(0.01);
  if (bestparm <= 0.02   || real_eq(bestparm, 0.02))    return(0.015);
  if (bestparm <= 0.025  || real_eq(bestparm, 0.025))   return(0.02);
  if (bestparm <= 0.03   || real_eq(bestparm, 0.03))    return(0.025);
  if (bestparm <= 0.035  || real_eq(bestparm, 0.035))   return(0.03);
  if (bestparm <= 0.04   || real_eq(bestparm, 0.04))    return(0.035);
  if (bestparm <= 0.045  || real_eq(bestparm, 0.045))   return(0.04);
  if (bestparm <= 0.05   || real_eq(bestparm, 0.05))    return(0.045);

  if (maxim_type == 2) {
    for (i=6; i<=95; i++) {
      if (bestparm <= ((double)i)/100.0 || real_eq(bestparm, ((double)i)/100.0))
        return(((double)(i-1))/100.0);
    }
    if (bestparm >= 0.955  || real_eq(bestparm, 0.955))   return(0.95);
    if (bestparm >= 0.96   || real_eq(bestparm, 0.96))    return(0.955);
    if (bestparm >= 0.965  || real_eq(bestparm, 0.965))   return(0.96);
    if (bestparm >= 0.97   || real_eq(bestparm, 0.97))    return(0.965);
    if (bestparm >= 0.975  || real_eq(bestparm, 0.975))   return(0.97);
    if (bestparm >= 0.98   || real_eq(bestparm, 0.98))    return(0.975);
    if (bestparm >= 0.985  || real_eq(bestparm, 0.985))   return(0.98);
    if (bestparm >= 0.99   || real_eq(bestparm, 0.99))    return(0.985);
    if (bestparm >= 0.992  || real_eq(bestparm, 0.992))   return(0.99);
    if (bestparm >= 0.994  || real_eq(bestparm, 0.994))   return(0.992);
    if (bestparm >= 0.995  || real_eq(bestparm, 0.995))   return(0.994);
    if (bestparm >= 0.996  || real_eq(bestparm, 0.996))   return(0.995);
    if (bestparm >= 0.997  || real_eq(bestparm, 0.997))   return(0.996);
    if (bestparm >= 0.998  || real_eq(bestparm, 0.998))   return(0.997);
    if (bestparm >= 0.999  || real_eq(bestparm, 0.999))   return(0.998);
    if (bestparm >= 0.9992 || real_eq(bestparm, 0.9992))  return(0.999);
    if (bestparm >= 0.9994 || real_eq(bestparm, 0.9994))  return(0.9992);
    if (bestparm >= 0.9995 || real_eq(bestparm, 0.9995))  return(0.9994);
    if (bestparm >= 0.9996 || real_eq(bestparm, 0.9996))  return(0.9995);
    if (bestparm >= 0.9997 || real_eq(bestparm, 0.9997))  return(0.9996);
    if (bestparm >= 0.9998 || real_eq(bestparm, 0.9998))  return(0.9997);
    if (bestparm >= 0.9999 || real_eq(bestparm, 0.9999))  return(0.9998);
    if (bestparm >= 0.99992 || real_eq(bestparm, 0.99992))  return(0.9999);
    if (bestparm >= 0.99994 || real_eq(bestparm, 0.99994))  return(0.99992);
    if (bestparm >= 0.99995 || real_eq(bestparm, 0.99995))  return(0.99994);
    if (bestparm >= 0.99996 || real_eq(bestparm, 0.99996))  return(0.99995);
    if (bestparm >= 0.99997 || real_eq(bestparm, 0.99997))  return(0.99996);
    if (bestparm >= 0.99998 || real_eq(bestparm, 0.99998))  return(0.99997);
    if (bestparm >= 0.99999 || real_eq(bestparm, 0.99999))  return(0.99998);
    if (bestparm >= 0.999992 || real_eq(bestparm, 0.999992))  return(0.99999);
    if (bestparm >= 0.999994 || real_eq(bestparm, 0.999994))  return(0.999992);
    if (bestparm >= 0.999995 || real_eq(bestparm, 0.999995))  return(0.999994);
    if (bestparm >= 0.999996 || real_eq(bestparm, 0.999996))  return(0.999995);
    if (bestparm >= 0.999997 || real_eq(bestparm, 0.999997))  return(0.999996);
    if (bestparm >= 0.999998 || real_eq(bestparm, 0.999998))  return(0.999997);
    if (bestparm >= 0.999999 || real_eq(bestparm, 0.999999))  return(0.999998);
    return(0.999999);
  } else {
    for (i=6; i<=99; i++) {
      if (bestparm <= ((double)i)/100.0 || real_eq(bestparm, ((double)i)/100.0))
        return(((double)(i-1))/100.0);
    }
    return(0.99);
  }
}


int penetrance_restriction_violated (void)
{
  if (penetrance_restriction &&
      (((penetrance[modliab][HET_AU] < penetrance[modliab][HOM_UNAFF])
	&& !real_eq(penetrance[modliab][HET_AU], penetrance[modliab][HOM_UNAFF])) ||
       ((penetrance[modliab][HET_UA] < penetrance[modliab][HOM_UNAFF])
	&& !real_eq(penetrance[modliab][HET_UA], penetrance[modliab][HOM_UNAFF])) ||
       ((penetrance[modliab][HET_AU] > penetrance[modliab][HOM_AFF])
	&& !real_eq(penetrance[modliab][HET_AU], penetrance[modliab][HOM_AFF])) ||
       ((penetrance[modliab][HET_UA] > penetrance[modliab][HOM_AFF])
	&& !real_eq(penetrance[modliab][HET_UA], penetrance[modliab][HOM_AFF]))))
    return (1);
  else return (0);
}


int user_penetrance_restriction_violated (double f1, double f2, double f3,
					  double f4)
{
  if (((f2 < f1) && !real_eq(f2, f1)) ||
      ((f3 < f1) && !real_eq(f3, f1)) ||
      ((f2 > f4) && !real_eq(f2, f4)) ||
      ((f3 > f4) && !real_eq(f3, f4)))
    return (1);
  else return (0);
}


int all_penetrances_equal (void)
{
  if (real_eq(penetrance[modliab][HOM_UNAFF], penetrance[modliab][HET_AU]) &&
      real_eq(penetrance[modliab][HOM_UNAFF], penetrance[modliab][HET_UA]) &&
      real_eq(penetrance[modliab][HOM_UNAFF], penetrance[modliab][HOM_AFF]))
    return (1);
  else return (0);
}


int all_penetrances_all_liab_equal (void)
{
  int i, equal=TRUE;

  for (i=0; i<num_liability_classes; i++)
    if (!real_eq(penetrance[i+1][HOM_UNAFF], penetrance[i+1][HET_AU]) ||
	!real_eq(penetrance[i+1][HOM_UNAFF], penetrance[i+1][HET_UA]) ||
	!real_eq(penetrance[i+1][HOM_UNAFF], penetrance[i+1][HOM_AFF])) {
      equal = FALSE;
      break;
    }
  return (equal);
}


int allfreq_restriction_violated (void)
{
  if (allfreq_restriction &&
      ((affected_freq > highest_allfreq) &&
       !real_eq(affected_freq, highest_allfreq)))
    return (1);
  else return (0);
}


int user_allfreq_restriction_violated (double p)
{
  if ((p > highest_allfreq) && !real_eq(p, highest_allfreq))
    return (1);
  else return (0);
}


int model_already_checked (void)
{
	int count;

	for (count=0; count<modcount; count++)
	{
		if(!real_eq(penetrance[modliab][HOM_UNAFF], modpen1[count])) continue;
		if(!real_eq(penetrance[modliab][HET_AU],    modpen2[count])) continue;
		if(!real_eq(penetrance[modliab][HET_UA],    modpen3[count])) continue;
		if(!real_eq(penetrance[modliab][HOM_AFF],   modpen4[count])) continue;
		if(!real_eq(affected_freq, modafreq[count])) continue;
		return (1);
	}
	return (0);
}

void error_too_many_models (void)
{
  nl();
  sf(ps, "number of tested models > MAX_MODELS (%d) - please edit npl.h",
     MAX_MODELS);
  error(ps);
}


int calculate_lod (void)
{
  int i, j, dis_geno, vec, new_vec, num_in_ped, old_mom;
  double this_prob, total, lodlike;
  double *num_par_score=NULL, *this_p_pos=NULL, *this_model_vector=NULL;
  PED_MEMBER *this_famtree;
  NUC_FAMILY *this_nuc_fam;
  double *disease_prob_old_mom = NULL, *disease_prob_new_mom = NULL, *this_par_score;
  int *place_allele_old_mom = NULL, *place_allele_new_mom = NULL;
  int count, quick_calc=FALSE, quick_model;
  int vec_class;
  int this_srnvecs, max_count;
  int *this_new_vec_to_alg_prob;

  if (modcalc == 1 || long_mod_output) {
    if (imprinting)
      sf(ps,"%8.6lf  (%6.4lf, %6.4lf, %6.4lf, %6.4lf)  ", affected_freq,
	 penetrance[modliab][HOM_UNAFF], penetrance[modliab][HET_AU],
	 penetrance[modliab][HET_UA], penetrance[modliab][HOM_AFF]);
    else
      sf(ps,"%8.6lf  (%6.4lf, %6.4lf, %6.4lf)  ", affected_freq,
	 penetrance[modliab][HOM_UNAFF], penetrance[modliab][HET_AU],
	 penetrance[modliab][HOM_AFF]);
    if (!pmod_calc_flag) {
      pr();
      flush();
    } else if (replicates_store_result && pmod_calc_flag) {
      fpr(replifp);
      fflush(replifp);
    }
  }

  if (modcalc == 1) 
    for (i=first_mod_pos; i<=last_mod_pos; i++) {
      for (famnum=0; famnum<num_pedigrees; famnum++)
        sub_lods[famnum][i] = 0.0;
      total_lod[i] = 0.0;
    }

  if (modcalc == 2) {
    for (famnum=0; famnum<num_pedigrees; famnum++)
      sub_lods[famnum][pos] = 0.0;
    total_lod[pos] = 0.0;  /* only pos; leave other elements unchanged */
  }

  if (all_penetrances_all_liab_equal()) {
    if (modcalc == 1) maxlodpos = first_mod_pos;
    if (modcalc == 2) maxlodpos = pos;
    maxlod = 0.0;
    if (modcalc == 1 || long_mod_output) {
      if (ind_sp==1) {
        sf(pos_str,"%7.2lf %7.2lf %7.2lf",0.5*(wpos_m[maxlodpos] + wpos_f[maxlodpos]), wpos_f[maxlodpos], wpos_m[maxlodpos]);
        sf(lod_str,"%10.6lf", maxlod);
        sf(ps,"%s %s", pos_str, lod_str);
        if (!pmod_calc_flag)
          pr();
        else if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      } else if (ind_sp==0) {
        sf(pos_str,"%7.2lf",wpos[maxlodpos]);
        sf(lod_str,"%10.6lf", maxlod);
        sf(ps,"%s %s", pos_str, lod_str);
	if (!pmod_calc_flag)
          pr();
        else if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
    }
    return (0);
  }

  /*** look if model vector is already in storage; yes --> quick_calc=TRUE ***/
 if (saved_models) /* changed, now with modcalc ==1 or ==2 */
 {       max_count = model_storage_full ? saved_models : current_saved_model;
	 for (count = 0; count < max_count; count++)
	 {
		 quick_calc = TRUE;
		 for (i=0; i<num_liability_classes; i++)
		 {
			 if (!real_eq(penetrance[i+1][HOM_UNAFF], saved_model_pen1[count][i]))
			 {
				 quick_calc = FALSE;
				 break;
			 }
			 if(!real_eq(penetrance[i+1][HET_AU],    saved_model_pen2[count][i]))
			 {
				 quick_calc = FALSE;
				 break;
			 }
			 if(!real_eq(penetrance[i+1][HET_UA],    saved_model_pen3[count][i]))
			 {
				 quick_calc = FALSE;
				 break;
			 }
			 if(!real_eq(penetrance[i+1][HOM_AFF],   saved_model_pen4[count][i]))
			 {
				 quick_calc = FALSE;
				 break;
			 }
		 }
		 if(quick_calc)
		 {
			 if (!real_eq(affected_freq, saved_model_afreq[count]))
				 quick_calc = FALSE;
		 }
		 if (quick_calc) {
			 quick_model = count;
			 break;
		 }

	 }
 }
  
  if (quick_calc) {    /*** calculate LOD with saved model vector ***/
    for (famnum=0; famnum<num_pedigrees; famnum++) {
	    this_srnvecs = save_real_num_vecs[famnum_pointer[famnum]];
      if (quick_model_famnum[famnum_pointer[famnum]][quick_model] == 0) {
        if (calculate_lod_famnum(quick_model))
          return(-1);
      }

      if (modcalc == 1) {
        for (i=first_mod_pos; i<=last_mod_pos; i++) {
	  lodlike = 0.0;
	  this_p_pos = save_p_all_pos[famnum][i-first_mod_pos];
	  this_model_vector = saved_model_vectors[famnum_pointer[famnum]][quick_model];
	  total = 0.0;
	  for (new_vec=0; new_vec<this_srnvecs; new_vec++)
	  {
	    lodlike += this_p_pos[new_vec] * this_model_vector[new_vec];
	    total += this_model_vector[new_vec]; /* MB - 08/12 */ 
	  }
	  lodlike /= total;
	  if (lodlike <= 0.0) sub_lods[famnum][i] = -10000.0;
	  else sub_lods[famnum][i] = log10(lodlike/save_unlinked[famnum_pointer[famnum]]);
	  total_lod[i] += sub_lods[famnum][i];
        }
      } else if (modcalc == 2) {
        lodlike = 0.0;
        this_p_pos = save_p_one_pos[famnum];
        this_model_vector = saved_model_vectors[famnum_pointer[famnum]][quick_model];
	total = 0.0;
	  for (new_vec=0; new_vec<this_srnvecs; new_vec++)
	  {
	    lodlike += this_p_pos[new_vec] * this_model_vector[new_vec];
	    total += this_model_vector[new_vec]; /* MB - 08/12 */ 
	  }
	lodlike /= total;
        if (lodlike <= 0.0) sub_lods[famnum][pos] = -10000.0;
        else sub_lods[famnum][pos] = log10(lodlike/save_unlinked[famnum_pointer[famnum]]);
        total_lod[pos] += sub_lods[famnum][pos];
      }
    }

    if (!best_model_flag) total_quick_models++;

    /*if (!pmod_calc_flag) {
       sf(ps, "Q%d", quick_model); pr(); 
    } */

  } else { /** model not in memory --> standard LOD calc **/

	  if(algebra)
	  {
		  algebra_power_prep(); /* prepare power vectors before looping over families and new_vec MB - 07/12 */
		  algebra_calculated(); /* algebra_matrix has already been called for all families in analyze_family; unique power structures can therefore be calculated and stored in vector */
		  for (vec_class=0; vec_class <= vec_classes; vec_class++){
			  if (num_fam_in_vec_class[vec_class] > 0){  
				  algebra_force(vec_class);
			  }
		  }
		  for (famnum=0; famnum<num_pedigrees; famnum++) {
			  this_famtree = famtree[famnum];
			  this_nuc_fam = nuc_fam[famnum];
			  num_in_ped = save_num_in_ped[famnum];
			  old_mom = save_old_mom[famnum];
			  this_par_score = par_score[algebra_pointer[famnum]];
			  this_srnvecs = save_real_num_vecs[famnum_pointer[famnum]];
			  this_new_vec_to_alg_prob = new_vec_to_alg_prob[algebra_pointer[famnum]];
			  algebra_cache_opt(this_par_score, this_srnvecs, this_new_vec_to_alg_prob);

			  total=0.0;
			  for (j=0; j<this_srnvecs; j++) 
				  total += this_par_score[j];
			  if (j != this_srnvecs) /* not really useful */
				  error("num_vecs don't match in modscore routine");

			  if (total == 0.0) {
				  if (modcalc==1 || long_mod_output) {
					  sf(ps, "\n  disease pvectors sum to 0.0 for pedigree %s - SKIPPING THIS MODEL", pedigree_names[famnum]);
					  if (!pmod_calc_flag)
						  pr();
					  else if (replicates_store_result && pmod_calc_flag)
						  fpr(replifp);
				  }
				  unarray(num_par_score, double);
				  if (saved_models &&
						  ((!model_storage_full) || (!initial_fill_storage))) {
					  /* same condition as below, for saving of model vector */
					  /* Model vector may already have been saved for previous families *
					   * - so we set the penetrances and disease allele frequency to 0  *
					   * here in order to prevent this partially corrupted saved model  *
					   * vector from being reused for quick calculation in the future.  *
					   * Since current_saved_model is not updated, this memory position *
					   * will be overwritten the next time a model vector is saved.     */
					  for (i=0; i<num_liability_classes; i++) {
						  saved_model_pen1[current_saved_model][i] = 0.0;
						  saved_model_pen2[current_saved_model][i] = 0.0;
						  saved_model_pen3[current_saved_model][i] = 0.0;
						  saved_model_pen4[current_saved_model][i] = 0.0;
					  }
					  saved_model_afreq[current_saved_model] = 0.0;
				  }
				  return(-1);
			  } 

			  if (modcalc == 1)
			  {
				  for (i=first_mod_pos; i<=last_mod_pos; i++) {
					  lodlike = 0.0;
					  this_p_pos = save_p_all_pos[famnum][i-first_mod_pos];
					  for (new_vec=0; new_vec<this_srnvecs; new_vec++)
						  lodlike += this_p_pos[new_vec] * this_par_score[new_vec];
					  lodlike /= total; /* MB - 08/12 */
					  if (lodlike <= 0.0) sub_lods[famnum][i] = -10000.0;
					  else sub_lods[famnum][i] = log10(lodlike/save_unlinked[famnum_pointer[famnum]]);
					  total_lod[i] += sub_lods[famnum][i];
				  }
			  } 

			  if (modcalc == 2) {
				  lodlike = 0.0;
				  this_p_pos = save_p_one_pos[famnum];
				  for (new_vec=0; new_vec<this_srnvecs; new_vec++)
					  lodlike += this_p_pos[new_vec] * this_par_score[new_vec];
				  lodlike /= total; /* MB - 08/12 */
				  if (lodlike <= 0.0) sub_lods[famnum][pos] = -10000.0;
				  else sub_lods[famnum][pos] = log10(lodlike/save_unlinked[famnum_pointer[famnum]]);
				  total_lod[pos] += sub_lods[famnum][pos];
			  } 

			  /** save model if enough memory **/
			  if (saved_models &&
					  ((!model_storage_full) || (!initial_fill_storage))) {
				  /* save if there is free storage left or if there is 
				     enough space to save also fine-maximization models */
				  this_model_vector = saved_model_vectors[famnum_pointer[famnum]][current_saved_model];
				  for (new_vec=0; new_vec<this_srnvecs; new_vec++)
					  this_model_vector[new_vec] = this_par_score[new_vec];
				  quick_model_famnum[famnum_pointer[famnum]][current_saved_model] = 1;
			  } 
		  }  /* for famnum */
	  } else{ /* if(!algebra) */
		  init_disease_info(); /* only creates penetrance[0] and apriori_disease, which is needed for peeling; the former is never used MB - 02/13 */
		  for (famnum=0; famnum<num_pedigrees; famnum++) {

			  array(num_par_score, save_real_num_vecs[famnum_pointer[famnum]], double);
			  this_famtree = famtree[famnum];
			  this_nuc_fam = nuc_fam[famnum];
			  num_in_ped = save_num_in_ped[famnum];
			  old_mom = save_old_mom[famnum];
			  this_srnvecs = save_real_num_vecs[famnum_pointer[famnum]];

			  if (save_num_inbred[famnum] >= 1) { /* num_inbred is 0 in algebraic mode */
				  disease_prob_old_mom = this_famtree[old_mom].disease_prob;
				  disease_prob_new_mom = this_famtree[num_in_ped].disease_prob;
				  place_allele_old_mom = this_famtree[old_mom].place_allele;
				  place_allele_new_mom = this_famtree[num_in_ped].place_allele;
			  }

			  vec = 0;

			  for (new_vec=0; new_vec<save_real_num_vecs[famnum_pointer[famnum]]; new_vec++) {

				  if (new_vec > 0) {
					  vec = 0;
					  for(i=0; i<save_npadMS[famnum]; i++)
						  vec |= (new_vec & save_padMask[famnum][i])
							  << save_padShift[famnum][i];
				  }
				  fill_placeholder_alleles_1(vec,save_real_non_originals[famnum],
						  this_famtree); 
				  if (save_num_inbred[famnum] >= 1) {
					  /* break the loop! */
					  num_par_score[new_vec] = 0.0;
					  for (dis_geno=0; dis_geno<4; dis_geno++) {
						  init_disease_probs(num_in_ped, this_famtree);

						  disease_prob_new_mom[dis_geno]=1.0;
						  j = dis_geno^1;
						  disease_prob_old_mom[j]=0.0;
						  disease_prob_new_mom[j]=0.0;
						  j ^= 2;
						  disease_prob_old_mom[j]=0.0;
						  disease_prob_new_mom[j]=0.0;
						  j ^= 1;
						  disease_prob_old_mom[j]=0.0;
						  disease_prob_new_mom[j]=0.0;

						  if (save_num_nuc_fams_to_peel[famnum] > 0)
							  peel(vec, save_num_bits[famnum], this_famtree, this_nuc_fam);
						  place_allele_new_mom[0] = place_allele_old_mom[0];
						  place_allele_new_mom[1] = place_allele_old_mom[1];
						  this_prob = brute_force_analyze(
								  save_bf_originals[famnum],
								  save_bf_original_list[famnum],
								  save_bf_non_originals[famnum],
								  save_bf_non_original_list[famnum],
								  this_famtree);
						  num_par_score[new_vec] += this_prob;
					  }
				  } else {
					  init_disease_probs(num_in_ped, this_famtree);
					  if (save_num_nuc_fams_to_peel[famnum] > 0)
						  peel(vec, save_num_bits[famnum], this_famtree, this_nuc_fam);
					  num_par_score[new_vec] =
						  brute_force_analyze(save_bf_originals[famnum], save_bf_original_list[famnum], save_bf_non_originals[famnum], save_bf_non_original_list[famnum], this_famtree);

				  }
			  }

			  if (new_vec != save_real_num_vecs[famnum_pointer[famnum]])
				  error("num_vecs don't match in modscore routine");
			  total=0.0;
			  for (j=0; j<save_real_num_vecs[famnum_pointer[famnum]]; j++) 
				  total += num_par_score[j];

			  if (total == 0.0) {
				  if (modcalc==1 || long_mod_output) {
					  sf(ps, "\n  disease pvectors sum to 0.0 for pedigree %s - SKIPPING THIS MODEL", pedigree_names[famnum]);
					  if (!pmod_calc_flag)
						  pr();
					  else if (replicates_store_result && pmod_calc_flag)
						  fpr(replifp);
				  }
				  unarray(num_par_score, double);
				  if (saved_models &&
						  ((!model_storage_full) || (!initial_fill_storage))) {
					  /* same condition as below, for saving of model vector */
					  /* Model vector may already have been saved for previous families *
					   * - so we set the penetrances and disease allele frequency to 0  *
					   * here in order to prevent this partially corrupted saved model  *
					   * vector from being reused for quick calculation in the future.  *
					   * Since current_saved_model is not updated, this memory position *
					   * will be overwritten the next time a model vector is saved.     */
					  for (i=0; i<num_liability_classes; i++) {
						  saved_model_pen1[current_saved_model][i] = 0.0;
						  saved_model_pen2[current_saved_model][i] = 0.0;
						  saved_model_pen3[current_saved_model][i] = 0.0;
						  saved_model_pen4[current_saved_model][i] = 0.0;
					  }
					  saved_model_afreq[current_saved_model] = 0.0;
				  }
				  return(-1);
			  } else {
				  for(j=0; j<save_real_num_vecs[famnum_pointer[famnum]]; j++)
					  num_par_score[j] /= total;
			  }

			  if (modcalc == 1)
			  {
				  for (i=first_mod_pos; i<=last_mod_pos; i++) {
					  lodlike = 0.0;
					  this_p_pos = save_p_all_pos[famnum][i-first_mod_pos];
					  for (new_vec=0; new_vec<this_srnvecs; new_vec++)
						  lodlike += this_p_pos[new_vec] * num_par_score[new_vec];
					  if (lodlike <= 0.0) sub_lods[famnum][i] = -10000.0;
					  else sub_lods[famnum][i] = log10(lodlike/save_unlinked[famnum_pointer[famnum]]);
					  total_lod[i] += sub_lods[famnum][i];
				  }
			  }

			  if (modcalc == 2) {
				  lodlike = 0.0;
				  this_p_pos = save_p_one_pos[famnum];
				  for (new_vec=0; new_vec<this_srnvecs; new_vec++)
					  lodlike += this_p_pos[new_vec] * num_par_score[new_vec];
				  if (lodlike <= 0.0) sub_lods[famnum][pos] = -10000.0;
				  else sub_lods[famnum][pos] = log10(lodlike/save_unlinked[famnum_pointer[famnum]]);
				  total_lod[pos] += sub_lods[famnum][pos];
			  }

			  /** save model if enough memory **/
			  if (saved_models &&
					  ((!model_storage_full) || (!initial_fill_storage))) {
				  /* save if there is free storage left or if there is 
				     enough space to save also fine-maximization models */
				  this_model_vector = saved_model_vectors[famnum_pointer[famnum]][current_saved_model];
				  for (new_vec=0; new_vec<this_srnvecs; new_vec++)
					  this_model_vector[new_vec] = num_par_score[new_vec];
				  quick_model_famnum[famnum_pointer[famnum]][current_saved_model] = 1;
			  }
			  unarray(num_par_score, double); 
		  }  /* for famnum */
	  } /* end of if(!algebra) */

    if (saved_models) {
      /** save penetrances / disease allele frequency and
	  update current_saved_model if model has been stored **/
      if ((!model_storage_full) || (!initial_fill_storage)) {
	  /* same condition as above, for saving of model vector */
	for (i=0; i<num_liability_classes; i++) {
	  saved_model_pen1[current_saved_model][i] =
	      penetrance[i+1][HOM_UNAFF];
	  saved_model_pen2[current_saved_model][i] =
	      penetrance[i+1][HET_AU];
	  saved_model_pen3[current_saved_model][i] =
	      penetrance[i+1][HET_UA];
	  saved_model_pen4[current_saved_model][i] =
	      penetrance[i+1][HOM_AFF];
	}
	saved_model_afreq[current_saved_model] = affected_freq;
	if (current_saved_model == saved_models-1) {
	  if (initial_count)
	    /* jump to right behind initial models, -1 for datafile model */
	    current_saved_model = initial_count-1;
	  else {
	    /* initial models completely fill storage, save no further model */
	    initial_fill_storage = TRUE;
	    current_saved_model = 0;
	  }
	  model_storage_full = TRUE;
	  /* if (long_mod_output) print("J"); */
	} else current_saved_model++;
	/* if (long_mod_output) {sf(ps, "S%d ", current_saved_model); pr();} */
      }
    }
  } /* end of standard LOD calculation (no quick_calc) */

  if (ind_sp==1) {
    if (modcalc == 1) {
      maxlod = -10000.0;
      maxlodpos = -1;
      for (i=first_mod_pos; i<=last_mod_pos; i++) if (total_lod[i] > maxlod) {
        maxlod = total_lod[i];
        maxlodpos = i;
      }
      sf(pos_str,"%7.2lf %7.2lf %7.2lf",0.5*(wpos_m[maxlodpos] + wpos_f[maxlodpos]), wpos_f[maxlodpos], wpos_m[maxlodpos]);
      if (maxlod < -9000.0) strcpy(lod_str," -INFINITY");
      else sf(lod_str,"%10.6lf", maxlod);
      sf(ps,"%s %s", pos_str, lod_str);
      if (!pmod_calc_flag)
        pr();
      else if (replicates_store_result && pmod_calc_flag)
        fpr(replifp);
    }
    if (modcalc == 2) {
      maxlod = total_lod[pos];
      maxlodpos = pos;
      if (long_mod_output) {
        sf(pos_str,"%7.2lf %7.2lf %7.2lf",0.5*(wpos_m[maxlodpos] + wpos_f[maxlodpos]), wpos_f[maxlodpos], wpos_m[maxlodpos]);
        if (maxlod < -9000.0) strcpy(lod_str," -INFINITY");
        else sf(lod_str,"%10.6lf", maxlod);
        sf(ps,"%s %s", pos_str, lod_str);
        if (!pmod_calc_flag)
          pr();
        else if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
    }
  } else if (ind_sp==0) {
    if (modcalc == 1) {
      maxlod = -10000.0;
      maxlodpos = -1;
      for (i=first_mod_pos; i<=last_mod_pos; i++) if (total_lod[i] > maxlod) {
        maxlod = total_lod[i];
        maxlodpos = i;
      }
      sf(pos_str,"%7.2lf",wpos[maxlodpos]);
      if (maxlod < -9000.0) strcpy(lod_str," -INFINITY");
      else sf(lod_str,"%10.6lf", maxlod);
      sf(ps,"%s %s", pos_str, lod_str);
      if (!pmod_calc_flag)
        pr();
      else if (replicates_store_result && pmod_calc_flag)
        fpr(replifp);
    }
    if (modcalc == 2) {
      maxlod = total_lod[pos];
      maxlodpos = pos;
      if (long_mod_output) {
        sf(pos_str,"%7.2lf",wpos[maxlodpos]);
        if (maxlod < -9000.0) strcpy(lod_str," -INFINITY");
        else sf(lod_str,"%10.6lf", maxlod);
        sf(ps,"%s %s", pos_str, lod_str);
	if (!pmod_calc_flag)
          pr();
        else if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
    }
  }
  return(0);
}


int calculate_lod_famnum(int quick_model)
{
  int i, j, dis_geno, vec, new_vec, num_in_ped, old_mom;
  double this_prob, total;
  double *num_par_score=NULL, *this_model_vector=NULL;
  PED_MEMBER *this_famtree;
  NUC_FAMILY *this_nuc_fam;
  double *disease_prob_old_mom = NULL, *disease_prob_new_mom = NULL;
  int *place_allele_old_mom = NULL, *place_allele_new_mom = NULL;
  int this_srnvecs;

  array(num_par_score, save_real_num_vecs[famnum_pointer[famnum]], double);
  init_disease_info();
  this_famtree = famtree[famnum];
  this_nuc_fam = nuc_fam[famnum];
  num_in_ped = save_num_in_ped[famnum];
  old_mom = save_old_mom[famnum];
  this_srnvecs = save_real_num_vecs[famnum_pointer[famnum]];
  if (save_num_inbred[famnum] >= 1) {
      disease_prob_old_mom = this_famtree[old_mom].disease_prob;
      disease_prob_new_mom = this_famtree[num_in_ped].disease_prob;
      place_allele_old_mom = this_famtree[old_mom].place_allele;
      place_allele_new_mom = this_famtree[num_in_ped].place_allele;
  }

  /* sf(ps,"now we are in calculate_lod_famnum, pedigree: %d\n", famnum_pointer[famnum]); pr(); */

  vec = 0;

  for (new_vec=0; new_vec<save_real_num_vecs[famnum_pointer[famnum]]; new_vec++) {
      if (new_vec > 0) {
          vec = 0;
          for (i=0; i<save_npadMS[famnum]; i++)
              vec |= (new_vec & save_padMask[famnum][i]) 
                  << save_padShift[famnum][i];
      }
      fill_placeholder_alleles_1(vec,save_real_non_originals[famnum],
              this_famtree); 
      if (save_num_inbred[famnum] >= 1) {
          /* break the loop! */
          num_par_score[new_vec] = 0.0; 
          for (dis_geno=0; dis_geno<4; dis_geno++) {
              init_disease_probs(num_in_ped, this_famtree);
              disease_prob_new_mom[dis_geno]=1.0;
              j = dis_geno^1;
              disease_prob_old_mom[j]=0.0;
              disease_prob_new_mom[j]=0.0;
              j ^= 2;
              disease_prob_old_mom[j]=0.0;
              disease_prob_new_mom[j]=0.0;
              j ^= 1;
              disease_prob_old_mom[j]=0.0;
              disease_prob_new_mom[j]=0.0;
              if (save_num_nuc_fams_to_peel[famnum] > 0)
                  peel(vec, save_num_bits[famnum], this_famtree, this_nuc_fam);
              place_allele_new_mom[0] = place_allele_old_mom[0];
              place_allele_new_mom[1] = place_allele_old_mom[1];
              this_prob = brute_force_analyze(
                      save_bf_originals[famnum],
                      save_bf_original_list[famnum],
                      save_bf_non_originals[famnum],
                      save_bf_non_original_list[famnum],
                      this_famtree);
              num_par_score[new_vec] += this_prob; 
          }
      } else {
          init_disease_probs(num_in_ped, this_famtree);
          if (save_num_nuc_fams_to_peel[famnum] > 0)
              peel(vec, save_num_bits[famnum], this_famtree, this_nuc_fam);
          num_par_score[new_vec] = 
              brute_force_analyze(
                      save_bf_originals[famnum], save_bf_original_list[famnum],
                      save_bf_non_originals[famnum], save_bf_non_original_list[famnum],
                      this_famtree); 
      }
  } /* end of new_vec loop */

  total=0.0;
  for (j=0; j<this_srnvecs; j++) total += num_par_score[j];
  if (j != this_srnvecs)
      error("num_vecs don't match in modscore routine");

  if (total == 0.0) {
      if (pmod_calc_flag && replicates_store_result) {
          sf(ps, "\n  disease pvectors sum to 0.0 for pedigree %s - SKIPPING THIS MODEL", pedigree_names[famnum]);
          fpr(replifp);
      }
      unarray(num_par_score, double); 
      quick_model_famnum[famnum_pointer[famnum]][current_saved_model] = 0;
      return(-1);
  } else {
      for(j=0; j<save_real_num_vecs[famnum_pointer[famnum]]; j++) num_par_score[j] /= total;
  }

  /** save model (memory already allocated) **/
  this_model_vector = saved_model_vectors[famnum_pointer[famnum]][quick_model];
  for (new_vec=0; new_vec<this_srnvecs; new_vec++)
      this_model_vector[new_vec] = num_par_score[new_vec];
  quick_model_famnum[famnum_pointer[famnum]][quick_model] = 1;

  unarray(num_par_score, double);
  return(0); 
}

void fine_maximization(void)
{
  int changed_parm;

  do {  /*** overall fine maximization loop ***/
    changed_parm = 0;
    /*** first, change only one parameter at a time ***/
    allele_frequency_flag = 0;
    if (modcalc==1 || long_mod_output) {
      sf(ps,"Optimized parameter:  f+/+ (liability class %d)\n", modliab); 
      if (!pmod_calc_flag)
        pr();
      if (replicates_store_result && pmod_calc_flag)
        fpr(replifp);
    }
    if (optimize_parameter(&penetrance[modliab][HOM_UNAFF])) changed_parm=1;
    if (imprinting) {
      if (modcalc==1 || long_mod_output) {
	sf(ps,"Optimized parameter:  fm/+ (liability class %d)\n", modliab);
	if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_parameter(&penetrance[modliab][HET_AU])) changed_parm = 1;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameter:  f+/m (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_parameter(&penetrance[modliab][HET_UA])) changed_parm = 1;
    } else {
      if (modcalc==1 || long_mod_output) {
	sf(ps,"Optimized parameter:  fHet (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_parameter(&penetrance[modliab][HET_AU])) changed_parm = 1;
    }
    if (modcalc==1 || long_mod_output) {
      sf(ps,"Optimized parameter:  fm/m (liability class %d)\n", modliab);
      if (!pmod_calc_flag)
        pr();
      if (replicates_store_result && pmod_calc_flag)
        fpr(replifp);
    }
    if (optimize_parameter(&penetrance[modliab][HOM_AFF])) changed_parm = 1;
    allele_frequency_flag = 1;
    if (modcalc==1 || long_mod_output) {
      sf(ps,"Optimized parameter:  p\n");
      if (!pmod_calc_flag)
        pr();
      if (replicates_store_result && pmod_calc_flag)
        fpr(replifp);
    }
    if (optimize_parameter(&affected_freq)) changed_parm = 1;
    if (changed_parm) continue;  /*** go back to beginning ***/
    if (dimensions == 1) break;  /*** userdefined: change only one parameter at a time ***/

    /*** then, change two parameters at a time ***/
    if (imprinting) {

      allele_frequency_flag = 0;
      if (modcalc==1 || long_mod_output) {
	sf(ps,"Optimized parameters:  f+/+, fm/+ (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_two_parameters(&penetrance[modliab][HOM_UNAFF], &penetrance[modliab][HET_AU])) continue;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  f+/+, f+/m (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_two_parameters(&penetrance[modliab][HOM_UNAFF], &penetrance[modliab][HET_UA])) continue;
      if (modcalc==1 || long_mod_output) {
	sf(ps,"Optimized parameters:  f+/+, fm/m (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_two_parameters(&penetrance[modliab][HOM_UNAFF], &penetrance[modliab][HOM_AFF])) continue;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  fm/+, f+/m (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_two_parameters(&penetrance[modliab][HET_AU], &penetrance[modliab][HET_UA])) continue;
      if (modcalc==1 || long_mod_output) {
	sf(ps,"Optimized parameters:  fm/+, fm/m (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_two_parameters(&penetrance[modliab][HET_AU], &penetrance[modliab][HOM_AFF])) continue;
      if (modcalc==1 || long_mod_output) {
	sf(ps,"Optimized parameters:  f+/m, fm/m (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_two_parameters(&penetrance[modliab][HET_UA], &penetrance[modliab][HOM_AFF])) continue;
      allele_frequency_flag = 1;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  p, f+/+ (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);

      }
      if (optimize_two_parameters(&affected_freq, &penetrance[modliab][HOM_UNAFF])) continue;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  p, fm/+ (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_two_parameters(&affected_freq, &penetrance[modliab][HET_AU])) continue;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  p, f+/m (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_two_parameters(&affected_freq, &penetrance[modliab][HET_UA])) continue;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  p, fm/m (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_two_parameters(&affected_freq, &penetrance[modliab][HOM_AFF])) continue;

    } else {  /* two parameters, no imprinting */

      allele_frequency_flag = 0;
      if (modcalc==1 || long_mod_output) {
	sf(ps,"Optimized parameters:  f+/+, fHet (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);

      }
      if (optimize_two_parameters(&penetrance[modliab][HOM_UNAFF], &penetrance[modliab][HET_AU])) continue;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  f+/+, fm/m (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);

      }
      if (optimize_two_parameters(&penetrance[modliab][HOM_UNAFF], &penetrance[modliab][HOM_AFF])) continue;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  fHet, fm/m (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_two_parameters(&penetrance[modliab][HET_AU], &penetrance[modliab][HOM_AFF])) continue;
      allele_frequency_flag = 1;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  p, f+/+ (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_two_parameters(&affected_freq, &penetrance[modliab][HOM_UNAFF])) continue;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  p, fHet (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_two_parameters(&affected_freq, &penetrance[modliab][HET_AU])) continue;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  p, fm/m (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_two_parameters(&affected_freq, &penetrance[modliab][HOM_AFF])) continue;
    }

    if (dimensions <= 2) break; /*** userdefined: change only up to two parameter at a time ***/

    /*** then, change three parameters at a time ***/
    if (imprinting) {
      allele_frequency_flag = 0;
      if (modcalc==1 || long_mod_output) {
        sf(ps, "Optimized parameters:  f+/+, fm/+, f+/m (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_three_parameters(&penetrance[modliab][HOM_UNAFF], 
				&penetrance[modliab][HET_AU],
				&penetrance[modliab][HET_UA])) continue;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  f+/+, fm/+, fm/m (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_three_parameters(&penetrance[modliab][HOM_UNAFF],
			      &penetrance[modliab][HET_AU],
			      &penetrance[modliab][HOM_AFF])) continue;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  f+/+, f+/m, fm/m (liability class %d)\n",modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_three_parameters(&penetrance[modliab][HOM_UNAFF],
			      &penetrance[modliab][HET_UA],
			      &penetrance[modliab][HOM_AFF])) continue;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  fm/+, f+/m, fm/m (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_three_parameters(&penetrance[modliab][HET_AU],
			      &penetrance[modliab][HET_UA],
			      &penetrance[modliab][HOM_AFF])) continue;
      allele_frequency_flag = 1;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  p, f+/+, fm/+ (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_three_parameters(&affected_freq,
			      &penetrance[modliab][HOM_UNAFF],
			      &penetrance[modliab][HET_AU])) continue;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  p, f+/+, f+/m (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_three_parameters(&affected_freq,
			      &penetrance[modliab][HOM_UNAFF],
			      &penetrance[modliab][HET_UA])) continue;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  p, f+/+, fm/m (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_three_parameters(&affected_freq,
			      &penetrance[modliab][HOM_UNAFF],
			      &penetrance[modliab][HOM_AFF])) continue;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  p, fm/+, f+/m (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
       }
      if (optimize_three_parameters(&affected_freq,
				      &penetrance[modliab][HET_AU],
			      &penetrance[modliab][HET_UA])) continue;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  p, fm/+, fm/m (liability class %d)\n", modliab); 
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);

       }
      if (optimize_three_parameters(&affected_freq,
			      &penetrance[modliab][HET_AU],
			      &penetrance[modliab][HOM_AFF])) continue;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  p, f+/m, fm/m (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);

      }
      if (optimize_three_parameters(&affected_freq,
			      &penetrance[modliab][HET_UA],
			      &penetrance[modliab][HOM_AFF])) continue;

    } else {  /* three parameters, no imprinting */
      allele_frequency_flag = 0;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  f+/+, fHet, fm/m (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_three_parameters(&penetrance[modliab][HOM_UNAFF],
			      &penetrance[modliab][HET_AU],
			      &penetrance[modliab][HOM_AFF])) continue;
      allele_frequency_flag = 1;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  p, f+/+, fHet (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_three_parameters(&affected_freq,
			      &penetrance[modliab][HOM_UNAFF],
			      &penetrance[modliab][HET_AU])) continue;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  p, f+/+, fm/m (liability class %d)\n", modliab); 
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_three_parameters(&affected_freq,
			      &penetrance[modliab][HOM_UNAFF],
			      &penetrance[modliab][HOM_AFF])) continue;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  p, fHet, fm/m (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_three_parameters(&affected_freq,
			      &penetrance[modliab][HET_AU],
			      &penetrance[modliab][HOM_AFF])) continue;
    }
    if (dimensions <= 3) break; /*** userdefined : change only up tp three parameter at a time ***/

    /*** then, change four parameters at a time ***/
    if (imprinting) {
      allele_frequency_flag = 0;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  f+/+, fm/+, f+/m, fm/m (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_four_parameters(&penetrance[modliab][HOM_UNAFF],
			     &penetrance[modliab][HET_AU],
			     &penetrance[modliab][HET_UA],
			     &penetrance[modliab][HOM_AFF])) continue;
      allele_frequency_flag = 1;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  p, f+/+, fm/+, f+/m (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_four_parameters(&affected_freq,
			     &penetrance[modliab][HOM_UNAFF],
			     &penetrance[modliab][HET_AU],
			     &penetrance[modliab][HET_UA])) continue;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  p, f+/+, fm/+, fm/m (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_four_parameters(&affected_freq,
			     &penetrance[modliab][HOM_UNAFF],
			     &penetrance[modliab][HET_AU],
			     &penetrance[modliab][HOM_AFF])) continue;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  p, f+/+, f+/m, fm/m (liability class %d)\n", modliab);
	if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_four_parameters(&affected_freq,
			     &penetrance[modliab][HOM_UNAFF],
			     &penetrance[modliab][HET_UA],
			     &penetrance[modliab][HOM_AFF])) continue;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  p, fm/+, f+/m, fm/m (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_four_parameters(&affected_freq,
			     &penetrance[modliab][HET_AU],
			     &penetrance[modliab][HET_UA],
			     &penetrance[modliab][HOM_AFF])) continue;

    } else {  /* four parameters, no imprinting */

      allele_frequency_flag = 1;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  p, f+/+, fHet, fm/m (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_four_parameters(&affected_freq,
			     &penetrance[modliab][HOM_UNAFF],
			     &penetrance[modliab][HET_AU],
			     &penetrance[modliab][HOM_AFF])) continue;
    }
    if (dimensions <= 4) break; /*** userdefined: change only up to four parameter at a time ***/

    /*** then, change all five parameters at a time, only with imprinting ***/
    if (imprinting) {
      allele_frequency_flag = 1;
      if (modcalc==1 || long_mod_output) {
        sf(ps,"Optimized parameters:  p, f+/+, fm/+, f+/m, fm/m (liability class %d)\n", modliab);
        if (!pmod_calc_flag)
          pr();
        if (replicates_store_result && pmod_calc_flag)
          fpr(replifp);
      }
      if (optimize_five_parameters(&affected_freq,
			     &penetrance[modliab][HOM_UNAFF],
			     &penetrance[modliab][HET_AU],
			     &penetrance[modliab][HET_UA],
			     &penetrance[modliab][HOM_AFF])) continue;
    }
    break;   /*** if we've made it up to here, we are done ***/
  } while (1);  /*** end of overall fine maximization loop ***/
}


void asymmetric_grid_calc(void)
{
  int a,b,c,d,e;

  for (a=1; a<=3; a++) {
    switch(a) {
      case 1: affected_freq = 0.01;   break;
      case 2: affected_freq = 0.05;   break;
      case 3: affected_freq = 0.1;    break;
      default: error("inconsistency in MOD-score routine");  break;
    }
    for (b=1; b<=3; b++) {
      switch(b) {
        case 1: penetrance[modliab][HOM_UNAFF] = 0.001; break;
        case 2: penetrance[modliab][HOM_UNAFF] = 0.01;  break;
        case 3: penetrance[modliab][HOM_UNAFF] = 0.05;  break;
        default: error("inconsistency in MOD-score routine");  break;
      }
      if (imprinting) for (c=1; c<=4; c++) for (d=1; d<=4; d++) for (e=1; e<=4; e++) {
       switch(c) {
         case 1:  penetrance[modliab][HET_AU]  = penetrance[modliab][HOM_UNAFF];  break;
         case 2:  penetrance[modliab][HET_AU]  = 0.10;  break;
         case 3:  penetrance[modliab][HET_AU]  = 0.50;  break;
         case 4:  penetrance[modliab][HET_AU]  = 0.95;  break;
         default: error("inconsistency in MOD-score routine");  break;
       }
       switch(d) {
         case 1:  penetrance[modliab][HET_UA]  = penetrance[modliab][HOM_UNAFF];  break;
         case 2:  penetrance[modliab][HET_UA]  = 0.10;  break;
         case 3:  penetrance[modliab][HET_UA]  = 0.50;  break;
         case 4:  penetrance[modliab][HET_UA]  = 0.95;  break;
         default: error("inconsistency in MOD-score routine");  break;
       }
       switch(e) {
         case 1:  penetrance[modliab][HOM_AFF]  = penetrance[modliab][HOM_UNAFF];  break;
         case 2:  penetrance[modliab][HOM_AFF]  = 0.10;  break;
         case 3:  penetrance[modliab][HOM_AFF]  = 0.50;  break;
         case 4:  penetrance[modliab][HOM_AFF]  = 0.95;  break;
         default: error("inconsistency in MOD-score routine");  break;
       }

       if (penetrance_restriction_violated()
           || allfreq_restriction_violated() || all_penetrances_equal()
           || model_already_checked()) continue;
       modscore_model_incompatible = calculate_lod();
       if (modcount >= MAX_MODELS) error_too_many_models();
       modpen1[modcount] = penetrance[modliab][HOM_UNAFF];
       modpen2[modcount] = penetrance[modliab][HET_AU];
       modpen3[modcount] = penetrance[modliab][HET_UA];
       modpen4[modcount] = penetrance[modliab][HOM_AFF];
       modafreq[modcount] = affected_freq;
       if (modscore_model_incompatible) {
         modpos[modcount] = -1;
         mod[modcount] = -10000.0;
       } else {
         modpos[modcount] = maxlodpos; /* for completeness; not needed */
         mod[modcount] = maxlod;
       }
       if ((!modscore_model_incompatible) &&
           (mod[modcount] > mod[best_model])) {
         best_model = modcount;         
         if (modcalc==1 || long_mod_output) {
           sf(ps,"   new max MOD\n"); 
           if (!pmod_calc_flag)
             pr();
           else if (replicates_store_result && pmod_calc_flag)
             fpr(replifp);
         }
       } else if (modcalc==1 || long_mod_output) {
         if (!pmod_calc_flag)
    	   nl();
         else if (replicates_store_result && pmod_calc_flag)
           fnl(replifp);
       }
       modcount++;
      } else for (c=1; c<=4; c++) for (e=1; e<=4; e++) { /* no imprinting */
        switch(c) {
          case 1:  penetrance[modliab][HET_AU]  = penetrance[modliab][HOM_UNAFF];  break;
          case 2:  penetrance[modliab][HET_AU]  = 0.10;  break;
          case 3:  penetrance[modliab][HET_AU]  = 0.50;  break;
          case 4:  penetrance[modliab][HET_AU]  = 0.95;  break;
          default: error("inconsistency in MOD-score routine");  break;
        }
        switch(e) {
          case 1:  penetrance[modliab][HOM_AFF]  = penetrance[modliab][HOM_UNAFF];  break;
          case 2:  penetrance[modliab][HOM_AFF]  = 0.10;  break;
          case 3:  penetrance[modliab][HOM_AFF]  = 0.50;  break;
          case 4:  penetrance[modliab][HOM_AFF]  = 0.95;  break;
          default: error("inconsistency in MOD-score routine");  break;
        }
        penetrance[modliab][HET_UA] = penetrance[modliab][HET_AU];
        if (penetrance_restriction_violated()
            || allfreq_restriction_violated() || all_penetrances_equal()
            || model_already_checked()) continue;
        modscore_model_incompatible = calculate_lod();
        if (modcount >= MAX_MODELS) error_too_many_models();
        modpen1[modcount] = penetrance[modliab][HOM_UNAFF];
        modpen2[modcount] = penetrance[modliab][HET_AU];
        modpen3[modcount] = penetrance[modliab][HET_UA];
        modpen4[modcount] = penetrance[modliab][HOM_AFF];
        modafreq[modcount] = affected_freq;
        if (modscore_model_incompatible) {
          modpos[modcount] = -1;
          mod[modcount] = -10000.0;
        } else {
          modpos[modcount] = maxlodpos; /* for completeness; not needed */
          mod[modcount] = maxlod;
        }
        if ((!modscore_model_incompatible) &&
            (mod[modcount] > mod[best_model])) {
          best_model = modcount;
          if (modcalc==1 || long_mod_output) {
            sf(ps,"   new max MOD\n"); 
            if (!pmod_calc_flag)
              pr();
            else if (replicates_store_result && pmod_calc_flag)
              fpr(replifp);
          }
        } else if (modcalc==1 || long_mod_output) {
          if (!pmod_calc_flag)
     	    nl();
          else if (replicates_store_result && pmod_calc_flag)
            fnl(replifp);
        }
        modcount++;
      } /* else if (imprinting) for c */
    }  /* for b */
  }
}


void symmetric_grid_calc(void)
{
  int a,b,c,d,e;

  for (a=1; a<=4; a++) {
    switch(a) {
      case 1: affected_freq = 0.01;   break;
      case 2: affected_freq = 0.05;   break;
      case 3: affected_freq = 0.1;    break;
      case 4: affected_freq = 0.5;    break;
      default: error("inconsistency in MOD-score routine");  break;
    }
    for (b=1; b<=7; b++) {
      switch(b) {
	    case 1: penetrance[modliab][HOM_UNAFF] = 0.001;  break;
	    case 2: penetrance[modliab][HOM_UNAFF] = 0.01;   break;
	    case 3: penetrance[modliab][HOM_UNAFF] = 0.05;   break;
	    case 4: penetrance[modliab][HOM_UNAFF] = 0.50;   break;
	    case 5: penetrance[modliab][HOM_UNAFF] = 0.95;   break;
	    case 6: penetrance[modliab][HOM_UNAFF] = 0.99;   break;
	    case 7: penetrance[modliab][HOM_UNAFF] = 0.999;  break;
	    default: error("inconsistency in MOD-score routine");  break;
      }
      if (imprinting) for (c=1; c<=6; c++) for (d=1; d<=6; d++) for (e=1; e<=6; e++) {
	    switch(c) {
	      case 1: penetrance[modliab][HET_AU]  = penetrance[modliab][HOM_UNAFF];  break;
	      case 2: penetrance[modliab][HET_AU]  = 0.05;  break;
	      case 3: penetrance[modliab][HET_AU]  = 0.10;  break;
	      case 4: penetrance[modliab][HET_AU]  = 0.50;  break;
	      case 5: penetrance[modliab][HET_AU]  = 0.90;  break;
	      case 6: penetrance[modliab][HET_AU]  = 0.95;  break;
	      default: error("inconsistency in MOD-score routine");  break;
	    }
	    switch(d) {
	      case 1: penetrance[modliab][HET_UA]  = penetrance[modliab][HOM_UNAFF];  break;
	      case 2: penetrance[modliab][HET_UA]  = 0.05;  break;
	      case 3: penetrance[modliab][HET_UA]  = 0.10;  break;
	      case 4: penetrance[modliab][HET_UA]  = 0.50;  break;
	      case 5: penetrance[modliab][HET_UA]  = 0.90;  break;
	      case 6: penetrance[modliab][HET_UA]  = 0.95;  break;
	      default: error("inconsistency in MOD-score routine");  break;
	    }
	    switch(e) {
	      case 1: penetrance[modliab][HOM_AFF] = penetrance[modliab][HOM_UNAFF];  break;
	      case 2: penetrance[modliab][HOM_AFF] = 0.05;  break;
	      case 3: penetrance[modliab][HOM_AFF] = 0.10;  break;
	      case 4: penetrance[modliab][HOM_AFF] = 0.50;  break;
	      case 5: penetrance[modliab][HOM_AFF] = 0.90;  break;
	      case 6: penetrance[modliab][HOM_AFF] = 0.95;  break;
	      default: error("inconsistency in MOD-score routine");  break;
	    }
	    if (penetrance_restriction_violated()
	        || allfreq_restriction_violated() || all_penetrances_equal()
	        || model_already_checked()) continue;
	    modscore_model_incompatible = calculate_lod();
	    if (modcount >= MAX_MODELS) error_too_many_models();
	    modpen1[modcount] = penetrance[modliab][HOM_UNAFF];
	    modpen2[modcount] = penetrance[modliab][HET_AU];
	    modpen3[modcount] = penetrance[modliab][HET_UA];
	    modpen4[modcount] = penetrance[modliab][HOM_AFF];
	    modafreq[modcount] = affected_freq;
	    if (modscore_model_incompatible) {
	      modpos[modcount] = -1;
	      mod[modcount] = -10000.0;
	    } else {
	      modpos[modcount] = maxlodpos; /* for completeness; not needed */
	      mod[modcount] = maxlod;
	    }
	    if ((!modscore_model_incompatible) &&
	        (mod[modcount] > mod[best_model])) {
	      best_model = modcount;
          if (modcalc==1 || long_mod_output) {
            sf(ps,"   new max MOD\n"); 
            if (!pmod_calc_flag)
              pr();
            else if (replicates_store_result && pmod_calc_flag)
              fpr(replifp);
          }
        } else if (modcalc==1 || long_mod_output) {
          if (!pmod_calc_flag)
            nl();
          else if (replicates_store_result && pmod_calc_flag)
            fnl(replifp);
        }
	    modcount++;
      } else for (c=1; c<=6; c++) for (e=1; e<=6; e++) { /* no imprinting */
	    switch(c) {
	      case 1: penetrance[modliab][HET_AU]  = penetrance[modliab][HOM_UNAFF];  break;
	      case 2: penetrance[modliab][HET_AU]  = 0.05;  break;
	      case 3: penetrance[modliab][HET_AU]  = 0.10;  break;
	      case 4: penetrance[modliab][HET_AU]  = 0.50;  break;
	      case 5: penetrance[modliab][HET_AU]  = 0.90;  break;
	      case 6: penetrance[modliab][HET_AU]  = 0.95;  break;
	      default: error("inconsistency in MOD-score routine");  break;
	    }
	    switch(e) {
	      case 1: penetrance[modliab][HOM_AFF] = penetrance[modliab][HOM_UNAFF];  break;
	      case 2: penetrance[modliab][HOM_AFF] = 0.05;  break;
	      case 3: penetrance[modliab][HOM_AFF] = 0.10;  break;
	      case 4: penetrance[modliab][HOM_AFF] = 0.50;  break;
	      case 5: penetrance[modliab][HOM_AFF] = 0.90;  break;
	      case 6: penetrance[modliab][HOM_AFF] = 0.95;  break;
	      default: error("inconsistency in MOD-score routine");  break;
	    }
	    penetrance[modliab][HET_UA] = penetrance[modliab][HET_AU];
	    if (penetrance_restriction_violated()
	        || allfreq_restriction_violated() || all_penetrances_equal()
	        || model_already_checked()) continue;
	    modscore_model_incompatible = calculate_lod();
	    if (modcount >= MAX_MODELS) error_too_many_models();
	    modpen1[modcount] = penetrance[modliab][HOM_UNAFF];
	    modpen2[modcount] = penetrance[modliab][HET_AU];
	    modpen3[modcount] = penetrance[modliab][HET_UA];
	    modpen4[modcount] = penetrance[modliab][HOM_AFF];
	    modafreq[modcount] = affected_freq;
	    if (modscore_model_incompatible) {
	      modpos[modcount] = -1;
	      mod[modcount] = -10000.0;
	    } else {
	      modpos[modcount] = maxlodpos; /* for completeness; not needed */
	      mod[modcount] = maxlod;
	    }
	    if ((!modscore_model_incompatible) &&
	        (mod[modcount] > mod[best_model])) {
	      best_model = modcount;
          if (modcalc==1 || long_mod_output) {
            sf(ps,"   new max MOD\n"); 
            if (!pmod_calc_flag)
              pr();
            else if (replicates_store_result && pmod_calc_flag)
              fpr(replifp);
          }
        } else if (modcalc==1 || long_mod_output) {
          if (!pmod_calc_flag)
            nl();
          else if (replicates_store_result && pmod_calc_flag)
            fnl(replifp);
        }
	    modcount++;
      } /* else if (imprinting) for c */
    }  /* for b */
  }  /* for a */
}
